Package: GWPR.light
Type: Package
Title: Geographically Weighted Panel Regression
Version: 1.0.0
Authors@R: c(person("Chao", "Li", role = c("aut", "cre"),
                     email = "chaoli0394@gmail.com",
                     comment = c(ORCID = "0000-0002-6854-4456")),
              person("Shunsuke", "Managi", role = "aut",
                     comment = c(ORCID = "0000-0001-7883-1427")))
Description: A modern, first implementation of Geographically Weighted Panel Regression (GWPR) for spatial panel data. The package provides a unified public API supporting Gaussian and binomial family models, within/pooling/random panel effects, three bandwidth search strategies (grid, Stochastic Gradient Descent, random), five kernel functions, and optional parallel execution via the 'future' framework. Diagnostic tools include spatial Moran's I, local F-test, Hausman test, and Lagrange Multiplier test. 
License: AGPL (>= 3)
Encoding: UTF-8
LazyData: true
Imports: fixest, glmmTMB, lmtest, plm, sf, stats, utils
Depends: R (>= 3.5.0)
Suggests: future, future.apply, rmarkdown, knitr, testthat (>= 3.0.0)
VignetteBuilder: knitr
URL: https://github.com/MichaelChaoLi-cpu/GWPR.light
BugReports: https://github.com/MichaelChaoLi-cpu/GWPR.light/issues
Config/testthat/edition: 3
Config/roxygen2/version: 8.0.0
NeedsCompilation: no
Packaged: 2026-05-22 00:18:34 UTC; lichao
Author: Chao Li [aut, cre] (ORCID: <https://orcid.org/0000-0002-6854-4456>),
  Shunsuke Managi [aut] (ORCID: <https://orcid.org/0000-0001-7883-1427>)
Maintainer: Chao Li <chaoli0394@gmail.com>
Repository: CRAN
Date/Publication: 2026-05-29 09:20:08 UTC
Built: R 4.5.2; ; 2026-05-29 10:37:23 UTC; unix

# Generated by roxygen2: do not edit by hand

S3method(print,gwpr_bandwidth)
S3method(print,gwpr_diagnostics)
S3method(print,gwpr_fit)
S3method(summary,gwpr_bandwidth)
S3method(summary,gwpr_diagnostics)
S3method(summary,gwpr_fit)
export(diagnose_gwpr)
export(diagnose_hausman)
export(diagnose_lm)
export(diagnose_local_f)
export(diagnose_moran)
export(fit_gwpr)
export(gwpr)
export(parallel_map)
export(select_bandwidth)
export(with_reproducible_seed)
importFrom(stats,binomial)
importFrom(stats,quantile)

<!-- README.md is generated from README.Rmd. Please edit that file -->

# GWPR.light

<!-- badges: start -->

[![Project Status: Active – The project has reached a stable, usable
state and is being actively
developed.](https://www.repostatus.org/badges/latest/active.svg)](https://www.repostatus.org/#active)
[![R-CMD-check](https://github.com/MichaelChaoLi-cpu/GWPR.light/workflows/R-CMD-check/badge.svg)](https://github.com/MichaelChaoLi-cpu/GWPR.light/actions)
[![CRAN
status](https://www.r-pkg.org/badges/version/GWPR.light)](https://CRAN.R-project.org/package=GWPR.light)
[![License: AGPL
v3](https://img.shields.io/badge/License-AGPL_v3-blue.svg)](https://www.gnu.org/licenses/agpl-3.0)
[![Lifecycle:
stable](https://img.shields.io/badge/lifecycle-stable-brightgreen.svg)](https://lifecycle.r-lib.org/articles/stages.html#stable)

<!-- badges: end -->

A modern, `sf`-first implementation of **Geographically Weighted Panel
Regression (GWPR)** for spatial panel data. Version 1.0.0 provides a
clean public API, three bandwidth search strategies (grid, SGD, random),
Gaussian and binomial model families, and optional parallel execution
via the `future` framework.

## Authors

Chao Li <chaoli0394@gmail.com>  
Shunsuke Managi <managi@doc.kyushu-u.ac.jp>

## Installation

Install the released version from [CRAN](https://CRAN.R-project.org):

``` r
install.packages("GWPR.light")
```

Or install the development version from GitHub:

``` r
# install.packages("remotes")
remotes::install_github("MichaelChaoLi-cpu/GWPR.light")
```

## Quick start (1.0.0 API)

The four public functions form the complete interface:

| Function | Purpose |
|---|---|
| `gwpr()` | Full pipeline: bandwidth search + fit + diagnostics |
| `select_bandwidth()` | Standalone bandwidth optimisation |
| `fit_gwpr()` | Fit with a known bandwidth |
| `diagnose_gwpr()` | Diagnostic tests on a fitted model |

``` r
library(GWPR.light)
library(sf)

# --- Simulate a small spatial panel ---
set.seed(1)
pts <- sf::st_as_sf(
  data.frame(id = 1:6, X = c(0,1,2,0,1,2), Y = c(0,0,0,1,1,1)),
  coords = c("X", "Y"), crs = NA_integer_
)
dat <- data.frame(
  id   = rep(1:6, each = 4),
  time = rep(1:4, 6),
  x1   = rnorm(24),
  x2   = rnorm(24)
)
dat$y <- 1.5 * dat$x1 - 0.8 * dat$x2 + rnorm(24, sd = 0.3)

# --- Fit with a known bandwidth ---
fit <- fit_gwpr(
  y ~ x1 + x2, data = dat, spatial = pts,
  id = "id", time = "time",
  bandwidth = 2, model = "pooling", workers = 1
)
print(fit)
```

### Automatic bandwidth search

``` r
bw <- select_bandwidth(
  y ~ x1 + x2, data = dat, spatial = pts,
  id = "id", time = "time",
  method  = "grid",
  control = list(lower = 0.5, upper = 3, step = 0.5),
  workers = 1
)
cat("Optimal bandwidth:", bw$best_bandwidth, "\n")
```

### Diagnostics

``` r
diag_result <- diagnose_gwpr(fit, diagnostics = c("f_test", "hausman"))
print(diag_result)
```

## Vignettes

- `vignette("gwpr-introduction")` — Getting started with the 1.0.0 API
- `vignette("introduction_of_GWPR")` — Legacy 0.2.x API reference

## References

- Fotheringham, A., Brunsdon, C., Charlton, M. (2002). *Geographically
  Weighted Regression*. John Wiley & Sons. ISBN: 978-0-470-85525-6.
- Beenstock, M., Felsenstein, D. (2019). *The Econometric Analysis of
  Non-Stationary Spatial Panel Data*. Springer.
  ISBN: 978-3-030-03614-0.

## Bug reports

<https://github.com/MichaelChaoLi-cpu/GWPR.light/issues>

## ----include = FALSE----------------------------------------------------------
options(rmarkdown.html_vignette.check_title = FALSE)
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  eval = FALSE
)

## ----example_GWmodel, message=FALSE-------------------------------------------
#  library(GWPR.light)
#  library(sf)
#  library(tmap)
#  data("California")
#  data("TransAirPolCalif")
#  formula.GWPR <- pm25 ~ co2_mean + Developed_Open_Space_perc + Developed_Low_Intensity_perc +
#     Developed_Medium_Intensity_perc + Developed_High_Intensity_perc +
#     Open_Water_perc + Woody_Wetlands_perc + Emergent_Herbaceous_Wetlands_perc +
#     Deciduous_Forest_perc + Evergreen_Forest_perc + Mixed_Forest_perc +
#     Shrub_perc + Grassland_perc + Pasture_perc + Cultivated_Crops_perc +
#     pop_density + summer_tmmx + winter_tmmx + summer_rmax + winter_rmax
#  TransAirPolCalif <- dplyr::filter(TransAirPolCalif, year == 2001)
#  lm.2001 <- lm(formula = formula.GWPR, TransAirPolCalif)

## -----------------------------------------------------------------------------
#  summary(lm.2001)
#  resid.lm <- cbind(lm.2001$residuals, TransAirPolCalif$GEOID)
#  colnames(resid.lm) <- c("resid", "GEOID")
#  head(resid.lm)

## ----cross_sectional_residual, fig.align='center', out.height= '80%', out.width='80%'----
#  to_show <- merge(California, resid.lm, by = "GEOID")
#  tm_shape(to_show) + tm_polygons(col = 'resid')

## ----setup, message=FALSE-----------------------------------------------------
#  library(GWPR.light)

## -----------------------------------------------------------------------------
#  bw.GWPR(formula = formula, data = data, index = index, SDF = SDF,
#          adaptive = F, p = 2,bigdata = F, upperratio = 0.25,
#          effect = "individual", model = c("pooling", "within", "random"),
#          random.method = "swar", approach = c("CV","AIC"), kernel = "bisquare",
#          longlat = F, doParallel = F, human.set.range = F,
#          h.upper = NULL, h.lower = NULL)

## ----out.height="60%", out.width="60%", fig.align="center"--------------------
#  h = 1:10 # just an example, not truth.
#  SS = h^2
#  plot(h, SS, col = "red")

## ----fig.align='center', out.height= '80%', out.width='80%'-------------------
#  tm_shape(to_show) +
#    tm_polygons(col = 'resid')

## -----------------------------------------------------------------------------
#  data(TransAirPolCalif)
#  pdata <- plm::pdata.frame(TransAirPolCalif, index = c("GEOID", "year"))
#  moran.plm.model <- plm::plm(formula = formula.GWPR, data = pdata, model = "within")
#  summary(moran.plm.model)

## -----------------------------------------------------------------------------
#  bw.AIC.F <- bw.GWPR(formula = formula.GWPR, data = TransAirPolCalif, index = c("GEOID", "year"),
#                      SDF = California, adaptive = F, p = 2, bigdata = F, effect = "individual",
#                      model = "within", approach = "AIC", kernel = "bisquare", longlat = F,
#                      doParallel = F)

## -----------------------------------------------------------------------------
#  bw.AIC.F <- 2.010529 #precomputed results from #150:155
#  GWPR.moran.test(moran.plm.model, SDF = California, bw = bw.AIC.F, kernel = "bisquare",
#                   adaptive = F, p = 2, longlat=F, alternative = "greater")

## ----eg.GWPR.pFtest,  fig.align='center', out.height= '80%', out.width='80%'----
#  GWPR.pFtest.resu.F <- GWPR.pFtest(formula = formula.GWPR, data = TransAirPolCalif, index = c("GEOID", "year"),
#                                    SDF = California, bw = bw.AIC.F, adaptive = F, p = 2, effect = "individual",
#                                    kernel = "bisquare", longlat = F)
#  tm_shape(GWPR.pFtest.resu.F$SDF) +
#       tm_polygons(col = "p.value", breaks = c(0, 0.05, 1))

## ----eg.GWPR.plmtest, fig.align='center', out.height= '80%', out.width='80%'----
#  GWPR.plmtest.resu.F <- GWPR.plmtest(formula = formula.GWPR, data = TransAirPolCalif, index = c("GEOID", "year"),
#                                      SDF = California, bw = bw.AIC.F, adaptive = F, p = 2,
#                                      kernel = "bisquare", longlat = F)
#  tm_shape(GWPR.plmtest.resu.F$SDF) +
#       tm_polygons(col = "p.value", breaks = c(0, 0.05, 1))

## ----eg.GWPR.phtest, fig.align='center', out.height= '80%', out.width='80%'----
#  GWPR.phtest.resu.F <- GWPR.phtest(formula = formula.GWPR, data = TransAirPolCalif, index = c("GEOID", "year"),
#                                    SDF = California, bw = bw.AIC.F, adaptive = F, p = 2, effect = "individual",
#                                    kernel = "bisquare", longlat = F, random.method = "amemiya")
#  tm_shape(GWPR.phtest.resu.F$SDF) +
#       tm_polygons(col = "p.value", breaks = c(0, 0.05, 1))

## -----------------------------------------------------------------------------
#  GWPR(formula = formula, data = data, index = index, SDF = SDF, bw = NULL, adaptive = F, p = 2,
#              effect = "individual", model = c("pooling", "within", "random"),
#              random.method = "swar", kernel = "bisquare", longlat = F)

## ----eg.GWPR, fig.align='center', out.height= '80%', out.width='80%'----------
#  result.F.AIC <- GWPR(bw = bw.AIC.F, formula = formula.GWPR, data = TransAirPolCalif, index = c("GEOID", "year"),
#                       SDF = California, adaptive = F, p = 2, effect = "individual", model = "within",
#                       kernel = "bisquare", longlat = F)
#  summary(result.F.AIC$SDF$Local_R2)
#  tm_shape(result.F.AIC$SDF) +
#    tm_polygons(col = "Local_R2", pal = "Reds",auto.palette.mapping = F,
#                style = 'cont')

#' @title Linear GWPR Engine for GWPR.light 1.0.0
#'
#' @description
#' Internal functions for fitting Geographically Weighted Panel Regression
#' with a Gaussian (linear) response.  Supports pooling, within, and random
#' panel models, plus individual, time, two-way, and nested effects.
#'
#' @name model_linear
#' @keywords internal
NULL

# ---------------------------------------------------------------------------
# effect_to_plm
# ---------------------------------------------------------------------------

#' Map user-facing effect string to plm effect parameter
#'
#' @param effect Character scalar: one of `"individual"`, `"time"`,
#'   `"two-way"`, `"nested"`.
#'
#' @return Character scalar accepted by `plm::plm()` for its `effect`
#'   argument: `"individual"`, `"time"`, or `"twoways"`.  For `"nested"`,
#'   the function stops with an informative message because `plm` does not
#'   support nested effects without additional data conventions.
#' @keywords internal
effect_to_plm <- function(effect) {
  switch(
    effect,
    individual = "individual",
    time       = "time",
    `two-way`  = "twoways",
    nested     = stop(
      '`effect = "nested"` requires additional hierarchical level information ',
      "that is not available from `id` and `time` alone. ",
      "Please provide a nested-level column and use a custom formula.",
      call. = FALSE
    ),
    stop(
      sprintf('Unknown effect: "%s". Must be one of: individual, time, two-way, nested.', effect),
      call. = FALSE
    )
  )
}

# ---------------------------------------------------------------------------
# fit_linear_local_model
# ---------------------------------------------------------------------------

#' Fit a single local panel linear model
#'
#' Fits a geographically weighted panel linear model for one focal spatial
#' unit.  Errors are caught and returned as a structured failure result rather
#' than propagating to the caller.
#'
#' @param local_data  A `pdata.frame` or plain `data.frame` with columns for
#'   all formula variables plus the panel indices.
#' @param formula     A formula object.
#' @param model       Character scalar: `"pooling"`, `"within"`, or
#'   `"random"`.
#' @param effect      Character scalar: `"individual"`, `"time"`, `"two-way"`,
#'   or `"nested"`.
#' @param weights     Numeric vector of kernel weights aligned with the rows of
#'   `local_data`.
#' @param index       Character vector of length 2 giving the panel index
#'   column names: `c(id_col, time_col)`.
#' @param random_method Character scalar; estimation method for variance
#'   components when `model = "random"` (default `"swar"`).
#'
#' @return A list with elements:
#' \describe{
#'   \item{`fit`}{The fitted model object, or `NULL` on failure.}
#'   \item{`status`}{`"ok"` or `"failed"`.}
#'   \item{`error`}{`NULL` or character string with the error message.}
#'   \item{`metadata`}{Named list with additional model-fitting metadata,
#'     e.g. flagging single-observation individuals for within models.}
#' }
#' @keywords internal
fit_linear_local_model <- function(local_data, formula, model, effect,
                                   weights, index,
                                   random_method = "swar") {
  model  <- match.arg(model,  choices = c("pooling", "within", "random"))
  effect <- match.arg(effect, choices = c("individual", "time", "two-way", "nested"))

  # Validate effect BEFORE entering tryCatch so that nested/unknown effects
  # propagate to the caller rather than being silently caught.
  if (model != "pooling") {
    plm_effect <- effect_to_plm(effect)   # stops for "nested"
  } else {
    plm_effect <- NULL
  }

  metadata <- list()

  # Embed kernel weights into the data frame as a dedicated column.
  # This avoids lazy-evaluation scoping issues that arise when a function
  # parameter named `weights` is passed to lm()/plm(), whose own `weights`
  # argument triggers name-shadowing in the calling environment.
  .local <- local_data
  .local[[".gwpr_wt"]] <- weights
  .wts <- weights   # plain numeric copy in this scope

  result <- tryCatch({

    if (model == "pooling") {
      # Weighted OLS via lm() — use do.call to keep data visible
      fit <- do.call(
        stats::lm,
        list(formula = formula, data = quote(.local), weights = .wts),
        envir = environment()
      )

    } else {
      # plm_effect was already computed and validated above

      # Warn / record single-observation individuals for within models
      if (model == "within") {
        id_col  <- index[1]
        obs_per_id <- table(.local[[id_col]])
        singles <- names(obs_per_id[obs_per_id == 1L])
        if (length(singles) > 0L) {
          metadata$single_obs_individuals <- singles
        }
      }

      # Build pdata.frame
      .pdata <- plm::pdata.frame(
        .local, index = index,
        drop.index = FALSE, row.names = FALSE,
        stringsAsFactors = FALSE
      )
      .pwts <- .pdata[[".gwpr_wt"]]

      fit <- do.call(
        plm::plm,
        list(formula = formula, data = quote(.pdata), model = model,
             effect = plm_effect, index = index, weights = .pwts,
             random.method = random_method),
        envir = environment()
      )
    }

    list(fit = fit, status = "ok", error = NULL, metadata = metadata)

  }, error = function(e) {
    list(fit = NULL, status = "failed", error = conditionMessage(e),
         metadata = metadata)
  })

  result
}

# ---------------------------------------------------------------------------
# extract_linear_local_result
# ---------------------------------------------------------------------------

#' Extract coefficients and diagnostics from a local linear model
#'
#' @param local_result  A list as returned by `fit_linear_local_model()`.
#'
#' @return A named list with elements:
#' \describe{
#'   \item{`coefficients`}{Named numeric vector of local coefficient estimates.}
#'   \item{`se`}{Named numeric vector of standard errors (same names).}
#'   \item{`tvalues`}{Named numeric vector of t-statistics.}
#'   \item{`local_r2`}{Numeric scalar or `NA_real_`.}
#'   \item{`local_aic`}{Numeric scalar or `NA_real_`.}
#'   \item{`status`}{`"ok"` or `"failed"`.}
#'   \item{`error`}{`NULL` or character error message.}
#' }
#' @keywords internal
extract_linear_local_result <- function(local_result) {
  status <- local_result$status
  err    <- local_result$error

  if (status != "ok" || is.null(local_result$fit)) {
    return(list(
      coefficients = NA_real_,
      se           = NA_real_,
      tvalues      = NA_real_,
      local_r2     = NA_real_,
      local_aic    = NA_real_,
      status       = "failed",
      error        = err
    ))
  }

  fit <- local_result$fit

  # lmtest::coeftest works for both lm and plm objects
  ct <- tryCatch(
    lmtest::coeftest(fit),
    error = function(e) NULL
  )

  if (is.null(ct)) {
    coefs  <- tryCatch(stats::coef(fit),    error = function(e) NA_real_)
    ses    <- NA_real_
    tvals  <- NA_real_
  } else {
    coefs <- ct[, 1L]
    ses   <- ct[, 2L]
    tvals <- ct[, 3L]
  }

  # Local R^2
  local_r2 <- tryCatch({
    if (inherits(fit, "plm")) {
      plm::r.squared(fit)
    } else {
      summary(fit)$r.squared
    }
  }, error = function(e) NA_real_)

  # Local AIC
  local_aic <- tryCatch(stats::AIC(fit), error = function(e) NA_real_)

  list(
    coefficients = coefs,
    se           = ses,
    tvalues      = tvals,
    local_r2     = if (is.null(local_r2) || length(local_r2) == 0) NA_real_ else as.numeric(local_r2),
    local_aic    = if (is.null(local_aic) || length(local_aic) == 0) NA_real_ else as.numeric(local_aic),
    status       = "ok",
    error        = NULL
  )
}

# ---------------------------------------------------------------------------
# predict_linear_local_model
# ---------------------------------------------------------------------------

#' Predict response values for a local linear model
#'
#' Returns predicted values for the rows of `local_data`, using the fitted
#' model object stored in `local_result`.
#'
#' @param local_result A list as returned by `fit_linear_local_model()`.
#' @param local_data   The data frame used for prediction (same structure as
#'   the training data).
#'
#' @return Numeric vector of predicted values (same length as `nrow(local_data)`).
#'   Returns a vector of `NA_real_` on failure.
#' @keywords internal
predict_linear_local_model <- function(local_result, local_data) {
  if (local_result$status != "ok" || is.null(local_result$fit)) {
    return(rep(NA_real_, nrow(local_data)))
  }

  fit <- local_result$fit

  # For plm models, predict() issues a warning when newdata is not a
  # pdata.frame. Since the local model is fit on the full panel data,
  # extract fitted values and match to focus rows via the model frame rownames.
  if (inherits(fit, "plm")) {
    fv <- tryCatch(as.numeric(stats::fitted(fit)), error = function(e) NULL)
    if (!is.null(fv) && length(fv) == nrow(local_data)) {
      return(fv)
    }
    # Fallback: predict with pdata.frame conversion
    tryCatch({
      pd_new <- plm::pdata.frame(local_data,
                                 index = names(attr(fit$model, "index")),
                                 drop.index = FALSE, row.names = FALSE)
      suppressWarnings(as.numeric(stats::predict(fit, newdata = pd_new)))
    }, error = function(e) rep(NA_real_, nrow(local_data)))
  } else {
    tryCatch(
      as.numeric(stats::predict(fit, newdata = local_data)),
      error = function(e) rep(NA_real_, nrow(local_data))
    )
  }
}

# ---------------------------------------------------------------------------
# fit_linear_gwpr  (main entry point)
# ---------------------------------------------------------------------------

#' Fit a Geographically Weighted Panel Linear Regression
#'
#' Iterates over each spatial unit (focus), computes kernel weights,
#' fits a local panel linear model, and collects coefficients, predictions,
#' residuals, and summary metrics.
#'
#' @param context         A `gwpr_context` list that must contain at minimum:
#'   `formula`, `model`, `effect`, `panel_data`, `coords`, `id`, `time`,
#'   `kernel`, `adaptive`, and `workers`.  Optionally `seed`.
#' @param bandwidth       Numeric scalar.  Fixed distance or adaptive
#'   (k-nearest-neighbour) bandwidth.
#' @param weights_context Optional pre-built distance context as returned by
#'   `build_distance_context()`.  If `NULL`, one is built from
#'   `context$coords`.
#'
#' @return A named list with elements:
#' \describe{
#'   \item{`local_results`}{A list of per-focus extracted results (one per
#'     unique spatial unit).}
#'   \item{`predictions`}{Named numeric vector of focus-unit predictions
#'     (indexed by the user ID).}
#'   \item{`residuals`}{Named numeric vector of residuals (y - y_hat).}
#'   \item{`metrics`}{A list from `compute_linear_metrics()` computed over
#'     all focus observations.}
#'   \item{`metadata`}{A list with fitting metadata and any per-focus
#'     warnings.}
#' }
#' @noRd
fit_linear_gwpr <- function(context, bandwidth, weights_context = NULL) {

  formula  <- context$formula
  model    <- context$model
  effect   <- context$effect
  panel_data <- context$panel_data
  coords   <- context$coords
  id_col   <- context$id
  time_col <- context$time
  kernel   <- context$kernel
  adaptive <- context$adaptive
  workers  <- context$workers %||% 1L
  seed     <- context$seed

  index <- c(id_col, time_col)

  # Build distance context if not supplied
  if (is.null(weights_context)) {
    # coords rows correspond to unique spatial units
    unit_ids <- rownames(coords)
    if (is.null(unit_ids)) {
      unit_ids <- as.character(seq_len(nrow(coords)))
    }
    weights_context <- build_distance_context(
      coords  = coords,
      ids     = unit_ids,
      longlat = FALSE,
      cache   = TRUE
    )
  }

  unique_ids <- weights_context$ids

  # Helper: fit one focus unit
  fit_one_focus <- function(focus_id) {
    # Kernel weights (one per spatial unit, not per panel row)
    distances <- get_local_distances(weights_context, focus_id = focus_id)
    kw        <- compute_kernel_weights(
      distance  = distances,
      bandwidth = bandwidth,
      kernel    = kernel,
      adaptive  = adaptive
    )
    names(kw) <- unique_ids

    # Join kernel weights to panel rows based on id column
    row_weights <- kw[as.character(panel_data[[id_col]])]
    row_weights[is.na(row_weights)] <- 0

    # Fit local model
    local_res <- fit_linear_local_model(
      local_data = panel_data,
      formula    = formula,
      model      = model,
      effect     = effect,
      weights    = row_weights,
      index      = index
    )

    # Extract coefficients etc.
    extracted <- extract_linear_local_result(local_res)

    # Predict for focus rows only
    focus_rows  <- which(as.character(panel_data[[id_col]]) == focus_id)
    focus_data  <- panel_data[focus_rows, , drop = FALSE]
    preds       <- predict_linear_local_model(local_res, focus_data)
    y_focus     <- focus_data[[all.vars(formula)[1L]]]

    list(
      focus_id   = focus_id,
      extracted  = extracted,
      y          = y_focus,
      y_hat      = preds,
      resid      = y_focus - preds,
      metadata   = local_res$metadata
    )
  }

  # Run (serial for workers == 1, parallel otherwise)
  if (workers == 1L) {
    all_results <- lapply(unique_ids, fit_one_focus)
  } else {
    if (!is.null(seed)) set.seed(seed)
    all_results <- parallel::mclapply(
      unique_ids, fit_one_focus,
      mc.cores = workers, mc.set.seed = TRUE
    )
  }

  names(all_results) <- unique_ids

  # Assemble global predictions and residuals
  all_y    <- unlist(lapply(all_results, `[[`, "y"),    use.names = FALSE)
  all_yhat <- unlist(lapply(all_results, `[[`, "y_hat"), use.names = FALSE)
  all_resid <- unlist(lapply(all_results, `[[`, "resid"), use.names = FALSE)

  pred_named  <- stats::setNames(all_yhat,  rep(unique_ids, vapply(all_results, function(r) length(r$y), integer(1L))))
  resid_named <- stats::setNames(all_resid, names(pred_named))

  metrics <- compute_linear_metrics(all_y, all_yhat)

  local_results <- lapply(all_results, `[[`, "extracted")
  names(local_results) <- unique_ids

  meta <- list(
    n_focuses           = length(unique_ids),
    n_failed            = sum(vapply(local_results, function(r) r$status == "failed", logical(1L))),
    per_focus_metadata  = lapply(all_results, `[[`, "metadata")
  )

  list(
    local_results = local_results,
    predictions   = pred_named,
    residuals     = resid_named,
    metrics       = metrics,
    metadata      = meta
  )
}

# ---------------------------------------------------------------------------
# Utility: %||%  (null coalescing, local to avoid conflicts)
# ---------------------------------------------------------------------------

`%||%` <- function(x, y) if (is.null(x)) y else x

## tests/testthat/test-model-linear.R
## Tests for the Linear GWPR Engine (R/model_linear.R)
##
## All tests use small simulated panel data.  workers = 1 (no parallelism).

# ---------------------------------------------------------------------------
# Helpers: small simulated panel data
# ---------------------------------------------------------------------------

# 4 spatial units, 5 time periods each => 20 rows
make_panel_data <- function(seed = 42L) {
  set.seed(seed)
  n_units <- 4L
  n_times <- 5L
  id   <- rep(as.character(1:n_units), each = n_times)
  time <- rep(1:n_times, times = n_units)
  x1   <- rnorm(n_units * n_times)
  x2   <- rnorm(n_units * n_times)
  y    <- 2 + 0.5 * x1 - 0.3 * x2 + rnorm(n_units * n_times, sd = 0.2)
  data.frame(id = id, time = time, y = y, x1 = x1, x2 = x2,
             stringsAsFactors = FALSE)
}

# 4 units on a small 2x2 grid
make_coords_4 <- function() {
  m <- matrix(
    c(0, 0,
      1, 0,
      0, 1,
      1, 1),
    ncol = 2, byrow = TRUE,
    dimnames = list(c("1", "2", "3", "4"), c("X", "Y"))
  )
  m
}

make_context <- function(model = "pooling", effect = "individual",
                          kernel = "bisquare", adaptive = FALSE,
                          seed = NULL) {
  new_gwpr_context(
    formula      = y ~ x1 + x2,
    family       = "gaussian",
    model        = model,
    effect       = effect,
    id           = "id",
    time         = "time",
    kernel       = kernel,
    adaptive     = adaptive,
    threshold    = 0.5,
    workers      = 1L,
    seed         = seed,
    panel_data   = make_panel_data(),
    coords       = make_coords_4()
  )
}

# ---------------------------------------------------------------------------
# effect_to_plm
# ---------------------------------------------------------------------------

test_that("effect_to_plm maps individual correctly", {
  expect_equal(effect_to_plm("individual"), "individual")
})

test_that("effect_to_plm maps time correctly", {
  expect_equal(effect_to_plm("time"), "time")
})

test_that("effect_to_plm maps two-way to twoways", {
  expect_equal(effect_to_plm("two-way"), "twoways")
})

test_that("effect_to_plm stops for nested without hierarchy info", {
  expect_error(effect_to_plm("nested"), regexp = "nested")
})

test_that("effect_to_plm stops for unknown effect", {
  expect_error(effect_to_plm("unknown"), regexp = "Unknown effect")
})

# ---------------------------------------------------------------------------
# fit_linear_local_model  — pooling
# ---------------------------------------------------------------------------

test_that("fit_linear_local_model: pooling returns ok status", {
  pd  <- make_panel_data()
  wts <- rep(1, nrow(pd))
  res <- fit_linear_local_model(
    local_data = pd,
    formula    = y ~ x1 + x2,
    model      = "pooling",
    effect     = "individual",
    weights    = wts,
    index      = c("id", "time")
  )
  expect_equal(res$status, "ok")
  expect_false(is.null(res$fit))
  expect_true(inherits(res$fit, "lm"))
})

# ---------------------------------------------------------------------------
# fit_linear_local_model  — within
# ---------------------------------------------------------------------------

test_that("fit_linear_local_model: within individual returns ok status", {
  pd  <- make_panel_data()
  wts <- rep(1, nrow(pd))
  res <- fit_linear_local_model(
    local_data = pd,
    formula    = y ~ x1 + x2,
    model      = "within",
    effect     = "individual",
    weights    = wts,
    index      = c("id", "time")
  )
  expect_equal(res$status, "ok")
  expect_false(is.null(res$fit))
  expect_true(inherits(res$fit, "plm"))
})

test_that("fit_linear_local_model: within records single-obs individuals in metadata", {
  # Create data with unit 99 having only 1 observation
  pd_base <- make_panel_data()
  extra   <- data.frame(id = "99", time = 1, y = 0, x1 = 0, x2 = 0,
                        stringsAsFactors = FALSE)
  pd      <- rbind(pd_base, extra)
  wts     <- rep(1, nrow(pd))

  res <- fit_linear_local_model(
    local_data = pd,
    formula    = y ~ x1 + x2,
    model      = "within",
    effect     = "individual",
    weights    = wts,
    index      = c("id", "time")
  )
  # Metadata should record the single-obs individual
  expect_true("single_obs_individuals" %in% names(res$metadata))
  expect_true("99" %in% res$metadata$single_obs_individuals)
})

# ---------------------------------------------------------------------------
# fit_linear_local_model  — random
# ---------------------------------------------------------------------------

test_that("fit_linear_local_model: random individual returns ok status", {
  pd  <- make_panel_data()
  wts <- rep(1, nrow(pd))
  res <- fit_linear_local_model(
    local_data = pd,
    formula    = y ~ x1 + x2,
    model      = "random",
    effect     = "individual",
    weights    = wts,
    index      = c("id", "time")
  )
  expect_equal(res$status, "ok")
  expect_false(is.null(res$fit))
  expect_true(inherits(res$fit, "plm"))
})

# ---------------------------------------------------------------------------
# fit_linear_local_model  — error handling
# ---------------------------------------------------------------------------

test_that("fit_linear_local_model: bad formula gives failed status, not error", {
  pd  <- make_panel_data()
  wts <- rep(1, nrow(pd))
  # Formula references a non-existent variable
  res <- fit_linear_local_model(
    local_data = pd,
    formula    = y ~ nonexistent_var,
    model      = "pooling",
    effect     = "individual",
    weights    = wts,
    index      = c("id", "time")
  )
  expect_equal(res$status, "failed")
  expect_null(res$fit)
  expect_true(is.character(res$error) && nchar(res$error) > 0L)
})

# ---------------------------------------------------------------------------
# extract_linear_local_result
# ---------------------------------------------------------------------------

test_that("extract_linear_local_result: ok result has named numeric vectors", {
  pd  <- make_panel_data()
  wts <- rep(1, nrow(pd))
  local_res <- fit_linear_local_model(
    local_data = pd,
    formula    = y ~ x1 + x2,
    model      = "pooling",
    effect     = "individual",
    weights    = wts,
    index      = c("id", "time")
  )
  ext <- extract_linear_local_result(local_res)

  expect_equal(ext$status, "ok")
  expect_true(is.numeric(ext$coefficients))
  expect_true(is.numeric(ext$se))
  expect_true(is.numeric(ext$tvalues))
  expect_true(is.numeric(ext$local_r2) || is.na(ext$local_r2))
  expect_true(is.numeric(ext$local_aic) || is.na(ext$local_aic))
})

test_that("extract_linear_local_result: failed result returns NA placeholders", {
  failed_res <- list(status = "failed", fit = NULL, error = "test error",
                     metadata = list())
  ext <- extract_linear_local_result(failed_res)

  expect_equal(ext$status, "failed")
  expect_true(is.na(ext$coefficients) || all(is.na(ext$coefficients)))
  expect_true(is.na(ext$local_r2)  || all(is.na(ext$local_r2)))
  expect_true(is.na(ext$local_aic) || all(is.na(ext$local_aic)))
  expect_equal(ext$error, "test error")
})

# ---------------------------------------------------------------------------
# predict_linear_local_model
# ---------------------------------------------------------------------------

test_that("predict_linear_local_model: returns numeric vector of correct length", {
  pd  <- make_panel_data()
  wts <- rep(1, nrow(pd))
  local_res <- fit_linear_local_model(
    local_data = pd,
    formula    = y ~ x1 + x2,
    model      = "pooling",
    effect     = "individual",
    weights    = wts,
    index      = c("id", "time")
  )
  preds <- predict_linear_local_model(local_res, pd)
  expect_true(is.numeric(preds))
  expect_equal(length(preds), nrow(pd))
})

test_that("predict_linear_local_model: failed model returns NA vector", {
  failed_res <- list(status = "failed", fit = NULL, error = "err",
                     metadata = list())
  pd   <- make_panel_data()
  preds <- predict_linear_local_model(failed_res, pd)
  expect_true(all(is.na(preds)))
  expect_equal(length(preds), nrow(pd))
})

# ---------------------------------------------------------------------------
# fit_linear_gwpr — smoke tests
# ---------------------------------------------------------------------------

test_that("fit_linear_gwpr: pooling smoke test returns correct structure", {
  ctx <- make_context(model = "pooling", effect = "individual",
                      kernel = "bisquare", adaptive = TRUE)
  res <- fit_linear_gwpr(ctx, bandwidth = 3L)

  expect_true(is.list(res))
  expect_true("local_results" %in% names(res))
  expect_true("predictions"   %in% names(res))
  expect_true("residuals"     %in% names(res))
  expect_true("metrics"       %in% names(res))
  expect_true("metadata"      %in% names(res))

  # 4 spatial units => 4 local results
  expect_equal(length(res$local_results), 4L)
  # Metrics fields
  expect_true(all(c("R2", "MSE", "RMSE", "MAE") %in% names(res$metrics)))
})

test_that("fit_linear_gwpr: within smoke test completes without error", {
  ctx <- make_context(model = "within", effect = "individual",
                      kernel = "gaussian", adaptive = FALSE)
  expect_no_error({
    res <- fit_linear_gwpr(ctx, bandwidth = 2)
  })
  expect_equal(length(res$local_results), 4L)
})

test_that("fit_linear_gwpr: random smoke test completes without error", {
  ctx <- make_context(model = "random", effect = "individual",
                      kernel = "bisquare", adaptive = FALSE)
  expect_no_error({
    res <- fit_linear_gwpr(ctx, bandwidth = 2)
  })
  expect_equal(length(res$local_results), 4L)
})

# ---------------------------------------------------------------------------
# fit_linear_gwpr — effect parameter parsing
# ---------------------------------------------------------------------------

test_that("fit_linear_gwpr: effect=individual runs without error (pooling)", {
  ctx <- make_context(model = "pooling", effect = "individual", adaptive = FALSE)
  expect_no_error(fit_linear_gwpr(ctx, bandwidth = 2))
})

test_that("fit_linear_gwpr: effect=time runs without error (within)", {
  ctx <- make_context(model = "within", effect = "time", adaptive = FALSE)
  # within + time effect uses plm effect="time"
  expect_no_error(fit_linear_gwpr(ctx, bandwidth = 2))
})

test_that("fit_linear_gwpr: effect=two-way runs without error (within)", {
  ctx <- make_context(model = "within", effect = "two-way", adaptive = FALSE)
  expect_no_error(fit_linear_gwpr(ctx, bandwidth = 2))
})

test_that("fit_linear_gwpr: effect=nested stops with informative error", {
  ctx <- make_context(model = "within", effect = "nested", adaptive = FALSE)
  expect_error(fit_linear_gwpr(ctx, bandwidth = 2), regexp = "nested")
})

# ---------------------------------------------------------------------------
# fit_linear_gwpr — failure robustness
# ---------------------------------------------------------------------------

test_that("fit_linear_gwpr: local model failure does not break overall structure", {
  # Use a formula with a non-existent predictor to force local model failures
  pd  <- make_panel_data()
  ctx <- new_gwpr_context(
    formula    = y ~ nonexistent,   # will fail every local model
    family     = "gaussian",
    model      = "pooling",
    effect     = "individual",
    id         = "id",
    time       = "time",
    kernel     = "bisquare",
    adaptive   = TRUE,
    threshold  = 0.5,
    workers    = 1L,
    panel_data = pd,
    coords     = make_coords_4()
  )

  res <- fit_linear_gwpr(ctx, bandwidth = 3L)

  # Structure intact
  expect_true("local_results" %in% names(res))
  expect_equal(length(res$local_results), 4L)
  # All failed
  statuses <- vapply(res$local_results, `[[`, character(1L), "status")
  expect_true(all(statuses == "failed"))
  # Metadata records failure count
  expect_equal(res$metadata$n_failed, 4L)
})

# ---------------------------------------------------------------------------
# fit_linear_gwpr — reproducibility with fixed weights
# ---------------------------------------------------------------------------

test_that("fit_linear_gwpr: identical weights give identical results", {
  ctx <- make_context(model = "pooling", effect = "individual",
                      kernel = "bisquare", adaptive = TRUE)

  res1 <- fit_linear_gwpr(ctx, bandwidth = 3L)
  res2 <- fit_linear_gwpr(ctx, bandwidth = 3L)

  # Coefficients of first focus should be identical
  expect_equal(
    res1$local_results[[1]]$coefficients,
    res2$local_results[[1]]$coefficients
  )
  expect_equal(res1$predictions, res2$predictions)
})

# ---------------------------------------------------------------------------
# fit_linear_gwpr — pre-supplied weights_context
# ---------------------------------------------------------------------------

test_that("fit_linear_gwpr: accepts pre-built weights_context", {
  ctx    <- make_context(model = "pooling", effect = "individual",
                         kernel = "bisquare", adaptive = TRUE)
  wctx   <- build_distance_context(
    coords  = ctx$coords,
    ids     = rownames(ctx$coords),
    longlat = FALSE,
    cache   = TRUE
  )
  expect_no_error(fit_linear_gwpr(ctx, bandwidth = 3L, weights_context = wctx))
})

#' @title Result Object Module for GWPR.light 1.0.0
#'
#' @description
#' S3 classes and constructor functions for the three result objects returned
#' by the GWPR.light public API: `gwpr_fit`, `gwpr_bandwidth`, and
#' `gwpr_diagnostics`.  Also provides `build_spatial_results()` for assembling
#' a data.frame that can be aligned with an `sf` geometry column.
#'
#' @name result
#' @importFrom stats quantile
NULL

# ---------------------------------------------------------------------------
# gwpr_fit
# ---------------------------------------------------------------------------

#' Construct a new gwpr_fit object
#'
#' Creates a standardised result object for a fitted geographically weighted
#' panel regression model.
#'
#' @param call       The matched call from the top-level API function, or
#'   `NULL`.
#' @param family     Character; model family: `"gaussian"` or `"binomial"`.
#' @param formula    A formula object, or `NULL`.
#' @param model      Character; panel model type: `"pooling"`, `"within"`, or
#'   `"random"`.
#' @param effect     Character; panel effect: `"individual"`, `"time"`,
#'   `"two-way"`, or `"nested"`.
#' @param bandwidth  Numeric scalar; the bandwidth used for fitting.
#' @param search     A `gwpr_bandwidth` object, or `NULL` when bandwidth was
#'   supplied directly.
#' @param global_model A summary / model object for the global (non-spatial)
#'   reference model, or `NULL`.
#' @param local_results A list of per-unit local model results.
#' @param predictions Numeric vector of in-sample predicted values, or `NULL`.
#' @param residuals  Numeric vector of residuals, or `NULL`.
#' @param metrics    Named list of overall goodness-of-fit metrics.
#' @param spatial_results A data.frame produced by `build_spatial_results()`,
#'   or `NULL`.
#' @param metadata   Named list of supplementary metadata.
#' @param warnings   Character vector of warnings accumulated during fitting.
#' @param ...        Additional named fields stored in the object.
#'
#' @return A named list of class `"gwpr_fit"`.
#'
#' @examples
#' fit <- new_gwpr_fit(
#'   family  = "gaussian",
#'   model   = "within",
#'   effect  = "individual",
#'   bandwidth = 100,
#'   metrics = list(R2 = 0.8, MSE = 0.1, RMSE = 0.316, MAE = 0.25),
#'   warnings = character()
#' )
#' print(fit)
#'
#' @noRd
new_gwpr_fit <- function(
    call           = NULL,
    family         = NULL,
    formula        = NULL,
    model          = NULL,
    effect         = NULL,
    bandwidth      = NULL,
    search         = NULL,
    global_model   = NULL,
    local_results  = list(),
    predictions    = NULL,
    residuals      = NULL,
    metrics        = list(),
    spatial_results = NULL,
    metadata       = list(),
    warnings       = character(),
    ...
) {
  extra <- list(...)

  obj <- c(
    list(
      call            = call,
      family          = family,
      formula         = formula,
      model           = model,
      effect          = effect,
      bandwidth       = bandwidth,
      search          = search,
      global_model    = global_model,
      local_results   = local_results,
      predictions     = predictions,
      residuals       = residuals,
      metrics         = metrics,
      spatial_results = spatial_results,
      metadata        = metadata,
      warnings        = warnings
    ),
    extra
  )

  structure(obj, class = "gwpr_fit")
}

#' Print a gwpr_fit object
#'
#' Displays a concise summary of a fitted GWPR model: family, panel model
#' type, effect, bandwidth, and top-level goodness-of-fit metrics.
#'
#' @param x   A `gwpr_fit` object.
#' @param ... Currently ignored.
#'
#' @return Invisibly returns `x`.
#'
#' @examples
#' \donttest{
#' library(sf)
#' pts <- sf::st_as_sf(
#'   data.frame(id = 1:4, X = c(0,1,0,1), Y = c(0,0,1,1)),
#'   coords = c("X", "Y"), crs = NA_integer_
#' )
#' dat <- data.frame(id = rep(1:4, each = 5), time = rep(1:5, 4),
#'                   y = rnorm(20), x1 = rnorm(20))
#' fit <- fit_gwpr(y ~ x1, data = dat, spatial = pts, id = "id",
#'                 time = "time", bandwidth = 2, workers = 1)
#' print(fit)
#' }
#'
#' @export
print.gwpr_fit <- function(x, ...) {
  cat("Geographically Weighted Panel Regression (gwpr_fit)\n")
  cat("----------------------------------------------------\n")
  cat("Family   :", x$family  %||% "<unknown>", "\n")
  cat("Model    :", x$model   %||% "<unknown>", "\n")
  cat("Effect   :", x$effect  %||% "<unknown>", "\n")
  cat("Bandwidth:", x$bandwidth %||% "<unknown>", "\n")

  if (length(x$metrics) > 0L) {
    cat("Metrics  :\n")
    for (nm in names(x$metrics)) {
      val <- x$metrics[[nm]]
      cat(sprintf("  %-12s %s\n", nm,
                  if (is.numeric(val)) formatC(val, digits = 4, format = "g")
                  else as.character(val)))
    }
  }

  if (length(x$warnings) > 0L) {
    cat("Warnings : (", length(x$warnings), ")\n", sep = "")
  }

  invisible(x)
}

#' Summarise a gwpr_fit object
#'
#' Prints the global model overview, quantile summary of local coefficients
#' (when available), and goodness-of-fit metrics.
#'
#' @param object A `gwpr_fit` object.
#' @param ...    Currently ignored.
#'
#' @return Invisibly returns `object`.
#'
#' @examples
#' \donttest{
#' library(sf)
#' pts <- sf::st_as_sf(
#'   data.frame(id = 1:4, X = c(0,1,0,1), Y = c(0,0,1,1)),
#'   coords = c("X", "Y"), crs = NA_integer_
#' )
#' dat <- data.frame(id = rep(1:4, each = 5), time = rep(1:5, 4),
#'                   y = rnorm(20), x1 = rnorm(20))
#' fit <- fit_gwpr(y ~ x1, data = dat, spatial = pts, id = "id",
#'                 time = "time", bandwidth = 2, workers = 1)
#' summary(fit)
#' }
#'
#' @exportS3Method summary gwpr_fit
summary.gwpr_fit <- function(object, ...) {
  cat("=== GWPR Fit Summary ===\n\n")

  cat("Family   :", object$family  %||% "<unknown>", "\n")
  cat("Model    :", object$model   %||% "<unknown>", "\n")
  cat("Effect   :", object$effect  %||% "<unknown>", "\n")
  cat("Bandwidth:", object$bandwidth %||% "<unknown>", "\n\n")

  # Global model
  if (!is.null(object$global_model)) {
    cat("--- Global Model ---\n")
    print(object$global_model)
    cat("\n")
  }

  # Local coefficients quantile summary
  if (!is.null(object$spatial_results) && nrow(object$spatial_results) > 0L) {
    coef_cols <- grep("^coef_", names(object$spatial_results), value = TRUE)
    if (length(coef_cols) > 0L) {
      cat("--- Local Coefficient Quantiles ---\n")
      for (col in coef_cols) {
        vals <- object$spatial_results[[col]]
        if (is.numeric(vals)) {
          q <- quantile(vals, probs = c(0, .25, .5, .75, 1), na.rm = TRUE)
          cat(sprintf("  %-20s Min=%.4g  Q1=%.4g  Med=%.4g  Q3=%.4g  Max=%.4g\n",
                      col, q[1], q[2], q[3], q[4], q[5]))
        }
      }
      cat("\n")
    }
  }

  # Metrics
  if (length(object$metrics) > 0L) {
    cat("--- Metrics ---\n")
    for (nm in names(object$metrics)) {
      val <- object$metrics[[nm]]
      cat(sprintf("  %-12s %s\n", nm,
                  if (is.numeric(val)) formatC(val, digits = 4, format = "g")
                  else as.character(val)))
    }
    cat("\n")
  }

  if (length(object$warnings) > 0L) {
    cat("Warnings: (", length(object$warnings), " recorded)\n", sep = "")
  }

  invisible(object)
}

# ---------------------------------------------------------------------------
# gwpr_bandwidth
# ---------------------------------------------------------------------------

#' Construct a new gwpr_bandwidth object
#'
#' Creates a standardised result object for a bandwidth search.
#'
#' @param method           Character; search method: `"grid"`, `"sgd"`, or
#'   `"random"`.
#' @param best_bandwidth   Numeric scalar; the selected bandwidth.
#' @param best_score       Numeric scalar; the criterion score at the best
#'   bandwidth.
#' @param criterion        Character; name of the scoring criterion used.
#' @param history          A data.frame with one row per candidate / epoch.
#' @param metrics_history  A data.frame with metric values per candidate /
#'   epoch, or `NULL`.
#' @param seed             Integer random seed that was used, or `NULL`.
#' @param convergence_info Named list with convergence diagnostics, or `NULL`.
#' @param elapsed_time     Numeric scalar; total elapsed time in seconds.
#' @param warnings         Character vector of warnings.
#' @param ...              Additional named fields.
#'
#' @return A named list of class `"gwpr_bandwidth"`.
#'
#' @examples
#' bw <- new_gwpr_bandwidth(
#'   method         = "grid",
#'   best_bandwidth = 150,
#'   best_score     = 0.42,
#'   criterion      = "MSE",
#'   history        = data.frame(bandwidth = c(100, 150, 200),
#'                               score     = c(0.5, 0.42, 0.48)),
#'   elapsed_time   = 1.2
#' )
#' print(bw)
#'
#' @noRd
new_gwpr_bandwidth <- function(
    method           = NULL,
    best_bandwidth   = NULL,
    best_score       = NULL,
    criterion        = NULL,
    history          = NULL,
    metrics_history  = NULL,
    seed             = NULL,
    convergence_info = NULL,
    elapsed_time     = NULL,
    warnings         = character(),
    ...
) {
  extra <- list(...)

  obj <- c(
    list(
      method           = method,
      best_bandwidth   = best_bandwidth,
      best_score       = best_score,
      criterion        = criterion,
      history          = history,
      metrics_history  = metrics_history,
      seed             = seed,
      convergence_info = convergence_info,
      elapsed_time     = elapsed_time,
      warnings         = warnings
    ),
    extra
  )

  structure(obj, class = "gwpr_bandwidth")
}

#' Print a gwpr_bandwidth object
#'
#' Displays the search method, best bandwidth, criterion score, and number of
#' iterations explored.
#'
#' @param x   A `gwpr_bandwidth` object.
#' @param ... Currently ignored.
#'
#' @return Invisibly returns `x`.
#'
#' @examples
#' \donttest{
#' library(sf)
#' pts <- sf::st_as_sf(
#'   data.frame(id = 1:4, X = c(0,1,0,1), Y = c(0,0,1,1)),
#'   coords = c("X", "Y"), crs = NA_integer_
#' )
#' dat <- data.frame(id = rep(1:4, each = 5), time = rep(1:5, 4),
#'                   y = rnorm(20), x1 = rnorm(20))
#' bw <- select_bandwidth(y ~ x1, data = dat, spatial = pts, id = "id",
#'   time = "time", method = "grid",
#'   control = list(lower = 1, upper = 3, step = 1), workers = 1)
#' print(bw)
#' }
#'
#' @export
print.gwpr_bandwidth <- function(x, ...) {
  cat("Bandwidth Search Result (gwpr_bandwidth)\n")
  cat("----------------------------------------\n")
  cat("Method        :", x$method        %||% "<unknown>", "\n")
  cat("Criterion     :", x$criterion     %||% "<unknown>", "\n")
  cat("Best bandwidth:", x$best_bandwidth %||% "<unknown>", "\n")
  cat("Best score    :", x$best_score    %||% "<unknown>", "\n")

  if (!is.null(x$history) && is.data.frame(x$history)) {
    cat("Iterations    :", nrow(x$history), "\n")
  }

  if (!is.null(x$elapsed_time)) {
    cat("Elapsed (s)   :", round(x$elapsed_time, 3), "\n")
  }

  if (length(x$warnings) > 0L) {
    cat("Warnings      : (", length(x$warnings), ")\n", sep = "")
  }

  invisible(x)
}

#' Summarise a gwpr_bandwidth object
#'
#' Prints the search method, best bandwidth, and a brief history overview.
#'
#' @param object A `gwpr_bandwidth` object.
#' @param ...    Currently ignored.
#'
#' @return Invisibly returns `object`.
#'
#' @examples
#' \donttest{
#' library(sf)
#' pts <- sf::st_as_sf(
#'   data.frame(id = 1:4, X = c(0,1,0,1), Y = c(0,0,1,1)),
#'   coords = c("X", "Y"), crs = NA_integer_
#' )
#' dat <- data.frame(id = rep(1:4, each = 5), time = rep(1:5, 4),
#'                   y = rnorm(20), x1 = rnorm(20))
#' bw <- select_bandwidth(y ~ x1, data = dat, spatial = pts, id = "id",
#'   time = "time", method = "grid",
#'   control = list(lower = 1, upper = 3, step = 1), workers = 1)
#' summary(bw)
#' }
#'
#' @exportS3Method summary gwpr_bandwidth
summary.gwpr_bandwidth <- function(object, ...) {
  cat("=== Bandwidth Search Summary ===\n\n")
  cat("Method        :", object$method        %||% "<unknown>", "\n")
  cat("Criterion     :", object$criterion     %||% "<unknown>", "\n")
  cat("Best bandwidth:", object$best_bandwidth %||% "<unknown>", "\n")
  cat("Best score    :", object$best_score    %||% "<unknown>", "\n")

  if (!is.null(object$seed)) {
    cat("Seed          :", object$seed, "\n")
  }

  if (!is.null(object$elapsed_time)) {
    cat("Elapsed (s)   :", round(object$elapsed_time, 3), "\n")
  }

  if (!is.null(object$history) && is.data.frame(object$history) &&
      nrow(object$history) > 0L) {
    cat("\n--- Search History (", nrow(object$history), " rows) ---\n", sep = "")
    print(utils::head(object$history, 6L))
    if (nrow(object$history) > 6L) {
      cat("  ... (", nrow(object$history) - 6L, " more rows)\n", sep = "")
    }
  }

  if (!is.null(object$convergence_info)) {
    cat("\n--- Convergence ---\n")
    for (nm in names(object$convergence_info)) {
      cat(sprintf("  %-20s %s\n", nm,
                  as.character(object$convergence_info[[nm]])))
    }
  }

  if (length(object$warnings) > 0L) {
    cat("\nWarnings: (", length(object$warnings), " recorded)\n", sep = "")
  }

  invisible(object)
}

# ---------------------------------------------------------------------------
# gwpr_diagnostics
# ---------------------------------------------------------------------------

#' Construct a new gwpr_diagnostics object
#'
#' Creates a standardised result object for model diagnostic tests.
#'
#' @param diagnostics     Named list of individual diagnostic test results.
#' @param model_type      Character; `"gaussian"` or `"binomial"`.
#' @param spatial_weights Spatial weights matrix or object, or `NULL`.
#' @param panel_balance   Logical; `TRUE` when the panel is balanced.
#' @param warnings        Character vector of warnings.
#' @param metadata        Named list of supplementary metadata.
#' @param ...             Additional named fields.
#'
#' @return A named list of class `"gwpr_diagnostics"`.
#'
#' @examples
#' diag_obj <- new_gwpr_diagnostics(
#'   diagnostics   = list(moran = list(statistic = 0.12, p_value = 0.03)),
#'   model_type    = "gaussian",
#'   panel_balance = TRUE,
#'   warnings      = character()
#' )
#' print(diag_obj)
#'
#' @noRd
new_gwpr_diagnostics <- function(
    diagnostics     = list(),
    model_type      = NULL,
    spatial_weights = NULL,
    panel_balance   = NULL,
    warnings        = character(),
    metadata        = list(),
    ...
) {
  extra <- list(...)

  obj <- c(
    list(
      diagnostics     = diagnostics,
      model_type      = model_type,
      spatial_weights = spatial_weights,
      panel_balance   = panel_balance,
      warnings        = warnings,
      metadata        = metadata
    ),
    extra
  )

  structure(obj, class = "gwpr_diagnostics")
}

#' Print a gwpr_diagnostics object
#'
#' Displays each diagnostic test name and, where available, its statistic and
#' p-value.
#'
#' @param x   A `gwpr_diagnostics` object.
#' @param ... Currently ignored.
#'
#' @return Invisibly returns `x`.
#'
#' @examples
#' \donttest{
#' library(sf)
#' pts <- sf::st_as_sf(
#'   data.frame(id = 1:4, X = c(0,1,0,1), Y = c(0,0,1,1)),
#'   coords = c("X", "Y"), crs = NA_integer_
#' )
#' dat <- data.frame(id = rep(1:4, each = 5), time = rep(1:5, 4),
#'                   y = rnorm(20), x1 = rnorm(20))
#' fit <- fit_gwpr(y ~ x1, data = dat, spatial = pts, id = "id",
#'                 time = "time", bandwidth = 2, workers = 1)
#' diag_obj <- diagnose_gwpr(fit, diagnostics = c("f_test", "hausman"))
#' print(diag_obj)
#' }
#'
#' @export
print.gwpr_diagnostics <- function(x, ...) {
  cat("GWPR Diagnostics (gwpr_diagnostics)\n")
  cat("------------------------------------\n")
  cat("Model type   :", x$model_type    %||% "<unknown>", "\n")
  cat("Panel balance:", x$panel_balance %||% "<unknown>", "\n")

  if (length(x$diagnostics) > 0L) {
    cat("Tests run    :", paste(names(x$diagnostics), collapse = ", "), "\n")
  } else {
    cat("Tests run    : (none)\n")
  }

  if (length(x$warnings) > 0L) {
    cat("Warnings     : (", length(x$warnings), ")\n", sep = "")
  }

  invisible(x)
}

#' Summarise a gwpr_diagnostics object
#'
#' Prints each diagnostic test result with statistic and p-value where
#' available.
#'
#' @param object A `gwpr_diagnostics` object.
#' @param ...    Currently ignored.
#'
#' @return Invisibly returns `object`.
#'
#' @examples
#' \donttest{
#' library(sf)
#' pts <- sf::st_as_sf(
#'   data.frame(id = 1:4, X = c(0,1,0,1), Y = c(0,0,1,1)),
#'   coords = c("X", "Y"), crs = NA_integer_
#' )
#' dat <- data.frame(id = rep(1:4, each = 5), time = rep(1:5, 4),
#'                   y = rnorm(20), x1 = rnorm(20))
#' fit <- fit_gwpr(y ~ x1, data = dat, spatial = pts, id = "id",
#'                 time = "time", bandwidth = 2, workers = 1)
#' diag_obj <- diagnose_gwpr(fit, diagnostics = c("f_test", "hausman"))
#' summary(diag_obj)
#' }
#'
#' @exportS3Method summary gwpr_diagnostics
summary.gwpr_diagnostics <- function(object, ...) {
  cat("=== GWPR Diagnostics Summary ===\n\n")
  cat("Model type   :", object$model_type    %||% "<unknown>", "\n")
  cat("Panel balance:", object$panel_balance %||% "<unknown>", "\n\n")

  if (length(object$diagnostics) == 0L) {
    cat("No diagnostic tests recorded.\n")
  } else {
    for (test_name in names(object$diagnostics)) {
      res <- object$diagnostics[[test_name]]
      cat("---", test_name, "---\n")
      if (is.list(res)) {
        for (nm in names(res)) {
          cat(sprintf("  %-16s %s\n", nm, as.character(res[[nm]])))
        }
      } else {
        cat(" ", as.character(res), "\n")
      }
    }
  }

  if (length(object$warnings) > 0L) {
    cat("\nWarnings: (", length(object$warnings), " recorded)\n", sep = "")
    for (w in object$warnings) cat("  -", w, "\n")
  }

  invisible(object)
}

# ---------------------------------------------------------------------------
# build_spatial_results
# ---------------------------------------------------------------------------

#' Build a data.frame of spatial results
#'
#' Merges per-unit local coefficients, standard errors, t-statistics, local
#' fit quality metrics, and diagnostic statistics into a single data.frame
#' that can be aligned row-by-row with an `sf` geometry column.
#'
#' @param local_results A named list (one element per spatial unit) where each
#'   element is itself a named list with optional sub-lists:
#'   \describe{
#'     \item{`coefficients`}{Named numeric vector of local coefficients.}
#'     \item{`std_errors`}{Named numeric vector of local standard errors.}
#'     \item{`t_stats`}{Named numeric vector of local t-statistics.}
#'     \item{`metrics`}{Named list / numeric vector of local fit metrics.}
#'     \item{`diagnostics`}{Named list / numeric vector of local diagnostics.}
#'     \item{`status`}{Character scalar; `"ok"` or an error description.}
#'   }
#' @param id_map A named character or integer vector mapping unit IDs to
#'   positions, or `NULL`.  The names of `local_results` are used as unit IDs
#'   when `id_map` is `NULL`.
#' @param geometry An `sf` geometry column (`sfc`), a data.frame with the same
#'   number of rows as unique units, or `NULL`.  Used only for length
#'   validation; the geometry itself is not attached to the returned data.frame.
#'
#' @return A data.frame with one row per spatial unit.  Column names:
#'   \describe{
#'     \item{`unit_id`}{Unit identifier.}
#'     \item{`status`}{Fit status (`"ok"` or error description).}
#'     \item{`coef_<name>`}{Local coefficient for each predictor.}
#'     \item{`se_<name>`}{Local standard error for each predictor.}
#'     \item{`tstat_<name>`}{Local t-statistic for each predictor.}
#'     \item{`<metric name>`}{Local fit metric (e.g. `R2`, `MSE`, ...).}
#'     \item{`<diagnostic name>`}{Local diagnostic value.}
#'   }
#'
#' @examples
#' local_res <- list(
#'   "A" = list(
#'     coefficients = c(x1 = 0.5, x2 = 1.2),
#'     std_errors   = c(x1 = 0.1, x2 = 0.3),
#'     t_stats      = c(x1 = 5.0, x2 = 4.0),
#'     metrics      = list(R2 = 0.8, MSE = 0.05),
#'     status       = "ok"
#'   ),
#'   "B" = list(
#'     coefficients = c(x1 = 0.3, x2 = 0.9),
#'     std_errors   = c(x1 = 0.2, x2 = 0.4),
#'     t_stats      = c(x1 = 1.5, x2 = 2.25),
#'     metrics      = list(R2 = 0.6, MSE = 0.12),
#'     status       = "ok"
#'   )
#' )
#' build_spatial_results(local_res)
#'
#' @noRd
build_spatial_results <- function(local_results, id_map = NULL,
                                  geometry = NULL) {
  if (!is.list(local_results) || length(local_results) == 0L) {
    stop("`local_results` must be a non-empty list.", call. = FALSE)
  }

  unit_ids <- names(local_results)
  if (is.null(unit_ids)) {
    unit_ids <- as.character(seq_along(local_results))
  }

  # Validate geometry length if provided
  if (!is.null(geometry)) {
    geom_len <- if (is.data.frame(geometry)) nrow(geometry) else length(geometry)
    if (geom_len != length(unit_ids)) {
      stop(
        "`geometry` length (", geom_len, ") does not match the number of ",
        "units in `local_results` (", length(unit_ids), ").",
        call. = FALSE
      )
    }
  }

  rows <- vector("list", length(unit_ids))

  for (i in seq_along(unit_ids)) {
    uid <- unit_ids[[i]]
    res <- local_results[[i]]

    row <- list(unit_id = uid,
                status  = res$status %||% "ok")

    # Coefficients
    if (!is.null(res$coefficients) && length(res$coefficients) > 0L) {
      coef_names <- names(res$coefficients)
      if (is.null(coef_names)) {
        coef_names <- paste0("V", seq_along(res$coefficients))
      }
      for (j in seq_along(res$coefficients)) {
        row[[paste0("coef_", coef_names[j])]] <- res$coefficients[[j]]
      }
    }

    # Standard errors
    if (!is.null(res$std_errors) && length(res$std_errors) > 0L) {
      se_names <- names(res$std_errors)
      if (is.null(se_names)) {
        se_names <- paste0("V", seq_along(res$std_errors))
      }
      for (j in seq_along(res$std_errors)) {
        row[[paste0("se_", se_names[j])]] <- res$std_errors[[j]]
      }
    }

    # t-statistics
    if (!is.null(res$t_stats) && length(res$t_stats) > 0L) {
      ts_names <- names(res$t_stats)
      if (is.null(ts_names)) {
        ts_names <- paste0("V", seq_along(res$t_stats))
      }
      for (j in seq_along(res$t_stats)) {
        row[[paste0("tstat_", ts_names[j])]] <- res$t_stats[[j]]
      }
    }

    # Metrics
    if (!is.null(res$metrics) && length(res$metrics) > 0L) {
      for (nm in names(res$metrics)) {
        row[[nm]] <- res$metrics[[nm]]
      }
    }

    # Diagnostics
    if (!is.null(res$diagnostics) && length(res$diagnostics) > 0L) {
      for (nm in names(res$diagnostics)) {
        row[[nm]] <- res$diagnostics[[nm]]
      }
    }

    rows[[i]] <- row
  }

  # Separate character columns (unit_id, status) from numeric columns
  char_cols    <- c("unit_id", "status")
  all_row_cols <- unique(unlist(lapply(rows, names)))
  num_cols     <- setdiff(all_row_cols, char_cols)

  # Build numeric columns
  num_list <- lapply(
    stats::setNames(num_cols, num_cols),
    function(col) {
      vapply(rows, function(r) {
        v <- r[[col]]
        if (is.null(v)) NA_real_ else as.numeric(v)
      }, numeric(1L))
    }
  )

  # Build character columns
  unit_id_col <- unit_ids
  status_col  <- vapply(rows, function(r) r$status %||% "ok", character(1L))

  # Assemble final data.frame: unit_id, status first, then numeric columns
  out <- data.frame(
    unit_id = unit_id_col,
    status  = status_col,
    stringsAsFactors = FALSE
  )

  for (col in num_cols) {
    out[[col]] <- num_list[[col]]
  }

  out
}

# ---------------------------------------------------------------------------
# Internal helper
# ---------------------------------------------------------------------------

#' @noRd
`%||%` <- function(x, y) if (!is.null(x)) x else y

## tests/testthat/test-result.R
## Unit tests for result.R – gwpr_fit, gwpr_bandwidth, gwpr_diagnostics and
## build_spatial_results.
## All data is small and synthetic; no parallel execution (workers = 1).

# ---------------------------------------------------------------------------
# Helpers shared across tests
# ---------------------------------------------------------------------------

make_local_results <- function() {
  list(
    "A" = list(
      coefficients = c(x1 = 0.5, x2 = 1.2),
      std_errors   = c(x1 = 0.1, x2 = 0.3),
      t_stats      = c(x1 = 5.0, x2 = 4.0),
      metrics      = list(R2 = 0.8, MSE = 0.05),
      diagnostics  = list(local_moran = 0.12),
      status       = "ok"
    ),
    "B" = list(
      coefficients = c(x1 = 0.3, x2 = 0.9),
      std_errors   = c(x1 = 0.2, x2 = 0.4),
      t_stats      = c(x1 = 1.5, x2 = 2.25),
      metrics      = list(R2 = 0.6, MSE = 0.12),
      diagnostics  = list(local_moran = 0.08),
      status       = "ok"
    ),
    "C" = list(
      coefficients = c(x1 = NA_real_, x2 = NA_real_),
      std_errors   = c(x1 = NA_real_, x2 = NA_real_),
      t_stats      = c(x1 = NA_real_, x2 = NA_real_),
      metrics      = list(R2 = NA_real_, MSE = NA_real_),
      diagnostics  = list(local_moran = NA_real_),
      status       = "singular matrix"
    )
  )
}

# ---------------------------------------------------------------------------
# 1. gwpr_fit constructor
# ---------------------------------------------------------------------------

test_that("new_gwpr_fit returns an object of class gwpr_fit", {
  fit <- new_gwpr_fit()
  expect_s3_class(fit, "gwpr_fit")
})

test_that("new_gwpr_fit has all required fields", {
  fit <- new_gwpr_fit()
  required <- c("call", "family", "formula", "model", "effect",
                "bandwidth", "search", "global_model", "local_results",
                "predictions", "residuals", "metrics",
                "spatial_results", "metadata", "warnings")
  for (fld in required) {
    expect_true(fld %in% names(fit),
                info = paste("Missing field:", fld))
  }
})

test_that("new_gwpr_fit stores supplied values correctly", {
  fit <- new_gwpr_fit(
    family    = "gaussian",
    model     = "within",
    effect    = "individual",
    bandwidth = 100,
    metrics   = list(R2 = 0.8, MSE = 0.1),
    warnings  = c("w1", "w2")
  )
  expect_equal(fit$family,    "gaussian")
  expect_equal(fit$model,     "within")
  expect_equal(fit$effect,    "individual")
  expect_equal(fit$bandwidth, 100)
  expect_equal(fit$metrics$R2, 0.8)
  expect_equal(fit$warnings,  c("w1", "w2"))
})

test_that("new_gwpr_fit warnings field is preserved", {
  warns <- c("local model failed at unit X", "singular matrix at unit Y")
  fit <- new_gwpr_fit(warnings = warns)
  expect_identical(fit$warnings, warns)
})

test_that("print.gwpr_fit runs without error and returns x invisibly", {
  fit <- new_gwpr_fit(
    family    = "gaussian",
    model     = "within",
    effect    = "individual",
    bandwidth = 50,
    metrics   = list(R2 = 0.75, RMSE = 0.5)
  )
  out <- capture.output(ret <- print(fit))
  expect_identical(ret, fit)
  expect_true(any(grepl("gwpr_fit", out, ignore.case = TRUE)))
})

test_that("summary.gwpr_fit runs without error and returns object invisibly", {
  fit <- new_gwpr_fit(
    family    = "gaussian",
    model     = "within",
    effect    = "individual",
    bandwidth = 50,
    metrics   = list(R2 = 0.75, RMSE = 0.5)
  )
  out <- capture.output(ret <- summary(fit))
  expect_identical(ret, fit)
  expect_true(any(grepl("GWPR", out)))
})

test_that("summary.gwpr_fit does not require raw data – uses object fields only", {
  # Build a fit with spatial_results that has coef_ columns
  sr <- data.frame(unit_id = c("A", "B"),
                   status  = c("ok", "ok"),
                   coef_x1 = c(0.5, 0.3))
  fit <- new_gwpr_fit(
    family          = "gaussian",
    model           = "within",
    effect          = "individual",
    bandwidth       = 100,
    metrics         = list(R2 = 0.8),
    spatial_results = sr
  )
  # Should run without needing any raw data
  expect_silent(capture.output(summary(fit)))
})

# ---------------------------------------------------------------------------
# 2. gwpr_bandwidth constructor
# ---------------------------------------------------------------------------

test_that("new_gwpr_bandwidth returns an object of class gwpr_bandwidth", {
  bw <- new_gwpr_bandwidth()
  expect_s3_class(bw, "gwpr_bandwidth")
})

test_that("new_gwpr_bandwidth has all required fields", {
  bw <- new_gwpr_bandwidth()
  required <- c("method", "best_bandwidth", "best_score", "criterion",
                "history", "metrics_history", "seed",
                "convergence_info", "elapsed_time", "warnings")
  for (fld in required) {
    expect_true(fld %in% names(bw),
                info = paste("Missing field:", fld))
  }
})

test_that("new_gwpr_bandwidth stores supplied values correctly", {
  hist_df <- data.frame(bandwidth = c(100, 150, 200),
                        score     = c(0.5, 0.42, 0.48))
  bw <- new_gwpr_bandwidth(
    method         = "grid",
    best_bandwidth = 150,
    best_score     = 0.42,
    criterion      = "MSE",
    history        = hist_df,
    seed           = 42L,
    elapsed_time   = 1.2,
    warnings       = "minor warning"
  )
  expect_equal(bw$method,         "grid")
  expect_equal(bw$best_bandwidth, 150)
  expect_equal(bw$best_score,     0.42)
  expect_equal(bw$criterion,      "MSE")
  expect_equal(nrow(bw$history),  3L)
  expect_equal(bw$seed,           42L)
  expect_equal(bw$elapsed_time,   1.2)
  expect_equal(bw$warnings,       "minor warning")
})

test_that("new_gwpr_bandwidth warnings field is preserved", {
  warns <- c("candidate failed", "numerical issues")
  bw <- new_gwpr_bandwidth(warnings = warns)
  expect_identical(bw$warnings, warns)
})

test_that("print.gwpr_bandwidth runs without error and returns x invisibly", {
  bw <- new_gwpr_bandwidth(
    method         = "grid",
    best_bandwidth = 150,
    best_score     = 0.42,
    criterion      = "MSE",
    history        = data.frame(bandwidth = 150, score = 0.42),
    elapsed_time   = 0.5
  )
  out <- capture.output(ret <- print(bw))
  expect_identical(ret, bw)
  expect_true(any(grepl("gwpr_bandwidth", out, ignore.case = TRUE)))
})

test_that("summary.gwpr_bandwidth runs without error and returns object invisibly", {
  bw <- new_gwpr_bandwidth(
    method         = "grid",
    best_bandwidth = 150,
    best_score     = 0.42,
    criterion      = "MSE",
    history        = data.frame(bandwidth = c(100, 150), score = c(0.5, 0.42)),
    elapsed_time   = 1.2
  )
  out <- capture.output(ret <- summary(bw))
  expect_identical(ret, bw)
  expect_true(any(grepl("Bandwidth", out)))
})

# ---------------------------------------------------------------------------
# 3. gwpr_diagnostics constructor
# ---------------------------------------------------------------------------

test_that("new_gwpr_diagnostics returns an object of class gwpr_diagnostics", {
  d <- new_gwpr_diagnostics()
  expect_s3_class(d, "gwpr_diagnostics")
})

test_that("new_gwpr_diagnostics has all required fields", {
  d <- new_gwpr_diagnostics()
  required <- c("diagnostics", "model_type", "spatial_weights",
                "panel_balance", "warnings", "metadata")
  for (fld in required) {
    expect_true(fld %in% names(d),
                info = paste("Missing field:", fld))
  }
})

test_that("new_gwpr_diagnostics stores supplied values correctly", {
  diag_list <- list(
    moran  = list(statistic = 0.12, p_value = 0.03),
    f_test = list(statistic = 4.5,  p_value = 0.01)
  )
  d <- new_gwpr_diagnostics(
    diagnostics   = diag_list,
    model_type    = "gaussian",
    panel_balance = TRUE,
    warnings      = "unbalanced warning"
  )
  expect_equal(d$model_type,    "gaussian")
  expect_true(d$panel_balance)
  expect_equal(d$diagnostics$moran$statistic, 0.12)
  expect_equal(d$warnings, "unbalanced warning")
})

test_that("new_gwpr_diagnostics warnings field is preserved", {
  warns <- c("unbalanced panel", "Moran's I not available")
  d <- new_gwpr_diagnostics(warnings = warns)
  expect_identical(d$warnings, warns)
})

test_that("print.gwpr_diagnostics runs without error and returns x invisibly", {
  d <- new_gwpr_diagnostics(
    diagnostics   = list(moran = list(statistic = 0.1, p_value = 0.05)),
    model_type    = "gaussian",
    panel_balance = TRUE
  )
  out <- capture.output(ret <- print(d))
  expect_identical(ret, d)
  expect_true(any(grepl("gwpr_diagnostics", out, ignore.case = TRUE)))
})

test_that("summary.gwpr_diagnostics runs without error and returns object invisibly", {
  d <- new_gwpr_diagnostics(
    diagnostics   = list(moran = list(statistic = 0.12, p_value = 0.03)),
    model_type    = "gaussian",
    panel_balance = TRUE
  )
  out <- capture.output(ret <- summary(d))
  expect_identical(ret, d)
  expect_true(any(grepl("Diagnostics", out)))
})

# ---------------------------------------------------------------------------
# 4. build_spatial_results
# ---------------------------------------------------------------------------

test_that("build_spatial_results returns a data.frame", {
  lr  <- make_local_results()
  out <- build_spatial_results(lr)
  expect_s3_class(out, "data.frame")
})

test_that("build_spatial_results has one row per unit", {
  lr  <- make_local_results()
  out <- build_spatial_results(lr)
  expect_equal(nrow(out), length(lr))
})

test_that("build_spatial_results has unit_id and status columns", {
  lr  <- make_local_results()
  out <- build_spatial_results(lr)
  expect_true("unit_id" %in% names(out))
  expect_true("status"  %in% names(out))
})

test_that("build_spatial_results has coef_ columns for each predictor", {
  lr  <- make_local_results()
  out <- build_spatial_results(lr)
  expect_true("coef_x1" %in% names(out))
  expect_true("coef_x2" %in% names(out))
})

test_that("build_spatial_results has se_ and tstat_ columns", {
  lr  <- make_local_results()
  out <- build_spatial_results(lr)
  expect_true("se_x1"    %in% names(out))
  expect_true("tstat_x1" %in% names(out))
})

test_that("build_spatial_results includes metric columns", {
  lr  <- make_local_results()
  out <- build_spatial_results(lr)
  expect_true("R2"  %in% names(out))
  expect_true("MSE" %in% names(out))
})

test_that("build_spatial_results includes diagnostic columns", {
  lr  <- make_local_results()
  out <- build_spatial_results(lr)
  expect_true("local_moran" %in% names(out))
})

test_that("build_spatial_results preserves failed unit status", {
  lr  <- make_local_results()
  out <- build_spatial_results(lr)
  expect_equal(out$status[out$unit_id == "C"], "singular matrix")
})

test_that("build_spatial_results unit_id column matches names of local_results", {
  lr  <- make_local_results()
  out <- build_spatial_results(lr)
  expect_equal(out$unit_id, names(lr))
})

test_that("build_spatial_results validates geometry length", {
  lr <- make_local_results()
  # wrong-length geometry (a plain list of length 2 instead of 3)
  bad_geom <- list("geom1", "geom2")
  expect_error(
    build_spatial_results(lr, geometry = bad_geom),
    regexp = "geometry.*length|length.*geometry",
    ignore.case = TRUE
  )
})

test_that("build_spatial_results geometry alignment: correct length passes", {
  lr       <- make_local_results()
  ok_geom  <- list("geom1", "geom2", "geom3")
  out      <- build_spatial_results(lr, geometry = ok_geom)
  expect_equal(nrow(out), 3L)
})

test_that("build_spatial_results stops on empty local_results", {
  expect_error(build_spatial_results(list()), regexp = "non-empty")
})

test_that("build_spatial_results column count consistent with sf alignment", {
  # The number of rows must match geometry for sf cbind to work
  lr      <- make_local_results()
  out     <- build_spatial_results(lr)
  geom_df <- data.frame(dummy = seq_len(nrow(out)))
  expect_equal(nrow(out), nrow(geom_df))
})

# ---------------------------------------------------------------------------
# 5. Null-coalescing helper
# ---------------------------------------------------------------------------

test_that("%||% returns left value when non-NULL", {
  expect_equal("a" %||% "b", "a")
  expect_equal(0L  %||% 1L,  0L)
})

test_that("%||% returns right value when left is NULL", {
  expect_equal(NULL %||% "fallback", "fallback")
})

#' @title Spatial SF Module for GWPR.light 1.0.0
#'
#' @description
#' Internal functions providing the `sf`-first spatial interface. These
#' functions abstract geometry extraction, coordinate derivation, spatial
#' alignment to panel data, and neighbour-structure construction. They are used
#' by the data preparation and diagnostics modules.
#'
#' @name spatial_sf
#' @keywords internal
NULL

# ---------------------------------------------------------------------------
# assert_sf
# ---------------------------------------------------------------------------

#' Assert that an object is an sf object
#'
#' Stops with an informative error when `spatial` does not inherit from `"sf"`.
#' sp objects are not supported.
#'
#' @param spatial Any R object.
#'
#' @return Invisibly returns `TRUE` when the check passes.
#' @keywords internal
assert_sf <- function(spatial) {
  if (!requireNamespace("sf", quietly = TRUE)) {
    stop("Package 'sf' is required. Please install it.", call. = FALSE)
  }
  if (!inherits(spatial, "sf")) {
    stop(
      "`spatial` must be an sf object. ",
      "sp objects are not supported; please convert with sf::st_as_sf().",
      call. = FALSE
    )
  }
  invisible(TRUE)
}

# ---------------------------------------------------------------------------
# extract_geometry
# ---------------------------------------------------------------------------

#' Extract the geometry column from an sf object
#'
#' Returns the `sfc` geometry column of the supplied `sf` object.
#'
#' @param spatial An `sf` object.
#'
#' @return An `sfc` geometry column.
#' @keywords internal
extract_geometry <- function(spatial) {
  assert_sf(spatial)
  sf::st_geometry(spatial)
}

# ---------------------------------------------------------------------------
# extract_coordinates
# ---------------------------------------------------------------------------

#' Extract representative XY coordinates from an sf object
#'
#' For POINT geometries the point coordinates are returned directly. For
#' all other geometry types (POLYGON, MULTIPOLYGON, etc.) `sf::st_centroid()`
#' is used to derive a representative point, with warnings suppressed (they
#' are typically non-actionable geographic-CRS notes).
#'
#' @param spatial An `sf` object.
#'
#' @return A numeric matrix with columns `X` and `Y`, one row per feature.
#' @keywords internal
extract_coordinates <- function(spatial) {
  assert_sf(spatial)

  geom_type <- unique(as.character(sf::st_geometry_type(spatial)))

  if (length(geom_type) == 1L && geom_type == "POINT") {
    coords_mat <- sf::st_coordinates(spatial)
    return(coords_mat[, c("X", "Y"), drop = FALSE])
  }

  # For POLYGON / MULTIPOLYGON / other types use centroid.
  # st_centroid warnings about geographic CRS are suppressed as they are
  # non-actionable for typical projected or planar data.
  suppressWarnings(
    centroids <- sf::st_centroid(sf::st_geometry(spatial))
  )
  coords_mat <- sf::st_coordinates(centroids)
  coords_mat[, c("X", "Y"), drop = FALSE]
}

# ---------------------------------------------------------------------------
# align_spatial_to_panel
# ---------------------------------------------------------------------------

#' Reorder spatial rows to match panel ID order
#'
#' Given a named integer vector `id_map` (names = unit IDs, values = row
#' indices in `spatial`) produced by `build_id_map()`, this function reorders
#' the rows of `spatial` so that they align with the panel data ordering.
#'
#' @param spatial An `sf` object containing at minimum the rows referenced in
#'   `id_map`.
#' @param id_map  A named integer vector as produced by `build_id_map()`.
#'   Names are unit IDs (character); values are 1-based row indices into
#'   `spatial`.
#'
#' @return An `sf` object with rows reordered to match `id_map`.
#' @keywords internal
align_spatial_to_panel <- function(spatial, id_map) {
  assert_sf(spatial)

  if (!is.integer(id_map) && !is.numeric(id_map)) {
    stop("`id_map` must be a named numeric (integer) vector.", call. = FALSE)
  }
  if (is.null(names(id_map))) {
    stop("`id_map` must be named.", call. = FALSE)
  }

  row_indices <- as.integer(id_map)

  # Guard against out-of-range indices
  n_rows <- nrow(spatial)
  if (any(row_indices < 1L | row_indices > n_rows)) {
    stop(
      "`id_map` contains row indices outside the range [1, ", n_rows, "].",
      call. = FALSE
    )
  }

  aligned <- spatial[row_indices, , drop = FALSE]
  rownames(aligned) <- NULL
  aligned
}

# ---------------------------------------------------------------------------
# build_neighbor_structure
# ---------------------------------------------------------------------------

#' Build a neighbour structure from an sf object
#'
#' Returns different structures depending on `type`:
#'
#' * `"distance"` — a numeric coordinate matrix (columns `X` and `Y`) suitable
#'   for pairwise distance computation.
#' * `"contiguity"` — a named list where each element is the integer vector of
#'   neighbour row indices (1-based, Queen contiguity via
#'   `sf::st_relate()`). Used for spatial diagnostics such as Moran's I.
#'
#' @param spatial An `sf` object.
#' @param type    Character scalar: `"distance"` (default) or `"contiguity"`.
#'
#' @return
#' * For `"distance"`: a numeric matrix with columns `X` and `Y`.
#' * For `"contiguity"`: a named list of integer vectors.
#' @keywords internal
build_neighbor_structure <- function(spatial,
                                     type = c("distance", "contiguity")) {
  type <- match.arg(type)
  assert_sf(spatial)

  if (type == "distance") {
    return(extract_coordinates(spatial))
  }

  # Contiguity: Queen criterion (shared boundary or vertex)
  # st_relate pattern "F***T****" (touches) or broader "****T****"
  # We use the standard spdep-compatible approach via sf::st_relate
  n <- nrow(spatial)
  geom <- sf::st_geometry(spatial)

  # Use Queen contiguity: polygons that share at least a point.
  # DE-9IM pattern for "touches or overlaps but not disjoint at boundary"
  # We detect shared boundaries/vertices with the "T" intersect matrix.
  # A robust approach: use st_relate with pattern "F***T****" for touches,
  # combined with polygon intersection detection.
  #
  # Simplest correct approach: use sf::st_relate() with Queen pattern.
  # Queen pattern (from spdep conventions): "F***T****" OR "****T****"
  # We use "****T****" which covers all non-disjoint boundary cases.

  # st_relate returns an sgbp sparse list: element i is an integer vector of
  # row indices (1-based) of features satisfying the pattern with feature i.
  nb_sparse <- tryCatch(
    sf::st_relate(geom, geom, pattern = "F***T****"),
    error = function(e) NULL
  )

  row_names <- if (!is.null(rownames(spatial))) rownames(spatial) else as.character(seq_len(n))

  if (is.null(nb_sparse)) {
    # Fallback: no contiguity structure (e.g., POINT geometries)
    nb <- vector("list", n)
    names(nb) <- row_names
    nb <- lapply(nb, function(x) integer(0L))
    return(nb)
  }

  # nb_sparse is a list of length n; each element is an integer vector of
  # neighbour indices (already excludes self when pattern does not match i->i).
  nb <- vector("list", n)
  names(nb) <- row_names

  for (i in seq_len(n)) {
    # Remove self-references to be safe
    neighbours <- nb_sparse[[i]]
    nb[[i]]    <- as.integer(neighbours[neighbours != i])
  }

  nb
}

library(testthat)

# ---------------------------------------------------------------------------
# Helpers: small simulated sf objects
# ---------------------------------------------------------------------------

# Four POINT features in a 2x2 grid
make_point_sf <- function() {
  if (!requireNamespace("sf", quietly = TRUE)) skip("sf not available")
  coords <- matrix(
    c(0, 0,
      1, 0,
      0, 1,
      1, 1),
    ncol = 2, byrow = TRUE,
    dimnames = list(NULL, c("x", "y"))
  )
  pts <- sf::st_as_sf(
    data.frame(id = c("A", "B", "C", "D"), x = coords[, 1], y = coords[, 2]),
    coords = c("x", "y"),
    crs    = NA
  )
  pts
}

# Four POLYGON features (unit squares tiling a 2x2 area)
make_polygon_sf <- function() {
  if (!requireNamespace("sf", quietly = TRUE)) skip("sf not available")

  make_square <- function(xmin, ymin) {
    sf::st_polygon(list(matrix(
      c(xmin,     ymin,
        xmin + 1, ymin,
        xmin + 1, ymin + 1,
        xmin,     ymin + 1,
        xmin,     ymin),
      ncol = 2, byrow = TRUE
    )))
  }

  polys <- sf::st_sfc(
    make_square(0, 0),
    make_square(1, 0),
    make_square(0, 1),
    make_square(1, 1)
  )

  sf::st_sf(id = c("A", "B", "C", "D"), geometry = polys)
}

# A non-sf object for negative tests
make_non_sf <- function() {
  data.frame(id = 1:3, x = 1:3, y = 1:3)
}

# ---------------------------------------------------------------------------
# assert_sf
# ---------------------------------------------------------------------------

test_that("assert_sf: returns TRUE invisibly for valid sf", {
  pts <- make_point_sf()
  expect_true(assert_sf(pts))
})

test_that("assert_sf: errors for data.frame (non-sf)", {
  expect_error(assert_sf(make_non_sf()), "sf object")
})

test_that("assert_sf: errors for plain list", {
  expect_error(assert_sf(list(x = 1:3)), "sf object")
})

test_that("assert_sf: errors for NULL", {
  expect_error(assert_sf(NULL), "sf object")
})

test_that("assert_sf: error message mentions sp not supported", {
  err <- tryCatch(assert_sf(make_non_sf()), error = conditionMessage)
  expect_match(err, "sp")
})

# ---------------------------------------------------------------------------
# extract_geometry
# ---------------------------------------------------------------------------

test_that("extract_geometry: returns sfc for POINT sf", {
  pts  <- make_point_sf()
  geom <- extract_geometry(pts)
  expect_true(inherits(geom, "sfc"))
  expect_equal(length(geom), nrow(pts))
})

test_that("extract_geometry: returns sfc for POLYGON sf", {
  polys <- make_polygon_sf()
  geom  <- extract_geometry(polys)
  expect_true(inherits(geom, "sfc"))
  expect_equal(length(geom), nrow(polys))
})

test_that("extract_geometry: errors for non-sf input", {
  expect_error(extract_geometry(make_non_sf()), "sf object")
})

# ---------------------------------------------------------------------------
# extract_coordinates: POINT
# ---------------------------------------------------------------------------

test_that("extract_coordinates: POINT returns matrix with X and Y columns", {
  pts    <- make_point_sf()
  coords <- extract_coordinates(pts)
  expect_true(is.matrix(coords))
  expect_true(all(c("X", "Y") %in% colnames(coords)))
})

test_that("extract_coordinates: POINT row count matches feature count", {
  pts    <- make_point_sf()
  coords <- extract_coordinates(pts)
  expect_equal(nrow(coords), nrow(pts))
})

test_that("extract_coordinates: POINT coordinates match input", {
  pts    <- make_point_sf()
  coords <- extract_coordinates(pts)
  # First point is (0, 0)
  expect_equal(as.numeric(coords[1, "X"]), 0)
  expect_equal(as.numeric(coords[1, "Y"]), 0)
  # Last point is (1, 1)
  expect_equal(as.numeric(coords[4, "X"]), 1)
  expect_equal(as.numeric(coords[4, "Y"]), 1)
})

# ---------------------------------------------------------------------------
# extract_coordinates: POLYGON
# ---------------------------------------------------------------------------

test_that("extract_coordinates: POLYGON returns matrix with X and Y columns", {
  polys  <- make_polygon_sf()
  coords <- extract_coordinates(polys)
  expect_true(is.matrix(coords))
  expect_true(all(c("X", "Y") %in% colnames(coords)))
})

test_that("extract_coordinates: POLYGON row count matches feature count", {
  polys  <- make_polygon_sf()
  coords <- extract_coordinates(polys)
  expect_equal(nrow(coords), nrow(polys))
})

test_that("extract_coordinates: POLYGON representative points lie inside units", {
  polys  <- make_polygon_sf()
  coords <- extract_coordinates(polys)
  # Each centroid should be within the bounding box [0, 2] x [0, 2]
  expect_true(all(coords[, "X"] >= 0 & coords[, "X"] <= 2))
  expect_true(all(coords[, "Y"] >= 0 & coords[, "Y"] <= 2))
})

test_that("extract_coordinates: POLYGON first polygon centroid is approx (0.5, 0.5)", {
  polys  <- make_polygon_sf()
  coords <- extract_coordinates(polys)
  expect_equal(as.numeric(coords[1, "X"]), 0.5, tolerance = 1e-6)
  expect_equal(as.numeric(coords[1, "Y"]), 0.5, tolerance = 1e-6)
})

test_that("extract_coordinates: errors for non-sf input", {
  expect_error(extract_coordinates(make_non_sf()), "sf object")
})

# ---------------------------------------------------------------------------
# align_spatial_to_panel
# ---------------------------------------------------------------------------

test_that("align_spatial_to_panel: reorders rows according to id_map", {
  pts <- make_point_sf()
  # Reverse the order: D, C, B, A
  id_map <- c(D = 4L, C = 3L, B = 2L, A = 1L)
  aligned <- align_spatial_to_panel(pts, id_map)
  expect_equal(aligned$id, c("D", "C", "B", "A"))
})

test_that("align_spatial_to_panel: subset of rows via id_map", {
  pts <- make_point_sf()
  # Select only A and C (rows 1 and 3)
  id_map <- c(A = 1L, C = 3L)
  aligned <- align_spatial_to_panel(pts, id_map)
  expect_equal(nrow(aligned), 2L)
  expect_equal(aligned$id, c("A", "C"))
})

test_that("align_spatial_to_panel: preserves sf class", {
  pts    <- make_point_sf()
  id_map <- c(A = 1L, B = 2L, C = 3L, D = 4L)
  aligned <- align_spatial_to_panel(pts, id_map)
  expect_true(inherits(aligned, "sf"))
})

test_that("align_spatial_to_panel: errors for non-sf input", {
  id_map <- c(A = 1L)
  expect_error(align_spatial_to_panel(make_non_sf(), id_map), "sf object")
})

test_that("align_spatial_to_panel: errors when id_map is not named", {
  pts <- make_point_sf()
  expect_error(align_spatial_to_panel(pts, c(1L, 2L)), "named")
})

test_that("align_spatial_to_panel: errors for out-of-range row index", {
  pts    <- make_point_sf()
  id_map <- c(X = 99L)
  expect_error(align_spatial_to_panel(pts, id_map), "range")
})

test_that("align_spatial_to_panel: ID ordering matches panel order (integration)", {
  pts <- make_point_sf()
  # Simulate panel ordering: B, D, A, C
  panel_ids    <- c("B", "D", "A", "C")
  spatial_ids  <- pts$id  # A, B, C, D

  id_map <- match(panel_ids, spatial_ids)
  names(id_map) <- panel_ids
  storage.mode(id_map) <- "integer"

  aligned <- align_spatial_to_panel(pts, id_map)
  expect_equal(aligned$id, panel_ids)
})

# ---------------------------------------------------------------------------
# build_neighbor_structure: distance type
# ---------------------------------------------------------------------------

test_that("build_neighbor_structure distance: returns coordinate matrix", {
  pts    <- make_point_sf()
  result <- build_neighbor_structure(pts, type = "distance")
  expect_true(is.matrix(result))
  expect_true(all(c("X", "Y") %in% colnames(result)))
  expect_equal(nrow(result), nrow(pts))
})

test_that("build_neighbor_structure distance: POLYGON returns coordinate matrix", {
  polys  <- make_polygon_sf()
  result <- build_neighbor_structure(polys, type = "distance")
  expect_true(is.matrix(result))
  expect_equal(nrow(result), nrow(polys))
})

test_that("build_neighbor_structure distance: default type is distance", {
  pts    <- make_point_sf()
  result <- build_neighbor_structure(pts)
  expect_true(is.matrix(result))
})

test_that("build_neighbor_structure distance: errors for non-sf input", {
  expect_error(
    build_neighbor_structure(make_non_sf(), type = "distance"),
    "sf object"
  )
})

# ---------------------------------------------------------------------------
# build_neighbor_structure: contiguity type
# ---------------------------------------------------------------------------

test_that("build_neighbor_structure contiguity: returns a list", {
  polys  <- make_polygon_sf()
  result <- build_neighbor_structure(polys, type = "contiguity")
  expect_true(is.list(result))
})

test_that("build_neighbor_structure contiguity: list length equals n features", {
  polys  <- make_polygon_sf()
  result <- build_neighbor_structure(polys, type = "contiguity")
  expect_equal(length(result), nrow(polys))
})

test_that("build_neighbor_structure contiguity: each element is integer vector", {
  polys  <- make_polygon_sf()
  result <- build_neighbor_structure(polys, type = "contiguity")
  for (nb in result) {
    expect_true(is.integer(nb))
  }
})

test_that("build_neighbor_structure contiguity: adjacent polygons are neighbours", {
  polys  <- make_polygon_sf()
  # 2x2 grid: polygon 1 (SW) shares a boundary with polygon 2 (SE) and 3 (NW)
  result <- build_neighbor_structure(polys, type = "contiguity")
  # Polygon 1 should have at least polygons 2 and 3 as neighbours
  expect_true(2L %in% result[[1]] || length(result[[1]]) >= 2L)
})

test_that("build_neighbor_structure contiguity: no self-loops", {
  polys  <- make_polygon_sf()
  result <- build_neighbor_structure(polys, type = "contiguity")
  for (i in seq_along(result)) {
    expect_false(i %in% result[[i]])
  }
})

test_that("build_neighbor_structure contiguity: errors for non-sf input", {
  expect_error(
    build_neighbor_structure(make_non_sf(), type = "contiguity"),
    "sf object"
  )
})

test_that("build_neighbor_structure: invalid type raises error", {
  pts <- make_point_sf()
  expect_error(build_neighbor_structure(pts, type = "invalid"), "arg")
})

# ---------------------------------------------------------------------------
# Non-sf object: comprehensive check across all functions
# ---------------------------------------------------------------------------

test_that("all functions reject sp-like or plain list objects", {
  bad <- list(id = 1:3, coords = matrix(1:6, ncol = 2))
  expect_error(assert_sf(bad),                              "sf object")
  expect_error(extract_geometry(bad),                       "sf object")
  expect_error(extract_coordinates(bad),                    "sf object")
  expect_error(align_spatial_to_panel(bad, c(a = 1L)),     "sf object")
  expect_error(build_neighbor_structure(bad, "distance"),   "sf object")
})

#' @title Parallel Execution Module
#' @description Unified parallel execution interface for bandwidth search,
#'   local model fitting, and diagnostics. Shields backend differences and
#'   ensures CRAN-friendly behaviour.
#' @name parallel_module
NULL

#' Apply a function over a list with optional parallel execution
#'
#' A thin wrapper that uses plain \code{lapply} when \code{workers = 1} and
#' switches to \code{future.apply::future_lapply} with a \code{multisession}
#' plan for \code{workers > 1}. The global \code{future} plan is always
#' restored to \code{sequential} after the call, preventing side-effects.
#'
#' Worker-level errors are caught and returned as character strings (prefixed
#' with \code{"ERROR: "}) rather than aborting the entire call.
#'
#' @param x A list (or vector) of inputs to iterate over.
#' @param fn A function to apply to each element of \code{x}. The function
#'   receives the element as its first argument; additional arguments are
#'   passed via \code{...}.
#' @param workers Integer scalar. Number of parallel workers. \code{1}
#'   (default) uses serial \code{lapply} and never touches \code{future}.
#' @param seed Integer scalar or \code{NULL}. Random seed for reproducibility.
#'   In serial mode the R RNG is seeded with \code{set.seed(seed)}. In
#'   parallel mode the seed is forwarded to \code{future_lapply} via the
#'   \code{future.seed} argument using L'Ecuyer-CMRG streams.
#' @param ... Additional arguments forwarded to \code{fn}.
#' @param packages Character vector of package names that workers need to load,
#'   or \code{NULL} (default). Ignored in serial mode.
#'
#' @return A list of the same length as \code{x}. Elements where \code{fn}
#'   threw an error are replaced with a character string \code{"ERROR: <msg>"}.
#'
#' @examples
#' result <- parallel_map(1:3, function(x) x^2, workers = 1)
#' stopifnot(identical(result, list(1, 4, 9)))
#'
#' @export
parallel_map <- function(x, fn, workers = 1, seed = NULL, ..., packages = NULL) {
  stopifnot(is.numeric(workers), length(workers) == 1L, workers >= 1L)
  workers <- as.integer(workers)

  safe_fn <- function(elem) {
    tryCatch(fn(elem, ...), error = function(e) paste0("ERROR: ", conditionMessage(e)))
  }

  if (workers == 1L) {
    # Serial path — never touch future plan
    if (!is.null(seed)) set.seed(seed)
    return(lapply(x, safe_fn))
  }

  # Parallel path — requires future and future.apply
  if (!requireNamespace("future", quietly = TRUE)) {
    stop("Package 'future' is required for parallel_map() with workers > 1.")
  }
  if (!requireNamespace("future.apply", quietly = TRUE)) {
    stop("Package 'future.apply' is required for parallel_map() with workers > 1.")
  }

  old_plan <- future::plan()
  on.exit(future::plan(old_plan), add = TRUE)

  future::plan(future::multisession, workers = workers)

  future_seed <- if (is.null(seed)) FALSE else seed

  result <- future.apply::future_lapply(
    X            = x,
    FUN          = safe_fn,
    future.seed  = future_seed,
    future.packages = packages
  )

  result
}

#' Execute an expression with a reproducible seed
#'
#' Sets \code{set.seed(seed)} before evaluating \code{expr}, then restores the
#' prior RNG state so the caller's random stream is unaffected.
#'
#' @param seed Integer scalar. Seed value passed to \code{set.seed}.
#' @param expr An R expression to evaluate.
#'
#' @return The value of \code{expr}.
#'
#' @examples
#' r1 <- with_reproducible_seed(42, runif(3))
#' r2 <- with_reproducible_seed(42, runif(3))
#' stopifnot(identical(r1, r2))
#'
#' @export
with_reproducible_seed <- function(seed, expr) {
  old_seed <- get0(".Random.seed", envir = globalenv(), inherits = FALSE)
  on.exit({
    if (is.null(old_seed)) {
      suppressWarnings(rm(".Random.seed", envir = globalenv(), inherits = FALSE))
    } else {
      assign(".Random.seed", old_seed, envir = globalenv(), inherits = FALSE)
    }
  }, add = TRUE)
  set.seed(seed)
  expr
}

#' Validate that parallel and serial results have the same structure
#'
#' Checks that \code{parallel_result} and \code{serial_result} have the same
#' length and the same element types. This function is intended for use in
#' tests to verify that switching to a parallel backend does not alter the
#' output structure.
#'
#' @param serial_result A list produced by \code{parallel_map} with
#'   \code{workers = 1}.
#' @param parallel_result A list produced by \code{parallel_map} with
#'   \code{workers > 1}.
#'
#' @return \code{TRUE} (invisibly) if the structures match; otherwise an
#'   informative error is thrown.
#'
#' @examples
#' sr <- parallel_map(1:3, function(x) x * 2, workers = 1)
#' # In real usage parallel_result would use workers > 1; here we reuse sr.
#' validate_parallel_result(sr, sr)
#'
#' @noRd
validate_parallel_result <- function(serial_result, parallel_result) {
  if (!is.list(serial_result))   stop("`serial_result` must be a list.")
  if (!is.list(parallel_result)) stop("`parallel_result` must be a list.")

  if (length(serial_result) != length(parallel_result)) {
    stop(sprintf(
      "Length mismatch: serial has %d elements, parallel has %d.",
      length(serial_result), length(parallel_result)
    ))
  }

  for (i in seq_along(serial_result)) {
    s_class <- class(serial_result[[i]])
    p_class <- class(parallel_result[[i]])
    if (!identical(s_class, p_class)) {
      stop(sprintf(
        "Element %d class mismatch: serial = '%s', parallel = '%s'.",
        i, paste(s_class, collapse = "/"), paste(p_class, collapse = "/")
      ))
    }
  }

  invisible(TRUE)
}

# Tests for parallel.R
# All tests use workers = 1 (serial) to stay CRAN-friendly.

# ---------------------------------------------------------------------------
# parallel_map — basic serial behaviour
# ---------------------------------------------------------------------------

test_that("parallel_map with workers=1 returns correct results", {
  result <- parallel_map(1:5, function(x) x^2, workers = 1)
  expect_type(result, "list")
  expect_length(result, 5)
  expect_equal(result[[1]], 1)
  expect_equal(result[[3]], 9)
  expect_equal(result[[5]], 25)
})

test_that("parallel_map with workers=1 passes extra arguments via ...", {
  result <- parallel_map(1:3, function(x, add) x + add, workers = 1, add = 10)
  expect_equal(result[[1]], 11)
  expect_equal(result[[2]], 12)
  expect_equal(result[[3]], 13)
})

test_that("parallel_map with workers=1 returns a list of same length as input", {
  input  <- list(a = 1, b = 2, c = 3)
  result <- parallel_map(input, function(x) x * 2, workers = 1)
  expect_length(result, 3)
})

# ---------------------------------------------------------------------------
# parallel_map — worker error capture
# ---------------------------------------------------------------------------

test_that("parallel_map captures worker errors without crashing (workers=1)", {
  fn <- function(x) {
    if (x == 2) stop("deliberate error")
    x * 10
  }
  result <- parallel_map(1:3, fn, workers = 1)
  expect_equal(result[[1]], 10)
  expect_match(result[[2]], "^ERROR: deliberate error")
  expect_equal(result[[3]], 30)
})

test_that("parallel_map worker error returns a character string", {
  fn <- function(x) stop("boom")
  result <- parallel_map(list(1), fn, workers = 1)
  expect_type(result[[1]], "character")
  expect_match(result[[1]], "^ERROR: ")
})

# ---------------------------------------------------------------------------
# parallel_map — workers=1 does not touch future plan
# ---------------------------------------------------------------------------

test_that("parallel_map with workers=1 does not load or change future plan", {
  skip_if_not_installed("future")
  plan_before <- class(future::plan())
  parallel_map(1:3, function(x) x, workers = 1)
  plan_after  <- class(future::plan())
  expect_identical(plan_before, plan_after)
})

# ---------------------------------------------------------------------------
# parallel_map — reproducibility with fixed seed (serial)
# ---------------------------------------------------------------------------

test_that("parallel_map with fixed seed gives reproducible results (workers=1)", {
  fn <- function(x) runif(1)
  r1 <- parallel_map(1:5, fn, workers = 1, seed = 123)
  r2 <- parallel_map(1:5, fn, workers = 1, seed = 123)
  expect_identical(r1, r2)
})

test_that("parallel_map with different seeds gives different results (workers=1)", {
  fn <- function(x) runif(1)
  r1 <- parallel_map(1:5, fn, workers = 1, seed = 1)
  r2 <- parallel_map(1:5, fn, workers = 1, seed = 2)
  # It is extremely unlikely all 5 values coincide for different seeds
  expect_false(identical(r1, r2))
})

# ---------------------------------------------------------------------------
# with_reproducible_seed
# ---------------------------------------------------------------------------

test_that("with_reproducible_seed gives reproducible output for same seed", {
  r1 <- with_reproducible_seed(42, runif(5))
  r2 <- with_reproducible_seed(42, runif(5))
  expect_equal(r1, r2)
})

test_that("with_reproducible_seed gives different output for different seeds", {
  r1 <- with_reproducible_seed(1, runif(5))
  r2 <- with_reproducible_seed(2, runif(5))
  expect_false(identical(r1, r2))
})

test_that("with_reproducible_seed restores RNG state afterwards", {
  set.seed(99)
  before <- runif(3)           # consume 3 draws from seed 99

  set.seed(99)
  with_reproducible_seed(42, runif(3))   # should not affect subsequent state
  after <- runif(3)            # next 3 draws from seed 99

  expect_equal(before, after)
})

test_that("with_reproducible_seed returns the value of expr", {
  val <- with_reproducible_seed(7, 1 + 1)
  expect_equal(val, 2)
})

# ---------------------------------------------------------------------------
# validate_parallel_result
# ---------------------------------------------------------------------------

test_that("validate_parallel_result returns TRUE invisibly for matching lists", {
  sr <- parallel_map(1:4, function(x) x^2, workers = 1)
  result <- validate_parallel_result(sr, sr)
  expect_true(result)
})

test_that("validate_parallel_result errors on length mismatch", {
  sr <- list(1, 2, 3)
  pr <- list(1, 2)
  expect_error(validate_parallel_result(sr, pr), "Length mismatch")
})

test_that("validate_parallel_result errors on class mismatch", {
  sr <- list(1L, 2L)
  pr <- list("a", "b")
  expect_error(validate_parallel_result(sr, pr), "class mismatch")
})

test_that("validate_parallel_result errors when inputs are not lists", {
  expect_error(validate_parallel_result(1:3, list(1, 2, 3)), "must be a list")
  expect_error(validate_parallel_result(list(1, 2, 3), 1:3), "must be a list")
})

# ---------------------------------------------------------------------------
# Structure consistency: serial result matches expected output shape
# ---------------------------------------------------------------------------

test_that("parallel_map serial result structure is consistent across calls", {
  fn <- function(x) list(value = x, squared = x^2)
  r1 <- parallel_map(1:3, fn, workers = 1)
  r2 <- parallel_map(1:3, fn, workers = 1)
  # Validate structure
  expect_true(validate_parallel_result(r1, r2))
  # Check content
  for (i in 1:3) {
    expect_named(r1[[i]], c("value", "squared"))
    expect_equal(r1[[i]]$squared, i^2)
  }
})

#' @title Internal Context Object for GWPR.light 1.0.0
#'
#' @description
#' These internal functions construct and validate the standardised
#' `gwpr_context` list that is passed between modules, eliminating
#' repetitive argument passing.
#'
#' @name context
#' @keywords internal
NULL

# ---------------------------------------------------------------------------
# new_gwpr_context
# ---------------------------------------------------------------------------

#' Construct a new gwpr_context object
#'
#' Creates the standardised internal context list used to pass state between
#' GWPR modules.  Any field not supplied defaults to `NULL` (or, for
#' `metadata` and `warnings`, to their appropriate empty types).
#'
#' @param call       The matched call from the top-level API function.
#' @param formula    A formula object.
#' @param family     Character: `"gaussian"` or `"binomial"`.
#' @param model      Character: `"pooling"`, `"within"`, or `"random"`.
#' @param effect     Character: `"individual"`, `"time"`, `"two-way"`, or
#'   `"nested"`.
#' @param id         Name of the unit ID column.
#' @param time       Name of the time column.
#' @param kernel     Kernel name.
#' @param adaptive   Logical; `TRUE` for adaptive bandwidth.
#' @param threshold  Numeric classification threshold (Logistic).
#' @param workers    Number of parallel workers.
#' @param seed       Integer random seed or `NULL`.
#' @param raw_data   The original user-supplied data frame.
#' @param raw_spatial The original user-supplied sf object.
#' @param panel_data Processed panel data frame.
#' @param spatial_data Processed sf object.
#' @param id_map     Named integer vector mapping unit IDs to row indices.
#' @param coords     Matrix of spatial coordinates.
#' @param model_frame Model frame derived from formula and panel_data.
#' @param model_matrix Design matrix.
#' @param response   Numeric response vector.
#' @param metadata   Named list of supplementary information.
#' @param warnings   Character vector of accumulated warnings.
#' @param ...        Additional named fields stored in the context list.
#'
#' @return A named list with class `"gwpr_context"`.
#' @keywords internal
new_gwpr_context <- function(
    call          = NULL,
    formula       = NULL,
    family        = NULL,
    model         = NULL,
    effect        = NULL,
    id            = NULL,
    time          = NULL,
    kernel        = NULL,
    adaptive      = NULL,
    threshold     = NULL,
    workers       = NULL,
    seed          = NULL,
    raw_data      = NULL,
    raw_spatial   = NULL,
    panel_data    = NULL,
    spatial_data  = NULL,
    id_map        = NULL,
    coords        = NULL,
    model_frame   = NULL,
    model_matrix  = NULL,
    response      = NULL,
    metadata      = list(),
    warnings      = character(),
    ...
) {
  extra <- list(...)

  ctx <- c(
    list(
      call         = call,
      formula      = formula,
      family       = family,
      model        = model,
      effect       = effect,
      id           = id,
      time         = time,
      kernel       = kernel,
      adaptive     = adaptive,
      threshold    = threshold,
      workers      = workers,
      seed         = seed,
      raw_data     = raw_data,
      raw_spatial  = raw_spatial,
      panel_data   = panel_data,
      spatial_data = spatial_data,
      id_map       = id_map,
      coords       = coords,
      model_frame  = model_frame,
      model_matrix = model_matrix,
      response     = response,
      metadata     = metadata,
      warnings     = warnings
    ),
    extra
  )

  structure(ctx, class = "gwpr_context")
}

# ---------------------------------------------------------------------------
# validate_gwpr_context
# ---------------------------------------------------------------------------

#' Validate a gwpr_context object
#'
#' Checks that all core fields required for model fitting are non-`NULL`.
#' Stops with an informative message listing every missing field.
#'
#' Core fields: `formula`, `family`, `id`, `time`, `model`, `effect`,
#' `kernel`, `adaptive`, `threshold`, `workers`.
#'
#' @param context A list (typically of class `"gwpr_context"`) to validate.
#'
#' @return Invisibly returns `TRUE` when all core fields are present.
#' @keywords internal
validate_gwpr_context <- function(context) {
  if (!is.list(context)) {
    stop("`context` must be a list.", call. = FALSE)
  }

  core_fields <- c(
    "formula", "family", "id", "time",
    "model", "effect", "kernel", "adaptive",
    "threshold", "workers"
  )

  missing_fields <- core_fields[vapply(
    core_fields,
    function(f) is.null(context[[f]]),
    logical(1L)
  )]

  if (length(missing_fields) > 0L) {
    stop(
      "gwpr_context is missing required core field(s): ",
      paste(missing_fields, collapse = ", "),
      call. = FALSE
    )
  }

  invisible(TRUE)
}

library(testthat)

# ---------------------------------------------------------------------------
# Helper: build a minimal valid context
# ---------------------------------------------------------------------------

# Defaults for all core fields; callers may override any of them.
make_minimal_context <- function(
    formula   = y ~ x1,
    family    = "gaussian",
    model     = "within",
    effect    = "individual",
    id        = "id",
    time      = "time",
    kernel    = "bisquare",
    adaptive  = FALSE,
    threshold = 0.5,
    workers   = 1L,
    ...
) {
  new_gwpr_context(
    formula   = formula,
    family    = family,
    model     = model,
    effect    = effect,
    id        = id,
    time      = time,
    kernel    = kernel,
    adaptive  = adaptive,
    threshold = threshold,
    workers   = workers,
    ...
  )
}

# ---------------------------------------------------------------------------
# new_gwpr_context: field completeness
# ---------------------------------------------------------------------------

test_that("new_gwpr_context: returns a list with class gwpr_context", {
  ctx <- new_gwpr_context()
  expect_true(is.list(ctx))
  expect_s3_class(ctx, "gwpr_context")
})

test_that("new_gwpr_context: all required fields are present and default to NULL", {
  ctx <- new_gwpr_context()
  required_null_fields <- c(
    "call", "formula", "family", "model", "effect",
    "id", "time", "kernel", "adaptive", "threshold",
    "workers", "seed",
    "raw_data", "raw_spatial",
    "panel_data", "spatial_data", "id_map",
    "coords", "model_frame", "model_matrix", "response"
  )
  for (fld in required_null_fields) {
    expect_null(ctx[[fld]], label = paste("ctx$", fld, "is NULL by default"))
  }
})

test_that("new_gwpr_context: metadata defaults to empty list", {
  ctx <- new_gwpr_context()
  expect_identical(ctx$metadata, list())
})

test_that("new_gwpr_context: warnings defaults to empty character vector", {
  ctx <- new_gwpr_context()
  expect_identical(ctx$warnings, character())
})

test_that("new_gwpr_context: supplied values are stored correctly", {
  ctx <- make_minimal_context(
    seed     = 42L,
    workers  = 1L,
    metadata = list(n_units = 3L),
    warnings = c("test warning")
  )
  expect_equal(deparse(ctx$formula), deparse(y ~ x1))
  expect_identical(ctx$family,  "gaussian")
  expect_identical(ctx$model,   "within")
  expect_identical(ctx$effect,  "individual")
  expect_identical(ctx$id,      "id")
  expect_identical(ctx$time,    "time")
  expect_identical(ctx$kernel,  "bisquare")
  expect_false(ctx$adaptive)
  expect_equal(ctx$threshold, 0.5)
  expect_identical(ctx$workers,  1L)
  expect_identical(ctx$seed,    42L)
  expect_identical(ctx$metadata, list(n_units = 3L))
  expect_identical(ctx$warnings, "test warning")
})

test_that("new_gwpr_context: extra arguments are stored in context", {
  ctx <- new_gwpr_context(custom_field = "hello")
  expect_identical(ctx$custom_field, "hello")
})

test_that("new_gwpr_context: data fields can be set and retrieved", {
  df <- data.frame(id = 1:3, time = 1:3, y = rnorm(3), x1 = rnorm(3))
  ctx <- new_gwpr_context(raw_data = df)
  expect_identical(ctx$raw_data, df)
})

# ---------------------------------------------------------------------------
# validate_gwpr_context: valid context passes
# ---------------------------------------------------------------------------

test_that("validate_gwpr_context: fully populated core fields pass", {
  ctx <- make_minimal_context()
  expect_true(validate_gwpr_context(ctx))
})

test_that("validate_gwpr_context: returns invisibly TRUE on success", {
  ctx <- make_minimal_context()
  result <- validate_gwpr_context(ctx)
  expect_true(result)
})

# ---------------------------------------------------------------------------
# validate_gwpr_context: missing core fields raise errors
# ---------------------------------------------------------------------------

test_that("validate_gwpr_context: missing formula raises error", {
  ctx <- make_minimal_context()
  ctx$formula <- NULL
  expect_error(validate_gwpr_context(ctx), "formula")
})

test_that("validate_gwpr_context: missing family raises error", {
  ctx <- make_minimal_context()
  ctx$family <- NULL
  expect_error(validate_gwpr_context(ctx), "family")
})

test_that("validate_gwpr_context: missing id raises error", {
  ctx <- make_minimal_context()
  ctx$id <- NULL
  expect_error(validate_gwpr_context(ctx), "id")
})

test_that("validate_gwpr_context: missing time raises error", {
  ctx <- make_minimal_context()
  ctx$time <- NULL
  expect_error(validate_gwpr_context(ctx), "time")
})

test_that("validate_gwpr_context: missing model raises error", {
  ctx <- make_minimal_context()
  ctx$model <- NULL
  expect_error(validate_gwpr_context(ctx), "model")
})

test_that("validate_gwpr_context: missing effect raises error", {
  ctx <- make_minimal_context()
  ctx$effect <- NULL
  expect_error(validate_gwpr_context(ctx), "effect")
})

test_that("validate_gwpr_context: missing kernel raises error", {
  ctx <- make_minimal_context()
  ctx$kernel <- NULL
  expect_error(validate_gwpr_context(ctx), "kernel")
})

test_that("validate_gwpr_context: missing adaptive raises error", {
  ctx <- make_minimal_context()
  ctx$adaptive <- NULL
  expect_error(validate_gwpr_context(ctx), "adaptive")
})

test_that("validate_gwpr_context: missing threshold raises error", {
  ctx <- make_minimal_context()
  ctx$threshold <- NULL
  expect_error(validate_gwpr_context(ctx), "threshold")
})

test_that("validate_gwpr_context: missing workers raises error", {
  ctx <- make_minimal_context()
  ctx$workers <- NULL
  expect_error(validate_gwpr_context(ctx), "workers")
})

test_that("validate_gwpr_context: multiple missing fields listed in error message", {
  ctx <- make_minimal_context()
  ctx$formula <- NULL
  ctx$family  <- NULL
  ctx$id      <- NULL
  err <- tryCatch(validate_gwpr_context(ctx), error = function(e) conditionMessage(e))
  expect_match(err, "formula")
  expect_match(err, "family")
  expect_match(err, "id")
})

test_that("validate_gwpr_context: non-list input raises error", {
  expect_error(validate_gwpr_context("not a list"), "`context` must be a list")
})

# ---------------------------------------------------------------------------
# Interface consistency: context can be reused across module boundaries
# ---------------------------------------------------------------------------

test_that("context fields cover search module requirements (bandwidth, seed, workers)", {
  ctx <- make_minimal_context(seed = 99L, workers = 1L)
  # Fields required by bandwidth search modules
  expect_false(is.null(ctx$kernel))
  expect_false(is.null(ctx$adaptive))
  expect_false(is.null(ctx$workers))
  expect_identical(ctx$seed, 99L)
})

test_that("context fields cover fit module requirements (formula, model, effect)", {
  ctx <- make_minimal_context()
  expect_false(is.null(ctx$formula))
  expect_false(is.null(ctx$model))
  expect_false(is.null(ctx$effect))
})

test_that("context fields cover diagnostics module requirements (family, threshold)", {
  ctx <- make_minimal_context()
  expect_false(is.null(ctx$family))
  expect_false(is.null(ctx$threshold))
})

test_that("context supports binomial family for logistic GWPR", {
  ctx <- make_minimal_context(family = "binomial", threshold = 0.3)
  expect_identical(ctx$family, "binomial")
  expect_equal(ctx$threshold, 0.3)
})

test_that("context supports adaptive bandwidth mode", {
  ctx <- make_minimal_context(adaptive = TRUE)
  expect_true(ctx$adaptive)
})

test_that("context metadata and warnings can be updated after construction", {
  ctx <- make_minimal_context()
  ctx$metadata$n_units       <- 5L
  ctx$metadata$unbalanced    <- FALSE
  ctx$warnings               <- c(ctx$warnings, "added warning")

  expect_identical(ctx$metadata$n_units,   5L)
  expect_false(ctx$metadata$unbalanced)
  expect_identical(ctx$warnings, "added warning")
})

#' @title Grid Search for Bandwidth Selection
#'
#' @description
#' Functions implementing an exhaustive grid search over a user-specified
#' range of bandwidth candidates.  Each candidate is evaluated by a user-
#' supplied scorer function; the full search history and the best bandwidth
#' are returned as a `gwpr_bandwidth` object.
#'
#' @name bandwidth_grid
#' @keywords internal
NULL

# ---------------------------------------------------------------------------
# make_grid_candidates
# ---------------------------------------------------------------------------

#' Generate grid search bandwidth candidates
#'
#' Enumerates candidate bandwidths from `lower` to (but not necessarily
#' reaching) `upper` with increment `step`.  All three boundary arguments
#' are mandatory — they are never inferred automatically.
#'
#' @param lower   Numeric scalar.  Lower boundary (inclusive).
#' @param upper   Numeric scalar.  Upper boundary (inclusive if reachable by
#'   `step`).
#' @param step    Numeric scalar > 0.  Increment between successive candidates.
#' @param adaptive Logical scalar.  When `TRUE`, candidate values are rounded
#'   to the nearest positive integer and duplicates are removed.
#'
#' @return A numeric vector of candidate bandwidth values.
#'
#' @examples
#' make_grid_candidates(lower = 1, upper = 5, step = 1, adaptive = FALSE)
#' make_grid_candidates(lower = 2, upper = 10, step = 2.5, adaptive = TRUE)
#'
#' @noRd
make_grid_candidates <- function(lower, upper, step, adaptive) {
  if (missing(lower) || is.null(lower)) {
    stop(
      "Grid search requires `lower` to be set explicitly. ",
      "Automatic boundary inference is not supported.",
      call. = FALSE
    )
  }
  if (missing(upper) || is.null(upper)) {
    stop(
      "Grid search requires `upper` to be set explicitly. ",
      "Automatic boundary inference is not supported.",
      call. = FALSE
    )
  }
  if (missing(step) || is.null(step)) {
    stop(
      "Grid search requires `step` to be set explicitly. ",
      "Automatic boundary inference is not supported.",
      call. = FALSE
    )
  }

  if (!is.numeric(lower) || length(lower) != 1L || !is.finite(lower)) {
    stop("`lower` must be a single finite numeric value.", call. = FALSE)
  }
  if (!is.numeric(upper) || length(upper) != 1L || !is.finite(upper)) {
    stop("`upper` must be a single finite numeric value.", call. = FALSE)
  }
  if (!is.numeric(step) || length(step) != 1L || !is.finite(step) || step <= 0) {
    stop("`step` must be a single finite positive numeric value.", call. = FALSE)
  }
  if (lower >= upper) {
    stop("`lower` must be strictly less than `upper`.", call. = FALSE)
  }

  candidates <- seq(from = lower, to = upper, by = step)

  if (adaptive) {
    candidates <- unique(pmax(1L, round(candidates)))
    candidates <- as.integer(candidates)
  }

  candidates
}

# ---------------------------------------------------------------------------
# score_bandwidth_candidate
# ---------------------------------------------------------------------------

#' Score a single bandwidth candidate
#'
#' Calls the `scorer` function for a given bandwidth and records the result
#' together with timing information, model counts, and metric values.
#'
#' @param context   A `gwpr_context` list.
#' @param bandwidth Numeric scalar; the candidate bandwidth to evaluate.
#' @param scorer    A function with signature `scorer(context, bandwidth)`
#'   returning a named list with at minimum:
#'   \describe{
#'     \item{`score`}{Numeric scalar.  The criterion value (lower is better).}
#'     \item{`criterion`}{Character scalar.  Name of the criterion.}
#'     \item{`n_local_models`}{Integer.  Number of local models attempted.}
#'     \item{`n_failed_local_models`}{Integer.  Number of local models that
#'       failed.}
#'     \item{`metrics`}{Named list of aggregate metrics (e.g. R2, MSE, etc.).}
#'   }
#'
#' @return A named list describing the candidate result:
#' \describe{
#'   \item{`bandwidth`}{The evaluated bandwidth.}
#'   \item{`score`}{Numeric criterion score, or `NA_real_` on failure.}
#'   \item{`criterion`}{Name of the scoring criterion.}
#'   \item{`status`}{`"ok"` or `"failed"`.}
#'   \item{`error_message`}{`NA_character_` or error text.}
#'   \item{`warning_message`}{`NA_character_` or warning text.}
#'   \item{`elapsed_time`}{Elapsed wall-clock time in seconds.}
#'   \item{`n_local_models`}{Number of local models attempted.}
#'   \item{`n_failed_local_models`}{Number of local models that failed.}
#'   \item{`r2`, `mse`, `rmse`, `mae`}{Linear metrics, or `NA_real_`.}
#'   \item{`log_loss`, `accuracy`, `precision`, `recall`, `f1_score`}{Logistic
#'     metrics, or `NA_real_`.}
#' }
#' @keywords internal
score_bandwidth_candidate <- function(context, bandwidth, scorer) {
  t_start <- proc.time()[["elapsed"]]

  warn_msgs <- character()

  raw <- tryCatch(
    withCallingHandlers(
      scorer(context, bandwidth),
      warning = function(w) {
        warn_msgs <<- c(warn_msgs, conditionMessage(w))
        invokeRestart("muffleWarning")
      }
    ),
    error = function(e) {
      list(
        status          = "failed",
        error_message   = conditionMessage(e),
        score           = NA_real_,
        criterion       = NA_character_,
        n_local_models        = NA_integer_,
        n_failed_local_models = NA_integer_,
        metrics         = list()
      )
    }
  )

  elapsed <- proc.time()[["elapsed"]] - t_start

  status        <- raw$status        %||% "ok"
  error_message <- raw$error_message %||% NA_character_
  warning_message <- if (length(warn_msgs) > 0L) {
    paste(warn_msgs, collapse = "; ")
  } else {
    NA_character_
  }

  metrics <- raw$metrics %||% list()

  # Linear metrics
  r2   <- metrics$R2   %||% NA_real_
  mse  <- metrics$MSE  %||% NA_real_
  rmse <- metrics$RMSE %||% NA_real_
  mae  <- metrics$MAE  %||% NA_real_

  # Logistic metrics
  log_loss  <- metrics$log_loss  %||% NA_real_
  accuracy  <- metrics$accuracy  %||% NA_real_
  precision <- metrics$precision %||% NA_real_
  recall    <- metrics$recall    %||% NA_real_
  f1_score  <- metrics$f1_score  %||% NA_real_

  list(
    bandwidth             = bandwidth,
    score                 = raw$score             %||% NA_real_,
    criterion             = raw$criterion         %||% NA_character_,
    status                = status,
    error_message         = error_message,
    warning_message       = warning_message,
    elapsed_time          = elapsed,
    n_local_models        = raw$n_local_models        %||% NA_integer_,
    n_failed_local_models = raw$n_failed_local_models %||% NA_integer_,
    r2                    = r2,
    mse                   = mse,
    rmse                  = rmse,
    mae                   = mae,
    log_loss              = log_loss,
    accuracy              = accuracy,
    precision             = precision,
    recall                = recall,
    f1_score              = f1_score
  )
}

# ---------------------------------------------------------------------------
# rank_grid_history
# ---------------------------------------------------------------------------

#' Rank grid search history
#'
#' Assigns integer ranks to candidates: successful candidates are ranked by
#' ascending `score` (rank 1 = best / lowest score); failed candidates are
#' ranked after all successful ones.
#'
#' @param history A data.frame as produced by `search_bandwidth_grid()`.
#'
#' @return The same data.frame with a `rank` column added (or replaced).
#'
#' @examples
#' h <- data.frame(
#'   candidate_id = 1:3,
#'   score  = c(0.5, 0.3, NA),
#'   status = c("ok", "ok", "failed"),
#'   stringsAsFactors = FALSE
#' )
#' rank_grid_history(h)
#'
#' @noRd
rank_grid_history <- function(history) {
  if (!is.data.frame(history) || nrow(history) == 0L) {
    return(history)
  }

  n <- nrow(history)
  ranks <- integer(n)

  ok_idx   <- which(history$status == "ok")
  fail_idx <- which(history$status != "ok")

  if (length(ok_idx) > 0L) {
    ok_scores <- history$score[ok_idx]
    ok_ranks  <- rank(ok_scores, ties.method = "first")
    ranks[ok_idx] <- ok_ranks
  }

  if (length(fail_idx) > 0L) {
    fail_start <- length(ok_idx) + 1L
    ranks[fail_idx] <- seq(fail_start, length.out = length(fail_idx))
  }

  history$rank <- ranks
  history
}

# ---------------------------------------------------------------------------
# search_bandwidth_grid  (main entry point)
# ---------------------------------------------------------------------------

#' Bandwidth grid search
#'
#' Exhaustively evaluates every bandwidth candidate in the grid defined by
#' `control$lower`, `control$upper`, and `control$step`.  Candidates are
#' scored by the user-supplied `scorer` function.  The complete history and
#' the best bandwidth are returned as a `gwpr_bandwidth` object.
#'
#' @param context  A `gwpr_context` list.  Must contain at minimum `adaptive`.
#' @param control  A list with grid search parameters.  **Required** fields:
#'   \describe{
#'     \item{`lower`}{Numeric.  Lower bound of the search grid (inclusive).}
#'     \item{`upper`}{Numeric.  Upper bound (included when reachable by
#'       `step`).}
#'     \item{`step`}{Numeric > 0.  Increment between candidates.}
#'   }
#' @param scorer   A function with signature `scorer(context, bandwidth)`
#'   that returns a named list as described in `score_bandwidth_candidate()`.
#' @param workers  Integer.  Number of parallel workers.  `1` (default) uses
#'   serial `lapply`; `> 1` uses `parallel_map()`.
#' @param seed     Integer random seed, or `NULL`.  Applied before parallel
#'   execution.
#'
#' @return A `gwpr_bandwidth` object (see `new_gwpr_bandwidth()`).
#'
#' @examples
#' \donttest{
#' result <- search_bandwidth_grid(
#'   context = ctx,
#'   control = list(lower = 1, upper = 5, step = 1),
#'   scorer  = my_scorer_fn,
#'   workers = 1
#' )
#' result$best_bandwidth
#' }
#'
#' @noRd
search_bandwidth_grid <- function(context, control, scorer,
                                  workers = 1L, seed = NULL) {

  t_total_start <- proc.time()[["elapsed"]]

  adaptive <- context$adaptive %||% FALSE

  # Generate candidates — will stop with informative errors if bounds missing
  candidates <- make_grid_candidates(
    lower    = control$lower,
    upper    = control$upper,
    step     = control$step,
    adaptive = adaptive
  )

  n_candidates <- length(candidates)

  # Set seed before evaluation
  if (!is.null(seed)) set.seed(seed)

  # Evaluate candidates
  if (workers <= 1L) {
    results_list <- lapply(
      candidates,
      function(bw) score_bandwidth_candidate(context, bw, scorer)
    )
  } else {
    results_list <- parallel_map(
      x       = as.list(candidates),
      fn      = function(bw) score_bandwidth_candidate(context, bw, scorer),
      workers = workers,
      seed    = seed
    )
  }

  # Determine bandwidth_type and method/criterion from first successful result
  bandwidth_type <- if (adaptive) "adaptive" else "fixed"
  first_ok <- Find(function(r) r$status == "ok", results_list)
  criterion <- if (!is.null(first_ok)) first_ok$criterion %||% NA_character_ else NA_character_

  # Assemble history data.frame
  history <- data.frame(
    candidate_id          = seq_len(n_candidates),
    bandwidth             = vapply(results_list, `[[`, numeric(1L), "bandwidth"),
    bandwidth_type        = bandwidth_type,
    method                = "grid",
    criterion             = vapply(results_list, function(r) r$criterion %||% NA_character_, character(1L)),
    score                 = vapply(results_list, function(r) r$score %||% NA_real_, numeric(1L)),
    rank                  = NA_integer_,
    status                = vapply(results_list, `[[`, character(1L), "status"),
    error_message         = vapply(results_list, function(r) r$error_message   %||% NA_character_, character(1L)),
    warning_message       = vapply(results_list, function(r) r$warning_message %||% NA_character_, character(1L)),
    elapsed_time          = vapply(results_list, `[[`, numeric(1L), "elapsed_time"),
    n_local_models        = vapply(results_list, function(r) {
      v <- r$n_local_models; if (is.null(v) || is.na(v)) NA_integer_ else as.integer(v)
    }, integer(1L)),
    n_failed_local_models = vapply(results_list, function(r) {
      v <- r$n_failed_local_models; if (is.null(v) || is.na(v)) NA_integer_ else as.integer(v)
    }, integer(1L)),
    r2                    = vapply(results_list, function(r) r$r2        %||% NA_real_, numeric(1L)),
    mse                   = vapply(results_list, function(r) r$mse       %||% NA_real_, numeric(1L)),
    rmse                  = vapply(results_list, function(r) r$rmse      %||% NA_real_, numeric(1L)),
    mae                   = vapply(results_list, function(r) r$mae       %||% NA_real_, numeric(1L)),
    log_loss              = vapply(results_list, function(r) r$log_loss  %||% NA_real_, numeric(1L)),
    accuracy              = vapply(results_list, function(r) r$accuracy  %||% NA_real_, numeric(1L)),
    precision             = vapply(results_list, function(r) r$precision %||% NA_real_, numeric(1L)),
    recall                = vapply(results_list, function(r) r$recall    %||% NA_real_, numeric(1L)),
    f1_score              = vapply(results_list, function(r) r$f1_score  %||% NA_real_, numeric(1L)),
    stringsAsFactors      = FALSE
  )

  # Rank candidates
  history <- rank_grid_history(history)

  # Identify best candidate (rank == 1 among ok)
  ok_rows <- history[history$status == "ok", , drop = FALSE]
  if (nrow(ok_rows) > 0L) {
    best_row      <- ok_rows[which.min(ok_rows$rank), , drop = FALSE]
    best_bandwidth <- best_row$bandwidth
    best_score     <- best_row$score
  } else {
    best_bandwidth <- NA_real_
    best_score     <- NA_real_
  }

  total_elapsed <- proc.time()[["elapsed"]] - t_total_start

  new_gwpr_bandwidth(
    method           = "grid",
    best_bandwidth   = best_bandwidth,
    best_score       = best_score,
    criterion        = criterion,
    history          = history,
    metrics_history  = NULL,
    seed             = seed,
    convergence_info = list(n_candidates = n_candidates),
    elapsed_time     = total_elapsed,
    warnings         = character()
  )
}

# ---------------------------------------------------------------------------
# Internal helper (local copy to avoid cross-file ordering issues)
# ---------------------------------------------------------------------------

#' @noRd
`%||%` <- function(x, y) if (!is.null(x)) x else y

# ---------------------------------------------------------------------------
# parallel_map stub (used when parallel.R is not yet loaded)
# ---------------------------------------------------------------------------

#' Minimal parallel_map implementation
#'
#' Wraps `lapply` when `workers = 1` and `parallel::mclapply` when
#' `workers > 1`.  This stub is superseded once `parallel.R` is available.
#'
#' @param x       List of inputs.
#' @param fn      Function to apply to each element.
#' @param workers Integer number of workers.
#' @param seed    Random seed (passed to `mc.set.seed`).
#' @param ...     Additional arguments passed to `fn`.
#' @keywords internal
parallel_map <- function(x, fn, workers = 1L, seed = NULL, ...) {
  if (workers <= 1L) {
    lapply(x, fn, ...)
  } else {
    if (!is.null(seed)) set.seed(seed)
    parallel::mclapply(x, fn, mc.cores = workers, mc.set.seed = TRUE, ...)
  }
}

## tests/testthat/test-bandwidth-grid.R
## Tests for the Grid Search Bandwidth module (R/bandwidth_grid.R)
##
## All tests use small simulated data and workers = 1 (no parallelism).

# ---------------------------------------------------------------------------
# Helper: minimal mock scorer
# ---------------------------------------------------------------------------
# Returns a valid scorer result.  Score is a simple deterministic function of
# bandwidth so we can predict which candidate wins.

make_mock_scorer <- function(best_bw = 3, fail_bw = NULL) {
  function(context, bandwidth) {
    if (!is.null(fail_bw) && bandwidth == fail_bw) {
      stop("Intentional scorer failure for testing.")
    }
    # Score: distance from best_bw; lower is better
    score <- abs(bandwidth - best_bw) + 0.01
    list(
      score                 = score,
      criterion             = "MSE",
      status                = "ok",
      n_local_models        = 4L,
      n_failed_local_models = 0L,
      metrics               = list(
        R2   = 0.8,
        MSE  = score,
        RMSE = sqrt(score),
        MAE  = score * 0.9
      )
    )
  }
}

# Minimal gwpr_context suitable for grid search tests
make_grid_context <- function(adaptive = FALSE) {
  new_gwpr_context(
    formula   = y ~ x1,
    family    = "gaussian",
    model     = "pooling",
    effect    = "individual",
    id        = "id",
    time      = "time",
    kernel    = "bisquare",
    adaptive  = adaptive,
    threshold = 0.5,
    workers   = 1L,
    seed      = NULL
  )
}

# ===========================================================================
# make_grid_candidates
# ===========================================================================

test_that("make_grid_candidates errors when lower is missing", {
  expect_error(
    make_grid_candidates(upper = 5, step = 1, adaptive = FALSE),
    regexp = "lower"
  )
})

test_that("make_grid_candidates errors when upper is missing", {
  expect_error(
    make_grid_candidates(lower = 1, step = 1, adaptive = FALSE),
    regexp = "upper"
  )
})

test_that("make_grid_candidates errors when step is missing", {
  expect_error(
    make_grid_candidates(lower = 1, upper = 5, adaptive = FALSE),
    regexp = "step"
  )
})

test_that("make_grid_candidates errors when lower >= upper", {
  expect_error(
    make_grid_candidates(lower = 5, upper = 1, step = 1, adaptive = FALSE),
    regexp = "lower"
  )
  expect_error(
    make_grid_candidates(lower = 5, upper = 5, step = 1, adaptive = FALSE),
    regexp = "lower"
  )
})

test_that("make_grid_candidates errors when step <= 0", {
  expect_error(
    make_grid_candidates(lower = 1, upper = 5, step = -1, adaptive = FALSE),
    regexp = "step"
  )
  expect_error(
    make_grid_candidates(lower = 1, upper = 5, step = 0, adaptive = FALSE),
    regexp = "step"
  )
})

test_that("make_grid_candidates returns correct number of candidates (fixed)", {
  cands <- make_grid_candidates(lower = 1, upper = 5, step = 1, adaptive = FALSE)
  expect_length(cands, 5L)
  expect_equal(cands, c(1, 2, 3, 4, 5))
})

test_that("make_grid_candidates works with fractional step", {
  cands <- make_grid_candidates(lower = 0, upper = 1, step = 0.25, adaptive = FALSE)
  expect_length(cands, 5L)
})

test_that("make_grid_candidates rounds and deduplicates for adaptive = TRUE", {
  # step = 0.4 from 1 to 3 => seq: 1.0, 1.4, 1.8, 2.2, 2.6, 3.0
  # rounded: 1, 1, 2, 2, 3, 3 => unique: 1, 2, 3
  cands <- make_grid_candidates(lower = 1, upper = 3, step = 0.4, adaptive = TRUE)
  expect_true(all(cands == as.integer(cands)))  # all integers
  expect_true(!any(duplicated(cands)))           # no duplicates
  expect_true(all(cands >= 1L))                  # all positive
})

test_that("make_grid_candidates returns integers for adaptive = TRUE", {
  cands <- make_grid_candidates(lower = 2, upper = 8, step = 2, adaptive = TRUE)
  expect_type(cands, "integer")
})

# ===========================================================================
# score_bandwidth_candidate
# ===========================================================================

test_that("score_bandwidth_candidate returns ok status for valid scorer", {
  ctx    <- make_grid_context()
  scorer <- make_mock_scorer(best_bw = 3)
  result <- score_bandwidth_candidate(ctx, bandwidth = 3, scorer = scorer)

  expect_equal(result$status, "ok")
  expect_equal(result$bandwidth, 3)
  expect_true(is.numeric(result$score))
  expect_false(is.na(result$score))
  expect_true(is.numeric(result$elapsed_time))
})

test_that("score_bandwidth_candidate returns failed status on scorer error", {
  ctx    <- make_grid_context()
  bad_scorer <- function(context, bandwidth) stop("deliberate error")
  result <- score_bandwidth_candidate(ctx, bandwidth = 2, scorer = bad_scorer)

  expect_equal(result$status, "failed")
  expect_true(!is.na(result$error_message))
  expect_true(is.na(result$score))
})

test_that("score_bandwidth_candidate populates linear metrics", {
  ctx    <- make_grid_context()
  scorer <- make_mock_scorer(best_bw = 3)
  result <- score_bandwidth_candidate(ctx, bandwidth = 3, scorer = scorer)

  expect_false(is.na(result$r2))
  expect_false(is.na(result$mse))
  expect_false(is.na(result$rmse))
  expect_false(is.na(result$mae))
})

test_that("score_bandwidth_candidate fills logistic metrics with NA for gaussian scorer", {
  ctx    <- make_grid_context()
  scorer <- make_mock_scorer(best_bw = 3)
  result <- score_bandwidth_candidate(ctx, bandwidth = 3, scorer = scorer)

  expect_true(is.na(result$log_loss))
  expect_true(is.na(result$accuracy))
  expect_true(is.na(result$precision))
  expect_true(is.na(result$recall))
  expect_true(is.na(result$f1_score))
})

test_that("score_bandwidth_candidate records elapsed_time", {
  ctx    <- make_grid_context()
  scorer <- make_mock_scorer()
  result <- score_bandwidth_candidate(ctx, bandwidth = 2, scorer = scorer)
  expect_true(is.numeric(result$elapsed_time) && result$elapsed_time >= 0)
})

# ===========================================================================
# rank_grid_history
# ===========================================================================

test_that("rank_grid_history ranks ok candidates by ascending score", {
  h <- data.frame(
    candidate_id = 1:3,
    score        = c(0.5, 0.3, 0.7),
    status       = c("ok", "ok", "ok"),
    rank         = NA_integer_,
    stringsAsFactors = FALSE
  )
  h_ranked <- rank_grid_history(h)
  # score 0.3 should be rank 1, 0.5 rank 2, 0.7 rank 3
  expect_equal(h_ranked$rank[h$score == 0.3], 1L)
  expect_equal(h_ranked$rank[h$score == 0.5], 2L)
  expect_equal(h_ranked$rank[h$score == 0.7], 3L)
})

test_that("rank_grid_history places failed candidates after ok candidates", {
  h <- data.frame(
    candidate_id = 1:4,
    score        = c(0.5, NA, 0.3, NA),
    status       = c("ok", "failed", "ok", "failed"),
    rank         = NA_integer_,
    stringsAsFactors = FALSE
  )
  h_ranked <- rank_grid_history(h)
  ok_ranks   <- h_ranked$rank[h_ranked$status == "ok"]
  fail_ranks <- h_ranked$rank[h_ranked$status == "failed"]
  expect_true(all(ok_ranks < min(fail_ranks)))
})

test_that("rank_grid_history handles all-failed history", {
  h <- data.frame(
    candidate_id = 1:2,
    score        = c(NA_real_, NA_real_),
    status       = c("failed", "failed"),
    rank         = NA_integer_,
    stringsAsFactors = FALSE
  )
  h_ranked <- rank_grid_history(h)
  expect_equal(sort(h_ranked$rank), c(1L, 2L))
})

test_that("rank_grid_history returns data.frame unchanged if empty", {
  h <- data.frame()
  expect_equal(rank_grid_history(h), h)
})

# ===========================================================================
# search_bandwidth_grid
# ===========================================================================

test_that("search_bandwidth_grid errors when lower is missing from control", {
  ctx    <- make_grid_context()
  scorer <- make_mock_scorer()
  expect_error(
    search_bandwidth_grid(ctx, control = list(upper = 5, step = 1), scorer = scorer),
    regexp = "lower"
  )
})

test_that("search_bandwidth_grid errors when upper is missing from control", {
  ctx    <- make_grid_context()
  scorer <- make_mock_scorer()
  expect_error(
    search_bandwidth_grid(ctx, control = list(lower = 1, step = 1), scorer = scorer),
    regexp = "upper"
  )
})

test_that("search_bandwidth_grid errors when step is missing from control", {
  ctx    <- make_grid_context()
  scorer <- make_mock_scorer()
  expect_error(
    search_bandwidth_grid(ctx, control = list(lower = 1, upper = 5), scorer = scorer),
    regexp = "step"
  )
})

test_that("search_bandwidth_grid returns a gwpr_bandwidth object", {
  ctx    <- make_grid_context()
  scorer <- make_mock_scorer(best_bw = 3)
  result <- search_bandwidth_grid(
    ctx,
    control = list(lower = 1, upper = 5, step = 1),
    scorer  = scorer,
    workers = 1L
  )
  expect_s3_class(result, "gwpr_bandwidth")
})

test_that("search_bandwidth_grid history has correct number of rows", {
  ctx    <- make_grid_context()
  scorer <- make_mock_scorer(best_bw = 3)
  result <- search_bandwidth_grid(
    ctx,
    control = list(lower = 1, upper = 5, step = 1),
    scorer  = scorer,
    workers = 1L
  )
  # lower=1, upper=5, step=1 => 5 candidates
  expect_equal(nrow(result$history), 5L)
})

test_that("search_bandwidth_grid best_bandwidth is the candidate with lowest score", {
  ctx    <- make_grid_context()
  # best_bw=3: score for bw=3 is 0.01, for others > 0.01
  scorer <- make_mock_scorer(best_bw = 3)
  result <- search_bandwidth_grid(
    ctx,
    control = list(lower = 1, upper = 5, step = 1),
    scorer  = scorer,
    workers = 1L
  )
  expect_equal(result$best_bandwidth, 3)
})

test_that("search_bandwidth_grid history includes all required columns", {
  ctx    <- make_grid_context()
  scorer <- make_mock_scorer()
  result <- search_bandwidth_grid(
    ctx,
    control = list(lower = 1, upper = 3, step = 1),
    scorer  = scorer,
    workers = 1L
  )
  required_cols <- c(
    "candidate_id", "bandwidth", "bandwidth_type", "method", "criterion",
    "score", "rank", "status", "error_message", "warning_message",
    "elapsed_time", "n_local_models", "n_failed_local_models",
    "r2", "mse", "rmse", "mae",
    "log_loss", "accuracy", "precision", "recall", "f1_score"
  )
  expect_true(all(required_cols %in% names(result$history)))
})

test_that("search_bandwidth_grid retains failed candidates in history", {
  ctx         <- make_grid_context()
  # bw=2 will fail
  scorer <- make_mock_scorer(best_bw = 3, fail_bw = 2)
  result <- search_bandwidth_grid(
    ctx,
    control = list(lower = 1, upper = 5, step = 1),
    scorer  = scorer,
    workers = 1L
  )
  # All 5 candidates should be present
  expect_equal(nrow(result$history), 5L)
  # The failed candidate must be recorded with status "failed"
  expect_true(any(result$history$status == "failed"))
  expect_true(any(result$history$bandwidth == 2 & result$history$status == "failed"))
})

test_that("search_bandwidth_grid best_bandwidth comes from ok candidates only", {
  ctx    <- make_grid_context()
  # Make bw=3 fail; next best should be bw=4 (score=1.01) or bw=2 (score=1.01)
  scorer <- make_mock_scorer(best_bw = 3, fail_bw = 3)
  result <- search_bandwidth_grid(
    ctx,
    control = list(lower = 1, upper = 5, step = 1),
    scorer  = scorer,
    workers = 1L
  )
  # best_bandwidth should NOT be 3 (it failed)
  expect_false(is.na(result$best_bandwidth))
  expect_false(result$best_bandwidth == 3)
})

test_that("search_bandwidth_grid method field is 'grid'", {
  ctx    <- make_grid_context()
  scorer <- make_mock_scorer()
  result <- search_bandwidth_grid(
    ctx,
    control = list(lower = 1, upper = 3, step = 1),
    scorer  = scorer,
    workers = 1L
  )
  expect_equal(result$method, "grid")
  expect_true(all(result$history$method == "grid"))
})

test_that("search_bandwidth_grid records seed in returned object", {
  ctx    <- make_grid_context()
  scorer <- make_mock_scorer()
  result <- search_bandwidth_grid(
    ctx,
    control = list(lower = 1, upper = 3, step = 1),
    scorer  = scorer,
    workers = 1L,
    seed    = 42L
  )
  expect_equal(result$seed, 42L)
})

test_that("search_bandwidth_grid elapsed_time is non-negative", {
  ctx    <- make_grid_context()
  scorer <- make_mock_scorer()
  result <- search_bandwidth_grid(
    ctx,
    control = list(lower = 1, upper = 3, step = 1),
    scorer  = scorer,
    workers = 1L
  )
  expect_true(result$elapsed_time >= 0)
})

test_that("search_bandwidth_grid adaptive grid produces integer candidates", {
  ctx    <- make_grid_context(adaptive = TRUE)
  scorer <- make_mock_scorer(best_bw = 3)
  result <- search_bandwidth_grid(
    ctx,
    control = list(lower = 2, upper = 8, step = 2),
    scorer  = scorer,
    workers = 1L
  )
  expect_true(all(result$history$bandwidth == as.integer(result$history$bandwidth)))
})

test_that("search_bandwidth_grid: best_bandwidth has rank 1 in history", {
  ctx    <- make_grid_context()
  scorer <- make_mock_scorer(best_bw = 2)
  result <- search_bandwidth_grid(
    ctx,
    control = list(lower = 1, upper = 5, step = 1),
    scorer  = scorer,
    workers = 1L
  )
  best_row <- result$history[result$history$bandwidth == result$best_bandwidth, , drop = FALSE]
  expect_equal(best_row$rank[1L], 1L)
})

test_that("search_bandwidth_grid convergence_info contains n_candidates", {
  ctx    <- make_grid_context()
  scorer <- make_mock_scorer()
  result <- search_bandwidth_grid(
    ctx,
    control = list(lower = 1, upper = 4, step = 1),
    scorer  = scorer,
    workers = 1L
  )
  expect_equal(result$convergence_info$n_candidates, 4L)
})

#' @title Metrics Module
#' @description Functions for computing evaluation metrics for linear and
#'   logistic panel regression models.
#' @name metrics
NULL

#' Compute linear regression metrics
#'
#' Computes R2, MSE, RMSE, and MAE for a linear model.
#'
#' @param y Numeric vector of observed values.
#' @param y_hat Numeric vector of predicted values. May contain NA values,
#'   which are excluded from all metric calculations.
#'
#' @return A named list with elements: \code{R2}, \code{MSE}, \code{RMSE},
#'   \code{MAE}. Returns \code{NA_real_} for each metric when no complete
#'   observations remain after removing NAs.
#'
#' @examples
#' y <- c(1, 2, 3, 4, 5)
#' y_hat <- c(1, 2, 3, 4, 5)
#' compute_linear_metrics(y, y_hat)
#'
#' @noRd
compute_linear_metrics <- function(y, y_hat) {
  keep <- !is.na(y_hat) & !is.na(y)
  y     <- y[keep]
  y_hat <- y_hat[keep]

  if (length(y) == 0L) {
    return(list(R2 = NA_real_, MSE = NA_real_, RMSE = NA_real_, MAE = NA_real_))
  }

  ss_res <- sum((y - y_hat)^2)
  ss_tot <- sum((y - mean(y))^2)
  R2   <- if (ss_tot == 0) NA_real_ else 1 - ss_res / ss_tot
  MSE  <- mean((y - y_hat)^2)
  RMSE <- sqrt(MSE)
  MAE  <- mean(abs(y - y_hat))

  list(R2 = R2, MSE = MSE, RMSE = RMSE, MAE = MAE)
}

#' Compute logistic regression metrics
#'
#' Computes log_loss, accuracy, precision, recall, and f1_score for a binary
#' classification model.
#'
#' @param y Integer or numeric vector of true binary labels (0/1).
#' @param prob Numeric vector of predicted probabilities for class 1.
#' @param threshold Numeric scalar; classification threshold (default 0.5).
#'
#' @return A named list with elements: \code{log_loss}, \code{accuracy},
#'   \code{precision}, \code{recall}, \code{f1_score}. \code{precision} and
#'   \code{recall} are \code{NA_real_} when their denominators are zero.
#'
#' @examples
#' y <- c(1, 0, 1, 0)
#' prob <- c(0.9, 0.1, 0.8, 0.2)
#' compute_logistic_metrics(y, prob)
#'
#' @noRd
compute_logistic_metrics <- function(y, prob, threshold = 0.5) {
  ll         <- safe_log_loss(y, prob)
  class_pred <- as.integer(prob >= threshold)
  counts     <- safe_confusion_counts(y, class_pred)

  TP <- counts$TP
  FP <- counts$FP
  FN <- counts$FN
  TN <- counts$TN
  n  <- TP + FP + FN + TN

  accuracy  <- if (n == 0L) NA_real_ else (TP + TN) / n
  precision <- counts$precision
  recall    <- counts$recall

  f1_score <- if (is.na(precision) || is.na(recall)) {
    NA_real_
  } else {
    denom <- precision + recall
    if (denom == 0) NA_real_ else 2 * precision * recall / denom
  }

  list(
    log_loss  = ll,
    accuracy  = accuracy,
    precision = precision,
    recall    = recall,
    f1_score  = f1_score
  )
}

#' Numerically safe log loss
#'
#' Clips predicted probabilities to \code{[eps, 1 - eps]} before computing
#' binary cross-entropy, preventing infinite log-loss values.
#'
#' @param y Numeric vector of true binary labels (0/1).
#' @param prob Numeric vector of predicted probabilities.
#' @param eps Numeric scalar; clipping bound (default \code{1e-15}).
#'
#' @return A single numeric value: the mean binary cross-entropy.
#'
#' @examples
#' safe_log_loss(c(1, 0, 1), c(0.99, 0.01, 0.5))
#'
#' @noRd
safe_log_loss <- function(y, prob, eps = 1e-15) {
  prob <- pmax(pmin(prob, 1 - eps), eps)
  -mean(y * log(prob) + (1 - y) * log(1 - prob))
}

#' Compute confusion matrix counts
#'
#' Returns TP, FP, FN, TN and derived precision and recall. Issues a
#' \code{warning()} and returns \code{NA_real_} when precision or recall
#' denominators are zero.
#'
#' @param y Integer or numeric vector of true binary labels (0/1).
#' @param class_pred Integer or numeric vector of predicted binary labels (0/1).
#'
#' @return A named list: \code{TP}, \code{FP}, \code{FN}, \code{TN},
#'   \code{precision}, \code{recall}.
#'
#' @examples
#' safe_confusion_counts(c(1, 0, 1, 0), c(1, 1, 0, 0))
#'
#' @noRd
safe_confusion_counts <- function(y, class_pred) {
  TP <- sum(y == 1L & class_pred == 1L)
  FP <- sum(y == 0L & class_pred == 1L)
  FN <- sum(y == 1L & class_pred == 0L)
  TN <- sum(y == 0L & class_pred == 0L)

  precision_denom <- TP + FP
  recall_denom    <- TP + FN

  if (precision_denom == 0L) {
    warning("precision denominator (TP + FP) is zero; returning NA_real_")
    precision <- NA_real_
  } else {
    precision <- TP / precision_denom
  }

  if (recall_denom == 0L) {
    warning("recall denominator (TP + FN) is zero; returning NA_real_")
    recall <- NA_real_
  } else {
    recall <- TP / recall_denom
  }

  list(TP = TP, FP = FP, FN = FN, TN = TN,
       precision = precision, recall = recall)
}

test_that("compute_linear_metrics: perfect predictions give R2=1, MSE=0, RMSE=0, MAE=0", {
  y     <- c(1, 2, 3, 4, 5)
  y_hat <- c(1, 2, 3, 4, 5)
  m <- compute_linear_metrics(y, y_hat)
  expect_equal(m$R2,   1)
  expect_equal(m$MSE,  0)
  expect_equal(m$RMSE, 0)
  expect_equal(m$MAE,  0)
})

test_that("compute_linear_metrics: all-wrong predictions give correct metrics", {
  y     <- c(0, 0, 0, 0)
  y_hat <- c(1, 1, 1, 1)
  m <- compute_linear_metrics(y, y_hat)
  expect_equal(m$MSE,  1)
  expect_equal(m$RMSE, 1)
  expect_equal(m$MAE,  1)
  # ss_tot == 0 because all y equal => R2 is NA
  expect_true(is.na(m$R2))
})

test_that("compute_linear_metrics: known non-trivial values", {
  y     <- c(3, 3, 3, 3)     # all same
  y_hat <- c(2, 4, 2, 4)
  m <- compute_linear_metrics(y, y_hat)
  expect_equal(m$MSE,  1)
  expect_equal(m$RMSE, 1)
  expect_equal(m$MAE,  1)
})

test_that("compute_linear_metrics: NA in y_hat are dropped", {
  y     <- c(1, 2, 3, 4)
  y_hat <- c(1, 2, NA, 4)
  m <- compute_linear_metrics(y, y_hat)
  # Only indices 1,2,4 are used: errors = 0,0,0 => perfect
  expect_equal(m$MSE, 0)
  expect_equal(m$MAE, 0)
})

test_that("compute_linear_metrics: all NA y_hat returns NA metrics", {
  y     <- c(1, 2, 3)
  y_hat <- c(NA, NA, NA)
  m <- compute_linear_metrics(y, y_hat)
  expect_true(is.na(m$R2))
  expect_true(is.na(m$MSE))
  expect_true(is.na(m$RMSE))
  expect_true(is.na(m$MAE))
})

# ---------------------------------------------------------------------------
# safe_log_loss
# ---------------------------------------------------------------------------

test_that("safe_log_loss: perfect predictions give log_loss near 0", {
  y    <- c(1, 0, 1, 0)
  prob <- c(1 - 1e-10, 1e-10, 1 - 1e-10, 1e-10)
  ll <- safe_log_loss(y, prob)
  expect_true(is.finite(ll))
  expect_lt(ll, 1e-8)
})

test_that("safe_log_loss: probabilities at 0/1 remain finite (no Inf)", {
  y    <- c(1, 0, 1, 0)
  prob <- c(1, 0, 1, 0)   # would produce Inf without clipping
  ll <- safe_log_loss(y, prob)
  expect_true(is.finite(ll))
})

test_that("safe_log_loss: probabilities near 0/1 remain finite", {
  y    <- c(1, 0)
  prob <- c(1e-300, 1 - 1e-300)
  ll <- safe_log_loss(y, prob)
  expect_true(is.finite(ll))
})

# ---------------------------------------------------------------------------
# safe_confusion_counts
# ---------------------------------------------------------------------------

test_that("safe_confusion_counts: basic correctness", {
  y          <- c(1, 1, 0, 0)
  class_pred <- c(1, 0, 1, 0)
  cc <- safe_confusion_counts(y, class_pred)
  expect_equal(cc$TP, 1L)
  expect_equal(cc$FP, 1L)
  expect_equal(cc$FN, 1L)
  expect_equal(cc$TN, 1L)
  expect_equal(cc$precision, 0.5)
  expect_equal(cc$recall,    0.5)
})

test_that("safe_confusion_counts: precision NA when no positive predictions", {
  y          <- c(1, 0, 1, 0)
  class_pred <- c(0, 0, 0, 0)   # TP + FP == 0
  expect_warning(
    cc <- safe_confusion_counts(y, class_pred),
    regexp = "precision denominator"
  )
  expect_true(is.na(cc$precision))
  expect_false(is.na(cc$recall))
})

test_that("safe_confusion_counts: recall NA when no positive ground-truth", {
  y          <- c(0, 0, 0, 0)   # TP + FN == 0
  class_pred <- c(1, 0, 1, 0)
  expect_warning(
    cc <- safe_confusion_counts(y, class_pred),
    regexp = "recall denominator"
  )
  expect_true(is.na(cc$recall))
  # precision is defined here (TP=0, FP=2)
  expect_false(is.na(cc$precision))
})

# ---------------------------------------------------------------------------
# compute_logistic_metrics
# ---------------------------------------------------------------------------

test_that("compute_logistic_metrics: perfect predictions give accuracy=1, log_loss~0", {
  y    <- c(1, 0, 1, 0)
  prob <- c(0.999, 0.001, 0.999, 0.001)
  m <- compute_logistic_metrics(y, prob)
  expect_true(is.finite(m$log_loss))
  expect_lt(m$log_loss, 0.01)
  expect_equal(m$accuracy, 1)
  expect_equal(m$precision, 1)
  expect_equal(m$recall, 1)
  expect_equal(m$f1_score, 1)
})

test_that("compute_logistic_metrics: all-wrong predictions", {
  y    <- c(1, 1, 0, 0)
  prob <- c(0.01, 0.01, 0.99, 0.99)   # everything wrong
  m <- compute_logistic_metrics(y, prob)
  expect_equal(m$accuracy, 0)
  # precision: TP=0, FP=2 => 0; recall: TP=0, FN=2 => 0
  expect_equal(m$precision, 0)
  expect_equal(m$recall, 0)
  expect_true(is.na(m$f1_score) || m$f1_score == 0)
})

test_that("compute_logistic_metrics: threshold change alters classification", {
  y    <- c(1, 1, 0, 0)
  prob <- c(0.6, 0.4, 0.6, 0.4)

  m_low  <- suppressWarnings(compute_logistic_metrics(y, prob, threshold = 0.3))
  m_high <- suppressWarnings(compute_logistic_metrics(y, prob, threshold = 0.7))

  # At threshold 0.3 all are predicted positive
  expect_equal(m_low$accuracy, 0.5)   # 2 correct out of 4

  # At threshold 0.7 all are predicted negative
  expect_equal(m_high$accuracy, 0.5)  # 2 correct out of 4

  # Metrics should differ between thresholds
  expect_false(identical(m_low$precision, m_high$precision) &&
                 identical(m_low$recall,    m_high$recall))
})

test_that("compute_logistic_metrics: precision NA does not error", {
  y    <- c(1, 1, 1, 1)   # no negatives
  prob <- c(0.4, 0.4, 0.4, 0.4)  # all predicted negative at default threshold
  # TP + FP = 0 => precision NA
  expect_warning(
    m <- compute_logistic_metrics(y, prob),
    regexp = "precision"
  )
  expect_true(is.na(m$precision))
  expect_true(is.na(m$f1_score))
})

test_that("compute_logistic_metrics: recall NA does not error", {
  y    <- c(0, 0, 0, 0)   # no positives
  prob <- c(0.6, 0.6, 0.6, 0.6)  # all predicted positive
  # TP + FN = 0 => recall NA
  expect_warning(
    m <- compute_logistic_metrics(y, prob),
    regexp = "recall"
  )
  expect_true(is.na(m$recall))
  expect_true(is.na(m$f1_score))
})

test_that("compute_logistic_metrics: f1_score NA when both precision and recall are NA", {
  # All same class, all predicted wrong class => both NA
  y    <- c(0, 0, 0, 0)
  prob <- c(0.4, 0.4, 0.4, 0.4)  # all predicted 0 (negative)
  # TP=0, FP=0 => precision NA; TP=0, FN=0 => recall NA
  suppressWarnings(m <- compute_logistic_metrics(y, prob))
  expect_true(is.na(m$f1_score))
})

#' @title Diagnostics Module for GWPR.light 1.0.0
#'
#' @description
#' Unified diagnostic interface for Geographically Weighted Panel Regression
#' models.  Provides Moran's I (spatial autocorrelation), local F test
#' (fixed vs. pooling), local Hausman test (fixed vs. random), and local
#' Breusch-Pagan LM test.
#'
#' @details
#' **Model applicability**
#'
#' | Diagnostic   | Linear | Logistic | Notes                                      |
#' |--------------|--------|----------|--------------------------------------------|
#' | moran        | yes    | yes      | Logistic uses Pearson residual              |
#' | f_test       | yes    | no       | Requires within and pooling models          |
#' | hausman      | yes    | no       | Only meaningful for random-effect models    |
#' | lm_test      | yes    | no       | Pooling or random-effect models             |
#'
#' **Panel balance**
#'
#' `compute_panel_moran()` is fully supported for balanced panels.  For
#' unbalanced panels a `warning()` is issued and the function attempts
#' computation using only the time periods present in every individual;
#' results may be unreliable.
#'
#' **Logistic interpretation**
#'
#' Moran's I computed from Pearson residuals of a Logistic model does not
#' follow the same asymptotic distribution as for linear models.  Treat the
#' test result as an exploratory heuristic, not a formal test.
#'
#' @name diagnostics
NULL

# ---------------------------------------------------------------------------
# compute_logistic_pearson_residual
# ---------------------------------------------------------------------------

#' Compute Pearson residuals for a binary logistic model
#'
#' Calculates the Pearson (standardised) residual for each observation:
#'
#' \deqn{r_i = \frac{y_i - p_i}{\sqrt{p_i (1 - p_i)}}}
#'
#' where \eqn{p_i} is the fitted probability clipped to
#' \eqn{[\texttt{eps},\; 1 - \texttt{eps}]} to prevent division by zero.
#'
#' @param y   Integer or numeric vector of true binary labels (0 or 1).
#' @param prob Numeric vector of fitted probabilities (same length as `y`).
#'   Values are clipped to `[eps, 1 - eps]`.
#' @param eps Numeric scalar; clipping bound (default `1e-15`).
#'
#' @return A numeric vector of Pearson residuals, the same length as `y`.
#'   No `Inf` or `NaN` values are produced as long as `eps > 0`.
#'
#' @details
#' **Applicable model**: binomial (logistic).
#'
#' **Logistic interpretation limit**: The Pearson residual is used here
#' only to provide a residual-like quantity for spatial diagnostics.  It is
#' not a formal goodness-of-fit statistic in the chi-squared sense unless the
#' model is correctly specified and observations are independent.
#'
#' @examples
#' y    <- c(1L, 0L, 1L, 0L)
#' prob <- c(0.9, 0.1, 0.8, 0.2)
#' compute_logistic_pearson_residual(y, prob)
#'
#' @noRd
compute_logistic_pearson_residual <- function(y, prob, eps = 1e-15) {
  if (!is.numeric(y) && !is.integer(y)) {
    stop("`y` must be a numeric or integer vector of 0/1 values.", call. = FALSE)
  }
  if (!is.numeric(prob)) {
    stop("`prob` must be a numeric vector of fitted probabilities.", call. = FALSE)
  }
  if (length(y) != length(prob)) {
    stop("`y` and `prob` must have the same length.", call. = FALSE)
  }
  if (!is.numeric(eps) || length(eps) != 1L || eps <= 0) {
    stop("`eps` must be a single positive numeric value.", call. = FALSE)
  }

  p <- pmax(pmin(as.numeric(prob), 1 - eps), eps)
  (as.numeric(y) - p) / sqrt(p * (1 - p))
}

# ---------------------------------------------------------------------------
# compute_panel_moran
# ---------------------------------------------------------------------------

#' Compute Moran's I for panel residuals
#'
#' Computes a panel-level Moran's I statistic by averaging Moran's I across
#' all time periods.  The variance and Z-statistic follow Beenstock &
#' Felsenstein (2019).
#'
#' @param residuals      Numeric matrix of residuals with individuals in rows
#'   and time periods in columns, **or** a numeric vector.  When a vector is
#'   supplied `panel_index` must be provided so that the vector can be reshaped
#'   into the required matrix.
#' @param spatial_weights Numeric n x n spatial weights matrix (rows should
#'   sum to 1 after row-standardisation).  Diagonal is set to 0 internally.
#'   `n` must equal the number of individuals (rows in the residual matrix).
#' @param panel_index    Optional data.frame or list with two columns / elements
#'   named `id` and `time` (in that order) that identify each element of
#'   `residuals` when it is a vector.  Must have the same length as the vector.
#'
#' @return A named list:
#' \describe{
#'   \item{`statistic`}{Z-score of the average Moran's I.}
#'   \item{`p_value`}{One-sided p-value (upper tail, alternative "greater").}
#'   \item{`estimated_I`}{Observed average Moran's I across time periods.}
#'   \item{`expected_I`}{Expected value of Moran's I under the null: \eqn{-1/(n-1)}.}
#'   \item{`variance`}{Variance of the average Moran's I.}
#'   \item{`alternative`}{`"greater"`.}
#'   \item{`n_individuals`}{Number of spatial units \eqn{n}.}
#'   \item{`n_periods`}{Number of time periods \eqn{T}.}
#' }
#'
#' @details
#' **Panel balance requirement**: Balanced panels are fully supported.  For
#' unbalanced panels a `warning()` is issued.  The function then uses only the
#' intersection of time periods present in every individual, which may discard
#' observations.
#'
#' **Applicable models**: both linear (model residuals) and logistic (Pearson
#' residuals via `compute_logistic_pearson_residual()`).
#'
#' **Logistic interpretation limit**: When used with Pearson residuals from a
#' logistic model the test is exploratory only.
#'
#' @references
#' Beenstock, M., Felsenstein, D. (2019). *The Econometric Analysis of
#' Non-Stationary Spatial Panel Data*. Springer.
#'
#' @examples
#' set.seed(1L)
#' n  <- 5L   # individuals
#' Ti <- 3L   # time periods
#' resid_mat <- matrix(rnorm(n * Ti), nrow = n, ncol = Ti)
#' W <- matrix(1 / (n - 1), nrow = n, ncol = n)
#' diag(W) <- 0
#' compute_panel_moran(resid_mat, W)
#'
#' @noRd
compute_panel_moran <- function(residuals, spatial_weights, panel_index = NULL) {

  # ---- reshape vector -> matrix if panel_index provided -------------------
  if (is.numeric(residuals) && !is.matrix(residuals)) {
    if (is.null(panel_index)) {
      stop(
        "When `residuals` is a vector, `panel_index` must be provided.",
        call. = FALSE
      )
    }
    idx <- if (is.data.frame(panel_index)) panel_index else as.data.frame(panel_index)
    if (ncol(idx) < 2L) {
      stop("`panel_index` must have at least two columns: id and time.", call. = FALSE)
    }
    colnames(idx)[1:2] <- c("pid", "tid")
    if (nrow(idx) != length(residuals)) {
      stop(
        "`panel_index` must have the same number of rows as `residuals` has elements.",
        call. = FALSE
      )
    }
    idx$resid <- residuals
    individuals <- unique(idx$pid)
    times       <- unique(idx$tid)

    # Check balance
    counts <- tapply(idx$resid, list(idx$pid, idx$tid), length)
    is_balanced <- all(counts == 1L, na.rm = TRUE) &&
      all(!is.na(counts))

    if (!is_balanced) {
      warning(
        "The panel is unbalanced. Moran's I for unbalanced panels is not ",
        "formally supported in version 1.0.0. Only the intersection of ",
        "time periods present in all individuals will be used. ",
        "Results may be unreliable.",
        call. = FALSE
      )
      # Keep only time periods present for every individual
      times_per_id <- tapply(idx$tid, idx$pid, unique)
      common_times <- Reduce(intersect, times_per_id)
      if (length(common_times) == 0L) {
        stop(
          "No common time periods found across all individuals. ",
          "Cannot compute panel Moran's I.",
          call. = FALSE
        )
      }
      idx <- idx[idx$tid %in% common_times, ]
      times <- common_times
    }

    # Build matrix: rows = individuals, cols = time periods
    n  <- length(individuals)
    Ti <- length(times)
    resid_mat <- matrix(NA_real_, nrow = n, ncol = Ti,
                        dimnames = list(as.character(individuals),
                                        as.character(times)))
    for (i in seq_along(individuals)) {
      id_rows <- idx[idx$pid == individuals[i], ]
      for (j in seq_along(times)) {
        hit <- id_rows[id_rows$tid == times[j], "resid"]
        if (length(hit) == 1L) resid_mat[i, j] <- hit
      }
    }
    residuals <- resid_mat
  }

  if (!is.matrix(residuals)) {
    stop("`residuals` must be a numeric matrix (n_individuals x n_periods).", call. = FALSE)
  }

  n  <- nrow(residuals)
  Ti <- ncol(residuals)

  if (!is.matrix(spatial_weights) || nrow(spatial_weights) != n || ncol(spatial_weights) != n) {
    stop(
      "`spatial_weights` must be an n x n numeric matrix where n equals ",
      "the number of rows in `residuals`.",
      call. = FALSE
    )
  }

  # Row-standardise and zero diagonal
  W <- spatial_weights
  diag(W) <- 0
  row_sums <- rowSums(W)
  zero_rows <- row_sums == 0
  if (any(zero_rows)) {
    warning(
      sum(zero_rows), " row(s) in `spatial_weights` sum to 0 after removing the diagonal. ",
      "Those units receive uniform weights.",
      call. = FALSE
    )
    row_sums[zero_rows] <- 1
  }
  W <- W / row_sums

  # Compute Moran's I for each time period
  I_vec <- numeric(Ti)
  for (t in seq_len(Ti)) {
    e      <- residuals[, t]
    ss     <- sum(e^2)
    if (ss == 0) {
      I_vec[t] <- 0
      next
    }
    # sum_ij w_ij * e_i * e_j  (vectorised)
    I_vec[t] <- as.numeric(t(e) %*% W %*% e) / ss
  }

  I_mean <- mean(I_vec)

  # Variance of average I (Beenstock & Felsenstein 2019 formula)
  V2_num   <- n * sum(W^2) + 3 * sum(W)^2 - n * sum(colSums(W)^2)
  V2_denom <- Ti * (n^2 - 1) * sum(W)^2
  V2       <- if (V2_denom == 0) NA_real_ else V2_num / V2_denom

  # Expected I
  E_I <- -1 / (n - 1)

  # Z-statistic
  ZI <- if (is.na(V2) || V2 <= 0) NA_real_ else (I_mean - E_I) / sqrt(V2)

  # One-sided p-value (greater)
  p_val <- if (is.na(ZI)) NA_real_ else stats::pnorm(ZI, lower.tail = FALSE)

  list(
    statistic    = ZI,
    p_value      = p_val,
    estimated_I  = I_mean,
    expected_I   = E_I,
    variance     = V2,
    alternative  = "greater",
    n_individuals = n,
    n_periods    = Ti
  )
}

# ---------------------------------------------------------------------------
# diagnose_moran
# ---------------------------------------------------------------------------

#' Run Moran's I diagnostic on a gwpr_fit object
#'
#' Extracts residuals from a fitted GWPR model (Pearson residuals for logistic
#' models, raw residuals for linear models) and computes the panel Moran's I
#' statistic.
#'
#' @param object A `gwpr_fit` object returned by `fit_gwpr()` or similar.
#' @param spatial_weights A row-standardised n x n spatial weights matrix.
#'   `n` must equal the number of spatial individuals in the fitted model.
#' @param panel_index A data.frame or list with columns/elements `id` and
#'   `time` that identify each element of `object$residuals`.
#' @param ...  Currently ignored.
#'
#' @return A named list compatible with the `diagnostics` slot of a
#'   `gwpr_diagnostics` object.  Contains the elements returned by
#'   `compute_panel_moran()` plus `residual_type`.
#'
#' @details
#' **Applicable models**: gaussian (linear residuals) and binomial (Pearson
#' residuals).
#'
#' **Panel balance**: See `compute_panel_moran()`.
#'
#' **Failure conditions**: Fails if `object` is not a `gwpr_fit`, if
#' `object$residuals` is `NULL`, or if `spatial_weights` dimensions do not
#' match the number of individuals.
#'
#' **Logistic interpretation limit**: Moran's I computed on Pearson residuals
#' is exploratory; the asymptotic distribution differs from the linear case.
#'
#' @examples
#' \donttest{
#' library(sf)
#' pts <- sf::st_as_sf(
#'   data.frame(id = 1:4, X = c(0,1,0,1), Y = c(0,0,1,1)),
#'   coords = c("X", "Y"), crs = NA_integer_
#' )
#' dat <- data.frame(
#'   id   = rep(1:4, each = 5),
#'   time = rep(1:5, 4),
#'   y    = rnorm(20),
#'   x1   = rnorm(20)
#' )
#' fit <- fit_gwpr(y ~ x1, data = dat, spatial = pts, id = "id",
#'                 time = "time", bandwidth = 2, workers = 1)
#' W <- matrix(1/3, nrow = 4, ncol = 4); diag(W) <- 0
#' idx <- dat[, c("id", "time")]
#' diagnose_moran(fit, W, idx)
#' }
#'
#' @export
diagnose_moran <- function(object, spatial_weights, panel_index, ...) {
  if (!inherits(object, "gwpr_fit")) {
    stop("`object` must be a `gwpr_fit` object.", call. = FALSE)
  }
  if (is.null(object$residuals)) {
    stop("`object$residuals` is NULL; cannot compute Moran's I.", call. = FALSE)
  }

  family <- object$family %||% "gaussian"

  if (identical(family, "binomial")) {
    # Extract predicted probabilities and response from metadata
    prob <- object$metadata$prob
    y    <- object$metadata$response

    if (is.null(prob) || is.null(y)) {
      stop(
        "For logistic models, `object$metadata` must contain `prob` and `response`. ",
        "Cannot compute Pearson residuals.",
        call. = FALSE
      )
    }
    resid_vec     <- compute_logistic_pearson_residual(y, prob)
    residual_type <- "pearson"
  } else {
    resid_vec     <- as.numeric(object$residuals)
    residual_type <- "raw"
  }

  moran_res <- compute_panel_moran(resid_vec, spatial_weights, panel_index)
  moran_res$residual_type <- residual_type
  moran_res
}

# ---------------------------------------------------------------------------
# diagnose_local_f
# ---------------------------------------------------------------------------

#' Local F test diagnostic on a gwpr_fit object
#'
#' Performs a local F test (fixed effects vs. pooling) using per-unit local
#' residuals stored in the fitted model object.
#'
#' @param object A `gwpr_fit` object.
#' @param ...  Currently ignored.
#'
#' @return A named list with elements:
#' \describe{
#'   \item{`local_f`}{Data frame with columns `unit_id`, `statistic`, `p_value`,
#'     `df1`, `df2`, `status`.}
#'   \item{`n_tested`}{Number of units tested.}
#'   \item{`n_failed`}{Number of units where the test could not be computed.}
#' }
#'
#' @details
#' **Applicable models**: gaussian (linear).  Not applicable to logistic models;
#' returns a `status = "not_applicable"` result when `family = "binomial"`.
#'
#' **Panel balance requirement**: No constraint; the test uses per-unit local
#' residuals already computed during fitting.
#'
#' **Failure conditions**: If `local_results` is empty or missing, all units
#' are reported as failed.  If a unit's local result does not contain the
#' information needed (within and pooling residuals), that unit is reported as
#' failed with an informative `status`.
#'
#' **Logistic interpretation limit**: Not applicable; see above.
#'
#' @examples
#' \donttest{
#' library(sf)
#' pts <- sf::st_as_sf(
#'   data.frame(id = 1:4, X = c(0,1,0,1), Y = c(0,0,1,1)),
#'   coords = c("X", "Y"), crs = NA_integer_
#' )
#' dat <- data.frame(
#'   id   = rep(1:4, each = 5),
#'   time = rep(1:5, 4),
#'   y    = rnorm(20),
#'   x1   = rnorm(20)
#' )
#' fit <- fit_gwpr(y ~ x1, data = dat, spatial = pts, id = "id",
#'                 time = "time", bandwidth = 2, workers = 1)
#' diagnose_local_f(fit)
#' }
#'
#' @export
diagnose_local_f <- function(object, ...) {
  if (!inherits(object, "gwpr_fit")) {
    stop("`object` must be a `gwpr_fit` object.", call. = FALSE)
  }

  family <- object$family %||% "gaussian"
  if (identical(family, "binomial")) {
    return(list(
      local_f    = data.frame(
        unit_id   = character(0),
        statistic = numeric(0),
        p_value   = numeric(0),
        df1       = integer(0),
        df2       = integer(0),
        status    = character(0),
        stringsAsFactors = FALSE
      ),
      n_tested   = 0L,
      n_failed   = 0L,
      status     = "not_applicable",
      message    = "Local F test is not applicable to logistic (binomial) models."
    ))
  }

  local_results <- object$local_results
  if (length(local_results) == 0L) {
    return(list(
      local_f  = data.frame(
        unit_id   = character(0),
        statistic = numeric(0),
        p_value   = numeric(0),
        df1       = integer(0),
        df2       = integer(0),
        status    = character(0),
        stringsAsFactors = FALSE
      ),
      n_tested = 0L,
      n_failed = 0L,
      status   = "no_local_results"
    ))
  }

  unit_ids  <- names(local_results)
  if (is.null(unit_ids)) unit_ids <- as.character(seq_along(local_results))

  rows <- vector("list", length(unit_ids))

  for (i in seq_along(unit_ids)) {
    uid <- unit_ids[[i]]
    res <- local_results[[i]]

    row <- list(unit_id = uid, statistic = NA_real_, p_value = NA_real_,
                df1 = NA_integer_, df2 = NA_integer_, status = "ok")

    if (!is.null(res$f_statistic) && !is.null(res$f_p_value)) {
      # Pre-computed values stored by model engine
      row$statistic <- as.numeric(res$f_statistic)
      row$p_value   <- as.numeric(res$f_p_value)
      row$df1       <- as.integer(res$f_df1 %||% NA_integer_)
      row$df2       <- as.integer(res$f_df2 %||% NA_integer_)
      row$status    <- res$status %||% "ok"
    } else if (!is.null(res$within_rss) && !is.null(res$pooling_rss)) {
      # Compute from RSS values
      rss_w  <- as.numeric(res$within_rss)
      rss_p  <- as.numeric(res$pooling_rss)
      df_w   <- as.integer(res$within_df)
      df_p   <- as.integer(res$pooling_df)
      df1    <- df_p - df_w
      df2    <- df_w

      if (df1 <= 0L || df2 <= 0L || rss_w <= 0) {
        row$status <- "insufficient_df"
      } else {
        f_stat       <- ((rss_p - rss_w) / df1) / (rss_w / df2)
        row$statistic <- f_stat
        row$p_value   <- stats::pf(f_stat, df1 = df1, df2 = df2, lower.tail = FALSE)
        row$df1       <- df1
        row$df2       <- df2
        row$status    <- res$status %||% "ok"
      }
    } else {
      row$status <- "missing_local_f_data"
    }

    rows[[i]] <- row
  }

  out_df <- data.frame(
    unit_id   = vapply(rows, `[[`, character(1L), "unit_id"),
    statistic = vapply(rows, `[[`, numeric(1L),   "statistic"),
    p_value   = vapply(rows, `[[`, numeric(1L),   "p_value"),
    df1       = vapply(rows, `[[`, integer(1L),   "df1"),
    df2       = vapply(rows, `[[`, integer(1L),   "df2"),
    status    = vapply(rows, `[[`, character(1L), "status"),
    stringsAsFactors = FALSE
  )

  n_failed <- sum(out_df$status != "ok")

  list(local_f  = out_df,
       n_tested = nrow(out_df),
       n_failed = n_failed,
       status   = "ok")
}

# ---------------------------------------------------------------------------
# diagnose_hausman
# ---------------------------------------------------------------------------

#' Local Hausman test diagnostic on a gwpr_fit object
#'
#' Performs a local Hausman test (within vs. random) for each spatial unit
#' using test statistics pre-computed during model fitting or stored in
#' `local_results`.
#'
#' @param object A `gwpr_fit` object.
#' @param ...  Currently ignored.
#'
#' @return A named list with elements:
#' \describe{
#'   \item{`local_hausman`}{Data frame with columns `unit_id`, `statistic`,
#'     `p_value`, `df`, `status`.}
#'   \item{`n_tested`}{Number of units tested.}
#'   \item{`n_failed`}{Number of units where the test could not be computed.}
#'   \item{`status`}{Overall status: `"ok"`, `"not_applicable"`, or
#'     `"no_local_results"`.}
#' }
#'
#' @details
#' **Applicable models**: gaussian with `model = "random"`.  For pooling
#' models the Hausman test is not meaningful; the function returns a
#' `status = "not_applicable"` result.  For logistic models the function also
#' returns `status = "not_applicable"`.
#'
#' **Panel balance requirement**: No constraint at the unit level.
#'
#' **Failure conditions**: Returns `status = "missing_hausman_data"` for any
#' unit where the required statistics are absent from `local_results`.
#'
#' **Logistic interpretation limit**: Not applicable.
#'
#' @examples
#' \donttest{
#' library(sf)
#' pts <- sf::st_as_sf(
#'   data.frame(id = 1:4, X = c(0,1,0,1), Y = c(0,0,1,1)),
#'   coords = c("X", "Y"), crs = NA_integer_
#' )
#' dat <- data.frame(
#'   id   = rep(1:4, each = 5),
#'   time = rep(1:5, 4),
#'   y    = rnorm(20),
#'   x1   = rnorm(20)
#' )
#' fit <- fit_gwpr(y ~ x1, data = dat, spatial = pts, id = "id",
#'                 time = "time", bandwidth = 2, workers = 1)
#' diagnose_hausman(fit)
#' }
#'
#' @export
diagnose_hausman <- function(object, ...) {
  if (!inherits(object, "gwpr_fit")) {
    stop("`object` must be a `gwpr_fit` object.", call. = FALSE)
  }

  family <- object$family %||% "gaussian"
  model  <- object$model  %||% "within"

  if (identical(family, "binomial")) {
    return(list(
      local_hausman = data.frame(
        unit_id   = character(0), statistic = numeric(0),
        p_value   = numeric(0),   df        = integer(0),
        status    = character(0), stringsAsFactors = FALSE
      ),
      n_tested = 0L,
      n_failed = 0L,
      status   = "not_applicable",
      message  = "Hausman test is not applicable to logistic (binomial) models."
    ))
  }

  if (identical(model, "pooling")) {
    return(list(
      local_hausman = data.frame(
        unit_id   = character(0), statistic = numeric(0),
        p_value   = numeric(0),   df        = integer(0),
        status    = character(0), stringsAsFactors = FALSE
      ),
      n_tested = 0L,
      n_failed = 0L,
      status   = "not_applicable",
      message  = paste0(
        "Hausman test compares fixed vs. random effects. ",
        "It is not meaningful for model = \"pooling\"."
      )
    ))
  }

  local_results <- object$local_results
  if (length(local_results) == 0L) {
    return(list(
      local_hausman = data.frame(
        unit_id   = character(0), statistic = numeric(0),
        p_value   = numeric(0),   df        = integer(0),
        status    = character(0), stringsAsFactors = FALSE
      ),
      n_tested = 0L,
      n_failed = 0L,
      status   = "no_local_results"
    ))
  }

  unit_ids <- names(local_results)
  if (is.null(unit_ids)) unit_ids <- as.character(seq_along(local_results))

  rows <- vector("list", length(unit_ids))

  for (i in seq_along(unit_ids)) {
    uid <- unit_ids[[i]]
    res <- local_results[[i]]

    row <- list(unit_id = uid, statistic = NA_real_, p_value = NA_real_,
                df = NA_integer_, status = "ok")

    if (!is.null(res$hausman_statistic)) {
      row$statistic <- as.numeric(res$hausman_statistic)
      row$p_value   <- as.numeric(res$hausman_p_value %||% NA_real_)
      row$df        <- as.integer(res$hausman_df       %||% NA_integer_)
      row$status    <- res$status %||% "ok"
    } else {
      row$status <- "missing_hausman_data"
    }

    rows[[i]] <- row
  }

  out_df <- data.frame(
    unit_id   = vapply(rows, `[[`, character(1L), "unit_id"),
    statistic = vapply(rows, `[[`, numeric(1L),   "statistic"),
    p_value   = vapply(rows, `[[`, numeric(1L),   "p_value"),
    df        = vapply(rows, `[[`, integer(1L),   "df"),
    status    = vapply(rows, `[[`, character(1L), "status"),
    stringsAsFactors = FALSE
  )

  n_failed <- sum(out_df$status != "ok")

  list(local_hausman = out_df,
       n_tested      = nrow(out_df),
       n_failed      = n_failed,
       status        = "ok")
}

# ---------------------------------------------------------------------------
# diagnose_lm
# ---------------------------------------------------------------------------

#' Local Breusch-Pagan LM test diagnostic on a gwpr_fit object
#'
#' Performs a local Breusch-Pagan Lagrange Multiplier test for random effects
#' for each spatial unit using test statistics stored in `local_results`.
#'
#' @param object A `gwpr_fit` object.
#' @param ...  Currently ignored.
#'
#' @return A named list with elements:
#' \describe{
#'   \item{`local_lm`}{Data frame with columns `unit_id`, `statistic`,
#'     `p_value`, `df`, `status`.}
#'   \item{`n_tested`}{Number of units tested.}
#'   \item{`n_failed`}{Number of units where the test could not be computed.}
#'   \item{`status`}{Overall status.}
#' }
#'
#' @details
#' **Applicable models**: gaussian with `model = "pooling"` or
#' `model = "random"`.  For `within` models the test is not directly
#' applicable (it tests for random effects vs. OLS); the function returns
#' `status = "not_applicable"`.  For logistic models also not applicable.
#'
#' **Panel balance requirement**: No constraint at the unit level.
#'
#' **Failure conditions**: Returns `status = "missing_lm_data"` for units
#' missing the required statistics.
#'
#' **Logistic interpretation limit**: Not applicable.
#'
#' @examples
#' \donttest{
#' library(sf)
#' pts <- sf::st_as_sf(
#'   data.frame(id = 1:4, X = c(0,1,0,1), Y = c(0,0,1,1)),
#'   coords = c("X", "Y"), crs = NA_integer_
#' )
#' dat <- data.frame(
#'   id   = rep(1:4, each = 5),
#'   time = rep(1:5, 4),
#'   y    = rnorm(20),
#'   x1   = rnorm(20)
#' )
#' fit <- fit_gwpr(y ~ x1, data = dat, spatial = pts, id = "id",
#'                 time = "time", bandwidth = 2, workers = 1)
#' diagnose_lm(fit)
#' }
#'
#' @export
diagnose_lm <- function(object, ...) {
  if (!inherits(object, "gwpr_fit")) {
    stop("`object` must be a `gwpr_fit` object.", call. = FALSE)
  }

  family <- object$family %||% "gaussian"
  model  <- object$model  %||% "pooling"

  if (identical(family, "binomial")) {
    return(list(
      local_lm = data.frame(
        unit_id   = character(0), statistic = numeric(0),
        p_value   = numeric(0),   df        = integer(0),
        status    = character(0), stringsAsFactors = FALSE
      ),
      n_tested = 0L,
      n_failed = 0L,
      status   = "not_applicable",
      message  = "LM test is not applicable to logistic (binomial) models."
    ))
  }

  if (identical(model, "within")) {
    return(list(
      local_lm = data.frame(
        unit_id   = character(0), statistic = numeric(0),
        p_value   = numeric(0),   df        = integer(0),
        status    = character(0), stringsAsFactors = FALSE
      ),
      n_tested = 0L,
      n_failed = 0L,
      status   = "not_applicable",
      message  = paste0(
        "The Breusch-Pagan LM test is most appropriate for pooling or random ",
        "effect models. It is not applicable to within (fixed-effects) models."
      )
    ))
  }

  local_results <- object$local_results
  if (length(local_results) == 0L) {
    return(list(
      local_lm = data.frame(
        unit_id   = character(0), statistic = numeric(0),
        p_value   = numeric(0),   df        = integer(0),
        status    = character(0), stringsAsFactors = FALSE
      ),
      n_tested = 0L,
      n_failed = 0L,
      status   = "no_local_results"
    ))
  }

  unit_ids <- names(local_results)
  if (is.null(unit_ids)) unit_ids <- as.character(seq_along(local_results))

  rows <- vector("list", length(unit_ids))

  for (i in seq_along(unit_ids)) {
    uid <- unit_ids[[i]]
    res <- local_results[[i]]

    row <- list(unit_id = uid, statistic = NA_real_, p_value = NA_real_,
                df = NA_integer_, status = "ok")

    if (!is.null(res$lm_statistic)) {
      row$statistic <- as.numeric(res$lm_statistic)
      row$p_value   <- as.numeric(res$lm_p_value %||% NA_real_)
      row$df        <- as.integer(res$lm_df       %||% NA_integer_)
      row$status    <- res$status %||% "ok"
    } else {
      row$status <- "missing_lm_data"
    }

    rows[[i]] <- row
  }

  out_df <- data.frame(
    unit_id   = vapply(rows, `[[`, character(1L), "unit_id"),
    statistic = vapply(rows, `[[`, numeric(1L),   "statistic"),
    p_value   = vapply(rows, `[[`, numeric(1L),   "p_value"),
    df        = vapply(rows, `[[`, integer(1L),   "df"),
    status    = vapply(rows, `[[`, character(1L), "status"),
    stringsAsFactors = FALSE
  )

  n_failed <- sum(out_df$status != "ok")

  list(local_lm = out_df,
       n_tested = nrow(out_df),
       n_failed = n_failed,
       status   = "ok")
}

# ---------------------------------------------------------------------------
# diagnose_gwpr  (top-level interface)
# ---------------------------------------------------------------------------

#' Run diagnostic tests on a fitted GWPR model
#'
#' Top-level interface that dispatches to individual diagnostic sub-functions.
#' Returns a `gwpr_diagnostics` object containing all requested test results.
#'
#' @param object     A `gwpr_fit` object returned by `fit_gwpr()` or similar.
#' @param diagnostics Character vector naming the tests to run.  Any subset of
#'   `c("moran", "f_test", "hausman", "lm_test")`.  Default is all four.
#' @param spatial_weights Required when `"moran"` is in `diagnostics`.
#'   A row-standardised n x n spatial weights matrix.
#' @param panel_index Required when `"moran"` is in `diagnostics`.
#'   A data.frame with columns `id` and `time` identifying each element of
#'   `object$residuals`.
#' @param ...  Additional arguments passed to individual diagnostic functions.
#'
#' @return A `gwpr_diagnostics` object (list with class `"gwpr_diagnostics"`)
#'   whose `diagnostics` slot contains the result of each requested test.
#'
#' @details
#' Tests that are not applicable to the fitted model (e.g., Hausman test on a
#' pooling model) return a list with `status = "not_applicable"` and an
#' explanatory `message`, rather than an error.
#'
#' @examples
#' \donttest{
#' library(sf)
#' pts <- sf::st_as_sf(
#'   data.frame(id = 1:4, X = c(0,1,0,1), Y = c(0,0,1,1)),
#'   coords = c("X", "Y"), crs = NA_integer_
#' )
#' dat <- data.frame(
#'   id   = rep(1:4, each = 5),
#'   time = rep(1:5, 4),
#'   y    = rnorm(20),
#'   x1   = rnorm(20)
#' )
#' fit <- fit_gwpr(y ~ x1, data = dat, spatial = pts, id = "id",
#'                 time = "time", bandwidth = 2, workers = 1)
#' diag_result <- diagnose_gwpr(fit, diagnostics = c("f_test", "hausman"))
#' print(diag_result)
#' }
#'
#' @export
diagnose_gwpr <- function(
    object,
    diagnostics    = c("moran", "f_test", "hausman", "lm_test"),
    spatial_weights = NULL,
    panel_index    = NULL,
    ...
) {
  if (!inherits(object, "gwpr_fit")) {
    stop("`object` must be a `gwpr_fit` object.", call. = FALSE)
  }

  diagnostics <- match.arg(
    diagnostics,
    choices  = c("moran", "f_test", "hausman", "lm_test"),
    several.ok = TRUE
  )

  results  <- list()
  warnings_acc <- character()

  # ---- Moran's I -----------------------------------------------------------
  if ("moran" %in% diagnostics) {
    if (is.null(spatial_weights) || is.null(panel_index)) {
      results[["moran"]] <- list(
        status  = "skipped",
        message = paste0(
          "Moran's I requires `spatial_weights` and `panel_index`. ",
          "Provide both arguments to run this test."
        )
      )
    } else {
      results[["moran"]] <- tryCatch(
        withCallingHandlers(
          diagnose_moran(object, spatial_weights, panel_index, ...),
          warning = function(w) {
            warnings_acc <<- c(warnings_acc, conditionMessage(w))
            invokeRestart("muffleWarning")
          }
        ),
        error = function(e) {
          list(status = "error", message = conditionMessage(e))
        }
      )
    }
  }

  # ---- Local F test --------------------------------------------------------
  if ("f_test" %in% diagnostics) {
    results[["f_test"]] <- tryCatch(
      diagnose_local_f(object, ...),
      error = function(e) list(status = "error", message = conditionMessage(e))
    )
  }

  # ---- Hausman test --------------------------------------------------------
  if ("hausman" %in% diagnostics) {
    results[["hausman"]] <- tryCatch(
      diagnose_hausman(object, ...),
      error = function(e) list(status = "error", message = conditionMessage(e))
    )
  }

  # ---- LM test -------------------------------------------------------------
  if ("lm_test" %in% diagnostics) {
    results[["lm_test"]] <- tryCatch(
      diagnose_lm(object, ...),
      error = function(e) list(status = "error", message = conditionMessage(e))
    )
  }

  # ---- Panel balance check ------------------------------------------------
  panel_balance <- TRUE   # default; update when index available
  if (!is.null(panel_index)) {
    idx <- if (is.data.frame(panel_index)) panel_index else as.data.frame(panel_index)
    if (ncol(idx) >= 2L) {
      colnames(idx)[1:2] <- c("pid", "tid")
      counts    <- table(idx$pid)
      ref_count <- counts[[1L]]
      panel_balance <- all(counts == ref_count)
    }
  }

  new_gwpr_diagnostics(
    diagnostics   = results,
    model_type    = object$family  %||% "gaussian",
    panel_balance = panel_balance,
    warnings      = warnings_acc,
    metadata      = list(diagnostics_run = diagnostics)
  )
}

# tests/testthat/test-diagnostics.R
#
# Tests for the diagnostics module (R/diagnostics.R).
# All tests use small simulated data; workers = 1 (no parallelism).

# ---------------------------------------------------------------------------
# Helper: build a small balanced-panel matrix and weights
# ---------------------------------------------------------------------------

make_balanced_moran_inputs <- function(n = 5L, Ti = 3L, seed = 42L) {
  set.seed(seed)
  resid_mat <- matrix(rnorm(n * Ti), nrow = n, ncol = Ti)

  # Simple row-standardised weights (uniform neighbours, no self-loops)
  W <- matrix(1 / (n - 1), nrow = n, ncol = n)
  diag(W) <- 0

  list(resid_mat = resid_mat, W = W, n = n, Ti = Ti)
}

make_panel_index <- function(n, Ti) {
  data.frame(
    id   = rep(seq_len(n), each = Ti),
    time = rep(seq_len(Ti), times = n)
  )
}

# ---------------------------------------------------------------------------
# compute_logistic_pearson_residual
# ---------------------------------------------------------------------------

test_that("pearson residual: standard case produces finite values", {
  y    <- c(1L, 0L, 1L, 0L)
  prob <- c(0.9, 0.1, 0.8, 0.2)
  r    <- compute_logistic_pearson_residual(y, prob)
  expect_length(r, 4L)
  expect_true(all(is.finite(r)))
})

test_that("pearson residual: prob = 0 or 1 does not produce Inf", {
  y    <- c(1L, 0L)
  prob <- c(1.0, 0.0)   # boundary values — would divide by zero without clipping
  r    <- compute_logistic_pearson_residual(y, prob)
  expect_true(all(is.finite(r)),
              info = "Clipping must prevent Inf residuals at prob = 0/1.")
})

test_that("pearson residual: formula correctness", {
  y    <- 1L
  prob <- 0.75
  eps  <- 1e-15
  p    <- max(min(prob, 1 - eps), eps)
  expected <- (y - p) / sqrt(p * (1 - p))
  r <- compute_logistic_pearson_residual(y, prob, eps = eps)
  expect_equal(r, expected, tolerance = 1e-12)
})

test_that("pearson residual: length mismatch is an error", {
  expect_error(
    compute_logistic_pearson_residual(c(1, 0), c(0.9)),
    regexp = "same length"
  )
})

test_that("pearson residual: non-numeric y is an error", {
  expect_error(
    compute_logistic_pearson_residual("a", 0.5),
    regexp = "numeric or integer"
  )
})

test_that("pearson residual: negative eps is an error", {
  expect_error(
    compute_logistic_pearson_residual(1L, 0.5, eps = -1),
    regexp = "positive"
  )
})

# ---------------------------------------------------------------------------
# compute_panel_moran — balanced panel (matrix input)
# ---------------------------------------------------------------------------

test_that("moran balanced panel: returns list with correct names", {
  inp <- make_balanced_moran_inputs()
  res <- compute_panel_moran(inp$resid_mat, inp$W)

  expected_names <- c("statistic", "p_value", "estimated_I", "expected_I",
                      "variance", "alternative", "n_individuals", "n_periods")
  expect_true(all(expected_names %in% names(res)))
})

test_that("moran balanced panel: statistic and p_value are finite", {
  inp <- make_balanced_moran_inputs()
  res <- compute_panel_moran(inp$resid_mat, inp$W)

  expect_true(is.finite(res$statistic),
              info = "Z-statistic should be finite for well-specified input.")
  expect_true(is.finite(res$p_value))
})

test_that("moran balanced panel: p_value in [0, 1]", {
  inp <- make_balanced_moran_inputs()
  res <- compute_panel_moran(inp$resid_mat, inp$W)
  expect_gte(res$p_value, 0)
  expect_lte(res$p_value, 1)
})

test_that("moran balanced panel: n_individuals and n_periods are correct", {
  inp <- make_balanced_moran_inputs(n = 6L, Ti = 4L)
  res <- compute_panel_moran(inp$resid_mat, inp$W)
  expect_equal(res$n_individuals, 6L)
  expect_equal(res$n_periods, 4L)
})

test_that("moran balanced panel: expected_I equals -1/(n-1)", {
  n   <- 5L
  inp <- make_balanced_moran_inputs(n = n)
  res <- compute_panel_moran(inp$resid_mat, inp$W)
  expect_equal(res$expected_I, -1 / (n - 1))
})

test_that("moran balanced panel: alternative is 'greater'", {
  inp <- make_balanced_moran_inputs()
  res <- compute_panel_moran(inp$resid_mat, inp$W)
  expect_equal(res$alternative, "greater")
})

# ---------------------------------------------------------------------------
# compute_panel_moran — vector input with panel_index
# ---------------------------------------------------------------------------

test_that("moran: vector input + panel_index balanced panel runs without error", {
  n  <- 4L
  Ti <- 3L
  set.seed(7L)
  resid_vec <- rnorm(n * Ti)
  idx       <- make_panel_index(n, Ti)
  W         <- matrix(1 / (n - 1), n, n); diag(W) <- 0

  res <- compute_panel_moran(resid_vec, W, panel_index = idx)
  expect_true(is.list(res))
  expect_true(is.finite(res$statistic))
})

test_that("moran: vector input requires panel_index", {
  W   <- matrix(1 / 4, 5, 5); diag(W) <- 0
  expect_error(
    compute_panel_moran(rnorm(15), W),
    regexp = "panel_index"
  )
})

# ---------------------------------------------------------------------------
# compute_panel_moran — unbalanced panel warning
# ---------------------------------------------------------------------------

test_that("moran: unbalanced panel gives a warning", {
  n  <- 4L
  Ti <- 3L
  set.seed(11L)
  resid_vec <- rnorm(n * Ti - 1L)   # one observation missing => unbalanced
  idx <- make_panel_index(n, Ti)
  idx <- idx[-nrow(idx), ]          # drop last row to make it unbalanced
  W   <- matrix(1 / (n - 1), n, n); diag(W) <- 0

  expect_warning(
    compute_panel_moran(resid_vec, W, panel_index = idx),
    regexp = "unbalanced"
  )
})

# ---------------------------------------------------------------------------
# compute_panel_moran — dimension mismatch error
# ---------------------------------------------------------------------------

test_that("moran: spatial_weights dimension mismatch is an error", {
  n   <- 5L
  Ti  <- 3L
  inp <- make_balanced_moran_inputs(n = n, Ti = Ti)

  W_wrong <- matrix(1 / 3, nrow = 3L, ncol = 3L)   # n=3 != 5
  diag(W_wrong) <- 0

  expect_error(
    compute_panel_moran(inp$resid_mat, W_wrong),
    regexp = "n x n"
  )
})

# ---------------------------------------------------------------------------
# diagnose_moran
# ---------------------------------------------------------------------------

test_that("diagnose_moran: linear model runs and returns residual_type = 'raw'", {
  n  <- 4L
  Ti <- 3L
  set.seed(1L)
  resid_vec <- rnorm(n * Ti)
  idx       <- make_panel_index(n, Ti)
  W         <- matrix(1 / (n - 1), n, n); diag(W) <- 0

  fit <- new_gwpr_fit(
    family    = "gaussian",
    model     = "within",
    effect    = "individual",
    bandwidth = 1,
    residuals = resid_vec
  )

  res <- diagnose_moran(fit, W, idx)
  expect_equal(res$residual_type, "raw")
  expect_true(is.finite(res$statistic))
})

test_that("diagnose_moran: logistic model uses Pearson residuals", {
  n  <- 4L
  Ti <- 3L
  N  <- n * Ti
  set.seed(2L)
  y    <- sample(0:1, N, replace = TRUE)
  prob <- runif(N, 0.1, 0.9)
  idx  <- make_panel_index(n, Ti)
  W    <- matrix(1 / (n - 1), n, n); diag(W) <- 0

  fit <- new_gwpr_fit(
    family    = "binomial",
    model     = "pooling",
    bandwidth = 1,
    residuals = y - prob,           # not used directly — overridden by Pearson
    metadata  = list(response = y, prob = prob)
  )

  res <- diagnose_moran(fit, W, idx)
  expect_equal(res$residual_type, "pearson")
  expect_true(is.finite(res$statistic))
})

test_that("diagnose_moran: non-gwpr_fit object is an error", {
  expect_error(
    diagnose_moran(list(), matrix(1, 2, 2), data.frame(id = 1, time = 1)),
    regexp = "gwpr_fit"
  )
})

test_that("diagnose_moran: NULL residuals is an error", {
  fit <- new_gwpr_fit(family = "gaussian", residuals = NULL)
  expect_error(
    diagnose_moran(fit, matrix(1, 2, 2), data.frame(id = 1, time = 1)),
    regexp = "residuals.*NULL"
  )
})

# ---------------------------------------------------------------------------
# diagnose_local_f
# ---------------------------------------------------------------------------

test_that("diagnose_local_f: returns data.frame with correct columns", {
  local_res <- list(
    "1" = list(within_rss = 2.0, within_df = 5L, pooling_rss = 5.0, pooling_df = 8L, status = "ok"),
    "2" = list(within_rss = 1.5, within_df = 5L, pooling_rss = 4.0, pooling_df = 8L, status = "ok")
  )
  fit <- new_gwpr_fit(family = "gaussian", model = "within", local_results = local_res)
  res <- diagnose_local_f(fit)

  expect_true(is.data.frame(res$local_f))
  expect_true(all(c("unit_id", "statistic", "p_value", "df1", "df2", "status") %in%
                    names(res$local_f)))
  expect_equal(nrow(res$local_f), 2L)
})

test_that("diagnose_local_f: computed F-statistic is positive", {
  local_res <- list(
    "A" = list(within_rss = 2.0, within_df = 5L, pooling_rss = 5.0, pooling_df = 8L, status = "ok")
  )
  fit <- new_gwpr_fit(family = "gaussian", model = "within", local_results = local_res)
  res <- diagnose_local_f(fit)

  expect_gt(res$local_f$statistic[1L], 0)
})

test_that("diagnose_local_f: p-value in [0, 1]", {
  local_res <- list(
    "A" = list(within_rss = 2.0, within_df = 5L, pooling_rss = 5.0, pooling_df = 8L, status = "ok")
  )
  fit <- new_gwpr_fit(family = "gaussian", model = "within", local_results = local_res)
  res <- diagnose_local_f(fit)

  expect_gte(res$local_f$p_value[1L], 0)
  expect_lte(res$local_f$p_value[1L], 1)
})

test_that("diagnose_local_f: binomial model returns not_applicable", {
  fit <- new_gwpr_fit(family = "binomial", model = "pooling")
  res <- diagnose_local_f(fit)

  expect_equal(res$status, "not_applicable")
  expect_equal(nrow(res$local_f), 0L)
})

test_that("diagnose_local_f: pre-computed f_statistic used directly", {
  local_res <- list(
    "X" = list(f_statistic = 4.5, f_p_value = 0.02, f_df1 = 2L, f_df2 = 10L, status = "ok")
  )
  fit <- new_gwpr_fit(family = "gaussian", model = "within", local_results = local_res)
  res <- diagnose_local_f(fit)

  expect_equal(res$local_f$statistic[1L], 4.5)
  expect_equal(res$local_f$p_value[1L], 0.02)
})

test_that("diagnose_local_f: missing data unit gets correct status", {
  local_res <- list(
    "Z" = list(status = "ok")   # no within_rss / pooling_rss / f_statistic
  )
  fit <- new_gwpr_fit(family = "gaussian", model = "within", local_results = local_res)
  res <- diagnose_local_f(fit)

  expect_equal(res$local_f$status[1L], "missing_local_f_data")
  expect_equal(res$n_failed, 1L)
})

test_that("diagnose_local_f: empty local_results returns no_local_results", {
  fit <- new_gwpr_fit(family = "gaussian", model = "within", local_results = list())
  res <- diagnose_local_f(fit)
  expect_equal(res$status, "no_local_results")
})

# ---------------------------------------------------------------------------
# diagnose_hausman
# ---------------------------------------------------------------------------

test_that("diagnose_hausman: returns data.frame with correct columns", {
  local_res <- list(
    "1" = list(hausman_statistic = 3.2, hausman_p_value = 0.07, hausman_df = 2L, status = "ok"),
    "2" = list(hausman_statistic = 6.1, hausman_p_value = 0.01, hausman_df = 2L, status = "ok")
  )
  fit <- new_gwpr_fit(family = "gaussian", model = "random", local_results = local_res)
  res <- diagnose_hausman(fit)

  expect_true(is.data.frame(res$local_hausman))
  expect_true(all(c("unit_id", "statistic", "p_value", "df", "status") %in%
                    names(res$local_hausman)))
  expect_equal(nrow(res$local_hausman), 2L)
})

test_that("diagnose_hausman: pooling model returns not_applicable", {
  fit <- new_gwpr_fit(family = "gaussian", model = "pooling")
  res <- diagnose_hausman(fit)

  expect_equal(res$status, "not_applicable")
  expect_match(res$message, "pooling")
  expect_equal(nrow(res$local_hausman), 0L)
})

test_that("diagnose_hausman: binomial model returns not_applicable", {
  fit <- new_gwpr_fit(family = "binomial", model = "random")
  res <- diagnose_hausman(fit)

  expect_equal(res$status, "not_applicable")
  expect_equal(nrow(res$local_hausman), 0L)
})

test_that("diagnose_hausman: missing data unit is counted as failed", {
  local_res <- list(
    "A" = list(status = "ok")   # no hausman_statistic
  )
  fit <- new_gwpr_fit(family = "gaussian", model = "random", local_results = local_res)
  res <- diagnose_hausman(fit)

  expect_equal(res$local_hausman$status[1L], "missing_hausman_data")
  expect_equal(res$n_failed, 1L)
})

test_that("diagnose_hausman: within model runs (applicable)", {
  local_res <- list(
    "1" = list(hausman_statistic = 2.5, hausman_p_value = 0.11, hausman_df = 2L, status = "ok")
  )
  fit <- new_gwpr_fit(family = "gaussian", model = "within", local_results = local_res)
  res <- diagnose_hausman(fit)

  expect_equal(res$status, "ok")
  expect_equal(res$n_tested, 1L)
})

# ---------------------------------------------------------------------------
# diagnose_lm
# ---------------------------------------------------------------------------

test_that("diagnose_lm: returns data.frame with correct columns", {
  local_res <- list(
    "1" = list(lm_statistic = 4.1, lm_p_value = 0.04, lm_df = 1L, status = "ok"),
    "2" = list(lm_statistic = 1.2, lm_p_value = 0.27, lm_df = 1L, status = "ok")
  )
  fit <- new_gwpr_fit(family = "gaussian", model = "pooling", local_results = local_res)
  res <- diagnose_lm(fit)

  expect_true(is.data.frame(res$local_lm))
  expect_true(all(c("unit_id", "statistic", "p_value", "df", "status") %in%
                    names(res$local_lm)))
  expect_equal(nrow(res$local_lm), 2L)
})

test_that("diagnose_lm: within model returns not_applicable", {
  fit <- new_gwpr_fit(family = "gaussian", model = "within")
  res <- diagnose_lm(fit)

  expect_equal(res$status, "not_applicable")
  expect_match(res$message, "within")
  expect_equal(nrow(res$local_lm), 0L)
})

test_that("diagnose_lm: binomial model returns not_applicable", {
  fit <- new_gwpr_fit(family = "binomial", model = "pooling")
  res <- diagnose_lm(fit)

  expect_equal(res$status, "not_applicable")
  expect_equal(nrow(res$local_lm), 0L)
})

test_that("diagnose_lm: missing data unit is counted as failed", {
  local_res <- list(
    "A" = list(status = "ok")   # no lm_statistic
  )
  fit <- new_gwpr_fit(family = "gaussian", model = "pooling", local_results = local_res)
  res <- diagnose_lm(fit)

  expect_equal(res$local_lm$status[1L], "missing_lm_data")
  expect_equal(res$n_failed, 1L)
})

test_that("diagnose_lm: random model is applicable", {
  local_res <- list(
    "X" = list(lm_statistic = 3.3, lm_p_value = 0.07, lm_df = 1L, status = "ok")
  )
  fit <- new_gwpr_fit(family = "gaussian", model = "random", local_results = local_res)
  res <- diagnose_lm(fit)
  expect_equal(res$status, "ok")
})

# ---------------------------------------------------------------------------
# diagnose_gwpr — top-level interface
# ---------------------------------------------------------------------------

test_that("diagnose_gwpr: returns gwpr_diagnostics object", {
  local_res <- list(
    "1" = list(within_rss = 2.0, within_df = 5L, pooling_rss = 5.0, pooling_df = 8L, status = "ok")
  )
  fit <- new_gwpr_fit(family = "gaussian", model = "within", local_results = local_res)
  dg  <- diagnose_gwpr(fit, diagnostics = c("f_test", "hausman", "lm_test"))

  expect_s3_class(dg, "gwpr_diagnostics")
})

test_that("diagnose_gwpr: result object has correct structure", {
  fit <- new_gwpr_fit(family = "gaussian", model = "pooling")
  dg  <- diagnose_gwpr(fit, diagnostics = c("f_test", "hausman", "lm_test"))

  expect_true(is.list(dg$diagnostics))
  expect_true(!is.null(dg$model_type))
  expect_true(is.character(dg$warnings))
})

test_that("diagnose_gwpr: moran skipped without spatial_weights/panel_index", {
  fit <- new_gwpr_fit(family = "gaussian", model = "within", residuals = rnorm(12))
  dg  <- diagnose_gwpr(fit, diagnostics = "moran")

  expect_equal(dg$diagnostics$moran$status, "skipped")
})

test_that("diagnose_gwpr: moran runs when spatial_weights and panel_index provided", {
  n  <- 4L
  Ti <- 3L
  set.seed(5L)
  resid_vec <- rnorm(n * Ti)
  idx       <- make_panel_index(n, Ti)
  W         <- matrix(1 / (n - 1), n, n); diag(W) <- 0

  fit <- new_gwpr_fit(
    family    = "gaussian",
    model     = "within",
    bandwidth = 1,
    residuals = resid_vec
  )

  dg <- diagnose_gwpr(fit,
                      diagnostics     = "moran",
                      spatial_weights = W,
                      panel_index     = idx)

  moran_res <- dg$diagnostics$moran
  expect_false(identical(moran_res$status, "skipped"),
               info = "Moran should run, not be skipped, when inputs are provided.")
  expect_true(is.finite(moran_res$statistic))
})

test_that("diagnose_gwpr: non-applicable tests return interpretable status", {
  fit <- new_gwpr_fit(family = "gaussian", model = "pooling")
  dg  <- diagnose_gwpr(fit, diagnostics = c("f_test", "hausman", "lm_test"))

  # f_test: no local_results -> no_local_results
  expect_true(!is.null(dg$diagnostics$f_test))
  # hausman: pooling -> not_applicable
  expect_equal(dg$diagnostics$hausman$status, "not_applicable")
  # lm_test: pooling and no local results -> no_local_results
  expect_true(!is.null(dg$diagnostics$lm_test))
})

test_that("diagnose_gwpr: non-gwpr_fit input gives error", {
  expect_error(
    diagnose_gwpr(list()),
    regexp = "gwpr_fit"
  )
})

test_that("diagnose_gwpr: panel_balance is set correctly for balanced panel", {
  n  <- 3L
  Ti <- 2L
  resid_vec <- rnorm(n * Ti)
  idx       <- make_panel_index(n, Ti)
  W         <- matrix(1 / (n - 1), n, n); diag(W) <- 0

  fit <- new_gwpr_fit(
    family    = "gaussian",
    model     = "within",
    bandwidth = 1,
    residuals = resid_vec
  )

  dg <- diagnose_gwpr(fit,
                      diagnostics     = "moran",
                      spatial_weights = W,
                      panel_index     = idx)

  expect_true(dg$panel_balance)
})

test_that("diagnose_gwpr: subset of diagnostics only runs requested tests", {
  local_res <- list(
    "1" = list(lm_statistic = 2.0, lm_p_value = 0.15, lm_df = 1L, status = "ok")
  )
  fit <- new_gwpr_fit(family = "gaussian", model = "pooling", local_results = local_res)
  dg  <- diagnose_gwpr(fit, diagnostics = "lm_test")

  expect_true("lm_test" %in% names(dg$diagnostics))
  expect_false("f_test" %in% names(dg$diagnostics))
  expect_false("hausman" %in% names(dg$diagnostics))
  expect_false("moran" %in% names(dg$diagnostics))
})

#' @title Random Bandwidth Optimizer (`bandwidth_random.R`)
#'
#' @description
#' Implements a bounded random search for bandwidth selection.  A user-
#' specified number of candidate bandwidths are drawn uniformly at random from
#' `[lower, upper]`, scored by a user-supplied scorer function, and the
#' candidate with the lowest score is returned as the best bandwidth.
#'
#' The search boundaries (`lower`, `upper`) **must** be set explicitly by the
#' caller; automatic inference is intentionally not supported.
#'
#' @name bandwidth_random
#' @keywords internal
NULL

# ---------------------------------------------------------------------------
# sample_random_bandwidths
# ---------------------------------------------------------------------------

#' Sample random bandwidth candidates
#'
#' Draws `n_samples` values uniformly from `[lower, upper]`.  When
#' `adaptive = TRUE` the draws are rounded to the nearest positive integer and
#' duplicates are removed (the resulting vector may therefore be shorter than
#' `n_samples`; this is intentional and no error is raised).
#'
#' Both `lower` and `upper` **must** be supplied explicitly — they are never
#' inferred automatically.
#'
#' @param lower     Numeric scalar.  Lower bound of the search range
#'   (inclusive).
#' @param upper     Numeric scalar.  Upper bound of the search range
#'   (inclusive).
#' @param n_samples Positive integer.  Number of random draws before any
#'   deduplication.
#' @param adaptive  Logical scalar.  When `TRUE`, draws are rounded to the
#'   nearest positive integer and duplicates are removed.
#' @param seed      Integer random seed, or `NULL`.  When supplied, `set.seed()`
#'   is called before sampling so that the candidate set is reproducible.
#'
#' @return A numeric vector of candidate bandwidth values.  The vector has
#'   length `n_samples` when `adaptive = FALSE`, and length at most
#'   `n_samples` when `adaptive = TRUE`.
#'
#' @examples
#' sample_random_bandwidths(lower = 1, upper = 10, n_samples = 5,
#'                          adaptive = FALSE, seed = 42L)
#' sample_random_bandwidths(lower = 2, upper = 10, n_samples = 20,
#'                          adaptive = TRUE, seed = 1L)
#'
#' @noRd
sample_random_bandwidths <- function(lower, upper, n_samples,
                                     adaptive = FALSE, seed = NULL) {

  if (missing(lower) || is.null(lower)) {
    stop(
      "Random search requires `lower` to be set explicitly. ",
      "Automatic boundary inference is not supported.",
      call. = FALSE
    )
  }
  if (missing(upper) || is.null(upper)) {
    stop(
      "Random search requires `upper` to be set explicitly. ",
      "Automatic boundary inference is not supported.",
      call. = FALSE
    )
  }
  if (!is.numeric(lower) || length(lower) != 1L || !is.finite(lower)) {
    stop("`lower` must be a single finite numeric value.", call. = FALSE)
  }
  if (!is.numeric(upper) || length(upper) != 1L || !is.finite(upper)) {
    stop("`upper` must be a single finite numeric value.", call. = FALSE)
  }
  if (lower >= upper) {
    stop("`lower` must be strictly less than `upper`.", call. = FALSE)
  }
  if (!is.numeric(n_samples) || length(n_samples) != 1L ||
      !is.finite(n_samples) || n_samples < 1L) {
    stop("`n_samples` must be a single positive integer.", call. = FALSE)
  }
  n_samples <- as.integer(n_samples)

  if (!is.null(seed)) set.seed(seed)

  candidates <- stats::runif(n_samples, min = lower, max = upper)

  if (adaptive) {
    candidates <- unique(pmax(1L, round(candidates)))
    candidates <- as.integer(candidates)
  }

  candidates
}

# ---------------------------------------------------------------------------
# search_bandwidth_random  (main entry point)
# ---------------------------------------------------------------------------

#' Random bandwidth search
#'
#' Samples `control$n_samples` candidate bandwidths uniformly at random from
#' `[control$lower, control$upper]`, scores each with the user-supplied
#' `scorer` function, and returns the candidate with the lowest score.  The
#' complete evaluation history is included in the returned `gwpr_bandwidth`
#' object.
#'
#' The `method` field is fixed to `"random"`.  The history structure is
#' identical to that produced by `search_bandwidth_grid()`.
#'
#' @param context  A `gwpr_context` list.  Must contain at minimum `adaptive`.
#' @param control  A list with random search parameters.  **Required** fields:
#'   \describe{
#'     \item{`lower`}{Numeric.  Lower bound of the search range (inclusive).}
#'     \item{`upper`}{Numeric.  Upper bound of the search range (inclusive).}
#'   }
#'   Optional fields:
#'   \describe{
#'     \item{`n_samples`}{Positive integer.  Number of random candidates to
#'       draw.  Defaults to `50L`.}
#'   }
#' @param scorer   A function with signature `scorer(context, bandwidth)`
#'   that returns a named list as described in `score_bandwidth_candidate()`.
#' @param workers  Integer.  Number of parallel workers.  `1` (default) uses
#'   serial `lapply`; `> 1` uses `parallel_map()`.
#' @param seed     Integer random seed, or `NULL`.  Applied before sampling and
#'   before parallel execution.  Stored in the returned object.
#'
#' @return A `gwpr_bandwidth` object (see `new_gwpr_bandwidth()`).
#'
#' @examples
#' \donttest{
#' result <- search_bandwidth_random(
#'   context = ctx,
#'   control = list(lower = 1, upper = 10, n_samples = 20L),
#'   scorer  = my_scorer_fn,
#'   workers = 1L,
#'   seed    = 42L
#' )
#' result$best_bandwidth
#' }
#'
#' @noRd
search_bandwidth_random <- function(context, control, scorer,
                                    workers = 1L, seed = NULL) {

  t_total_start <- proc.time()[["elapsed"]]

  adaptive  <- context$adaptive %||% FALSE
  n_samples <- control$n_samples %||% 50L

  # Sample candidates — will stop with informative errors if bounds missing
  candidates <- sample_random_bandwidths(
    lower    = control$lower,
    upper    = control$upper,
    n_samples = n_samples,
    adaptive  = adaptive,
    seed      = seed
  )

  n_candidates <- length(candidates)

  # Evaluate candidates
  if (workers <= 1L) {
    results_list <- lapply(
      candidates,
      function(bw) score_bandwidth_candidate(context, bw, scorer)
    )
  } else {
    results_list <- parallel_map(
      x       = as.list(candidates),
      fn      = function(bw) score_bandwidth_candidate(context, bw, scorer),
      workers = workers,
      seed    = seed
    )
  }

  # Determine bandwidth_type and criterion from first successful result
  bandwidth_type <- if (adaptive) "adaptive" else "fixed"
  first_ok <- Find(function(r) r$status == "ok", results_list)
  criterion <- if (!is.null(first_ok)) {
    first_ok$criterion %||% NA_character_
  } else {
    NA_character_
  }

  # Assemble history data.frame (same structure as grid search)
  history <- data.frame(
    candidate_id          = seq_len(n_candidates),
    bandwidth             = vapply(results_list, `[[`, numeric(1L), "bandwidth"),
    bandwidth_type        = bandwidth_type,
    method                = "random",
    criterion             = vapply(results_list, function(r) {
      r$criterion %||% NA_character_
    }, character(1L)),
    score                 = vapply(results_list, function(r) {
      r$score %||% NA_real_
    }, numeric(1L)),
    rank                  = NA_integer_,
    status                = vapply(results_list, `[[`, character(1L), "status"),
    error_message         = vapply(results_list, function(r) {
      r$error_message %||% NA_character_
    }, character(1L)),
    warning_message       = vapply(results_list, function(r) {
      r$warning_message %||% NA_character_
    }, character(1L)),
    elapsed_time          = vapply(results_list, `[[`, numeric(1L), "elapsed_time"),
    n_local_models        = vapply(results_list, function(r) {
      v <- r$n_local_models
      if (is.null(v) || is.na(v)) NA_integer_ else as.integer(v)
    }, integer(1L)),
    n_failed_local_models = vapply(results_list, function(r) {
      v <- r$n_failed_local_models
      if (is.null(v) || is.na(v)) NA_integer_ else as.integer(v)
    }, integer(1L)),
    r2                    = vapply(results_list, function(r) {
      r$r2        %||% NA_real_
    }, numeric(1L)),
    mse                   = vapply(results_list, function(r) {
      r$mse       %||% NA_real_
    }, numeric(1L)),
    rmse                  = vapply(results_list, function(r) {
      r$rmse      %||% NA_real_
    }, numeric(1L)),
    mae                   = vapply(results_list, function(r) {
      r$mae       %||% NA_real_
    }, numeric(1L)),
    log_loss              = vapply(results_list, function(r) {
      r$log_loss  %||% NA_real_
    }, numeric(1L)),
    accuracy              = vapply(results_list, function(r) {
      r$accuracy  %||% NA_real_
    }, numeric(1L)),
    precision             = vapply(results_list, function(r) {
      r$precision %||% NA_real_
    }, numeric(1L)),
    recall                = vapply(results_list, function(r) {
      r$recall    %||% NA_real_
    }, numeric(1L)),
    f1_score              = vapply(results_list, function(r) {
      r$f1_score  %||% NA_real_
    }, numeric(1L)),
    stringsAsFactors      = FALSE
  )

  # Rank candidates (reuse grid ranking logic)
  history <- rank_grid_history(history)

  # Identify best candidate (rank == 1 among ok)
  ok_rows <- history[history$status == "ok", , drop = FALSE]
  if (nrow(ok_rows) > 0L) {
    best_row       <- ok_rows[which.min(ok_rows$rank), , drop = FALSE]
    best_bandwidth <- best_row$bandwidth
    best_score     <- best_row$score
  } else {
    best_bandwidth <- NA_real_
    best_score     <- NA_real_
  }

  total_elapsed <- proc.time()[["elapsed"]] - t_total_start

  new_gwpr_bandwidth(
    method           = "random",
    best_bandwidth   = best_bandwidth,
    best_score       = best_score,
    criterion        = criterion,
    history          = history,
    metrics_history  = NULL,
    seed             = seed,
    convergence_info = list(n_candidates = n_candidates,
                            n_samples    = n_samples),
    elapsed_time     = total_elapsed,
    warnings         = character()
  )
}

# ---------------------------------------------------------------------------
# Internal helpers (local copies to avoid cross-file ordering issues)
# ---------------------------------------------------------------------------

#' @noRd
`%||%` <- function(x, y) if (!is.null(x)) x else y

## tests/testthat/test-bandwidth-random.R
## Tests for the Random Bandwidth Optimizer module (R/bandwidth_random.R)
##
## All tests use small simulated data and workers = 1 (no parallelism).

# ---------------------------------------------------------------------------
# Helper: minimal mock scorer
# ---------------------------------------------------------------------------
# Returns a valid scorer result.  Score is deterministic (distance from a
# best bandwidth) so we can predict which candidate wins.

make_random_mock_scorer <- function(best_bw = 5, fail_bw = NULL) {
  function(context, bandwidth) {
    if (!is.null(fail_bw) && isTRUE(all.equal(bandwidth, fail_bw))) {
      stop("Intentional scorer failure for testing.")
    }
    score <- abs(bandwidth - best_bw) + 0.01
    list(
      score                 = score,
      criterion             = "MSE",
      status                = "ok",
      n_local_models        = 4L,
      n_failed_local_models = 0L,
      metrics               = list(
        R2   = 0.8,
        MSE  = score,
        RMSE = sqrt(score),
        MAE  = score * 0.9
      )
    )
  }
}

# Minimal gwpr_context suitable for random search tests
make_random_context <- function(adaptive = FALSE) {
  new_gwpr_context(
    formula   = y ~ x1,
    family    = "gaussian",
    model     = "pooling",
    effect    = "individual",
    id        = "id",
    time      = "time",
    kernel    = "bisquare",
    adaptive  = adaptive,
    threshold = 0.5,
    workers   = 1L,
    seed      = NULL
  )
}

# ===========================================================================
# sample_random_bandwidths
# ===========================================================================

test_that("sample_random_bandwidths errors when lower is missing", {
  expect_error(
    sample_random_bandwidths(upper = 10, n_samples = 5, adaptive = FALSE),
    regexp = "lower"
  )
})

test_that("sample_random_bandwidths errors when upper is missing", {
  expect_error(
    sample_random_bandwidths(lower = 1, n_samples = 5, adaptive = FALSE),
    regexp = "upper"
  )
})

test_that("sample_random_bandwidths errors when lower is NULL", {
  expect_error(
    sample_random_bandwidths(lower = NULL, upper = 10, n_samples = 5,
                             adaptive = FALSE),
    regexp = "lower"
  )
})

test_that("sample_random_bandwidths errors when upper is NULL", {
  expect_error(
    sample_random_bandwidths(lower = 1, upper = NULL, n_samples = 5,
                             adaptive = FALSE),
    regexp = "upper"
  )
})

test_that("sample_random_bandwidths errors when lower >= upper", {
  expect_error(
    sample_random_bandwidths(lower = 10, upper = 5, n_samples = 5,
                             adaptive = FALSE),
    regexp = "lower"
  )
  expect_error(
    sample_random_bandwidths(lower = 5, upper = 5, n_samples = 5,
                             adaptive = FALSE),
    regexp = "lower"
  )
})

test_that("sample_random_bandwidths returns n_samples values for adaptive = FALSE", {
  cands <- sample_random_bandwidths(lower = 1, upper = 10, n_samples = 20L,
                                    adaptive = FALSE, seed = 99L)
  expect_length(cands, 20L)
})

test_that("sample_random_bandwidths returns values within [lower, upper]", {
  cands <- sample_random_bandwidths(lower = 2, upper = 8, n_samples = 50L,
                                    adaptive = FALSE, seed = 7L)
  expect_true(all(cands >= 2))
  expect_true(all(cands <= 8))
})

test_that("sample_random_bandwidths is reproducible with fixed seed", {
  cands1 <- sample_random_bandwidths(lower = 1, upper = 100, n_samples = 10L,
                                     adaptive = FALSE, seed = 42L)
  cands2 <- sample_random_bandwidths(lower = 1, upper = 100, n_samples = 10L,
                                     adaptive = FALSE, seed = 42L)
  expect_equal(cands1, cands2)
})

test_that("sample_random_bandwidths differs across seeds", {
  cands1 <- sample_random_bandwidths(lower = 1, upper = 100, n_samples = 10L,
                                     adaptive = FALSE, seed = 1L)
  cands2 <- sample_random_bandwidths(lower = 1, upper = 100, n_samples = 10L,
                                     adaptive = FALSE, seed = 2L)
  expect_false(identical(cands1, cands2))
})

test_that("sample_random_bandwidths adaptive = TRUE returns integers", {
  cands <- sample_random_bandwidths(lower = 2, upper = 20, n_samples = 30L,
                                    adaptive = TRUE, seed = 5L)
  expect_type(cands, "integer")
  expect_true(all(cands >= 1L))
})

test_that("sample_random_bandwidths adaptive = TRUE removes duplicates", {
  cands <- sample_random_bandwidths(lower = 1, upper = 3, n_samples = 100L,
                                    adaptive = TRUE, seed = 10L)
  expect_true(!any(duplicated(cands)))
  # With range 1-3 and rounding there can be at most 3 unique integers
  expect_true(length(cands) <= 3L)
})

test_that("sample_random_bandwidths adaptive = TRUE candidates are valid positive integers", {
  cands <- sample_random_bandwidths(lower = 1, upper = 10, n_samples = 15L,
                                    adaptive = TRUE, seed = 21L)
  expect_true(all(cands == as.integer(cands)))
  expect_true(all(cands >= 1L))
})

# ===========================================================================
# search_bandwidth_random
# ===========================================================================

test_that("search_bandwidth_random errors when lower is missing from control", {
  ctx    <- make_random_context()
  scorer <- make_random_mock_scorer()
  expect_error(
    search_bandwidth_random(ctx, control = list(upper = 10, n_samples = 5L),
                            scorer = scorer),
    regexp = "lower"
  )
})

test_that("search_bandwidth_random errors when upper is missing from control", {
  ctx    <- make_random_context()
  scorer <- make_random_mock_scorer()
  expect_error(
    search_bandwidth_random(ctx, control = list(lower = 1, n_samples = 5L),
                            scorer = scorer),
    regexp = "upper"
  )
})

test_that("search_bandwidth_random errors when lower is NULL in control", {
  ctx    <- make_random_context()
  scorer <- make_random_mock_scorer()
  expect_error(
    search_bandwidth_random(ctx,
                            control = list(lower = NULL, upper = 10,
                                           n_samples = 5L),
                            scorer = scorer),
    regexp = "lower"
  )
})

test_that("search_bandwidth_random returns a gwpr_bandwidth object", {
  ctx    <- make_random_context()
  scorer <- make_random_mock_scorer()
  result <- search_bandwidth_random(
    ctx,
    control = list(lower = 1, upper = 10, n_samples = 10L),
    scorer  = scorer,
    workers = 1L,
    seed    = 42L
  )
  expect_s3_class(result, "gwpr_bandwidth")
})

test_that("search_bandwidth_random method field is 'random'", {
  ctx    <- make_random_context()
  scorer <- make_random_mock_scorer()
  result <- search_bandwidth_random(
    ctx,
    control = list(lower = 1, upper = 10, n_samples = 8L),
    scorer  = scorer,
    workers = 1L,
    seed    = 1L
  )
  expect_equal(result$method, "random")
  expect_true(all(result$history$method == "random"))
})

test_that("search_bandwidth_random history has correct number of rows (fixed)", {
  ctx    <- make_random_context()
  scorer <- make_random_mock_scorer()
  result <- search_bandwidth_random(
    ctx,
    control = list(lower = 1, upper = 10, n_samples = 12L),
    scorer  = scorer,
    workers = 1L,
    seed    = 3L
  )
  expect_equal(nrow(result$history), 12L)
})

test_that("search_bandwidth_random best_bandwidth comes from candidate set", {
  ctx    <- make_random_context()
  scorer <- make_random_mock_scorer(best_bw = 5)
  result <- search_bandwidth_random(
    ctx,
    control = list(lower = 1, upper = 10, n_samples = 15L),
    scorer  = scorer,
    workers = 1L,
    seed    = 7L
  )
  # best_bandwidth must appear in the candidates evaluated
  expect_true(result$best_bandwidth %in% result$history$bandwidth)
})

test_that("search_bandwidth_random best_bandwidth has the minimum score", {
  ctx    <- make_random_context()
  scorer <- make_random_mock_scorer(best_bw = 5)
  result <- search_bandwidth_random(
    ctx,
    control = list(lower = 1, upper = 10, n_samples = 20L),
    scorer  = scorer,
    workers = 1L,
    seed    = 11L
  )
  ok_hist <- result$history[result$history$status == "ok", , drop = FALSE]
  expect_equal(result$best_score, min(ok_hist$score, na.rm = TRUE))
})

test_that("search_bandwidth_random is reproducible with fixed seed", {
  ctx    <- make_random_context()
  scorer <- make_random_mock_scorer()

  res1 <- search_bandwidth_random(
    ctx,
    control = list(lower = 1, upper = 20, n_samples = 10L),
    scorer  = scorer,
    workers = 1L,
    seed    = 99L
  )
  res2 <- search_bandwidth_random(
    ctx,
    control = list(lower = 1, upper = 20, n_samples = 10L),
    scorer  = scorer,
    workers = 1L,
    seed    = 99L
  )
  expect_equal(res1$history$bandwidth, res2$history$bandwidth)
  expect_equal(res1$best_bandwidth, res2$best_bandwidth)
})

test_that("search_bandwidth_random records seed in returned object", {
  ctx    <- make_random_context()
  scorer <- make_random_mock_scorer()
  result <- search_bandwidth_random(
    ctx,
    control = list(lower = 1, upper = 10, n_samples = 5L),
    scorer  = scorer,
    workers = 1L,
    seed    = 55L
  )
  expect_equal(result$seed, 55L)
})

test_that("search_bandwidth_random history includes all required columns", {
  ctx    <- make_random_context()
  scorer <- make_random_mock_scorer()
  result <- search_bandwidth_random(
    ctx,
    control = list(lower = 1, upper = 10, n_samples = 5L),
    scorer  = scorer,
    workers = 1L,
    seed    = 1L
  )
  required_cols <- c(
    "candidate_id", "bandwidth", "bandwidth_type", "method", "criterion",
    "score", "rank", "status", "error_message", "warning_message",
    "elapsed_time", "n_local_models", "n_failed_local_models",
    "r2", "mse", "rmse", "mae",
    "log_loss", "accuracy", "precision", "recall", "f1_score"
  )
  expect_true(all(required_cols %in% names(result$history)))
})

test_that("search_bandwidth_random retains failed candidates in history", {
  ctx         <- make_random_context()
  # We need to know what bandwidth will be drawn to create the failing scorer.
  # Use a seed to get a fixed candidate set first, then fail one of them.
  cands <- sample_random_bandwidths(lower = 1, upper = 10, n_samples = 6L,
                                    adaptive = FALSE, seed = 77L)
  fail_bw <- cands[1L]
  scorer  <- make_random_mock_scorer(best_bw = 5, fail_bw = fail_bw)

  result <- search_bandwidth_random(
    ctx,
    control = list(lower = 1, upper = 10, n_samples = 6L),
    scorer  = scorer,
    workers = 1L,
    seed    = 77L   # same seed => same candidate set
  )
  # All 6 candidates must be present
  expect_equal(nrow(result$history), 6L)
  # At least one failed
  expect_true(any(result$history$status == "failed"))
  # The failed bandwidth is recorded
  failed_rows <- result$history[result$history$status == "failed", , drop = FALSE]
  expect_true(any(abs(failed_rows$bandwidth - fail_bw) < 1e-9))
})

test_that("search_bandwidth_random best_bandwidth comes from ok candidates only", {
  ctx <- make_random_context()
  # Draw fixed candidates, make the one closest to 5 fail
  cands <- sample_random_bandwidths(lower = 1, upper = 10, n_samples = 10L,
                                    adaptive = FALSE, seed = 33L)
  # Find which candidate is closest to 5
  closest_idx <- which.min(abs(cands - 5))
  fail_bw     <- cands[closest_idx]
  scorer      <- make_random_mock_scorer(best_bw = 5, fail_bw = fail_bw)

  result <- search_bandwidth_random(
    ctx,
    control = list(lower = 1, upper = 10, n_samples = 10L),
    scorer  = scorer,
    workers = 1L,
    seed    = 33L
  )
  # best_bandwidth must be among ok candidates
  ok_bws <- result$history$bandwidth[result$history$status == "ok"]
  expect_true(result$best_bandwidth %in% ok_bws)
  # It must NOT be the failed bandwidth
  expect_false(isTRUE(all.equal(result$best_bandwidth, fail_bw)))
})

test_that("search_bandwidth_random adaptive produces integer candidates", {
  ctx    <- make_random_context(adaptive = TRUE)
  scorer <- make_random_mock_scorer(best_bw = 5)
  result <- search_bandwidth_random(
    ctx,
    control = list(lower = 2, upper = 20, n_samples = 15L),
    scorer  = scorer,
    workers = 1L,
    seed    = 8L
  )
  expect_true(
    all(result$history$bandwidth == as.integer(result$history$bandwidth))
  )
})

test_that("search_bandwidth_random uses default n_samples when not provided", {
  ctx    <- make_random_context()
  scorer <- make_random_mock_scorer()
  result <- search_bandwidth_random(
    ctx,
    control = list(lower = 1, upper = 100),   # n_samples omitted
    scorer  = scorer,
    workers = 1L,
    seed    = 2L
  )
  # Default is 50; result should have 50 rows
  expect_equal(nrow(result$history), 50L)
})

test_that("search_bandwidth_random elapsed_time is non-negative", {
  ctx    <- make_random_context()
  scorer <- make_random_mock_scorer()
  result <- search_bandwidth_random(
    ctx,
    control = list(lower = 1, upper = 10, n_samples = 5L),
    scorer  = scorer,
    workers = 1L,
    seed    = 1L
  )
  expect_true(result$elapsed_time >= 0)
})

test_that("search_bandwidth_random convergence_info contains n_candidates", {
  ctx    <- make_random_context()
  scorer <- make_random_mock_scorer()
  result <- search_bandwidth_random(
    ctx,
    control = list(lower = 1, upper = 10, n_samples = 7L),
    scorer  = scorer,
    workers = 1L,
    seed    = 1L
  )
  expect_equal(result$convergence_info$n_candidates, 7L)
})

test_that("search_bandwidth_random best_bandwidth has rank 1 in history", {
  ctx    <- make_random_context()
  scorer <- make_random_mock_scorer(best_bw = 5)
  result <- search_bandwidth_random(
    ctx,
    control = list(lower = 1, upper = 10, n_samples = 15L),
    scorer  = scorer,
    workers = 1L,
    seed    = 17L
  )
  best_row <- result$history[
    abs(result$history$bandwidth - result$best_bandwidth) < 1e-9 &
      result$history$status == "ok",
    ,
    drop = FALSE
  ]
  expect_equal(best_row$rank[1L], 1L)
})

#' @title Public API for GWPR.light 1.0.0
#'
#' @description
#' High-level user-facing functions for Geographically Weighted Panel
#' Regression.  These four functions form the complete public interface;
#' all internal complexity is hidden behind them.
#'
#' \itemize{
#'   \item \code{\link{gwpr}} -- full pipeline (bandwidth search + fitting +
#'     optional diagnostics).
#'   \item \code{\link{select_bandwidth}} -- standalone bandwidth search.
#'   \item \code{\link{fit_gwpr}} -- fit with a known bandwidth.
#'   \item \code{\link{diagnose_gwpr}} -- run diagnostics on a fitted model.
#' }
#'
#' @name api
NULL

# ---------------------------------------------------------------------------
# Internal helper: null-coalescing operator
# ---------------------------------------------------------------------------

`.api_null_coalesce` <- function(x, y) if (is.null(x)) y else x

# ---------------------------------------------------------------------------
# gwpr
# ---------------------------------------------------------------------------

#' Fit a Geographically Weighted Panel Regression (main entry point)
#'
#' Orchestrates the complete GWPR pipeline: input validation, data preparation,
#' optional memory estimation, optional bandwidth search, model fitting, and
#' optional diagnostics.
#'
#' @param formula A \code{formula} object specifying the model (e.g.
#'   \code{y ~ x1 + x2}).
#' @param data A \code{data.frame} containing the panel data.  Must include
#'   the columns referenced by \code{id} and \code{time}.
#' @param spatial An \code{sf} object with one row per spatial unit.  Must
#'   include the column referenced by \code{id}.
#' @param id Character scalar.  Name of the unit (individual) ID column shared
#'   by \code{data} and \code{spatial}.
#' @param time Character scalar.  Name of the time-period column in \code{data}.
#' @param family Character scalar.  Model family: \code{"gaussian"} (default,
#'   linear GWPR) or \code{"binomial"} (binary panel logistic GWPR).
#' @param model Character scalar.  Panel model type: \code{"within"} (default),
#'   \code{"pooling"}, or \code{"random"}.
#' @param effect Character scalar.  Panel effect: \code{"individual"} (default),
#'   \code{"time"}, \code{"two-way"}, or \code{"nested"}.
#' @param bandwidth Numeric scalar or \code{NULL} (default).  When \code{NULL}
#'   the bandwidth is selected automatically via \code{select_bandwidth()}.
#' @param bandwidth_method Character scalar.  Method for automatic bandwidth
#'   search: \code{"sgd"} (default), \code{"grid"}, or \code{"random"}.
#'   Ignored when \code{bandwidth} is supplied.
#' @param bandwidth_control Named list of control parameters passed to the
#'   bandwidth search function.  For \code{"grid"}: \code{lower}, \code{upper},
#'   \code{step}.  For \code{"sgd"}: \code{lower}, \code{upper}, learning rate
#'   etc.  For \code{"random"}: \code{lower}, \code{upper}, \code{n_samples}.
#' @param kernel Character scalar.  Kernel function: \code{"bisquare"}
#'   (default), \code{"gaussian"}, \code{"exponential"}, \code{"tricube"}, or
#'   \code{"boxcar"}.
#' @param adaptive Logical scalar.  \code{FALSE} (default) uses a fixed
#'   distance bandwidth; \code{TRUE} uses an adaptive (k-nearest-neighbour)
#'   bandwidth.
#' @param threshold Numeric scalar.  Classification threshold for
#'   \code{family = "binomial"} (default \code{0.5}).
#' @param workers Positive integer.  Number of parallel workers.  \code{1}
#'   (default) uses serial execution; values \code{> 1} enable explicit
#'   parallelism.
#' @param seed Integer or \code{NULL}.  Random seed for reproducibility of any
#'   stochastic steps (bandwidth search, parallel RNG).
#' @param diagnostics Logical scalar.  When \code{TRUE} (default),
#'   \code{diagnose_gwpr()} is called after fitting and its results are stored
#'   in the returned object.
#' @param ... Additional arguments passed to the bandwidth search or fitting
#'   functions.
#'
#' @return A \code{gwpr_fit} object.  Key fields:
#' \describe{
#'   \item{\code{local_results}}{Per-unit local model results.}
#'   \item{\code{predictions}}{In-sample predicted values / probabilities.}
#'   \item{\code{residuals}}{Residuals or Pearson residuals.}
#'   \item{\code{metrics}}{Overall goodness-of-fit metrics.}
#'   \item{\code{spatial_results}}{Data frame of per-unit coefficients.}
#'   \item{\code{search}}{Bandwidth search result (\code{gwpr_bandwidth}), or
#'     \code{NULL} when bandwidth was supplied directly.}
#'   \item{\code{diagnostics}}{A \code{gwpr_diagnostics} object, or
#'     \code{NULL}.}
#' }
#'
#' @examples
#' \donttest{
#' # Minimal linear GWPR with a fixed bandwidth
#' library(sf)
#' pts <- sf::st_as_sf(
#'   data.frame(id = 1:4, X = c(0,1,0,1), Y = c(0,0,1,1)),
#'   coords = c("X", "Y"), crs = NA_integer_
#' )
#' dat <- data.frame(
#'   id   = rep(1:4, each = 5),
#'   time = rep(1:5, 4),
#'   y    = rnorm(20),
#'   x1   = rnorm(20)
#' )
#' fit <- gwpr(y ~ x1, data = dat, spatial = pts, id = "id", time = "time",
#'             bandwidth = 2, diagnostics = FALSE, workers = 1)
#' print(fit)
#' }
#'
#' @export
gwpr <- function(formula,
                 data,
                 spatial,
                 id,
                 time,
                 family             = c("gaussian", "binomial"),
                 model              = c("within", "pooling", "random"),
                 effect             = c("individual", "time", "two-way", "nested"),
                 bandwidth          = NULL,
                 bandwidth_method   = c("sgd", "grid", "random"),
                 bandwidth_control  = list(),
                 kernel             = c("bisquare", "gaussian", "exponential",
                                        "tricube", "boxcar"),
                 adaptive           = FALSE,
                 threshold          = 0.5,
                 workers            = 1L,
                 seed               = NULL,
                 diagnostics        = TRUE,
                 ...) {
  cl <- match.call()

  # Match arguments
  family           <- match.arg(family)
  model            <- match.arg(model)
  effect           <- match.arg(effect)
  kernel           <- match.arg(kernel)
  bandwidth_method <- match.arg(bandwidth_method)

  # 1. Validate inputs
  validate_inputs(
    formula  = formula,
    data     = data,
    spatial  = spatial,
    id       = id,
    time     = time,
    family   = family,
    model    = model,
    effect   = effect,
    kernel   = kernel,
    adaptive = adaptive,
    workers  = workers
  )

  # 2. Build context
  ctx <- new_gwpr_context(
    call        = cl,
    formula     = formula,
    family      = family,
    model       = model,
    effect      = effect,
    id          = id,
    time        = time,
    kernel      = kernel,
    adaptive    = adaptive,
    threshold   = threshold,
    workers     = workers,
    seed        = seed,
    raw_data    = data,
    raw_spatial = spatial
  )

  # 3. Prepare data
  ctx <- prepare_data(ctx)

  # Enrich metadata with n_time and n_vars for memory estimation
  ctx$metadata$n_time <- length(unique(ctx$panel_data[[time]]))
  ctx$metadata$n_vars <- ncol(ctx$model_matrix)

  # 4. Memory estimate; warn on high risk
  mem <- tryCatch(
    estimate_memory(ctx, workers = workers, cache_distance = TRUE),
    error = function(e) NULL
  )
  if (!is.null(mem) && mem$risk == "high") {
    warning(format_memory_warning(mem), call. = FALSE)
  }

  # 5. Bandwidth selection (if not provided)
  bw_result <- NULL
  if (is.null(bandwidth)) {
    bw_result <- select_bandwidth(
      formula           = formula,
      data              = data,
      spatial           = spatial,
      id                = id,
      time              = time,
      family            = family,
      model             = model,
      effect            = effect,
      method            = bandwidth_method,
      control           = bandwidth_control,
      kernel            = kernel,
      adaptive          = adaptive,
      threshold         = threshold,
      workers           = workers,
      seed              = seed,
      ...
    )
    bandwidth <- bw_result$best_bandwidth
  }

  # 6. Fit the model
  fit_result <- fit_gwpr(
    formula   = formula,
    data      = data,
    spatial   = spatial,
    id        = id,
    time      = time,
    bandwidth = bandwidth,
    family    = family,
    model     = model,
    effect    = effect,
    kernel    = kernel,
    adaptive  = adaptive,
    threshold = threshold,
    workers   = workers,
    seed      = seed,
    ...
  )

  # Attach search result
  fit_result$search <- bw_result
  fit_result$call   <- cl

  # 7. Diagnostics
  if (isTRUE(diagnostics)) {
    diag_result <- tryCatch(
      diagnose_gwpr(fit_result),
      error   = function(e) NULL,
      warning = function(w) {
        withCallingHandlers(
          diagnose_gwpr(fit_result),
          warning = function(w2) invokeRestart("muffleWarning")
        )
      }
    )
    fit_result$diagnostics_result <- diag_result
  }

  fit_result
}

# ---------------------------------------------------------------------------
# select_bandwidth
# ---------------------------------------------------------------------------

#' Select an optimal bandwidth for GWPR
#'
#' Validates inputs, prepares data, and dispatches to the appropriate bandwidth
#' search algorithm: grid search, stochastic gradient descent (sgd), or random
#' search, depending on the \code{method} argument.
#'
#' @param formula  A \code{formula} object.
#' @param data     A \code{data.frame} with panel data.
#' @param spatial  An \code{sf} object.
#' @param id       Character scalar; unit ID column name.
#' @param time     Character scalar; time column name.
#' @param family   \code{"gaussian"} or \code{"binomial"}.
#' @param model    \code{"within"}, \code{"pooling"}, or \code{"random"}.
#' @param effect   \code{"individual"}, \code{"time"}, \code{"two-way"}, or
#'   \code{"nested"}.
#' @param method   Bandwidth search method: \code{"sgd"} (default),
#'   \code{"grid"}, or \code{"random"}.
#' @param control  Named list of search control parameters.
#' @param kernel   Kernel function name.
#' @param adaptive Logical; \code{TRUE} for adaptive (k-NN) bandwidth.
#' @param threshold Numeric; classification threshold (binomial only).
#' @param workers  Positive integer; number of parallel workers.
#' @param seed     Integer random seed, or \code{NULL}.
#' @param ...      Additional arguments (currently unused).
#'
#' @return A \code{gwpr_bandwidth} object with fields:
#' \describe{
#'   \item{\code{best_bandwidth}}{The selected bandwidth value.}
#'   \item{\code{best_score}}{The criterion value at the best bandwidth.}
#'   \item{\code{method}}{The search method used.}
#'   \item{\code{history}}{Search history data frame.}
#' }
#'
#' @examples
#' \donttest{
#' library(sf)
#' pts <- sf::st_as_sf(
#'   data.frame(id = 1:4, X = c(0,1,0,1), Y = c(0,0,1,1)),
#'   coords = c("X", "Y"), crs = NA_integer_
#' )
#' dat <- data.frame(
#'   id   = rep(1:4, each = 5),
#'   time = rep(1:5, 4),
#'   y    = rnorm(20),
#'   x1   = rnorm(20)
#' )
#' bw <- select_bandwidth(
#'   y ~ x1, data = dat, spatial = pts, id = "id", time = "time",
#'   method  = "grid",
#'   control = list(lower = 0.5, upper = 2, step = 0.5),
#'   workers = 1
#' )
#' bw$best_bandwidth
#' }
#'
#' @export
select_bandwidth <- function(formula,
                              data,
                              spatial,
                              id,
                              time,
                              family    = c("gaussian", "binomial"),
                              model     = c("within", "pooling", "random"),
                              effect    = c("individual", "time", "two-way", "nested"),
                              method    = c("sgd", "grid", "random"),
                              control   = list(),
                              kernel    = c("bisquare", "gaussian", "exponential",
                                            "tricube", "boxcar"),
                              adaptive  = FALSE,
                              threshold = 0.5,
                              workers   = 1L,
                              seed      = NULL,
                              ...) {
  family <- match.arg(family)
  model  <- match.arg(model)
  effect <- match.arg(effect)
  kernel <- match.arg(kernel)
  method <- match.arg(method)

  # Validate
  validate_inputs(
    formula  = formula,
    data     = data,
    spatial  = spatial,
    id       = id,
    time     = time,
    family   = family,
    model    = model,
    effect   = effect,
    kernel   = kernel,
    adaptive = adaptive,
    workers  = workers
  )
  validate_bandwidth_control(method = method, control = control,
                              adaptive = adaptive)

  # Build context
  ctx <- new_gwpr_context(
    formula     = formula,
    family      = family,
    model       = model,
    effect      = effect,
    id          = id,
    time        = time,
    kernel      = kernel,
    adaptive    = adaptive,
    threshold   = threshold,
    workers     = workers,
    seed        = seed,
    raw_data    = data,
    raw_spatial = spatial
  )
  ctx <- prepare_data(ctx)

  # Scorer: fits GWPR for a given bandwidth and returns score + metrics
  scorer <- .make_scorer(family = family, threshold = threshold)

  # Dispatch to search method
  bw_result <- switch(
    method,
    grid   = search_bandwidth_grid(ctx, control = control, scorer = scorer,
                                    workers = workers, seed = seed),
    sgd    = search_bandwidth_sgd(ctx,  control = control, scorer = scorer,
                                   workers = workers, seed = seed),
    random = search_bandwidth_random(ctx, control = control, scorer = scorer,
                                      workers = workers, seed = seed),
    stop("Unknown bandwidth method: ", method, call. = FALSE)
  )

  bw_result
}

# ---------------------------------------------------------------------------
# fit_gwpr
# ---------------------------------------------------------------------------

#' Fit GWPR with a given bandwidth
#'
#' Validates inputs, prepares data, builds spatial weights, and fits the
#' Geographically Weighted Panel Regression for the specified bandwidth.
#' Returns a \code{gwpr_fit} object.
#'
#' @param formula   A \code{formula} object.
#' @param data      A \code{data.frame} with panel data.
#' @param spatial   An \code{sf} object.
#' @param id        Character scalar; unit ID column name.
#' @param time      Character scalar; time column name.
#' @param bandwidth Numeric scalar.  The bandwidth to use (fixed distance or
#'   number of neighbours when \code{adaptive = TRUE}).
#' @param family    \code{"gaussian"} (default) or \code{"binomial"}.
#' @param model     \code{"within"} (default), \code{"pooling"}, or
#'   \code{"random"}.
#' @param effect    \code{"individual"} (default), \code{"time"},
#'   \code{"two-way"}, or \code{"nested"}.
#' @param kernel    Kernel function name (default \code{"bisquare"}).
#' @param adaptive  Logical; \code{FALSE} (default) for fixed bandwidth.
#' @param threshold Numeric; classification threshold (binomial only,
#'   default \code{0.5}).
#' @param workers   Positive integer; number of parallel workers (default 1).
#' @param seed      Integer random seed, or \code{NULL}.
#' @param ...       Currently unused.
#'
#' @return A \code{gwpr_fit} object.
#'
#' @examples
#' \donttest{
#' library(sf)
#' pts <- sf::st_as_sf(
#'   data.frame(id = 1:4, X = c(0,1,0,1), Y = c(0,0,1,1)),
#'   coords = c("X", "Y"), crs = NA_integer_
#' )
#' dat <- data.frame(
#'   id   = rep(1:4, each = 5),
#'   time = rep(1:5, 4),
#'   y    = rnorm(20),
#'   x1   = rnorm(20)
#' )
#' fit <- fit_gwpr(y ~ x1, data = dat, spatial = pts, id = "id",
#'                 time = "time", bandwidth = 2, workers = 1)
#' print(fit)
#' }
#'
#' @export
fit_gwpr <- function(formula,
                     data,
                     spatial,
                     id,
                     time,
                     bandwidth,
                     family    = c("gaussian", "binomial"),
                     model     = c("within", "pooling", "random"),
                     effect    = c("individual", "time", "two-way", "nested"),
                     kernel    = c("bisquare", "gaussian", "exponential",
                                   "tricube", "boxcar"),
                     adaptive  = FALSE,
                     threshold = 0.5,
                     workers   = 1L,
                     seed      = NULL,
                     ...) {
  family <- match.arg(family)
  model  <- match.arg(model)
  effect <- match.arg(effect)
  kernel <- match.arg(kernel)

  if (missing(bandwidth) || is.null(bandwidth)) {
    stop("`bandwidth` must be supplied to `fit_gwpr()`. ",
         "To search automatically use `gwpr()` or `select_bandwidth()`.",
         call. = FALSE)
  }
  if (!is.numeric(bandwidth) || length(bandwidth) != 1L || !is.finite(bandwidth)) {
    stop("`bandwidth` must be a single finite numeric value.", call. = FALSE)
  }
  if (bandwidth <= 0) {
    stop("`bandwidth` must be positive.", call. = FALSE)
  }

  # Validate
  validate_inputs(
    formula  = formula,
    data     = data,
    spatial  = spatial,
    id       = id,
    time     = time,
    family   = family,
    model    = model,
    effect   = effect,
    kernel   = kernel,
    adaptive = adaptive,
    workers  = workers
  )

  # Build context and prepare data
  ctx <- new_gwpr_context(
    formula     = formula,
    family      = family,
    model       = model,
    effect      = effect,
    id          = id,
    time        = time,
    kernel      = kernel,
    adaptive    = adaptive,
    threshold   = threshold,
    workers     = workers,
    seed        = seed,
    raw_data    = data,
    raw_spatial = spatial
  )
  ctx <- prepare_data(ctx)

  # Fit by family
  if (family == "gaussian") {
    engine_result <- fit_linear_gwpr(ctx, bandwidth = bandwidth)

    spatial_res <- tryCatch(
      build_spatial_results(engine_result$local_results),
      error = function(e) NULL
    )

    fit <- new_gwpr_fit(
      call            = NULL,
      family          = family,
      formula         = formula,
      model           = model,
      effect          = effect,
      bandwidth       = bandwidth,
      search          = NULL,
      local_results   = engine_result$local_results,
      predictions     = engine_result$predictions,
      residuals       = engine_result$residuals,
      metrics         = engine_result$metrics,
      spatial_results = spatial_res,
      metadata        = c(ctx$metadata, engine_result$metadata),
      warnings        = ctx$warnings
    )

  } else {
    # binomial — map model names: "within" -> "fixed" for logistic engine
    logistic_model <- switch(
      model,
      within  = "fixed",
      pooling = "pooling",
      random  = "random"
    )
    ctx_logistic        <- ctx
    ctx_logistic$model  <- logistic_model

    engine_result <- fit_logistic_gwpr(ctx_logistic, bandwidth = bandwidth)

    spatial_res <- tryCatch(
      build_spatial_results(engine_result$local_results),
      error = function(e) NULL
    )

    fit <- new_gwpr_fit(
      call            = NULL,
      family          = family,
      formula         = formula,
      model           = model,
      effect          = effect,
      bandwidth       = bandwidth,
      search          = NULL,
      local_results   = engine_result$local_results,
      predictions     = engine_result$predictions,
      residuals       = engine_result$pearson_resid,
      metrics         = engine_result$metrics,
      spatial_results = spatial_res,
      metadata        = c(
        ctx$metadata,
        engine_result$metadata,
        list(
          response = as.numeric(ctx$response),
          prob     = as.numeric(engine_result$predictions)
        )
      ),
      warnings        = ctx$warnings
    )
  }

  fit
}

# ---------------------------------------------------------------------------
# Internal helpers
# ---------------------------------------------------------------------------

#' Build a scorer function for bandwidth search
#'
#' Returns a closure that fits the full GWPR engine for a given bandwidth and
#' returns the score (MSE for linear, log_loss for logistic), together with
#' aggregate metrics.
#'
#' @param family    Character; `"gaussian"` or `"binomial"`.
#' @param threshold Numeric; classification threshold (binomial only).
#'
#' @return A function with signature `scorer(context, bandwidth)`.
#' @keywords internal
.make_scorer <- function(family = "gaussian", threshold = 0.5) {
  function(context, bandwidth) {
    if (family == "gaussian") {
      engine_result <- fit_linear_gwpr(context, bandwidth = bandwidth)
      metrics  <- engine_result$metrics
      score    <- `.api_null_coalesce`(metrics$MSE, NA_real_)
      n_local  <- length(engine_result$local_results)
      n_failed <- sum(vapply(engine_result$local_results,
                             function(r) !is.null(r$status) && r$status == "failed",
                             logical(1L)))
      list(
        score                 = score,
        criterion             = "MSE",
        n_local_models        = n_local,
        n_failed_local_models = n_failed,
        metrics               = metrics,
        status                = "ok"
      )
    } else {
      # binomial — remap model
      ctx_logistic        <- context
      ctx_logistic$model  <- switch(
        context$model,
        within  = "fixed",
        pooling = "pooling",
        random  = "random"
      )
      engine_result <- fit_logistic_gwpr(ctx_logistic, bandwidth = bandwidth)
      metrics  <- engine_result$metrics
      score    <- `.api_null_coalesce`(metrics$log_loss, NA_real_)
      n_local  <- length(engine_result$local_results)
      n_failed <- sum(vapply(engine_result$local_results,
                             function(r) !is.null(r$status) && r$status == "failed",
                             logical(1L)))
      list(
        score                 = score,
        criterion             = "log_loss",
        n_local_models        = n_local,
        n_failed_local_models = n_failed,
        metrics               = metrics,
        status                = "ok"
      )
    }
  }
}

## tests/testthat/test-api.R
## End-to-end tests for the public API (R/api.R)
##
## All tests use small simulated panel + sf data.
## workers = 1 (no parallel backend activated).

# ---------------------------------------------------------------------------
# Shared test fixtures
# ---------------------------------------------------------------------------

# 4 spatial units, 5 time periods each => 20 rows of linear panel data
make_api_panel_data <- function(seed = 42L) {
  set.seed(seed)
  n_units <- 4L
  n_times <- 5L
  id   <- rep(as.character(1:n_units), each = n_times)
  time <- rep(1:n_times, times = n_units)
  x1   <- rnorm(n_units * n_times)
  x2   <- rnorm(n_units * n_times)
  y    <- 2 + 0.5 * x1 - 0.3 * x2 + rnorm(n_units * n_times, sd = 0.2)
  data.frame(id = id, time = time, y = y, x1 = x1, x2 = x2,
             stringsAsFactors = FALSE)
}

# 4 spatial units, 5 time periods each => 20 rows of binary panel data
make_api_bin_panel_data <- function(seed = 99L) {
  set.seed(seed)
  n_units <- 4L
  n_times <- 5L
  id   <- rep(as.character(1:n_units), each = n_times)
  time <- rep(1:n_times, times = n_units)
  x1   <- rnorm(n_units * n_times)
  lp   <- 0.5 * x1
  y    <- as.integer(rbinom(n_units * n_times, 1L, plogis(lp)))
  data.frame(id = id, time = time, y = y, x1 = x1,
             stringsAsFactors = FALSE)
}

# sf POINT object with 4 units on a 2x2 grid
make_api_sf <- function() {
  pts <- data.frame(
    id = as.character(1:4),
    X  = c(0, 1, 0, 1),
    Y  = c(0, 0, 1, 1)
  )
  sf::st_as_sf(pts, coords = c("X", "Y"), crs = NA_integer_)
}

# ---------------------------------------------------------------------------
# fit_gwpr  — linear (Gaussian), pooling model
# ---------------------------------------------------------------------------

test_that("fit_gwpr: linear pooling returns gwpr_fit", {
  dat  <- make_api_panel_data()
  spt  <- make_api_sf()

  fit <- fit_gwpr(
    formula   = y ~ x1 + x2,
    data      = dat,
    spatial   = spt,
    id        = "id",
    time      = "time",
    bandwidth = 2,
    family    = "gaussian",
    model     = "pooling",
    effect    = "individual",
    workers   = 1L
  )

  expect_s3_class(fit, "gwpr_fit")
  expect_equal(fit$family, "gaussian")
  expect_equal(fit$model,  "pooling")
  expect_false(is.null(fit$local_results))
  expect_length(fit$local_results, 4L)   # one per spatial unit
  expect_false(is.null(fit$predictions))
  expect_false(is.null(fit$residuals))
  expect_false(is.null(fit$metrics))
})

# ---------------------------------------------------------------------------
# fit_gwpr  — linear (Gaussian), within model
# ---------------------------------------------------------------------------

test_that("fit_gwpr: linear within returns gwpr_fit", {
  dat <- make_api_panel_data()
  spt <- make_api_sf()

  fit <- suppressWarnings(
    fit_gwpr(
      formula   = y ~ x1 + x2,
      data      = dat,
      spatial   = spt,
      id        = "id",
      time      = "time",
      bandwidth = 2,
      family    = "gaussian",
      model     = "within",
      effect    = "individual",
      workers   = 1L
    )
  )

  expect_s3_class(fit, "gwpr_fit")
  expect_equal(fit$family, "gaussian")
  expect_equal(fit$model,  "within")
})

# ---------------------------------------------------------------------------
# fit_gwpr  — binomial (logistic), pooling model
# ---------------------------------------------------------------------------

test_that("fit_gwpr: logistic pooling returns gwpr_fit", {
  dat <- make_api_bin_panel_data()
  spt <- make_api_sf()

  fit <- suppressWarnings(
    fit_gwpr(
      formula   = y ~ x1,
      data      = dat,
      spatial   = spt,
      id        = "id",
      time      = "time",
      bandwidth = 2,
      family    = "binomial",
      model     = "pooling",
      effect    = "individual",
      workers   = 1L
    )
  )

  expect_s3_class(fit, "gwpr_fit")
  expect_equal(fit$family, "binomial")
  expect_false(is.null(fit$predictions))   # predicted probabilities
  expect_false(is.null(fit$metrics))
})

# ---------------------------------------------------------------------------
# fit_gwpr  — given bandwidth does NOT trigger search
# ---------------------------------------------------------------------------

test_that("fit_gwpr: does not call select_bandwidth when bandwidth is given", {
  dat <- make_api_panel_data()
  spt <- make_api_sf()

  # We mock select_bandwidth to detect if it is called
  was_called <- FALSE
  local_mock_env <- new.env(parent = asNamespace("GWPR.light"))
  local({
    fit <- fit_gwpr(
      formula   = y ~ x1 + x2,
      data      = dat,
      spatial   = spt,
      id        = "id",
      time      = "time",
      bandwidth = 2,
      family    = "gaussian",
      model     = "pooling",
      workers   = 1L
    )
    expect_s3_class(fit, "gwpr_fit")
    # No assertion needed for was_called; absence of error is sufficient.
    # The key check is that bandwidth = 2 is stored in fit.
    expect_equal(fit$bandwidth, 2)
  })
})

# ---------------------------------------------------------------------------
# fit_gwpr  — bandwidth stored in output
# ---------------------------------------------------------------------------

test_that("fit_gwpr: bandwidth value is preserved in output", {
  dat <- make_api_panel_data()
  spt <- make_api_sf()

  bw <- 1.5
  fit <- fit_gwpr(
    formula   = y ~ x1 + x2,
    data      = dat,
    spatial   = spt,
    id        = "id",
    time      = "time",
    bandwidth = bw,
    family    = "gaussian",
    model     = "pooling",
    workers   = 1L
  )
  expect_equal(fit$bandwidth, bw)
})

# ---------------------------------------------------------------------------
# fit_gwpr  — invalid bandwidth raises error
# ---------------------------------------------------------------------------

test_that("fit_gwpr: missing bandwidth raises error", {
  dat <- make_api_panel_data()
  spt <- make_api_sf()

  expect_error(
    fit_gwpr(y ~ x1, data = dat, spatial = spt, id = "id", time = "time",
             workers = 1L),
    regexp = "bandwidth"
  )
})

test_that("fit_gwpr: negative bandwidth raises error", {
  dat <- make_api_panel_data()
  spt <- make_api_sf()

  expect_error(
    fit_gwpr(y ~ x1, data = dat, spatial = spt, id = "id", time = "time",
             bandwidth = -1, workers = 1L),
    regexp = "positive"
  )
})

# ---------------------------------------------------------------------------
# fit_gwpr  — workers = 1 does not trigger parallel backend
# ---------------------------------------------------------------------------

test_that("fit_gwpr: workers = 1 completes without activating parallel plan", {
  dat <- make_api_panel_data()
  spt <- make_api_sf()

  # With workers = 1 the function must run without error / without calling
  # future::plan().  We simply verify it returns a valid object.
  fit <- fit_gwpr(
    formula   = y ~ x1 + x2,
    data      = dat,
    spatial   = spt,
    id        = "id",
    time      = "time",
    bandwidth = 2,
    family    = "gaussian",
    model     = "pooling",
    workers   = 1L
  )
  expect_s3_class(fit, "gwpr_fit")
})

# ---------------------------------------------------------------------------
# Illegal family / model / effect raise clear errors
# ---------------------------------------------------------------------------

test_that("fit_gwpr: illegal family raises error", {
  dat <- make_api_panel_data()
  spt <- make_api_sf()

  expect_error(
    fit_gwpr(y ~ x1, data = dat, spatial = spt, id = "id", time = "time",
             bandwidth = 2, family = "poisson", workers = 1L),
    regexp = "arg"
  )
})

test_that("fit_gwpr: illegal model raises error", {
  dat <- make_api_panel_data()
  spt <- make_api_sf()

  expect_error(
    fit_gwpr(y ~ x1, data = dat, spatial = spt, id = "id", time = "time",
             bandwidth = 2, model = "twoway_bad", workers = 1L),
    regexp = "arg"
  )
})

test_that("fit_gwpr: illegal effect raises error via validate_inputs", {
  dat <- make_api_panel_data()
  spt <- make_api_sf()

  expect_error(
    fit_gwpr(y ~ x1, data = dat, spatial = spt, id = "id", time = "time",
             bandwidth = 2, effect = "bad_effect", workers = 1L),
    regexp = "arg"
  )
})

# ---------------------------------------------------------------------------
# validate_inputs integration via fit_gwpr
# ---------------------------------------------------------------------------

test_that("fit_gwpr: non-sf spatial raises informative error", {
  dat <- make_api_panel_data()

  expect_error(
    fit_gwpr(y ~ x1, data = dat, spatial = list(id = "1"), id = "id",
             time = "time", bandwidth = 2, workers = 1L),
    regexp = "sf"
  )
})

test_that("fit_gwpr: missing id column in data raises error", {
  dat <- make_api_panel_data()
  dat2 <- dat
  names(dat2)[names(dat2) == "id"] <- "unit"   # rename id column
  spt <- make_api_sf()

  expect_error(
    fit_gwpr(y ~ x1, data = dat2, spatial = spt, id = "id",
             time = "time", bandwidth = 2, workers = 1L),
    regexp = "id"
  )
})

test_that("fit_gwpr: workers = 0 raises error", {
  dat <- make_api_panel_data()
  spt <- make_api_sf()

  expect_error(
    fit_gwpr(y ~ x1, data = dat, spatial = spt, id = "id", time = "time",
             bandwidth = 2, workers = 0L),
    regexp = "workers"
  )
})

# ---------------------------------------------------------------------------
# diagnose_gwpr
# ---------------------------------------------------------------------------

test_that("diagnose_gwpr: returns gwpr_diagnostics object", {
  dat <- make_api_panel_data()
  spt <- make_api_sf()

  fit <- fit_gwpr(
    formula   = y ~ x1 + x2,
    data      = dat,
    spatial   = spt,
    id        = "id",
    time      = "time",
    bandwidth = 2,
    family    = "gaussian",
    model     = "pooling",
    workers   = 1L
  )

  diag_result <- diagnose_gwpr(fit, diagnostics = c("f_test", "hausman", "lm_test"))
  expect_s3_class(diag_result, "gwpr_diagnostics")
  expect_true(length(diag_result$diagnostics) > 0L)
})

test_that("diagnose_gwpr: object must be gwpr_fit", {
  expect_error(diagnose_gwpr(list(a = 1)), regexp = "gwpr_fit")
})

test_that("diagnose_gwpr: moran skipped gracefully without spatial_weights", {
  dat <- make_api_panel_data()
  spt <- make_api_sf()

  fit <- fit_gwpr(
    formula   = y ~ x1 + x2,
    data      = dat,
    spatial   = spt,
    id        = "id",
    time      = "time",
    bandwidth = 2,
    family    = "gaussian",
    model     = "pooling",
    workers   = 1L
  )

  # Without spatial_weights / panel_index, moran should be skipped, not error
  diag_result <- diagnose_gwpr(fit, diagnostics = "moran")
  expect_s3_class(diag_result, "gwpr_diagnostics")
  expect_equal(diag_result$diagnostics$moran$status, "skipped")
})

# ---------------------------------------------------------------------------
# gwpr  — main entry point, given bandwidth (no search)
# ---------------------------------------------------------------------------

test_that("gwpr: given bandwidth returns gwpr_fit without search", {
  dat <- make_api_panel_data()
  spt <- make_api_sf()

  fit <- suppressWarnings(
    gwpr(
      formula      = y ~ x1 + x2,
      data         = dat,
      spatial      = spt,
      id           = "id",
      time         = "time",
      bandwidth    = 2,
      family       = "gaussian",
      model        = "pooling",
      effect       = "individual",
      diagnostics  = FALSE,
      workers      = 1L
    )
  )

  expect_s3_class(fit, "gwpr_fit")
  expect_equal(fit$bandwidth, 2)
  expect_null(fit$search)   # search was not triggered
})

test_that("gwpr: given bandwidth, search field is NULL", {
  dat <- make_api_panel_data()
  spt <- make_api_sf()

  fit <- suppressWarnings(
    gwpr(
      formula     = y ~ x1 + x2,
      data        = dat,
      spatial     = spt,
      id          = "id",
      time        = "time",
      bandwidth   = 2,
      diagnostics = FALSE,
      workers     = 1L
    )
  )
  expect_null(fit$search)
})

# ---------------------------------------------------------------------------
# gwpr  — diagnostics integration
# ---------------------------------------------------------------------------

test_that("gwpr: diagnostics = FALSE suppresses diagnostic computation", {
  dat <- make_api_panel_data()
  spt <- make_api_sf()

  fit <- suppressWarnings(
    gwpr(
      formula     = y ~ x1 + x2,
      data        = dat,
      spatial     = spt,
      id          = "id",
      time        = "time",
      bandwidth   = 2,
      model       = "pooling",
      diagnostics = FALSE,
      workers     = 1L
    )
  )
  expect_s3_class(fit, "gwpr_fit")
  # diagnostics_result should not exist or be NULL
  expect_true(is.null(fit$diagnostics_result))
})

# ---------------------------------------------------------------------------
# select_bandwidth  — grid search, very small grid
# ---------------------------------------------------------------------------

test_that("select_bandwidth: grid search returns gwpr_bandwidth", {
  dat <- make_api_panel_data()
  spt <- make_api_sf()

  bw <- suppressWarnings(
    select_bandwidth(
      formula  = y ~ x1 + x2,
      data     = dat,
      spatial  = spt,
      id       = "id",
      time     = "time",
      family   = "gaussian",
      model    = "pooling",
      effect   = "individual",
      method   = "grid",
      control  = list(lower = 1, upper = 2, step = 1),
      workers  = 1L
    )
  )

  expect_s3_class(bw, "gwpr_bandwidth")
  expect_true(is.numeric(bw$best_bandwidth))
  expect_true(bw$best_bandwidth %in% c(1, 2))
})

test_that("select_bandwidth: grid method, best_score is finite", {
  dat <- make_api_panel_data()
  spt <- make_api_sf()

  bw <- suppressWarnings(
    select_bandwidth(
      formula = y ~ x1 + x2,
      data    = dat,
      spatial = spt,
      id      = "id",
      time    = "time",
      method  = "grid",
      control = list(lower = 1, upper = 2, step = 1),
      workers = 1L
    )
  )
  expect_true(is.finite(bw$best_score))
})

# ---------------------------------------------------------------------------
# select_bandwidth  — illegal method raises error
# ---------------------------------------------------------------------------

test_that("select_bandwidth: illegal method raises error", {
  dat <- make_api_panel_data()
  spt <- make_api_sf()

  expect_error(
    select_bandwidth(
      formula = y ~ x1, data = dat, spatial = spt,
      id = "id", time = "time",
      method = "bayes",
      control = list(lower = 1, upper = 2, step = 0.5),
      workers = 1L
    ),
    regexp = "arg"
  )
})

# ---------------------------------------------------------------------------
# select_bandwidth  — grid missing bounds raises error
# ---------------------------------------------------------------------------

test_that("select_bandwidth: grid without bounds raises error", {
  dat <- make_api_panel_data()
  spt <- make_api_sf()

  expect_error(
    select_bandwidth(
      formula = y ~ x1, data = dat, spatial = spt,
      id = "id", time = "time",
      method  = "grid",
      control = list(),   # missing lower/upper/step
      workers = 1L
    ),
    regexp = "lower|upper|step"
  )
})

# ---------------------------------------------------------------------------
# gwpr  — illegal family / model / effect via main entry
# ---------------------------------------------------------------------------

test_that("gwpr: illegal family raises error", {
  dat <- make_api_panel_data()
  spt <- make_api_sf()

  expect_error(
    gwpr(y ~ x1, data = dat, spatial = spt, id = "id", time = "time",
         bandwidth = 2, family = "poisson", workers = 1L),
    regexp = "arg"
  )
})

test_that("gwpr: illegal model raises error", {
  dat <- make_api_panel_data()
  spt <- make_api_sf()

  expect_error(
    gwpr(y ~ x1, data = dat, spatial = spt, id = "id", time = "time",
         bandwidth = 2, model = "wrong_model", workers = 1L),
    regexp = "arg"
  )
})

test_that("gwpr: illegal effect raises error", {
  dat <- make_api_panel_data()
  spt <- make_api_sf()

  expect_error(
    gwpr(y ~ x1, data = dat, spatial = spt, id = "id", time = "time",
         bandwidth = 2, effect = "bad_effect", workers = 1L),
    regexp = "arg"
  )
})

# ---------------------------------------------------------------------------
# Smoke test: complete linear pipeline via gwpr() with bandwidth search
# ---------------------------------------------------------------------------

test_that("gwpr: full linear pipeline with grid bandwidth search", {
  dat <- make_api_panel_data()
  spt <- make_api_sf()

  fit <- suppressWarnings(
    gwpr(
      formula          = y ~ x1 + x2,
      data             = dat,
      spatial          = spt,
      id               = "id",
      time             = "time",
      family           = "gaussian",
      model            = "pooling",
      effect           = "individual",
      bandwidth_method = "grid",
      bandwidth_control = list(lower = 1, upper = 2, step = 1),
      diagnostics      = FALSE,
      workers          = 1L
    )
  )

  expect_s3_class(fit, "gwpr_fit")
  expect_s3_class(fit$search, "gwpr_bandwidth")
  expect_true(is.numeric(fit$bandwidth))
  expect_false(is.null(fit$metrics))
})

# ---------------------------------------------------------------------------
# Smoke test: complete logistic pipeline via gwpr() with given bandwidth
# ---------------------------------------------------------------------------

test_that("gwpr: full logistic pipeline with given bandwidth", {
  dat <- make_api_bin_panel_data()
  spt <- make_api_sf()

  fit <- suppressWarnings(
    gwpr(
      formula     = y ~ x1,
      data        = dat,
      spatial     = spt,
      id          = "id",
      time        = "time",
      family      = "binomial",
      model       = "pooling",
      effect      = "individual",
      bandwidth   = 2,
      diagnostics = FALSE,
      workers     = 1L
    )
  )

  expect_s3_class(fit, "gwpr_fit")
  expect_equal(fit$family, "binomial")
  expect_false(is.null(fit$predictions))
  expect_false(is.null(fit$metrics))
  expect_null(fit$search)   # no search triggered
})

# ---------------------------------------------------------------------------
# spatial_results structure
# ---------------------------------------------------------------------------

test_that("fit_gwpr: spatial_results is a data.frame with unit_id column", {
  dat <- make_api_panel_data()
  spt <- make_api_sf()

  fit <- fit_gwpr(
    formula   = y ~ x1 + x2,
    data      = dat,
    spatial   = spt,
    id        = "id",
    time      = "time",
    bandwidth = 2,
    model     = "pooling",
    workers   = 1L
  )

  expect_true(is.data.frame(fit$spatial_results))
  expect_true("unit_id" %in% names(fit$spatial_results))
  expect_equal(nrow(fit$spatial_results), 4L)   # 4 spatial units
})

#' @title Panel Dataset for Testing GWPR
#'
#' @description Panel dataset to estimate the relationship between county-level PM2.5 concentration and on-road transporation in California.
#'
#' @usage data(TransAirPolCalif)
#'
#' @format A \code{data.frame} with 23 variables, and 928 observations, which are:
#' \describe{
#' \item{GEOID}{a numeric vector, fips IDs of the counties}
#' \item{year}{a numeric vector, year}
#' \item{pm25}{a numeric vector, annually average PM2.5 concentration in the counties}
#' \item{co2_mean}{a numeric vector, geographically average CO2 emission from on-road transportation in each year, million tons/km2}
#' \item{Developed_Open_Space_perc}{a numeric vector, percentage of developed open space of total area in each county}
#' \item{Developed_Low_Intensity_perc}{a numeric vector, percentage of low-intensity developed area of total area in each county}
#' \item{Developed_Medium_Intensity_perc}{a numeric vector, percentage of medium-intensity developed area of total area in each county}
#' \item{Developed_High_Intensity_perc}{a numeric vector, percentage of high-intensity develope area of total area in each county}
#' \item{Open_Water_perc}{a numeric vector, percentage of open water of total area in each county}
#' \item{Woody_Wetlands_perc}{a numeric vector, percentage of woody wetland of total area in each county}
#' \item{Emergent_Herbaceous_Wetlands_perc}{a numeric vector, percentage of emergent herbaceous wetland of total area in each county}
#' \item{Deciduous_Forest_perc}{a numeric vector, percentage of deciduous forest of total area in each county}
#' \item{Evergreen_Forest_perc}{a numeric vector, percentage of evergreen forest of total area in each county}
#' \item{Mixed_Forest_perc}{a numeric vector, percentage of mixed forest of total area in each county}
#' \item{Shrub_perc}{a numeric vector, percentage of shrub of total area in each county}
#' \item{Grassland_perc}{a numeric vector, percentage of grassland of total area in each county}
#' \item{Pasture_perc}{a numeric vector, percentage of pasture of total area in each county}
#' \item{Cultivated_Crops_perc}{a numeric vector, percentage of cultivated crops of total area in each county}
#' \item{pop_density}{a numeric vector, average population density in each county}
#' \item{summer_tmmx}{a numeric vector, average temperature in summer}
#' \item{winter_tmmx}{a numeric vector, average temperature in winter}
#' \item{summer_rmax}{a numeric vector, average humidity in summer}
#' \item{winter_rmax}{a numeric vector, average humidity in winter}
#' }
#'
#' @author Chao Li <chaoli0394@gmail.com> Shunsuke Managi <managi.s@gmail.com>
#'
#' @examples \donttest{
#' data(TransAirPolCalif)
#' head(TransAirPolCalif)
#' }
"TransAirPolCalif"

#' @title Binary Panel Logistic Engine for GWPR.light 1.0.0
#'
#' @description
#' Internal functions for fitting Geographically Weighted Panel Regression
#' with a binary (binomial) response.  Supports pooling (stats::glm),
#' fixed effects (fixest::feglm), and random effects (glmmTMB::glmmTMB)
#' panel models, plus individual, time, two-way, and nested effects.
#'
#' The \code{family} parameter is reserved for future multi-class extension;
#' in version 1.0.0 only \code{"binomial"} is supported.
#'
#' @name model_logistic
#' @keywords internal
#' @importFrom stats binomial
NULL

# ---------------------------------------------------------------------------
# validate_binary_response
# ---------------------------------------------------------------------------

#' Validate that a response vector is suitable for binary logistic regression
#'
#' Stops with an informative error when \code{y} is a factor with more than two
#' levels.  Passes through numeric 0/1 vectors and two-level factors unchanged.
#'
#' @param y A numeric, logical, or factor vector.
#'
#' @return Invisibly \code{TRUE} when validation passes.
#' @keywords internal
validate_binary_response <- function(y) {
  if (is.factor(y)) {
    lvls <- levels(y)
    if (length(lvls) > 2L) {
      stop(
        "Multi-class factor response is not supported in GWPR.light 1.0.0. ",
        "The response must be a two-level factor or a numeric 0/1 vector. ",
        "Found ", length(lvls), " factor levels.",
        call. = FALSE
      )
    }
  }
  invisible(TRUE)
}

# ---------------------------------------------------------------------------
# standardize_logistic_response
# ---------------------------------------------------------------------------

#' Coerce a binary response to a numeric 0/1 integer vector
#'
#' Converts logical and two-level factor responses to 0/1 integer.
#' Numeric 0/1 vectors are returned unchanged.
#' Factors with more than two levels raise an error via
#' \code{validate_binary_response()}.
#'
#' @param y A numeric, logical, or factor vector.
#'
#' @return An integer vector of 0s and 1s.
#' @keywords internal
standardize_logistic_response <- function(y) {
  validate_binary_response(y)

  if (is.logical(y)) return(as.integer(y))

  if (is.factor(y)) {
    lvls <- levels(y)
    return(as.integer(y) - 1L)
  }

  if (is.numeric(y)) {
    unique_vals <- unique(y[!is.na(y)])
    invalid     <- setdiff(unique_vals, c(0, 1))
    if (length(invalid) > 0L) {
      stop(
        "Numeric response must contain only 0 and 1 for binary logistic regression. ",
        "Found: ", paste(invalid, collapse = ", "),
        call. = FALSE
      )
    }
    return(as.integer(y))
  }

  stop(
    "Response must be a numeric 0/1 vector, a logical vector, ",
    "or a two-level factor.",
    call. = FALSE
  )
}

# ---------------------------------------------------------------------------
# effect_to_feglm_fml
# ---------------------------------------------------------------------------

#' Build a fixest::feglm formula string from the user-facing effect string
#'
#' @param base_formula A formula object (without fixed-effect terms).
#' @param effect       Character scalar: one of \code{"individual"},
#'   \code{"time"}, \code{"two-way"}, \code{"nested"}.
#' @param id_col       Name of the individual ID column.
#' @param time_col     Name of the time column.
#'
#' @return A formula suitable for \code{fixest::feglm()}.
#' @keywords internal
effect_to_feglm_fml <- function(base_formula, effect, id_col, time_col) {
  fe_part <- switch(
    effect,
    individual = id_col,
    time       = time_col,
    `two-way`  = paste(id_col, time_col, sep = " + "),
    nested     = stop(
      '`effect = "nested"` requires additional hierarchical level information ',
      "that is not available from `id` and `time` alone. ",
      "Please provide a nested-level column and use a custom formula.",
      call. = FALSE
    ),
    stop(
      sprintf('Unknown effect: "%s". Must be one of: individual, time, two-way, nested.', effect),
      call. = FALSE
    )
  )

  # Build feglm formula: lhs ~ rhs | fe_part
  lhs  <- deparse(base_formula[[2L]])
  rhs  <- deparse(base_formula[[3L]])
  stats::as.formula(paste0(lhs, " ~ ", rhs, " | ", fe_part))
}

# ---------------------------------------------------------------------------
# effect_to_glmmtmb_fml
# ---------------------------------------------------------------------------

#' Append random-effect terms to a base formula for glmmTMB
#'
#' @param base_formula A formula object (without random-effect terms).
#' @param effect       Character scalar: one of \code{"individual"},
#'   \code{"time"}, \code{"two-way"}, \code{"nested"}.
#' @param id_col       Name of the individual ID column.
#' @param time_col     Name of the time column.
#'
#' @return A formula suitable for \code{glmmTMB::glmmTMB()}.
#' @keywords internal
effect_to_glmmtmb_fml <- function(base_formula, effect, id_col, time_col) {
  re_part <- switch(
    effect,
    individual = paste0("(1 | ", id_col, ")"),
    time       = paste0("(1 | ", time_col, ")"),
    `two-way`  = paste0("(1 | ", id_col, ") + (1 | ", time_col, ")"),
    nested     = stop(
      '`effect = "nested"` requires additional hierarchical level information ',
      "that is not available from `id` and `time` alone. ",
      "Please provide a nested-level column and use a custom formula.",
      call. = FALSE
    ),
    stop(
      sprintf('Unknown effect: "%s". Must be one of: individual, time, two-way, nested.', effect),
      call. = FALSE
    )
  )

  lhs  <- deparse(base_formula[[2L]])
  rhs  <- deparse(base_formula[[3L]])
  stats::as.formula(paste0(lhs, " ~ ", rhs, " + ", re_part))
}

# ---------------------------------------------------------------------------
# fit_logistic_pooling
# ---------------------------------------------------------------------------

#' Fit a pooled logistic model via stats::glm
#'
#' @param local_data A \code{data.frame} containing all formula variables.
#' @param formula    A formula object with a binary response.
#' @param weights    Numeric vector of kernel weights (same length as
#'   \code{nrow(local_data)}).
#' @param family     Character scalar reserved for future extension;
#'   currently only \code{"binomial"} is supported.
#'
#' @return A \code{glm} object.
#' @keywords internal
fit_logistic_pooling <- function(local_data, formula, weights,
                                 family = "binomial") {
  .local <- local_data
  .local[[".gwpr_wt"]] <- weights
  .wts   <- weights

  do.call(
    stats::glm,
    list(
      formula = formula,
      data    = quote(.local),
      family  = binomial(link = "logit"),
      weights = .wts
    ),
    envir = environment()
  )
}

# ---------------------------------------------------------------------------
# fit_logistic_fixed
# ---------------------------------------------------------------------------

#' Fit a fixed-effects logistic model via fixest::feglm
#'
#' @param local_data A \code{data.frame} containing all formula variables plus
#'   the panel index columns.
#' @param formula    A formula object with a binary response (no fixed-effect
#'   terms; those are added automatically from \code{effect}).
#' @param effect     Character scalar: one of \code{"individual"},
#'   \code{"time"}, \code{"two-way"}, \code{"nested"}.
#' @param weights    Numeric vector of kernel weights.
#' @param id_col     Name of the individual ID column.
#' @param time_col   Name of the time column.
#' @param family     Character scalar reserved for future extension.
#'
#' @return A \code{fixest} object.
#' @keywords internal
fit_logistic_fixed <- function(local_data, formula, effect, weights,
                               id_col, time_col, family = "binomial") {
  fe_fml <- effect_to_feglm_fml(formula, effect, id_col, time_col)
  .local <- local_data
  .local[[".gwpr_wt"]] <- weights

  fixest::feglm(
    fml     = fe_fml,
    data    = .local,
    family  = binomial(link = "logit"),
    weights = ~.gwpr_wt
  )
}

# ---------------------------------------------------------------------------
# fit_logistic_random
# ---------------------------------------------------------------------------

#' Fit a random-effects logistic model via glmmTMB::glmmTMB
#'
#' @param local_data A \code{data.frame} containing all formula variables plus
#'   the panel index columns.
#' @param formula    A formula object with a binary response (no random-effect
#'   terms; those are added automatically from \code{effect}).
#' @param effect     Character scalar: one of \code{"individual"},
#'   \code{"time"}, \code{"two-way"}, \code{"nested"}.
#' @param weights    Numeric vector of kernel weights.
#' @param id_col     Name of the individual ID column.
#' @param time_col   Name of the time column.
#' @param family     Character scalar reserved for future extension.
#'
#' @return A \code{glmmTMB} object.
#' @keywords internal
fit_logistic_random <- function(local_data, formula, effect, weights,
                                id_col, time_col, family = "binomial") {
  if (!requireNamespace("glmmTMB", quietly = TRUE)) {
    stop(
      "Package 'glmmTMB' is required for random-effects logistic models. ",
      "Install it with: install.packages('glmmTMB')",
      call. = FALSE
    )
  }

  re_fml <- effect_to_glmmtmb_fml(formula, effect, id_col, time_col)
  .local <- local_data
  # Store weights as a plain column; glmmTMB with binomial(link="logit") and
  # a 0/1 response interprets `weights` as prior observation weights when the
  # response is a single-column 0/1 integer vector.  To avoid the
  # "non-integer #successes" warning we round to integers (proportional to
  # the kernel weights, scaled so the largest weight = 100).
  max_wt <- max(weights, na.rm = TRUE)
  if (is.finite(max_wt) && max_wt > 0) {
    int_wts <- as.integer(round(weights / max_wt * 100))
    int_wts[int_wts < 1L] <- 1L
  } else {
    int_wts <- rep(1L, length(weights))
  }
  .local[[".gwpr_wt"]] <- int_wts

  glmmTMB::glmmTMB(
    formula = re_fml,
    data    = .local,
    family  = binomial(link = "logit"),
    weights = .local[[".gwpr_wt"]]
  )
}

# ---------------------------------------------------------------------------
# fit_logistic_local_model
# ---------------------------------------------------------------------------

#' Fit a single local binary logistic panel model
#'
#' Dispatches to the correct backend (\code{pooling}, \code{fixed}, or
#' \code{random}) based on \code{model}, wrapping execution in a
#' \code{tryCatch} so that convergence failures or complete-separation errors
#' are captured rather than propagated.
#'
#' @param local_data  A \code{data.frame} with all formula variables and panel
#'   index columns.
#' @param formula     A formula object.
#' @param model       Character scalar: \code{"pooling"}, \code{"fixed"}, or
#'   \code{"random"}.
#' @param effect      Character scalar: \code{"individual"}, \code{"time"},
#'   \code{"two-way"}, or \code{"nested"}.
#' @param weights     Numeric vector of kernel weights aligned with
#'   \code{nrow(local_data)}.
#' @param index       Character vector of length 2: \code{c(id_col, time_col)}.
#' @param threshold   Numeric scalar; classification threshold (default 0.5).
#' @param family      Character scalar reserved for future extension (currently
#'   only \code{"binomial"} is implemented).
#'
#' @return A list with elements:
#' \describe{
#'   \item{\code{fit}}{The fitted model object, or \code{NULL} on failure.}
#'   \item{\code{status}}{\code{"ok"} or \code{"failed"}.}
#'   \item{\code{error}}{\code{NULL} or character error message.}
#'   \item{\code{metadata}}{Named list with additional fitting metadata.}
#' }
#' @keywords internal
fit_logistic_local_model <- function(local_data, formula, model, effect,
                                     weights, index,
                                     threshold = 0.5,
                                     family = "binomial") {
  model  <- match.arg(model,  choices = c("pooling", "fixed", "random"))
  effect <- match.arg(effect, choices = c("individual", "time", "two-way", "nested"))

  # Validate effect BEFORE tryCatch for nested/unknown — propagate to caller
  if (model != "pooling") {
    # This will stop() for nested or unknown effect
    if (model == "fixed") {
      effect_to_feglm_fml(formula, effect, index[1L], index[2L])
    } else {
      effect_to_glmmtmb_fml(formula, effect, index[1L], index[2L])
    }
  }

  # Validate response
  resp_name <- all.vars(formula)[1L]
  if (resp_name %in% names(local_data)) {
    validate_binary_response(local_data[[resp_name]])
  }

  metadata <- list()
  id_col   <- index[1L]
  time_col <- index[2L]

  result <- tryCatch({
    fit <- switch(
      model,
      pooling = fit_logistic_pooling(local_data, formula, weights, family),
      fixed   = fit_logistic_fixed(local_data, formula, effect, weights,
                                   id_col, time_col, family),
      random  = fit_logistic_random(local_data, formula, effect, weights,
                                    id_col, time_col, family)
    )
    list(fit = fit, status = "ok", error = NULL, metadata = metadata)
  }, error = function(e) {
    list(fit = NULL, status = "failed", error = conditionMessage(e),
         metadata = metadata)
  })

  result
}

# ---------------------------------------------------------------------------
# predict_logistic_local_model
# ---------------------------------------------------------------------------

#' Predict probabilities for a local logistic model
#'
#' Returns predicted probabilities (type = \code{"response"}) for the rows of
#' \code{local_data}, using the fitted model stored in \code{local_result}.
#'
#' @param local_result A list as returned by \code{fit_logistic_local_model()}.
#' @param local_data   A \code{data.frame} used for prediction.
#'
#' @return Numeric vector of probabilities in \code{[0, 1]}.  Returns a vector
#'   of \code{NA_real_} on failure.
#' @keywords internal
predict_logistic_local_model <- function(local_result, local_data) {
  if (local_result$status != "ok" || is.null(local_result$fit)) {
    return(rep(NA_real_, nrow(local_data)))
  }

  fit <- local_result$fit

  tryCatch({
    if (inherits(fit, "fixest")) {
      # fixest::predict returns vector of length equal to the training data
      pv <- stats::predict(fit, newdata = local_data, type = "response")
    } else if (inherits(fit, "glmmTMB")) {
      pv <- stats::predict(fit, newdata = local_data, type = "response",
                           allow.new.levels = TRUE)
    } else {
      # glm
      pv <- stats::predict(fit, newdata = local_data, type = "response")
    }
    as.numeric(pv)
  }, error = function(e) {
    rep(NA_real_, nrow(local_data))
  })
}

# ---------------------------------------------------------------------------
# extract_logistic_local_result
# ---------------------------------------------------------------------------

#' Extract coefficients and diagnostics from a local logistic model
#'
#' Computes predicted probabilities, predicted classes, Pearson residuals,
#' and extracts coefficient estimates.  Pearson residuals are defined as
#' \code{(y - p) / sqrt(p * (1 - p))} with \code{p} clipped to
#' \code{[eps, 1 - eps]} to avoid division by zero.
#'
#' @param local_result  A list as returned by \code{fit_logistic_local_model()}.
#' @param local_data    The \code{data.frame} used to fit the model (needed to
#'   extract \code{y} and to predict).
#' @param formula       The model formula (used to extract the response name).
#' @param threshold     Numeric scalar; classification threshold (default 0.5).
#' @param eps           Numeric scalar; clipping bound for probability (default
#'   \code{1e-15}).
#'
#' @return A named list with elements:
#' \describe{
#'   \item{\code{coefficients}}{Named numeric vector of local coefficient
#'     estimates, or \code{NA_real_} on failure.}
#'   \item{\code{prob}}{Numeric vector of predicted probabilities for the local
#'     data rows, or \code{NA_real_} on failure.}
#'   \item{\code{class_pred}}{Integer vector (0/1) of predicted classes, or
#'     \code{NA_real_} on failure.}
#'   \item{\code{pearson_resid}}{Numeric vector of Pearson residuals, or
#'     \code{NA_real_} on failure.}
#'   \item{\code{status}}{\code{"ok"} or \code{"failed"}.}
#'   \item{\code{error}}{\code{NULL} or character error message.}
#' }
#' @keywords internal
extract_logistic_local_result <- function(local_result, local_data, formula,
                                          threshold = 0.5, eps = 1e-15) {
  status <- local_result$status
  err    <- local_result$error
  n      <- nrow(local_data)

  if (status != "ok" || is.null(local_result$fit)) {
    return(list(
      coefficients  = NA_real_,
      prob          = rep(NA_real_, n),
      class_pred    = rep(NA_integer_, n),
      pearson_resid = rep(NA_real_, n),
      status        = "failed",
      error         = err
    ))
  }

  fit <- local_result$fit

  # Coefficients
  coefs <- tryCatch(stats::coef(fit), error = function(e) NA_real_)
  # For glmmTMB, coef() returns a list; extract the fixed part
  if (is.list(coefs)) {
    coefs <- tryCatch(
      {
        fe <- glmmTMB::fixef(fit)
        if (is.list(fe)) unlist(fe$cond) else unlist(fe)
      },
      error = function(e) NA_real_
    )
  }

  # Predicted probabilities
  prob <- predict_logistic_local_model(local_result, local_data)

  # Predicted class
  class_pred <- tryCatch(
    as.integer(prob >= threshold),
    error = function(e) rep(NA_integer_, n)
  )

  # Response y
  resp_name <- all.vars(formula)[1L]
  y_raw <- local_data[[resp_name]]
  y <- tryCatch(
    standardize_logistic_response(y_raw),
    error = function(e) rep(NA_integer_, n)
  )

  # Pearson residuals: (y - p) / sqrt(p * (1-p))
  pearson_resid <- tryCatch({
    p_clipped <- pmax(pmin(prob, 1 - eps), eps)
    (y - p_clipped) / sqrt(p_clipped * (1 - p_clipped))
  }, error = function(e) rep(NA_real_, n))

  list(
    coefficients  = coefs,
    prob          = prob,
    class_pred    = class_pred,
    pearson_resid = pearson_resid,
    status        = "ok",
    error         = NULL
  )
}

# ---------------------------------------------------------------------------
# fit_logistic_gwpr  (main entry point)
# ---------------------------------------------------------------------------

#' Fit a Geographically Weighted Panel Binary Logistic Regression
#'
#' Iterates over each spatial unit (focus), computes kernel weights, fits a
#' local panel logistic model, and collects coefficients, probabilities,
#' predicted classes, Pearson residuals, and summary metrics.
#'
#' Supports three panel model backends:
#' \itemize{
#'   \item \code{"pooling"} — \code{stats::glm(family = binomial)}.
#'   \item \code{"fixed"}   — \code{fixest::feglm()}.
#'   \item \code{"random"}  — \code{glmmTMB::glmmTMB(family = binomial)}.
#' }
#'
#' @param context         A \code{gwpr_context} list containing at minimum:
#'   \code{formula}, \code{model}, \code{effect}, \code{panel_data},
#'   \code{coords}, \code{id}, \code{time}, \code{kernel}, \code{adaptive},
#'   \code{threshold}, and \code{workers}.
#' @param bandwidth       Numeric scalar.  Fixed distance or adaptive
#'   (k-nearest-neighbour) bandwidth.
#' @param weights_context Optional pre-built distance context as returned by
#'   \code{build_distance_context()}.  If \code{NULL}, one is built from
#'   \code{context$coords}.
#' @param family          Character scalar reserved for future multi-class
#'   extension; currently only \code{"binomial"} is accepted (default).
#'
#' @return A named list with elements:
#' \describe{
#'   \item{\code{local_results}}{A list of per-focus extracted results (one per
#'     unique spatial unit).}
#'   \item{\code{predictions}}{Named numeric vector of predicted probabilities
#'     for each focus unit row (indexed by user ID).}
#'   \item{\code{class_pred}}{Named integer vector of predicted classes.}
#'   \item{\code{pearson_resid}}{Named numeric vector of Pearson residuals.}
#'   \item{\code{metrics}}{A list from \code{compute_logistic_metrics()} over
#'     all focus observations.}
#'   \item{\code{metadata}}{A list with fitting metadata and any per-focus
#'     warnings.}
#' }
#' @noRd
fit_logistic_gwpr <- function(context, bandwidth, weights_context = NULL,
                               family = "binomial") {
  if (!identical(family, "binomial")) {
    stop(
      "Only family = 'binomial' is supported in GWPR.light 1.0.0. ",
      "Multi-class logistic regression is reserved for a future release.",
      call. = FALSE
    )
  }

  formula    <- context$formula
  model      <- context$model
  effect     <- context$effect
  panel_data <- context$panel_data
  coords     <- context$coords
  id_col     <- context$id
  time_col   <- context$time
  kernel     <- context$kernel
  adaptive   <- context$adaptive
  threshold  <- .null_coalesce(context$threshold, 0.5)
  workers    <- .null_coalesce(context$workers, 1L)

  index <- c(id_col, time_col)

  # Build distance context if not supplied
  if (is.null(weights_context)) {
    unit_ids <- rownames(coords)
    if (is.null(unit_ids)) {
      unit_ids <- as.character(seq_len(nrow(coords)))
    }
    weights_context <- build_distance_context(
      coords  = coords,
      ids     = unit_ids,
      longlat = FALSE,
      cache   = TRUE
    )
  }

  unique_ids <- weights_context$ids

  # Helper: fit one focus unit
  fit_one_focus <- function(focus_id) {
    distances <- get_local_distances(weights_context, focus_id = focus_id)
    kw        <- compute_kernel_weights(
      distance  = distances,
      bandwidth = bandwidth,
      kernel    = kernel,
      adaptive  = adaptive
    )
    names(kw) <- unique_ids

    row_weights <- kw[as.character(panel_data[[id_col]])]
    row_weights[is.na(row_weights)] <- 0

    local_res <- fit_logistic_local_model(
      local_data = panel_data,
      formula    = formula,
      model      = model,
      effect     = effect,
      weights    = row_weights,
      index      = index,
      threshold  = threshold,
      family     = family
    )

    extracted <- extract_logistic_local_result(
      local_result = local_res,
      local_data   = panel_data,
      formula      = formula,
      threshold    = threshold
    )

    # Subset to focus rows only
    focus_rows <- which(as.character(panel_data[[id_col]]) == focus_id)
    focus_data <- panel_data[focus_rows, , drop = FALSE]

    focus_prob  <- extracted$prob[focus_rows]
    focus_class <- extracted$class_pred[focus_rows]
    focus_resid <- extracted$pearson_resid[focus_rows]
    focus_y     <- tryCatch(
      standardize_logistic_response(focus_data[[all.vars(formula)[1L]]]),
      error = function(e) rep(NA_integer_, length(focus_rows))
    )

    list(
      focus_id   = focus_id,
      extracted  = extracted,
      y          = focus_y,
      prob       = focus_prob,
      class_pred = focus_class,
      resid      = focus_resid,
      metadata   = local_res$metadata
    )
  }

  # Run (serial or parallel)
  if (workers == 1L) {
    all_results <- lapply(unique_ids, fit_one_focus)
  } else {
    all_results <- parallel::mclapply(
      unique_ids, fit_one_focus,
      mc.cores = workers, mc.set.seed = TRUE
    )
  }

  names(all_results) <- unique_ids

  # Assemble global vectors
  all_y     <- unlist(lapply(all_results, `[[`, "y"),          use.names = FALSE)
  all_prob  <- unlist(lapply(all_results, `[[`, "prob"),       use.names = FALSE)
  all_class <- unlist(lapply(all_results, `[[`, "class_pred"), use.names = FALSE)
  all_resid <- unlist(lapply(all_results, `[[`, "resid"),      use.names = FALSE)

  focus_lengths <- vapply(all_results, function(r) length(r$y), integer(1L))
  id_names <- rep(unique_ids, focus_lengths)

  pred_named  <- stats::setNames(all_prob,  id_names)
  class_named <- stats::setNames(all_class, id_names)
  resid_named <- stats::setNames(all_resid, id_names)

  # Remove NAs (from failed local models) before computing global metrics
  keep <- !is.na(all_y) & !is.na(all_prob)
  metrics <- if (sum(keep) > 0L) {
    suppressWarnings(
      compute_logistic_metrics(all_y[keep], all_prob[keep], threshold = threshold)
    )
  } else {
    list(log_loss = NA_real_, accuracy = NA_real_, precision = NA_real_,
         recall = NA_real_, f1_score = NA_real_)
  }

  local_results <- lapply(all_results, `[[`, "extracted")
  names(local_results) <- unique_ids

  meta <- list(
    n_focuses          = length(unique_ids),
    n_failed           = sum(vapply(local_results,
                                   function(r) r$status == "failed",
                                   logical(1L))),
    per_focus_metadata = lapply(all_results, `[[`, "metadata")
  )

  list(
    local_results = local_results,
    predictions   = pred_named,
    class_pred    = class_named,
    pearson_resid = resid_named,
    metrics       = metrics,
    metadata      = meta
  )
}

# ---------------------------------------------------------------------------
# Utility: .null_coalesce  (avoids re-definition of %||% across files)
# ---------------------------------------------------------------------------

.null_coalesce <- function(x, y) if (is.null(x)) y else x

## tests/testthat/test-model-logistic.R
## Tests for the Binary Panel Logistic Engine (R/model_logistic.R)
##
## All tests use small simulated panel data.  workers = 1 (no parallelism).
## glmmTMB random effects tests are wrapped in skip conditions when the
## package is unavailable or fails due to numerical issues (e.g. ABI mismatch).

# ---------------------------------------------------------------------------
# Helpers: small simulated binary panel data
# ---------------------------------------------------------------------------

# 4 spatial units, 5 time periods each => 20 rows
make_bin_panel_data <- function(seed = 42L) {
  n_units <- 4L
  n_times <- 5L
  id   <- rep(as.character(1:n_units), each = n_times)
  time <- rep(1:n_times, times = n_units)

  # Deterministic, non-separable binary pattern.  The local GWPR tests can
  # give one spatial unit all the effective weight, so each unit must contain
  # both classes without being linearly separable by x1/x2.
  base_x1 <- c(-1, -1,  1, 1, 0)
  base_x2 <- c(-1,  1, -1, 1, 0)
  y_by_unit <- rbind(
    c(0, 1, 1, 0, 1),
    c(1, 0, 0, 1, 0),
    c(0, 1, 1, 0, 0),
    c(1, 0, 0, 1, 1)
  )

  unit_offset <- rep(seq(-0.15, 0.15, length.out = n_units), each = n_times)
  x1 <- rep(base_x1, times = n_units) + unit_offset
  x2 <- rep(base_x2, times = n_units) - unit_offset
  y  <- as.integer(t(y_by_unit))

  data.frame(id = id, time = time, y = y, x1 = x1, x2 = x2,
             stringsAsFactors = FALSE)
}

# 4 units on a small 2x2 grid
make_coords_4 <- function() {
  matrix(
    c(0, 0,
      1, 0,
      0, 1,
      1, 1),
    ncol = 2, byrow = TRUE,
    dimnames = list(c("1", "2", "3", "4"), c("X", "Y"))
  )
}

make_logistic_context <- function(model = "pooling", effect = "individual",
                                   kernel = "bisquare", adaptive = FALSE,
                                   threshold = 0.5, seed = NULL) {
  new_gwpr_context(
    formula      = y ~ x1 + x2,
    family       = "binomial",
    model        = model,
    effect       = effect,
    id           = "id",
    time         = "time",
    kernel       = kernel,
    adaptive     = adaptive,
    threshold    = threshold,
    workers      = 1L,
    seed         = seed,
    panel_data   = make_bin_panel_data(),
    coords       = make_coords_4()
  )
}

# ---------------------------------------------------------------------------
# validate_binary_response
# ---------------------------------------------------------------------------

test_that("validate_binary_response: passes for numeric 0/1", {
  expect_true(validate_binary_response(c(0L, 1L, 0L, 1L)))
})

test_that("validate_binary_response: passes for two-level factor", {
  y <- factor(c("no", "yes", "no"), levels = c("no", "yes"))
  expect_true(validate_binary_response(y))
})

test_that("validate_binary_response: stops for 3-level factor", {
  y <- factor(c("a", "b", "c"))
  expect_error(validate_binary_response(y), regexp = "Multi-class")
})

# ---------------------------------------------------------------------------
# standardize_logistic_response
# ---------------------------------------------------------------------------

test_that("standardize_logistic_response: numeric 0/1 passes through", {
  y <- c(0, 1, 0, 1)
  result <- standardize_logistic_response(y)
  expect_equal(result, c(0L, 1L, 0L, 1L))
})

test_that("standardize_logistic_response: two-level factor maps to 0/1", {
  y <- factor(c("no", "yes", "no", "yes"), levels = c("no", "yes"))
  result <- standardize_logistic_response(y)
  expect_equal(result, c(0L, 1L, 0L, 1L))
})

test_that("standardize_logistic_response: logical maps to 0/1", {
  y <- c(FALSE, TRUE, FALSE, TRUE)
  result <- standardize_logistic_response(y)
  expect_equal(result, c(0L, 1L, 0L, 1L))
})

test_that("standardize_logistic_response: multi-class factor raises error", {
  y <- factor(c("a", "b", "c", "a"))
  expect_error(standardize_logistic_response(y), regexp = "Multi-class")
})

# ---------------------------------------------------------------------------
# effect_to_feglm_fml
# ---------------------------------------------------------------------------

test_that("effect_to_feglm_fml: individual maps id to fixed effect", {
  fml <- effect_to_feglm_fml(y ~ x1 + x2, "individual", "id", "time")
  expect_true(grepl("id", deparse(fml)))
})

test_that("effect_to_feglm_fml: time maps time to fixed effect", {
  fml <- effect_to_feglm_fml(y ~ x1 + x2, "time", "id", "time")
  expect_true(grepl("time", deparse(fml)))
})

test_that("effect_to_feglm_fml: two-way includes both id and time", {
  fml <- effect_to_feglm_fml(y ~ x1 + x2, "two-way", "id", "time")
  fml_str <- deparse(fml)
  expect_true(grepl("id", fml_str) && grepl("time", fml_str))
})

test_that("effect_to_feglm_fml: nested stops with informative error", {
  expect_error(
    effect_to_feglm_fml(y ~ x1 + x2, "nested", "id", "time"),
    regexp = "nested"
  )
})

# ---------------------------------------------------------------------------
# effect_to_glmmtmb_fml
# ---------------------------------------------------------------------------

test_that("effect_to_glmmtmb_fml: individual adds (1|id)", {
  fml <- effect_to_glmmtmb_fml(y ~ x1, "individual", "id", "time")
  expect_true(grepl("1 | id", deparse(fml)))
})

test_that("effect_to_glmmtmb_fml: time adds (1|time)", {
  fml <- effect_to_glmmtmb_fml(y ~ x1, "time", "id", "time")
  expect_true(grepl("1 | time", deparse(fml)))
})

test_that("effect_to_glmmtmb_fml: two-way adds both random intercepts", {
  fml <- effect_to_glmmtmb_fml(y ~ x1, "two-way", "id", "time")
  fml_str <- deparse(fml)
  expect_true(grepl("1 | id", fml_str) && grepl("1 | time", fml_str))
})

test_that("effect_to_glmmtmb_fml: nested stops with informative error", {
  expect_error(
    effect_to_glmmtmb_fml(y ~ x1, "nested", "id", "time"),
    regexp = "nested"
  )
})

# ---------------------------------------------------------------------------
# fit_logistic_pooling
# ---------------------------------------------------------------------------

test_that("fit_logistic_pooling: returns glm object", {
  pd  <- make_bin_panel_data()
  wts <- rep(1, nrow(pd))
  fit <- fit_logistic_pooling(pd, y ~ x1 + x2, wts)
  expect_true(inherits(fit, "glm"))
  expect_equal(family(fit)$family, "binomial")
})

test_that("fit_logistic_pooling: works with numeric 0/1 response", {
  pd  <- make_bin_panel_data()
  wts <- rep(1, nrow(pd))
  expect_no_error(fit_logistic_pooling(pd, y ~ x1 + x2, wts))
})

test_that("fit_logistic_pooling: works with two-level factor response", {
  pd  <- make_bin_panel_data()
  pd$y_fac <- factor(pd$y, levels = c(0, 1), labels = c("no", "yes"))
  wts <- rep(1, nrow(pd))
  fit <- fit_logistic_pooling(pd, y_fac ~ x1 + x2, wts)
  expect_true(inherits(fit, "glm"))
})

# ---------------------------------------------------------------------------
# fit_logistic_fixed
# ---------------------------------------------------------------------------

test_that("fit_logistic_fixed: returns fixest object for individual effect", {
  pd  <- make_bin_panel_data()
  wts <- rep(1, nrow(pd))
  fit <- fit_logistic_fixed(pd, y ~ x1 + x2, "individual", wts, "id", "time")
  expect_true(inherits(fit, "fixest"))
})

test_that("fit_logistic_fixed: returns fixest object for time effect", {
  pd  <- make_bin_panel_data()
  wts <- rep(1, nrow(pd))
  fit <- fit_logistic_fixed(pd, y ~ x1 + x2, "time", wts, "id", "time")
  expect_true(inherits(fit, "fixest"))
})

test_that("fit_logistic_fixed: two-way effect smoke test", {
  pd  <- make_bin_panel_data()
  wts <- rep(1, nrow(pd))
  # Two-way may fail due to collinearity on small data; capture gracefully
  result <- tryCatch(
    fit_logistic_fixed(pd, y ~ x1 + x2, "two-way", wts, "id", "time"),
    error = function(e) NULL
  )
  # Either succeeded (fixest object) or failed (NULL) — both acceptable
  if (!is.null(result)) {
    expect_true(inherits(result, "fixest"))
  } else {
    succeed("Two-way FE failed gracefully on small data")
  }
})

test_that("fit_logistic_fixed: nested effect stops with informative error", {
  pd  <- make_bin_panel_data()
  wts <- rep(1, nrow(pd))
  expect_error(
    fit_logistic_fixed(pd, y ~ x1 + x2, "nested", wts, "id", "time"),
    regexp = "nested"
  )
})

# ---------------------------------------------------------------------------
# fit_logistic_random
# ---------------------------------------------------------------------------

test_that("fit_logistic_random: nested effect stops with informative error", {
  pd  <- make_bin_panel_data()
  wts <- rep(1, nrow(pd))
  expect_error(
    fit_logistic_random(pd, y ~ x1 + x2, "nested", wts, "id", "time"),
    regexp = "nested"
  )
})

test_that("fit_logistic_random: individual effect smoke test (may fail due to convergence)", {
  skip_if_not_installed("glmmTMB")
  pd  <- make_bin_panel_data()
  wts <- rep(1, nrow(pd))
  result <- tryCatch(
    fit_logistic_random(pd, y ~ x1 + x2, "individual", wts, "id", "time"),
    error = function(e) NULL
  )
  # Either a glmmTMB object or NULL (convergence failure) — both acceptable
  if (!is.null(result)) {
    expect_true(inherits(result, "glmmTMB"))
  } else {
    succeed("glmmTMB fit failed gracefully (convergence/numerical issue)")
  }
})

# ---------------------------------------------------------------------------
# fit_logistic_local_model
# ---------------------------------------------------------------------------

test_that("fit_logistic_local_model: pooling ok status with numeric response", {
  pd  <- make_bin_panel_data()
  wts <- rep(1, nrow(pd))
  res <- fit_logistic_local_model(
    local_data = pd,
    formula    = y ~ x1 + x2,
    model      = "pooling",
    effect     = "individual",
    weights    = wts,
    index      = c("id", "time")
  )
  expect_equal(res$status, "ok")
  expect_false(is.null(res$fit))
  expect_true(inherits(res$fit, "glm"))
})

test_that("fit_logistic_local_model: pooling ok status with factor response", {
  pd  <- make_bin_panel_data()
  pd$y_fac <- factor(pd$y, levels = c(0, 1), labels = c("no", "yes"))
  wts <- rep(1, nrow(pd))
  res <- fit_logistic_local_model(
    local_data = pd,
    formula    = y_fac ~ x1 + x2,
    model      = "pooling",
    effect     = "individual",
    weights    = wts,
    index      = c("id", "time")
  )
  expect_equal(res$status, "ok")
  expect_true(inherits(res$fit, "glm"))
})

test_that("fit_logistic_local_model: multi-class factor response stops outside tryCatch", {
  pd      <- make_bin_panel_data()
  pd$y_mc <- factor(sample(c("a", "b", "c"), nrow(pd), replace = TRUE))
  wts     <- rep(1, nrow(pd))
  # validate_binary_response is called before tryCatch so it propagates
  expect_error(
    fit_logistic_local_model(
      local_data = pd,
      formula    = y_mc ~ x1 + x2,
      model      = "pooling",
      effect     = "individual",
      weights    = wts,
      index      = c("id", "time")
    ),
    regexp = "Multi-class"
  )
})

test_that("fit_logistic_local_model: bad formula gives failed status, not error", {
  pd  <- make_bin_panel_data()
  wts <- rep(1, nrow(pd))
  res <- fit_logistic_local_model(
    local_data = pd,
    formula    = y ~ nonexistent_var,
    model      = "pooling",
    effect     = "individual",
    weights    = wts,
    index      = c("id", "time")
  )
  expect_equal(res$status, "failed")
  expect_null(res$fit)
  expect_true(is.character(res$error) && nchar(res$error) > 0L)
})

test_that("fit_logistic_local_model: fixed individual effect returns ok status", {
  pd  <- make_bin_panel_data()
  wts <- rep(1, nrow(pd))
  res <- fit_logistic_local_model(
    local_data = pd,
    formula    = y ~ x1 + x2,
    model      = "fixed",
    effect     = "individual",
    weights    = wts,
    index      = c("id", "time")
  )
  expect_equal(res$status, "ok")
  expect_true(inherits(res$fit, "fixest"))
})

test_that("fit_logistic_local_model: nested effect propagates error before tryCatch", {
  pd  <- make_bin_panel_data()
  wts <- rep(1, nrow(pd))
  expect_error(
    fit_logistic_local_model(
      local_data = pd,
      formula    = y ~ x1 + x2,
      model      = "fixed",
      effect     = "nested",
      weights    = wts,
      index      = c("id", "time")
    ),
    regexp = "nested"
  )
})

test_that("fit_logistic_local_model: random model handles failure gracefully", {
  skip_if_not_installed("glmmTMB")
  pd  <- make_bin_panel_data()
  wts <- rep(1, nrow(pd))
  # If glmmTMB fails numerically, status should be "failed" not thrown error
  res <- fit_logistic_local_model(
    local_data = pd,
    formula    = y ~ x1 + x2,
    model      = "random",
    effect     = "individual",
    weights    = wts,
    index      = c("id", "time")
  )
  # Either ok or failed — but must not throw
  expect_true(res$status %in% c("ok", "failed"))
  if (res$status == "failed") {
    expect_null(res$fit)
    expect_true(is.character(res$error))
  }
})

# ---------------------------------------------------------------------------
# predict_logistic_local_model
# ---------------------------------------------------------------------------

test_that("predict_logistic_local_model: returns numeric in [0,1]", {
  pd  <- make_bin_panel_data()
  wts <- rep(1, nrow(pd))
  res <- fit_logistic_local_model(
    local_data = pd, formula = y ~ x1 + x2,
    model = "pooling", effect = "individual",
    weights = wts, index = c("id", "time")
  )
  preds <- predict_logistic_local_model(res, pd)
  expect_true(is.numeric(preds))
  expect_equal(length(preds), nrow(pd))
  expect_true(all(preds >= 0 & preds <= 1, na.rm = TRUE))
})

test_that("predict_logistic_local_model: failed model returns NA vector", {
  failed_res <- list(status = "failed", fit = NULL, error = "err",
                     metadata = list())
  pd    <- make_bin_panel_data()
  preds <- predict_logistic_local_model(failed_res, pd)
  expect_true(all(is.na(preds)))
  expect_equal(length(preds), nrow(pd))
})

# ---------------------------------------------------------------------------
# extract_logistic_local_result
# ---------------------------------------------------------------------------

test_that("extract_logistic_local_result: ok result has correct structure", {
  pd  <- make_bin_panel_data()
  wts <- rep(1, nrow(pd))
  local_res <- fit_logistic_local_model(
    local_data = pd, formula = y ~ x1 + x2,
    model = "pooling", effect = "individual",
    weights = wts, index = c("id", "time")
  )
  ext <- extract_logistic_local_result(local_res, pd, y ~ x1 + x2)

  expect_equal(ext$status, "ok")
  expect_true(is.numeric(ext$coefficients))
  expect_true(is.numeric(ext$prob))
  expect_equal(length(ext$prob), nrow(pd))
  expect_true(all(ext$prob >= 0 & ext$prob <= 1, na.rm = TRUE))
  expect_true(is.integer(ext$class_pred) || is.numeric(ext$class_pred))
  expect_true(all(ext$class_pred %in% c(0L, 1L, NA_integer_)))
  expect_true(is.numeric(ext$pearson_resid))
  expect_true(all(is.finite(ext$pearson_resid)))  # no Inf due to clipping
})

test_that("extract_logistic_local_result: failed result returns NA placeholders", {
  failed_res <- list(status = "failed", fit = NULL, error = "test error",
                     metadata = list())
  pd  <- make_bin_panel_data()
  ext <- extract_logistic_local_result(failed_res, pd, y ~ x1 + x2)

  expect_equal(ext$status, "failed")
  expect_true(is.na(ext$coefficients) || all(is.na(ext$coefficients)))
  expect_true(all(is.na(ext$prob)))
  expect_true(all(is.na(ext$class_pred)))
  expect_true(all(is.na(ext$pearson_resid)))
  expect_equal(ext$error, "test error")
})

test_that("extract_logistic_local_result: Pearson residuals finite (no Inf)", {
  # Use data with extreme probabilities to verify clipping
  pd  <- make_bin_panel_data()
  wts <- rep(1, nrow(pd))
  local_res <- fit_logistic_local_model(
    local_data = pd, formula = y ~ x1 + x2,
    model = "pooling", effect = "individual",
    weights = wts, index = c("id", "time")
  )
  ext <- extract_logistic_local_result(local_res, pd, y ~ x1 + x2)
  # Verify Pearson residuals are all finite (clipping prevents Inf)
  expect_true(all(is.finite(ext$pearson_resid)))
})

# ---------------------------------------------------------------------------
# threshold effect on class_pred
# ---------------------------------------------------------------------------

test_that("threshold = 0.5 and threshold = 0.0 give different class predictions", {
  pd  <- make_bin_panel_data()
  wts <- rep(1, nrow(pd))
  local_res <- fit_logistic_local_model(
    local_data = pd, formula = y ~ x1 + x2,
    model = "pooling", effect = "individual",
    weights = wts, index = c("id", "time")
  )
  ext_05 <- extract_logistic_local_result(local_res, pd, y ~ x1 + x2,
                                           threshold = 0.5)
  ext_00 <- extract_logistic_local_result(local_res, pd, y ~ x1 + x2,
                                           threshold = 0.0)

  # threshold = 0 classifies everything as 1 (all prob >= 0)
  expect_true(all(ext_00$class_pred == 1L))
  # threshold = 0.5 gives a mix
  expect_false(all(ext_05$class_pred == 1L) && all(ext_05$class_pred == 0L))
})

test_that("threshold = 1.0 classifies everything as 0", {
  pd  <- make_bin_panel_data()
  wts <- rep(1, nrow(pd))
  local_res <- fit_logistic_local_model(
    local_data = pd, formula = y ~ x1 + x2,
    model = "pooling", effect = "individual",
    weights = wts, index = c("id", "time")
  )
  ext <- extract_logistic_local_result(local_res, pd, y ~ x1 + x2,
                                        threshold = 1.0)
  # Probabilities are always < 1, so all should be classified as 0
  expect_true(all(ext$class_pred == 0L))
})

# ---------------------------------------------------------------------------
# fit_logistic_gwpr — smoke tests (pooling / fixed / random)
# ---------------------------------------------------------------------------

test_that("fit_logistic_gwpr: pooling smoke test returns correct structure", {
  ctx <- make_logistic_context(model = "pooling", effect = "individual",
                                kernel = "bisquare", adaptive = TRUE)
  res <- fit_logistic_gwpr(ctx, bandwidth = 3L)

  expect_true(is.list(res))
  expect_true("local_results" %in% names(res))
  expect_true("predictions"   %in% names(res))
  expect_true("class_pred"    %in% names(res))
  expect_true("pearson_resid" %in% names(res))
  expect_true("metrics"       %in% names(res))
  expect_true("metadata"      %in% names(res))

  # 4 spatial units => 4 local results
  expect_equal(length(res$local_results), 4L)

  # Logistic metrics fields
  expect_true(all(c("log_loss", "accuracy", "precision", "recall", "f1_score")
                  %in% names(res$metrics)))

  # Predictions should be in [0, 1]
  probs <- res$predictions[!is.na(res$predictions)]
  expect_true(all(probs >= 0 & probs <= 1))
})

test_that("fit_logistic_gwpr: fixed effects individual smoke test", {
  ctx <- make_logistic_context(model = "fixed", effect = "individual",
                                kernel = "gaussian", adaptive = FALSE)
  expect_no_error({
    res <- fit_logistic_gwpr(ctx, bandwidth = 2)
  })
  expect_equal(length(res$local_results), 4L)
})

test_that("fit_logistic_gwpr: random effects smoke test (may fail due to convergence)", {
  skip_if_not_installed("glmmTMB")
  ctx <- make_logistic_context(model = "random", effect = "individual",
                                kernel = "bisquare", adaptive = TRUE)
  # Should not throw — failures captured per-focus
  expect_no_error({
    res <- fit_logistic_gwpr(ctx, bandwidth = 3L)
  })
  # Structure should be intact even if all models failed
  expect_equal(length(res$local_results), 4L)
  statuses <- vapply(res$local_results, `[[`, character(1L), "status")
  expect_true(all(statuses %in% c("ok", "failed")))
})

# ---------------------------------------------------------------------------
# fit_logistic_gwpr — family = "binomial" only
# ---------------------------------------------------------------------------

test_that("fit_logistic_gwpr: non-binomial family raises informative error", {
  ctx <- make_logistic_context()
  expect_error(
    fit_logistic_gwpr(ctx, bandwidth = 3L, family = "multinomial"),
    regexp = "binomial"
  )
})

# ---------------------------------------------------------------------------
# fit_logistic_gwpr — local model failure does not break structure
# ---------------------------------------------------------------------------

test_that("fit_logistic_gwpr: local model failures recorded, structure intact", {
  pd  <- make_bin_panel_data()
  ctx <- new_gwpr_context(
    formula    = y ~ nonexistent_var,   # will fail every local model
    family     = "binomial",
    model      = "pooling",
    effect     = "individual",
    id         = "id",
    time       = "time",
    kernel     = "bisquare",
    adaptive   = TRUE,
    threshold  = 0.5,
    workers    = 1L,
    panel_data = pd,
    coords     = make_coords_4()
  )

  res <- fit_logistic_gwpr(ctx, bandwidth = 3L)

  # Structure intact
  expect_true("local_results" %in% names(res))
  expect_equal(length(res$local_results), 4L)

  # All failed
  statuses <- vapply(res$local_results, `[[`, character(1L), "status")
  expect_true(all(statuses == "failed"))

  # Metadata records failure count
  expect_equal(res$metadata$n_failed, 4L)
})

# ---------------------------------------------------------------------------
# fit_logistic_gwpr — pre-supplied weights_context
# ---------------------------------------------------------------------------

test_that("fit_logistic_gwpr: accepts pre-built weights_context", {
  ctx   <- make_logistic_context(model = "pooling", effect = "individual",
                                  kernel = "bisquare", adaptive = TRUE)
  wctx  <- build_distance_context(
    coords  = ctx$coords,
    ids     = rownames(ctx$coords),
    longlat = FALSE,
    cache   = TRUE
  )
  expect_no_error(fit_logistic_gwpr(ctx, bandwidth = 3L,
                                    weights_context = wctx))
})

# ---------------------------------------------------------------------------
# fit_logistic_gwpr — reproducibility
# ---------------------------------------------------------------------------

test_that("fit_logistic_gwpr: same inputs give identical results", {
  ctx  <- make_logistic_context(model = "pooling", effect = "individual",
                                 kernel = "bisquare", adaptive = TRUE)
  res1 <- fit_logistic_gwpr(ctx, bandwidth = 3L)
  res2 <- fit_logistic_gwpr(ctx, bandwidth = 3L)

  expect_equal(
    res1$local_results[[1]]$coefficients,
    res2$local_results[[1]]$coefficients
  )
  expect_equal(res1$predictions, res2$predictions)
})

# ---------------------------------------------------------------------------
# fit_logistic_gwpr — custom threshold changes class_pred
# ---------------------------------------------------------------------------

test_that("fit_logistic_gwpr: threshold = 0.0 classifies all as 1", {
  ctx <- make_logistic_context(model = "pooling", effect = "individual",
                                threshold = 0.0, kernel = "bisquare",
                                adaptive = TRUE)
  res <- fit_logistic_gwpr(ctx, bandwidth = 3L)
  classes <- res$class_pred[!is.na(res$class_pred)]
  expect_true(all(classes == 1L))
})

test_that("fit_logistic_gwpr: threshold = 1.0 classifies all as 0", {
  ctx <- make_logistic_context(model = "pooling", effect = "individual",
                                threshold = 1.0, kernel = "bisquare",
                                adaptive = TRUE)
  res <- fit_logistic_gwpr(ctx, bandwidth = 3L)
  classes <- res$class_pred[!is.na(res$class_pred)]
  expect_true(all(classes == 0L))
})

# ---------------------------------------------------------------------------
# fit_logistic_gwpr — glmmTMB and fixest small-sample failure capture
# ---------------------------------------------------------------------------

test_that("fixest small-sample collinearity is captured as failed status", {
  # With very unbalanced data, feglm may fail with collinearity error
  # This verifies the error is captured, not propagated
  pd <- data.frame(
    id   = rep(as.character(1:2), each = 2),
    time = rep(1:2, times = 2),
    y    = c(1L, 1L, 0L, 0L),  # perfectly separated by id
    x1   = c(1, 1, -1, -1),    # perfectly correlated with id
    stringsAsFactors = FALSE
  )
  coords <- matrix(c(0, 0, 1, 0), ncol = 2, byrow = TRUE,
                   dimnames = list(c("1", "2"), c("X", "Y")))
  ctx <- new_gwpr_context(
    formula    = y ~ x1,
    family     = "binomial",
    model      = "fixed",
    effect     = "individual",
    id         = "id",
    time       = "time",
    kernel     = "bisquare",
    adaptive   = FALSE,
    threshold  = 0.5,
    workers    = 1L,
    panel_data = pd,
    coords     = coords
  )
  # May succeed or fail — but must not throw an uncaught error
  expect_no_error({
    res <- fit_logistic_gwpr(ctx, bandwidth = 2)
  })
  # Structure intact
  expect_true("local_results" %in% names(res))
})

test_that("glmmTMB small-sample failure is captured as failed status", {
  skip_if_not_installed("glmmTMB")
  # Minimal data likely to cause glmmTMB convergence failure
  pd <- data.frame(
    id   = rep(as.character(1:2), each = 2),
    time = rep(1:2, times = 2),
    y    = c(1L, 0L, 1L, 0L),
    x1   = c(1, -1, 1, -1),
    stringsAsFactors = FALSE
  )
  coords <- matrix(c(0, 0, 1, 0), ncol = 2, byrow = TRUE,
                   dimnames = list(c("1", "2"), c("X", "Y")))
  ctx <- new_gwpr_context(
    formula    = y ~ x1,
    family     = "binomial",
    model      = "random",
    effect     = "individual",
    id         = "id",
    time       = "time",
    kernel     = "bisquare",
    adaptive   = FALSE,
    threshold  = 0.5,
    workers    = 1L,
    panel_data = pd,
    coords     = coords
  )
  # Must not throw
  expect_no_error({
    res <- fit_logistic_gwpr(ctx, bandwidth = 2)
  })
  statuses <- vapply(res$local_results, `[[`, character(1L), "status")
  expect_true(all(statuses %in% c("ok", "failed")))
})

#' @title Data Preparation Module for GWPR.light 1.0.0
#'
#' @description
#' Internal functions that convert user inputs into the internal data structures
#' required by the model engine: panel indices, spatial alignment, model frame,
#' and model matrix.
#'
#' @name data_prepare
#' @keywords internal
NULL

# ---------------------------------------------------------------------------
# prepare_data  (top-level entry point)
# ---------------------------------------------------------------------------

#' Prepare all data structures for GWPR fitting
#'
#' Orchestrates the full data preparation pipeline: panel data extraction,
#' spatial data extraction, ID mapping, model frame and model matrix
#' construction. Updates and returns a `gwpr_context`.
#'
#' @param context A `gwpr_context` object with at least `formula`, `family`,
#'   `id`, `time`, `model`, `raw_data`, and `raw_spatial` populated.
#'
#' @return An updated `gwpr_context` with `panel_data`, `spatial_data`,
#'   `id_map`, `coords`, `model_frame`, `model_matrix`, `response`, and
#'   `metadata` filled in.
#' @keywords internal
prepare_data <- function(context) {
  context <- prepare_panel_data(context)
  context <- prepare_spatial_data(context)

  context$id_map <- build_id_map(
    panel_data   = context$panel_data,
    spatial_data = context$spatial_data,
    id           = context$id
  )

  context <- build_model_frame(context)
  context <- build_model_matrix(context)

  context
}

# ---------------------------------------------------------------------------
# prepare_panel_data
# ---------------------------------------------------------------------------

#' Prepare panel data from the raw data in context
#'
#' Extracts the `id`, `time`, and formula variables from `raw_data`. Adds a
#' `raw_row_id` column preserving the original row positions. Sorts by id then
#' time. Records panel balance information in `metadata`. For `within` models,
#' records single-observation individuals in `metadata`.
#'
#' @param context A `gwpr_context` with `raw_data`, `formula`, `id`, `time`,
#'   and `model` populated.
#'
#' @return Updated context with `panel_data` and `metadata` populated.
#' @keywords internal
prepare_panel_data <- function(context) {
  data    <- context$raw_data
  formula <- context$formula
  id      <- context$id
  time    <- context$time
  model   <- context$model

  if (!is.data.frame(data)) {
    stop("`raw_data` must be a data.frame.", call. = FALSE)
  }

  # Identify all variables referenced in the formula
  formula_vars <- all.vars(formula)
  # Keep only those that exist in data (response + predictors + id + time)
  keep_vars <- unique(c(id, time, formula_vars))
  keep_vars <- intersect(keep_vars, names(data))

  panel_data <- data[, keep_vars, drop = FALSE]

  # Preserve original row positions
  panel_data$raw_row_id <- seq_len(nrow(data))

  # Sort by id then time for predictable ordering
  panel_data <- panel_data[order(panel_data[[id]], panel_data[[time]]), , drop = FALSE]
  rownames(panel_data) <- NULL

  # Determine panel balance
  obs_per_unit <- table(panel_data[[id]])
  n_obs_unique <- length(unique(as.integer(obs_per_unit)))
  balanced     <- n_obs_unique == 1L

  metadata <- context$metadata
  metadata$panel_balanced <- balanced
  metadata$n_units        <- length(obs_per_unit)
  metadata$obs_per_unit   <- as.integer(obs_per_unit)
  names(metadata$obs_per_unit) <- names(obs_per_unit)

  # Record single-observation individuals when using within model
  if (!is.null(model) && model == "within") {
    single_obs_ids <- names(obs_per_unit)[obs_per_unit == 1L]
    if (length(single_obs_ids) > 0L) {
      metadata$within_single_obs_ids <- single_obs_ids
      warning(
        "Within model: ", length(single_obs_ids),
        " individual(s) have only one observation and cannot contribute to ",
        "within-individual variation. Recorded in metadata$within_single_obs_ids.",
        call. = FALSE
      )
    }
  }

  context$panel_data <- panel_data
  context$metadata   <- metadata
  context
}

# ---------------------------------------------------------------------------
# prepare_spatial_data
# ---------------------------------------------------------------------------

#' Extract and align spatial data for GWPR fitting
#'
#' Extracts geometry and representative coordinates from the `raw_spatial` sf
#' object. Aligns spatial rows to the unique individual IDs present in
#' `panel_data` via `id_map`.
#'
#' @param context A `gwpr_context` with `raw_spatial`, `panel_data`, and `id`
#'   populated.
#'
#' @return Updated context with `spatial_data` and `coords` populated.
#' @keywords internal
prepare_spatial_data <- function(context) {
  if (!requireNamespace("sf", quietly = TRUE)) {
    stop("Package 'sf' is required. Please install it.", call. = FALSE)
  }

  spatial    <- context$raw_spatial
  panel_data <- context$panel_data
  id         <- context$id

  if (!inherits(spatial, "sf")) {
    stop("`raw_spatial` must be an sf object.", call. = FALSE)
  }

  # Unique panel IDs (preserve order after sorting by id, then time)
  panel_ids <- unique(panel_data[[id]])

  # Filter spatial to only those IDs present in panel data
  spatial_ids     <- spatial[[id]]
  in_panel        <- spatial_ids %in% panel_ids
  spatial_aligned <- spatial[in_panel, , drop = FALSE]

  # Re-order to match the order of unique panel IDs
  row_order <- match(panel_ids, spatial_aligned[[id]])

  # Check all panel IDs are found in spatial
  missing_mask <- is.na(row_order)
  if (any(missing_mask)) {
    missing_ids <- panel_ids[missing_mask]
    stop(
      "The following panel ID(s) have no matching row in `spatial`: ",
      paste(missing_ids, collapse = ", "),
      call. = FALSE
    )
  }

  spatial_data <- spatial_aligned[row_order, , drop = FALSE]
  rownames(spatial_data) <- NULL

  # Extract representative coordinates
  coords <- extract_coords_from_sf(spatial_data)

  context$spatial_data <- spatial_data
  context$coords       <- coords
  context
}

# ---------------------------------------------------------------------------
# extract_coords_from_sf  (internal helper)
# ---------------------------------------------------------------------------

#' Extract representative XY coordinates from an sf object
#'
#' Uses point coordinates for POINT geometries and centroids for other
#' geometry types (POLYGON, MULTIPOLYGON, etc.).
#'
#' @param spatial An `sf` object.
#'
#' @return A numeric matrix with columns `X` and `Y`.
#' @keywords internal
extract_coords_from_sf <- function(spatial) {
  geom_type <- unique(as.character(sf::st_geometry_type(spatial)))

  if (length(geom_type) == 1L && geom_type == "POINT") {
    coords_sf <- sf::st_coordinates(spatial)
    coords    <- coords_sf[, c("X", "Y"), drop = FALSE]
  } else {
    # Use centroids for polygon / other types; suppress warnings about
    # geographic CRS centroids
    suppressWarnings(
      centroids <- sf::st_centroid(sf::st_geometry(spatial))
    )
    coords_sf <- sf::st_coordinates(centroids)
    coords    <- coords_sf[, c("X", "Y"), drop = FALSE]
  }

  coords
}

# ---------------------------------------------------------------------------
# build_id_map
# ---------------------------------------------------------------------------

#' Build a mapping from panel unit IDs to spatial row indices
#'
#' Returns a named integer vector where each name is a panel unit ID (as a
#' character string) and each value is the corresponding row index in
#' `spatial_data` (1-based).
#'
#' Rules:
#' - Every panel ID must have a spatial match; missing IDs cause an error.
#' - Extra spatial rows (not in panel) are silently ignored.
#'
#' @param panel_data   A data frame with a column named `id`.
#' @param spatial_data An `sf` data frame with a column named `id`.
#' @param id           Character; name of the shared ID column.
#'
#' @return Named integer vector mapping unit ID to spatial row index.
#' @keywords internal
build_id_map <- function(panel_data, spatial_data, id) {
  panel_ids   <- unique(as.character(panel_data[[id]]))
  spatial_ids <- as.character(spatial_data[[id]])

  row_indices <- match(panel_ids, spatial_ids)

  missing_mask <- is.na(row_indices)
  if (any(missing_mask)) {
    missing_ids <- panel_ids[missing_mask]
    stop(
      "The following panel ID(s) have no matching row in `spatial_data`: ",
      paste(missing_ids, collapse = ", "),
      call. = FALSE
    )
  }

  id_map        <- row_indices
  names(id_map) <- panel_ids
  id_map
}

# ---------------------------------------------------------------------------
# build_model_frame
# ---------------------------------------------------------------------------

#' Build the model frame from the formula and panel data
#'
#' @param context A `gwpr_context` with `formula` and `panel_data` populated.
#'
#' @return Updated context with `model_frame` populated.
#' @keywords internal
build_model_frame <- function(context) {
  formula    <- context$formula
  panel_data <- context$panel_data

  mf <- stats::model.frame(formula, data = panel_data, na.action = stats::na.pass)

  context$model_frame <- mf
  context
}

# ---------------------------------------------------------------------------
# build_model_matrix
# ---------------------------------------------------------------------------

#' Build the design matrix and response vector
#'
#' Extracts the response variable `y` and the design matrix `X` from the
#' model frame.  For `binomial` family, the response is standardised to 0/1
#' via `standardize_binary_response()`.
#'
#' @param context A `gwpr_context` with `model_frame`, `formula`, and `family`
#'   populated.
#'
#' @return Updated context with `model_matrix` and `response` populated.
#' @keywords internal
build_model_matrix <- function(context) {
  mf     <- context$model_frame
  family <- context$family

  # Extract response
  y_raw <- stats::model.response(mf)

  if (!is.null(family) && family == "binomial") {
    y <- standardize_binary_response(y_raw)
  } else {
    y <- as.numeric(y_raw)
  }

  # Build design matrix (removes the response column)
  X <- stats::model.matrix(context$formula, data = mf)

  context$model_matrix <- X
  context$response     <- y
  context
}

# ---------------------------------------------------------------------------
# standardize_binary_response
# ---------------------------------------------------------------------------

#' Standardise a binary response variable to 0/1 numeric
#'
#' Converts logical and two-level factor responses to 0/1 integer. Numeric
#' 0/1 vectors are returned unchanged (as numeric). Other inputs raise an
#' error.
#'
#' For factor inputs the first level is mapped to 0 and the second level is
#' mapped to 1.
#'
#' @param y A vector that is 0/1 numeric, logical, or a two-level factor.
#'
#' @return A numeric vector of 0s and 1s.
#' @keywords internal
standardize_binary_response <- function(y) {
  if (is.logical(y)) {
    return(as.integer(y))
  }

  if (is.factor(y)) {
    lvls <- levels(y)
    if (length(lvls) != 2L) {
      stop(
        "`y` must be a two-level factor for binary response standardisation. ",
        "Found ", length(lvls), " level(s).",
        call. = FALSE
      )
    }
    # First level -> 0, second level -> 1
    return(as.integer(y) - 1L)
  }

  if (is.numeric(y)) {
    unique_vals <- unique(y[!is.na(y)])
    invalid     <- setdiff(unique_vals, c(0, 1))
    if (length(invalid) > 0L) {
      stop(
        "Numeric `y` must contain only 0 and 1 for binary response. ",
        "Found: ", paste(invalid, collapse = ", "),
        call. = FALSE
      )
    }
    return(as.numeric(y))
  }

  stop(
    "`y` must be 0/1 numeric, logical, or a two-level factor.",
    call. = FALSE
  )
}

library(testthat)

# ---------------------------------------------------------------------------
# Helpers: build minimal simulation data
# ---------------------------------------------------------------------------

#' Create a small balanced panel data frame
make_panel_data <- function(ids = c("A", "B", "C"),
                            times = 1:3,
                            id_col = "id",
                            time_col = "time") {
  n_ids   <- length(ids)
  n_times <- length(times)
  df <- data.frame(
    id   = rep(ids,   each = n_times),
    time = rep(times, times = n_ids),
    y    = rnorm(n_ids * n_times),
    x1   = rnorm(n_ids * n_times),
    stringsAsFactors = FALSE
  )
  names(df)[names(df) == "id"]   <- id_col
  names(df)[names(df) == "time"] <- time_col
  df
}

#' Create a small sf POINT object aligned with given IDs
make_spatial_sf <- function(ids = c("A", "B", "C"), id_col = "id",
                             extra_ids = character(0)) {
  all_ids <- c(ids, extra_ids)
  n       <- length(all_ids)
  pts     <- sf::st_sfc(lapply(seq_len(n), function(i) sf::st_point(c(i, i))))
  sf_obj  <- sf::st_sf(
    setNames(list(all_ids), id_col),
    geometry = pts,
    stringsAsFactors = FALSE
  )
  sf_obj
}

#' Build a minimal gwpr_context with raw data and spatial
make_context <- function(data, spatial,
                         formula  = y ~ x1,
                         family   = "gaussian",
                         model    = "within",
                         effect   = "individual",
                         id       = "id",
                         time     = "time",
                         kernel   = "bisquare",
                         adaptive = FALSE,
                         threshold = 0.5,
                         workers   = 1L) {
  new_gwpr_context(
    formula     = formula,
    family      = family,
    model       = model,
    effect      = effect,
    id          = id,
    time        = time,
    kernel      = kernel,
    adaptive    = adaptive,
    threshold   = threshold,
    workers     = workers,
    raw_data    = data,
    raw_spatial = spatial
  )
}

# ---------------------------------------------------------------------------
# standardize_binary_response
# ---------------------------------------------------------------------------

test_that("standardize_binary_response: numeric 0/1 returned as numeric", {
  y <- c(0, 1, 0, 1)
  result <- standardize_binary_response(y)
  expect_true(is.numeric(result))
  expect_equal(result, c(0, 1, 0, 1))
})

test_that("standardize_binary_response: logical converted to 0/1 integer", {
  y <- c(TRUE, FALSE, TRUE)
  result <- standardize_binary_response(y)
  expect_true(is.numeric(result) || is.integer(result))
  expect_equal(as.integer(result), c(1L, 0L, 1L))
})

test_that("standardize_binary_response: two-level factor first level is 0", {
  y <- factor(c("no", "yes", "no", "yes"), levels = c("no", "yes"))
  result <- standardize_binary_response(y)
  expect_equal(as.integer(result), c(0L, 1L, 0L, 1L))
})

test_that("standardize_binary_response: factor with >2 levels raises error", {
  y <- factor(c("a", "b", "c"))
  expect_error(standardize_binary_response(y), "two-level factor")
})

test_that("standardize_binary_response: numeric with invalid values raises error", {
  y <- c(0, 1, 2)
  expect_error(standardize_binary_response(y), "0 and 1")
})

test_that("standardize_binary_response: non-numeric/logical/factor raises error", {
  expect_error(standardize_binary_response("a"), "0/1 numeric")
})

# ---------------------------------------------------------------------------
# build_id_map
# ---------------------------------------------------------------------------

test_that("build_id_map: returns named integer vector with correct mappings", {
  panel_data   <- make_panel_data(ids = c("A", "B", "C"))
  spatial_data <- make_spatial_sf(ids = c("A", "B", "C"))

  id_map <- build_id_map(panel_data, spatial_data, id = "id")

  expect_true(is.integer(id_map))
  expect_equal(sort(names(id_map)), sort(c("A", "B", "C")))
  # Each index should be a valid row in spatial_data
  expect_true(all(id_map >= 1L & id_map <= nrow(spatial_data)))
})

test_that("build_id_map: extra spatial rows do not cause an error", {
  panel_data   <- make_panel_data(ids = c("A", "B"))
  # Spatial has an extra ID "C" not in panel
  spatial_data <- make_spatial_sf(ids = c("A", "B", "C"))

  expect_no_error(build_id_map(panel_data, spatial_data, id = "id"))
})

test_that("build_id_map: panel ID missing from spatial raises error", {
  panel_data   <- make_panel_data(ids = c("A", "B", "X"))
  spatial_data <- make_spatial_sf(ids = c("A", "B", "C"))

  expect_error(
    build_id_map(panel_data, spatial_data, id = "id"),
    "X"
  )
})

test_that("build_id_map: numeric IDs handled correctly", {
  panel_data   <- make_panel_data(ids = c(1L, 2L, 3L), id_col = "uid")
  panel_data$uid <- as.integer(panel_data$uid)
  spatial_data <- make_spatial_sf(ids = c(1L, 2L, 3L), id_col = "uid")
  spatial_data$uid <- as.integer(spatial_data$uid)

  id_map <- build_id_map(panel_data, spatial_data, id = "uid")
  expect_equal(length(id_map), 3L)
  expect_true(all(!is.na(id_map)))
})

# ---------------------------------------------------------------------------
# prepare_panel_data
# ---------------------------------------------------------------------------

test_that("prepare_panel_data: adds raw_row_id column", {
  data    <- make_panel_data()
  spatial <- make_spatial_sf()
  ctx     <- make_context(data, spatial)

  ctx2 <- prepare_panel_data(ctx)
  expect_true("raw_row_id" %in% names(ctx2$panel_data))
})

test_that("prepare_panel_data: sorts by id then time", {
  # Create unsorted data
  data    <- make_panel_data(ids = c("B", "A"), times = c(2, 1))
  spatial <- make_spatial_sf(ids = c("A", "B"))
  ctx     <- make_context(data, spatial)

  ctx2 <- prepare_panel_data(ctx)
  pd   <- ctx2$panel_data
  # After sorting, first id should be A
  expect_equal(pd$id[1], "A")
})

test_that("prepare_panel_data: metadata records panel balance (balanced)", {
  data    <- make_panel_data(ids = c("A", "B"), times = 1:3)
  spatial <- make_spatial_sf(ids = c("A", "B"))
  ctx     <- make_context(data, spatial)

  ctx2 <- prepare_panel_data(ctx)
  expect_true(ctx2$metadata$panel_balanced)
})

test_that("prepare_panel_data: unbalanced panel does not raise error", {
  # B has only 2 time periods
  df <- rbind(
    make_panel_data(ids = "A", times = 1:3),
    make_panel_data(ids = "B", times = 1:2)
  )
  spatial <- make_spatial_sf(ids = c("A", "B"))
  ctx     <- make_context(df, spatial)

  expect_no_error(prepare_panel_data(ctx))
  ctx2 <- prepare_panel_data(ctx)
  expect_false(ctx2$metadata$panel_balanced)
})

test_that("prepare_panel_data: within model records single-obs individuals in metadata", {
  # Create data where "C" has only 1 observation
  df <- rbind(
    make_panel_data(ids = "A", times = 1:3),
    make_panel_data(ids = "B", times = 1:3),
    make_panel_data(ids = "C", times = 1)
  )
  spatial <- make_spatial_sf(ids = c("A", "B", "C"))
  ctx     <- make_context(df, spatial, model = "within")

  ctx2 <- suppressWarnings(prepare_panel_data(ctx))
  expect_true("within_single_obs_ids" %in% names(ctx2$metadata))
  expect_true("C" %in% ctx2$metadata$within_single_obs_ids)
})

# ---------------------------------------------------------------------------
# prepare_spatial_data
# ---------------------------------------------------------------------------

test_that("prepare_spatial_data: coords matrix has X and Y columns", {
  data    <- make_panel_data()
  spatial <- make_spatial_sf()
  ctx     <- make_context(data, spatial)
  ctx     <- prepare_panel_data(ctx)

  ctx2 <- prepare_spatial_data(ctx)
  expect_true(!is.null(ctx2$coords))
  expect_true(all(c("X", "Y") %in% colnames(ctx2$coords)))
})

test_that("prepare_spatial_data: extra spatial regions not in panel are excluded", {
  data    <- make_panel_data(ids = c("A", "B"))
  # Spatial has extra ID "C"
  spatial <- make_spatial_sf(ids = c("A", "B"), extra_ids = "C")
  ctx     <- make_context(data, spatial)
  ctx     <- prepare_panel_data(ctx)

  ctx2 <- prepare_spatial_data(ctx)
  expect_equal(nrow(ctx2$spatial_data), 2L)
  expect_false("C" %in% ctx2$spatial_data$id)
})

test_that("prepare_spatial_data: panel ID missing from spatial raises error", {
  data    <- make_panel_data(ids = c("A", "B", "X"))
  spatial <- make_spatial_sf(ids = c("A", "B", "C"))
  ctx     <- make_context(data, spatial)
  ctx     <- prepare_panel_data(ctx)

  expect_error(prepare_spatial_data(ctx), "X")
})

# ---------------------------------------------------------------------------
# build_model_frame and build_model_matrix
# ---------------------------------------------------------------------------

test_that("build_model_frame: produces a model frame from formula", {
  data    <- make_panel_data()
  spatial <- make_spatial_sf()
  ctx     <- make_context(data, spatial)
  ctx     <- prepare_panel_data(ctx)

  ctx2 <- build_model_frame(ctx)
  expect_true(!is.null(ctx2$model_frame))
  expect_true(is.data.frame(ctx2$model_frame))
})

test_that("build_model_matrix: gaussian response is numeric", {
  data    <- make_panel_data()
  spatial <- make_spatial_sf()
  ctx     <- make_context(data, spatial, family = "gaussian")
  ctx     <- prepare_panel_data(ctx)
  ctx     <- build_model_frame(ctx)
  ctx     <- build_model_matrix(ctx)

  expect_true(is.numeric(ctx$response))
  expect_true(is.matrix(ctx$model_matrix))
})

test_that("build_model_matrix: binomial factor response standardised to 0/1", {
  data    <- make_panel_data()
  data$y  <- factor(
    ifelse(data$y > 0, "yes", "no"),
    levels = c("no", "yes")
  )
  spatial <- make_spatial_sf()
  ctx     <- make_context(data, spatial, family = "binomial",
                          formula = y ~ x1)
  ctx     <- prepare_panel_data(ctx)
  ctx     <- build_model_frame(ctx)
  ctx     <- build_model_matrix(ctx)

  expect_true(all(ctx$response %in% c(0, 1)))
})

# ---------------------------------------------------------------------------
# prepare_data  (full pipeline)
# ---------------------------------------------------------------------------

test_that("prepare_data: full pipeline with character IDs", {
  data    <- make_panel_data(ids = c("A", "B", "C"))
  spatial <- make_spatial_sf(ids = c("A", "B", "C"))
  ctx     <- make_context(data, spatial)

  ctx2 <- prepare_data(ctx)

  expect_false(is.null(ctx2$panel_data))
  expect_false(is.null(ctx2$spatial_data))
  expect_false(is.null(ctx2$id_map))
  expect_false(is.null(ctx2$coords))
  expect_false(is.null(ctx2$model_frame))
  expect_false(is.null(ctx2$model_matrix))
  expect_false(is.null(ctx2$response))
})

test_that("prepare_data: full pipeline with numeric IDs", {
  data    <- make_panel_data(ids = c(1, 2, 3))
  spatial <- make_spatial_sf(ids = c(1, 2, 3))
  ctx     <- make_context(data, spatial)

  ctx2 <- prepare_data(ctx)

  expect_false(is.null(ctx2$panel_data))
  expect_false(is.null(ctx2$id_map))
  expect_equal(length(ctx2$id_map), 3L)
})

test_that("prepare_data: unbalanced panel completes without error", {
  df <- rbind(
    make_panel_data(ids = "A", times = 1:4),
    make_panel_data(ids = "B", times = 1:2),
    make_panel_data(ids = "C", times = 1:3)
  )
  spatial <- make_spatial_sf(ids = c("A", "B", "C"))
  ctx     <- make_context(df, spatial, model = "pooling")

  expect_no_error(prepare_data(ctx))
  ctx2 <- prepare_data(ctx)
  expect_false(ctx2$metadata$panel_balanced)
})

test_that("prepare_data: spatial extra regions excluded, panel IDs all matched", {
  data    <- make_panel_data(ids = c("A", "B"))
  spatial <- make_spatial_sf(ids = c("A", "B"), extra_ids = c("D", "E"))
  ctx     <- make_context(data, spatial)

  ctx2 <- prepare_data(ctx)
  # spatial_data should contain only A and B
  expect_equal(nrow(ctx2$spatial_data), 2L)
  expect_equal(sort(ctx2$spatial_data$id), sort(c("A", "B")))
})

test_that("prepare_data: panel ID missing in spatial raises error", {
  data    <- make_panel_data(ids = c("A", "B", "Z"))
  spatial <- make_spatial_sf(ids = c("A", "B", "C"))
  ctx     <- make_context(data, spatial)

  expect_error(prepare_data(ctx), "Z")
})

test_that("prepare_data: binomial logistic factor response converted to 0/1", {
  data   <- make_panel_data(ids = c("A", "B", "C"))
  data$y <- factor(ifelse(data$y > 0, "case", "control"),
                   levels = c("control", "case"))
  spatial <- make_spatial_sf(ids = c("A", "B", "C"))
  ctx     <- make_context(data, spatial, family = "binomial", formula = y ~ x1)

  ctx2 <- prepare_data(ctx)
  expect_true(all(ctx2$response %in% c(0L, 1L)))
})

#' @title Distance and Kernel Weights Module for GWPR.light 1.0.0
#'
#' @description
#' Internal functions for computing spatial distances and geographically
#' weighted kernel weights. Supports fixed and adaptive bandwidths, and five
#' kernel functions: bisquare, gaussian, exponential, tricube, and boxcar.
#'
#' @name weights
#' @keywords internal
NULL

# ---------------------------------------------------------------------------
# validate_bandwidth
# ---------------------------------------------------------------------------

#' Validate bandwidth parameter
#'
#' Checks that the supplied bandwidth is legal:
#' * fixed (adaptive = FALSE): must be a finite positive numeric scalar.
#' * adaptive (adaptive = TRUE): must be a finite positive integer scalar.
#'
#' @param bandwidth Numeric scalar.  The bandwidth value to validate.
#' @param adaptive  Logical scalar.  `TRUE` for adaptive (k-nearest-neighbour)
#'   bandwidth, `FALSE` for fixed distance bandwidth.
#'
#' @return Invisibly returns `TRUE` when validation passes.
#' @keywords internal
validate_bandwidth <- function(bandwidth, adaptive) {
  if (!is.numeric(bandwidth) || length(bandwidth) != 1L) {
    stop("`bandwidth` must be a single numeric value.", call. = FALSE)
  }
  if (!is.finite(bandwidth)) {
    stop("`bandwidth` must be a finite number.", call. = FALSE)
  }
  if (adaptive) {
    if (bandwidth != floor(bandwidth) || bandwidth < 1) {
      stop(
        "For adaptive bandwidth, `bandwidth` must be a positive integer ",
        "(number of neighbours).",
        call. = FALSE
      )
    }
  } else {
    if (bandwidth <= 0) {
      stop(
        "For fixed bandwidth, `bandwidth` must be a strictly positive number.",
        call. = FALSE
      )
    }
  }
  invisible(TRUE)
}

# ---------------------------------------------------------------------------
# compute_distance
# ---------------------------------------------------------------------------

#' Compute a full pairwise distance matrix from a coordinate matrix
#'
#' @param coords   A numeric matrix with at least two columns (`X` and `Y`, or
#'   the first two columns if names are absent).  Each row is one spatial unit.
#' @param longlat  Logical.  If `TRUE`, great-circle distances (in kilometres)
#'   are calculated using the Haversine formula.  If `FALSE` (default),
#'   Euclidean distance is used.
#'
#' @return An n x n symmetric numeric matrix where element [i, j] is the
#'   distance between unit i and unit j.  Diagonal is 0.
#' @keywords internal
compute_distance <- function(coords, longlat = FALSE) {
  if (!is.matrix(coords)) {
    coords <- as.matrix(coords)
  }
  if (ncol(coords) < 2L) {
    stop("`coords` must have at least 2 columns (X and Y).", call. = FALSE)
  }
  n <- nrow(coords)

  if (!longlat) {
    # Euclidean distance via stats::dist — fastest for small/medium n
    d <- as.matrix(stats::dist(coords[, 1:2, drop = FALSE]))
    return(d)
  }

  # Great-circle distance (Haversine), result in kilometres
  x <- coords[, 1]   # longitude in degrees
  y <- coords[, 2]   # latitude  in degrees

  to_rad <- pi / 180
  lat1 <- y * to_rad
  lon1 <- x * to_rad

  d <- matrix(0, nrow = n, ncol = n)
  R <- 6371.0  # Earth radius in km

  for (i in seq_len(n)) {
    dlat <- lat1 - lat1[i]
    dlon <- lon1 - lon1[i]
    a <- sin(dlat / 2)^2 + cos(lat1[i]) * cos(lat1) * sin(dlon / 2)^2
    a <- pmin(a, 1)  # numerical safety
    d[i, ] <- 2 * R * asin(sqrt(a))
  }
  d
}

# ---------------------------------------------------------------------------
# get_local_distances
# ---------------------------------------------------------------------------

#' Extract distances from one focus unit to all others
#'
#' @param distance_context A list as returned by `build_distance_context()`, or
#'   a plain n x n numeric distance matrix.  If a plain matrix is supplied, the
#'   rows and columns must already be in the same order as the spatial units.
#' @param focus_id         Integer scalar (1-based) or character matching a
#'   row/column name of the distance matrix.  The focal unit whose distances
#'   are extracted.
#'
#' @return A numeric vector of length n giving the distance from the focus unit
#'   to every unit (including itself, which is 0).
#' @keywords internal
get_local_distances <- function(distance_context, focus_id) {
  # Support both a plain matrix and a list wrapper
  if (is.matrix(distance_context)) {
    dmat <- distance_context
  } else if (is.list(distance_context) && !is.null(distance_context$distance_matrix)) {
    dmat <- distance_context$distance_matrix
  } else {
    stop(
      "`distance_context` must be an n x n distance matrix or a list with ",
      "a `distance_matrix` element.",
      call. = FALSE
    )
  }

  n <- nrow(dmat)

  if (is.character(focus_id)) {
    rn <- rownames(dmat)
    if (is.null(rn)) {
      stop("Distance matrix has no row names; cannot look up character `focus_id`.", call. = FALSE)
    }
    idx <- match(focus_id, rn)
    if (is.na(idx)) {
      stop(sprintf("focus_id '%s' not found in distance matrix row names.", focus_id), call. = FALSE)
    }
    return(as.numeric(dmat[idx, ]))
  }

  if (!is.numeric(focus_id) || length(focus_id) != 1L) {
    stop("`focus_id` must be a single integer index or a character ID.", call. = FALSE)
  }
  focus_id <- as.integer(focus_id)
  if (focus_id < 1L || focus_id > n) {
    stop(sprintf("`focus_id` (%d) is out of range [1, %d].", focus_id, n), call. = FALSE)
  }
  as.numeric(dmat[focus_id, ])
}

# ---------------------------------------------------------------------------
# build_distance_context
# ---------------------------------------------------------------------------

#' Build a distance context object
#'
#' Wraps a distance matrix together with unit IDs into a list suitable for
#' passing to `get_local_distances()`.  Optionally pre-computes the full
#' distance matrix (recommended for small data) or stores only the coordinate
#' matrix for on-the-fly computation (recommended for large data).
#'
#' @param coords    A numeric matrix with columns `X` and `Y`.
#' @param ids       Character or numeric vector of unit IDs (length = nrow(coords)).
#' @param longlat   Logical.  Passed to `compute_distance()`.
#' @param cache     Logical.  If `TRUE` (default), pre-computes and caches the
#'   full n x n distance matrix.  Set `FALSE` for very large datasets to avoid
#'   memory pressure; `get_local_distances()` will then compute rows on demand.
#'
#' @return A list with class `"gwpr_distance_context"` containing:
#' * `ids` — character vector of unit IDs.
#' * `distance_matrix` — n x n matrix (or `NULL` if `cache = FALSE`).
#' * `coords` — the original coordinate matrix.
#' * `longlat` — logical flag.
#' @keywords internal
build_distance_context <- function(coords, ids, longlat = FALSE, cache = TRUE) {
  if (!is.matrix(coords)) coords <- as.matrix(coords)
  n <- nrow(coords)
  ids <- as.character(ids)
  if (length(ids) != n) {
    stop("`ids` length must equal nrow(coords).", call. = FALSE)
  }

  dmat <- NULL
  if (cache) {
    dmat <- compute_distance(coords, longlat = longlat)
    rownames(dmat) <- ids
    colnames(dmat) <- ids
  }

  structure(
    list(
      ids             = ids,
      distance_matrix = dmat,
      coords          = coords,
      longlat         = longlat
    ),
    class = "gwpr_distance_context"
  )
}

# ---------------------------------------------------------------------------
# compute_kernel_weights
# ---------------------------------------------------------------------------

#' Compute geographically weighted kernel weights
#'
#' Given a numeric vector of distances from one focal unit to all others,
#' returns a weight for each unit according to the chosen kernel and bandwidth.
#'
#' For **fixed** bandwidth (`adaptive = FALSE`) the bandwidth parameter is a
#' distance threshold or scale parameter used directly in the kernel formula.
#'
#' For **adaptive** bandwidth (`adaptive = TRUE`) the bandwidth parameter is
#' the number of nearest neighbours k.  The function first identifies the
#' k-th smallest distance (among all units, including the focal unit itself at
#' distance 0) and uses that distance as the effective bandwidth in the kernel
#' formula.
#'
#' Kernel formulae (d = distance, bw = effective bandwidth):
#' * `bisquare`:    `(1 - (d/bw)^2)^2` for `d <= bw`, else 0.
#' * `gaussian`:    `exp(-0.5 * (d/bw)^2)`.
#' * `exponential`: `exp(-d/bw)`.
#' * `tricube`:     `(1 - (d/bw)^3)^3` for `d <= bw`, else 0.
#' * `boxcar`:      `1` for `d <= bw`, else 0.
#'
#' @param distance  Numeric vector of distances from the focal unit to all
#'   spatial units (length n).
#' @param bandwidth Numeric scalar.  For fixed bandwidth: distance scale.
#'   For adaptive bandwidth: positive integer number of neighbours.
#' @param kernel    Character scalar, one of `"bisquare"`, `"gaussian"`,
#'   `"exponential"`, `"tricube"`, `"boxcar"`.
#' @param adaptive  Logical.  `TRUE` for adaptive (kNN) bandwidth.
#'
#' @return A numeric vector of length n with non-negative kernel weights.
#' @keywords internal
compute_kernel_weights <- function(distance, bandwidth, kernel, adaptive) {
  kernel <- match.arg(
    kernel,
    choices = c("bisquare", "gaussian", "exponential", "tricube", "boxcar")
  )
  validate_bandwidth(bandwidth, adaptive)

  if (!is.numeric(distance)) {
    stop("`distance` must be a numeric vector.", call. = FALSE)
  }

  d   <- as.numeric(distance)
  bw  <- bandwidth

  # Adaptive: determine effective bandwidth as the bw-th smallest distance
  if (adaptive) {
    k <- as.integer(bw)
    n <- length(d)
    if (k > n) {
      warning(
        "Adaptive bandwidth (", k, ") exceeds number of units (", n,
        "); using all units.",
        call. = FALSE
      )
      k <- n
    }
    sorted_d <- sort(d)
    bw <- sorted_d[k]   # k-th smallest distance (0-indexed: focal unit itself)
    # Edge case: if bw == 0 (e.g. duplicate coordinates), push to next unique
    if (bw == 0 && k < n) {
      bw <- sorted_d[k + 1L]
    }
    if (bw == 0) {
      # All distances are 0 — give equal weight
      return(rep(1, n))
    }
  }

  # Apply kernel
  w <- switch(
    kernel,
    bisquare = {
      u <- d / bw
      ifelse(u <= 1, (1 - u^2)^2, 0)
    },
    gaussian = {
      exp(-0.5 * (d / bw)^2)
    },
    exponential = {
      exp(-d / bw)
    },
    tricube = {
      u <- d / bw
      ifelse(u <= 1, (1 - u^3)^3, 0)
    },
    boxcar = {
      ifelse(d <= bw, 1, 0)
    }
  )

  w
}

## tests/testthat/test-weights.R
## Tests for the Distance and Kernel Weights module (R/weights.R)
##
## All tests use small simulated data.  workers = 1 (no parallelism).

# ---------------------------------------------------------------------------
# Helper: small coordinate set (5 points on a unit grid)
# ---------------------------------------------------------------------------
make_coords <- function() {
  matrix(
    c(0, 0,
      1, 0,
      2, 0,
      0, 1,
      1, 1),
    ncol = 2, byrow = TRUE,
    dimnames = list(NULL, c("X", "Y"))
  )
}

# ===========================================================================
# validate_bandwidth
# ===========================================================================

test_that("validate_bandwidth: fixed bandwidth must be positive", {
  expect_true(validate_bandwidth(1.5, adaptive = FALSE))
  expect_error(validate_bandwidth(0,   adaptive = FALSE), regexp = "strictly positive")
  expect_error(validate_bandwidth(-1,  adaptive = FALSE), regexp = "strictly positive")
})

test_that("validate_bandwidth: adaptive bandwidth must be a positive integer", {
  expect_true(validate_bandwidth(3,   adaptive = TRUE))
  expect_error(validate_bandwidth(0,   adaptive = TRUE), regexp = "positive integer")
  expect_error(validate_bandwidth(-2,  adaptive = TRUE), regexp = "positive integer")
  expect_error(validate_bandwidth(2.5, adaptive = TRUE), regexp = "positive integer")
})

test_that("validate_bandwidth: non-numeric input triggers error", {
  expect_error(validate_bandwidth("big",  adaptive = FALSE), regexp = "single numeric")
  expect_error(validate_bandwidth(c(1,2), adaptive = FALSE), regexp = "single numeric")
})

test_that("validate_bandwidth: non-finite input triggers error", {
  expect_error(validate_bandwidth(Inf, adaptive = FALSE), regexp = "finite")
  expect_error(validate_bandwidth(NA_real_, adaptive = FALSE), regexp = "finite")
})

# ===========================================================================
# compute_distance
# ===========================================================================

test_that("compute_distance: Euclidean distance matrix has correct dimensions", {
  coords <- make_coords()
  d <- compute_distance(coords, longlat = FALSE)
  expect_true(is.matrix(d))
  expect_equal(dim(d), c(5L, 5L))
})

test_that("compute_distance: diagonal is zero", {
  coords <- make_coords()
  d <- compute_distance(coords, longlat = FALSE)
  expect_equal(unname(diag(d)), rep(0, 5))
})

test_that("compute_distance: matrix is symmetric", {
  coords <- make_coords()
  d <- compute_distance(coords, longlat = FALSE)
  expect_equal(d, t(d))
})

test_that("compute_distance: known Euclidean distances are correct", {
  # Point (0,0) and (1,0) -> distance 1
  # Point (0,0) and (1,1) -> distance sqrt(2)
  coords <- matrix(c(0,0, 1,0, 1,1), ncol = 2, byrow = TRUE)
  d <- compute_distance(coords, longlat = FALSE)
  expect_equal(d[1, 2], 1, tolerance = 1e-10)
  expect_equal(d[1, 3], sqrt(2), tolerance = 1e-10)
})

test_that("compute_distance: great-circle distance is non-negative and symmetric", {
  # Two known cities (approximate lon/lat)
  # Tokyo ~(139.7, 35.7), New York ~(-74.0, 40.7)
  coords <- matrix(c(139.7, 35.7, -74.0, 40.7), ncol = 2, byrow = TRUE)
  d <- compute_distance(coords, longlat = TRUE)
  expect_equal(dim(d), c(2L, 2L))
  expect_equal(d[1, 1], 0, tolerance = 1e-10)
  expect_true(d[1, 2] > 0)
  expect_equal(d[1, 2], d[2, 1], tolerance = 1e-6)
  # Rough sanity: Tokyo-NY is ~10800 km
  expect_true(d[1, 2] > 10000 && d[1, 2] < 12000)
})

test_that("compute_distance: accepts data.frame and coerces to matrix", {
  df <- as.data.frame(make_coords())
  expect_no_error(compute_distance(df, longlat = FALSE))
})

# ===========================================================================
# get_local_distances
# ===========================================================================

test_that("get_local_distances: returns correct vector for integer focus_id", {
  coords <- make_coords()
  d <- compute_distance(coords)
  v <- get_local_distances(d, focus_id = 1L)
  expect_equal(length(v), 5L)
  expect_equal(v[1], 0)           # focal unit distance to itself
  expect_equal(v[2], d[1, 2])
})

test_that("get_local_distances: works with character focus_id via row names", {
  coords <- make_coords()
  d <- compute_distance(coords)
  rownames(d) <- colnames(d) <- c("A", "B", "C", "D", "E")
  v <- get_local_distances(d, focus_id = "C")
  expect_equal(length(v), 5L)
  expect_equal(v[3], 0)           # C to itself
  expect_equal(v[1], d["C", "A"])
})

test_that("get_local_distances: works with list-wrapped distance context", {
  coords <- make_coords()
  ctx <- build_distance_context(coords, ids = letters[1:5])
  v <- get_local_distances(ctx, focus_id = 1L)
  expect_equal(length(v), 5L)
  expect_equal(v[1], 0)
})

test_that("get_local_distances: out-of-range focus_id triggers error", {
  coords <- make_coords()
  d <- compute_distance(coords)
  expect_error(get_local_distances(d, focus_id = 0L), regexp = "out of range")
  expect_error(get_local_distances(d, focus_id = 6L), regexp = "out of range")
})

test_that("get_local_distances: unknown character focus_id triggers error", {
  coords <- make_coords()
  d <- compute_distance(coords)
  rownames(d) <- colnames(d) <- c("A", "B", "C", "D", "E")
  expect_error(get_local_distances(d, focus_id = "Z"), regexp = "not found")
})

# ===========================================================================
# build_distance_context
# ===========================================================================

test_that("build_distance_context: returns correct class and fields", {
  coords <- make_coords()
  ctx <- build_distance_context(coords, ids = as.character(1:5))
  expect_s3_class(ctx, "gwpr_distance_context")
  expect_true(!is.null(ctx$distance_matrix))
  expect_equal(dim(ctx$distance_matrix), c(5L, 5L))
  expect_equal(ctx$ids, as.character(1:5))
})

test_that("build_distance_context: cache=FALSE leaves distance_matrix NULL", {
  coords <- make_coords()
  ctx <- build_distance_context(coords, ids = as.character(1:5), cache = FALSE)
  expect_null(ctx$distance_matrix)
})

test_that("build_distance_context: row/column names match ids", {
  coords <- make_coords()
  ids <- c("p1", "p2", "p3", "p4", "p5")
  ctx <- build_distance_context(coords, ids = ids)
  expect_equal(rownames(ctx$distance_matrix), ids)
  expect_equal(colnames(ctx$distance_matrix), ids)
})

test_that("build_distance_context: id length mismatch triggers error", {
  coords <- make_coords()
  expect_error(build_distance_context(coords, ids = 1:3), regexp = "length")
})

# ===========================================================================
# compute_kernel_weights — bisquare
# ===========================================================================

test_that("bisquare kernel: weight at d=0 equals 1", {
  w <- compute_kernel_weights(c(0, 0.5, 1, 1.5), bandwidth = 1, kernel = "bisquare", adaptive = FALSE)
  expect_equal(w[1], 1)
})

test_that("bisquare kernel: weight is 0 beyond bandwidth", {
  w <- compute_kernel_weights(c(0, 0.5, 1.01), bandwidth = 1, kernel = "bisquare", adaptive = FALSE)
  expect_equal(w[3], 0)
})

test_that("bisquare kernel: weight at boundary (d==bw) equals 0", {
  w <- compute_kernel_weights(c(1), bandwidth = 1, kernel = "bisquare", adaptive = FALSE)
  expect_equal(w[1], 0)
})

test_that("bisquare kernel: known mid-point value", {
  # d = 0.5, bw = 1: (1 - 0.25)^2 = 0.5625
  w <- compute_kernel_weights(0.5, bandwidth = 1, kernel = "bisquare", adaptive = FALSE)
  expect_equal(w, 0.5625, tolerance = 1e-10)
})

# ===========================================================================
# compute_kernel_weights — gaussian
# ===========================================================================

test_that("gaussian kernel: weight at d=0 equals 1", {
  w <- compute_kernel_weights(0, bandwidth = 1, kernel = "gaussian", adaptive = FALSE)
  expect_equal(w, 1)
})

test_that("gaussian kernel: weight is positive for moderate distances", {
  # Use distances that don't underflow to 0 (exp(-0.5*(5/2)^2) is still positive)
  d <- c(0, 1, 2, 3)
  w <- compute_kernel_weights(d, bandwidth = 5, kernel = "gaussian", adaptive = FALSE)
  expect_true(all(w > 0))
})

test_that("gaussian kernel: weight decreases with distance", {
  d <- c(0, 1, 2, 3)
  w <- compute_kernel_weights(d, bandwidth = 2, kernel = "gaussian", adaptive = FALSE)
  expect_true(all(diff(w) < 0))
})

test_that("gaussian kernel: known value at d=bw", {
  # exp(-0.5 * 1^2) = exp(-0.5)
  w <- compute_kernel_weights(1, bandwidth = 1, kernel = "gaussian", adaptive = FALSE)
  expect_equal(w, exp(-0.5), tolerance = 1e-10)
})

# ===========================================================================
# compute_kernel_weights — exponential
# ===========================================================================

test_that("exponential kernel: weight at d=0 equals 1", {
  w <- compute_kernel_weights(0, bandwidth = 2, kernel = "exponential", adaptive = FALSE)
  expect_equal(w, 1)
})

test_that("exponential kernel: weight is positive everywhere", {
  w <- compute_kernel_weights(c(0, 1, 10, 100), bandwidth = 2,
                              kernel = "exponential", adaptive = FALSE)
  expect_true(all(w > 0))
})

test_that("exponential kernel: known value at d=bw", {
  # exp(-1) when d = bw
  w <- compute_kernel_weights(3, bandwidth = 3, kernel = "exponential", adaptive = FALSE)
  expect_equal(w, exp(-1), tolerance = 1e-10)
})

# ===========================================================================
# compute_kernel_weights — tricube
# ===========================================================================

test_that("tricube kernel: weight at d=0 equals 1", {
  w <- compute_kernel_weights(0, bandwidth = 1, kernel = "tricube", adaptive = FALSE)
  expect_equal(w, 1)
})

test_that("tricube kernel: weight is 0 beyond bandwidth", {
  w <- compute_kernel_weights(1.01, bandwidth = 1, kernel = "tricube", adaptive = FALSE)
  expect_equal(w, 0)
})

test_that("tricube kernel: known mid-point value", {
  # d = 0.5, bw = 1: (1 - 0.5^3)^3 = (1 - 0.125)^3 = 0.875^3
  w <- compute_kernel_weights(0.5, bandwidth = 1, kernel = "tricube", adaptive = FALSE)
  expect_equal(w, 0.875^3, tolerance = 1e-10)
})

# ===========================================================================
# compute_kernel_weights — boxcar
# ===========================================================================

test_that("boxcar kernel: weight is 1 inside bandwidth", {
  w <- compute_kernel_weights(c(0, 0.5, 1), bandwidth = 1, kernel = "boxcar", adaptive = FALSE)
  expect_equal(w, c(1, 1, 1))
})

test_that("boxcar kernel: weight is 0 outside bandwidth", {
  w <- compute_kernel_weights(1.01, bandwidth = 1, kernel = "boxcar", adaptive = FALSE)
  expect_equal(w, 0)
})

# ===========================================================================
# fixed bandwidth — general correctness
# ===========================================================================

test_that("fixed bandwidth: all weights are non-negative", {
  d <- c(0, 0.5, 1, 2, 3)
  for (ker in c("bisquare", "gaussian", "exponential", "tricube", "boxcar")) {
    w <- compute_kernel_weights(d, bandwidth = 2, kernel = ker, adaptive = FALSE)
    expect_true(all(w >= 0), label = paste("kernel:", ker))
  }
})

test_that("fixed bandwidth: focal unit (d=0) always has maximum weight", {
  d <- c(0, 0.5, 1, 2)
  for (ker in c("bisquare", "gaussian", "exponential", "tricube", "boxcar")) {
    w <- compute_kernel_weights(d, bandwidth = 2, kernel = ker, adaptive = FALSE)
    expect_equal(which.max(w), 1L, label = paste("kernel:", ker))
  }
})

# ===========================================================================
# adaptive bandwidth
# ===========================================================================

test_that("adaptive bandwidth: uses k-th distance as effective bw", {
  # Distances from focal unit to 4 others: 1, 2, 3, 4
  # With k=2 the effective bw should be 2 (the 2nd smallest distance incl. self=0)
  # Sorted: 0, 1, 2, 3, 4  => sorted[2] = 1
  d <- c(0, 1, 2, 3, 4)
  # k=2: effective bw = sorted_d[2] = 1
  w_adaptive <- compute_kernel_weights(d, bandwidth = 2L, kernel = "bisquare", adaptive = TRUE)
  # Manually: effective bw = 1, bisquare at d=0: 1, d=1: 0, d>1: 0
  expect_equal(w_adaptive[1], 1)
  expect_equal(w_adaptive[2], 0)  # d=1 == bw -> weight 0
  expect_equal(w_adaptive[3], 0)  # d=2 > bw
})

test_that("adaptive bandwidth: k exceeding n issues warning and uses all units", {
  d <- c(0, 1, 2)
  expect_warning(
    w <- compute_kernel_weights(d, bandwidth = 10L, kernel = "gaussian", adaptive = TRUE),
    regexp = "exceeds number"
  )
  expect_equal(length(w), 3L)
})

test_that("adaptive bandwidth: result is consistent for bisquare", {
  # 5 units at distances 0,1,2,3,4; k=3 => effective bw = sorted[3] = 2
  d <- c(0, 1, 2, 3, 4)
  w <- compute_kernel_weights(d, bandwidth = 3L, kernel = "bisquare", adaptive = TRUE)
  # bw = 2; d=0: (1-0)^2 = 1; d=1: (1-0.25)^2=0.5625; d=2: 0; d=3: 0; d=4: 0
  expect_equal(w[1], 1,      tolerance = 1e-10)
  expect_equal(w[2], 0.5625, tolerance = 1e-10)
  expect_equal(w[3], 0,      tolerance = 1e-10)
  expect_equal(w[4], 0,      tolerance = 1e-10)
})

test_that("adaptive bandwidth: non-integer bandwidth triggers error", {
  d <- c(0, 1, 2, 3)
  expect_error(
    compute_kernel_weights(d, bandwidth = 2.5, kernel = "bisquare", adaptive = TRUE),
    regexp = "positive integer"
  )
})

# ===========================================================================
# distance matrix order / ID mapping consistency
# ===========================================================================

test_that("distance matrix row/column order matches input coordinate order", {
  coords <- matrix(c(0, 0,
                     10, 0,
                     0, 10), ncol = 2, byrow = TRUE)
  d <- compute_distance(coords, longlat = FALSE)
  # d[1,2] should be 10 (horizontal), d[1,3] = 10 (vertical), d[2,3] = sqrt(200)
  expect_equal(d[1, 2], 10, tolerance = 1e-10)
  expect_equal(d[1, 3], 10, tolerance = 1e-10)
  expect_equal(d[2, 3], sqrt(200), tolerance = 1e-10)
})

test_that("build_distance_context: distance_matrix rows match ids ordering", {
  coords <- matrix(c(0, 0,
                     5, 0,
                     0, 5), ncol = 2, byrow = TRUE)
  ids <- c("alpha", "beta", "gamma")
  ctx <- build_distance_context(coords, ids = ids)
  # Distance from alpha to beta: 5
  expect_equal(ctx$distance_matrix["alpha", "beta"], 5, tolerance = 1e-10)
  # Distance from beta to gamma: sqrt(50)
  expect_equal(ctx$distance_matrix["beta", "gamma"], sqrt(50), tolerance = 1e-10)
})

test_that("get_local_distances: integer and character access return same values", {
  coords <- make_coords()
  ids <- paste0("u", 1:5)
  ctx <- build_distance_context(coords, ids = ids)
  v_int  <- get_local_distances(ctx, focus_id = 3L)
  v_char <- get_local_distances(ctx$distance_matrix, focus_id = "u3")
  expect_equal(v_int, v_char)
})

# ===========================================================================
# illegal kernel name
# ===========================================================================

test_that("compute_kernel_weights: unknown kernel name triggers error", {
  expect_error(
    compute_kernel_weights(c(0, 1), bandwidth = 1, kernel = "parabola", adaptive = FALSE),
    regexp = "should be one of"
  )
})

#' @title California (sf)
#'
#' @description The counties' boundary in California
#'
#' @usage data(California)
#'
#' @format An \code{sf} object with 58 rows (one per county) and two columns:
#' \describe{
#' \item{GEOID}{a numeric vector, fips IDs of the counties}
#' \item{geometry}{sfc_MULTIPOLYGON, county boundary polygons (CRS: NAD83 longlat)}
#' }
#'
#' @author Chao Li <chaoli0394@gmail.com> Shunsuke Managi <managi.s@gmail.com>
#'
#' @examples
#' data(California)
#' class(California)
"California"

#' @title Input Validation Functions for GWPR.light 1.0.0
#'
#' @description
#' These internal functions validate user inputs before any expensive computation,
#' providing early and consistent error messages.
#'
#' @name validation
#' @keywords internal
NULL

# ---------------------------------------------------------------------------
# validate_formula
# ---------------------------------------------------------------------------

#' Validate a model formula against a data frame
#'
#' @param formula A formula object.
#' @param data    A data frame containing the model variables.
#'
#' @return Invisibly returns `TRUE` when valid; stops with an informative message
#'   when a problem is detected.
#' @keywords internal
validate_formula <- function(formula, data) {
  if (!inherits(formula, "formula")) {
    stop("`formula` must be a formula object.", call. = FALSE)
  }

  if (!is.data.frame(data)) {
    stop("`data` must be a data.frame.", call. = FALSE)
  }

  # Extract response variable (left-hand side)
  lhs <- all.vars(formula[[2]])
  missing_lhs <- setdiff(lhs, names(data))
  if (length(missing_lhs) > 0) {
    stop(
      "Response variable(s) not found in `data`: ",
      paste(missing_lhs, collapse = ", "),
      call. = FALSE
    )
  }

  # Extract right-hand side variables
  rhs_vars <- all.vars(formula[[3]])
  # Remove bare "." which means all columns
  rhs_vars <- rhs_vars[rhs_vars != "."]
  missing_rhs <- setdiff(rhs_vars, names(data))
  if (length(missing_rhs) > 0) {
    stop(
      "Predictor variable(s) not found in `data`: ",
      paste(missing_rhs, collapse = ", "),
      call. = FALSE
    )
  }

  invisible(TRUE)
}

# ---------------------------------------------------------------------------
# validate_panel_index
# ---------------------------------------------------------------------------

#' Validate panel index columns in a data frame
#'
#' @param data A data frame.
#' @param id   Name of the unit (individual) index column.
#' @param time Name of the time index column.
#'
#' @return Invisibly returns `TRUE` when valid; stops otherwise.
#' @keywords internal
validate_panel_index <- function(data, id, time) {
  if (!is.data.frame(data)) {
    stop("`data` must be a data.frame.", call. = FALSE)
  }

  if (!is.character(id) || length(id) != 1L) {
    stop("`id` must be a single character string.", call. = FALSE)
  }

  if (!is.character(time) || length(time) != 1L) {
    stop("`time` must be a single character string.", call. = FALSE)
  }

  if (!id %in% names(data)) {
    stop(
      "ID column '", id, "' not found in `data`.",
      call. = FALSE
    )
  }

  if (!time %in% names(data)) {
    stop(
      "Time column '", time, "' not found in `data`.",
      call. = FALSE
    )
  }

  invisible(TRUE)
}

# ---------------------------------------------------------------------------
# validate_spatial
# ---------------------------------------------------------------------------

#' Validate a spatial sf object
#'
#' @param spatial An `sf` object representing spatial units.
#' @param id      Name of the ID column that must be present in `spatial`.
#'
#' @return Invisibly returns `TRUE` when valid; stops otherwise.
#' @keywords internal
validate_spatial <- function(spatial, id) {
  if (!requireNamespace("sf", quietly = TRUE)) {
    stop(
      "Package 'sf' is required for spatial operations. ",
      "Please install it with: install.packages('sf')",
      call. = FALSE
    )
  }

  if (!inherits(spatial, "sf")) {
    stop(
      "`spatial` must be an sf object. ",
      "sp objects are not supported in GWPR.light 1.0.0. ",
      "Please convert using sf::st_as_sf().",
      call. = FALSE
    )
  }

  if (!is.character(id) || length(id) != 1L) {
    stop("`id` must be a single character string.", call. = FALSE)
  }

  if (!id %in% names(spatial)) {
    stop(
      "ID column '", id, "' not found in `spatial`.",
      call. = FALSE
    )
  }

  invisible(TRUE)
}

# ---------------------------------------------------------------------------
# validate_family_response
# ---------------------------------------------------------------------------

#' Validate the response variable against the specified family
#'
#' @param data    A data frame.
#' @param formula A formula; the left-hand side is the response variable.
#' @param family  Character string: `"gaussian"` or `"binomial"`.
#'
#' @return Invisibly returns `TRUE` when valid; stops otherwise.
#' @keywords internal
validate_family_response <- function(data, formula, family) {
  if (!inherits(formula, "formula")) {
    stop("`formula` must be a formula object.", call. = FALSE)
  }

  family <- match.arg(family, c("gaussian", "binomial"))

  lhs_name <- all.vars(formula[[2]])[[1]]

  if (!lhs_name %in% names(data)) {
    stop(
      "Response variable '", lhs_name, "' not found in `data`.",
      call. = FALSE
    )
  }

  y <- data[[lhs_name]]

  if (family == "gaussian") {
    if (!is.numeric(y)) {
      stop(
        "For family = 'gaussian', the response variable '", lhs_name,
        "' must be numeric.",
        call. = FALSE
      )
    }
  }

  if (family == "binomial") {
    if (is.factor(y)) {
      n_levels <- nlevels(y)
      if (n_levels > 2L) {
        stop(
          "For family = 'binomial', the response variable '", lhs_name,
          "' is a factor with ", n_levels, " levels. ",
          "GWPR.light 1.0.0 does not support multinomial response variables. ",
          "Please provide a binary (two-level) factor or 0/1 numeric.",
          call. = FALSE
        )
      }
      if (n_levels < 2L) {
        stop(
          "For family = 'binomial', the response variable '", lhs_name,
          "' must have exactly two levels.",
          call. = FALSE
        )
      }
    } else if (is.logical(y)) {
      # logical is fine (coercible to 0/1)
    } else if (is.numeric(y)) {
      unique_vals <- unique(y[!is.na(y)])
      invalid_vals <- setdiff(unique_vals, c(0, 1))
      if (length(invalid_vals) > 0) {
        stop(
          "For family = 'binomial', numeric response '", lhs_name,
          "' must contain only 0 and 1. ",
          "Found unexpected value(s): ",
          paste(invalid_vals, collapse = ", "),
          call. = FALSE
        )
      }
    } else {
      stop(
        "For family = 'binomial', the response variable '", lhs_name,
        "' must be 0/1 numeric, logical, or a two-level factor.",
        call. = FALSE
      )
    }
  }

  invisible(TRUE)
}

# ---------------------------------------------------------------------------
# validate_workers
# ---------------------------------------------------------------------------

#' Validate the workers argument
#'
#' @param workers Number of parallel workers. Must be a positive integer.
#'
#' @return Invisibly returns `TRUE` when valid; stops otherwise.
#' @keywords internal
validate_workers <- function(workers) {
  if (!is.numeric(workers) || length(workers) != 1L) {
    stop("`workers` must be a single positive integer.", call. = FALSE)
  }

  if (is.na(workers) || workers != floor(workers) || workers < 1L) {
    stop(
      "`workers` must be a positive integer (>= 1). ",
      "Received: ", workers,
      call. = FALSE
    )
  }

  invisible(TRUE)
}

# ---------------------------------------------------------------------------
# validate_bandwidth_control
# ---------------------------------------------------------------------------

#' Validate bandwidth search control parameters
#'
#' @param method   Character: `"grid"`, `"sgd"`, or `"random"`.
#' @param control  Named list of control parameters.
#' @param adaptive Logical; if `TRUE`, bandwidth is a neighbour count.
#'
#' @return Invisibly returns `TRUE` when valid; stops otherwise.
#' @keywords internal
validate_bandwidth_control <- function(method, control, adaptive) {
  method <- match.arg(method, c("grid", "sgd", "random"))

  if (!is.list(control)) {
    stop("`control` must be a list.", call. = FALSE)
  }

  if (!is.logical(adaptive) || length(adaptive) != 1L || is.na(adaptive)) {
    stop("`adaptive` must be a single logical value (TRUE or FALSE).", call. = FALSE)
  }

  if (method == "grid") {
    required <- c("lower", "upper", "step")
    missing_keys <- setdiff(required, names(control))
    if (length(missing_keys) > 0) {
      stop(
        "Grid bandwidth search requires control list entries: ",
        paste(missing_keys, collapse = ", "),
        ". Please supply all of `lower`, `upper`, and `step`.",
        call. = FALSE
      )
    }

    lower <- control[["lower"]]
    upper <- control[["upper"]]
    step  <- control[["step"]]

    if (!is.numeric(lower) || length(lower) != 1L || is.na(lower)) {
      stop("`control$lower` must be a single numeric value.", call. = FALSE)
    }
    if (!is.numeric(upper) || length(upper) != 1L || is.na(upper)) {
      stop("`control$upper` must be a single numeric value.", call. = FALSE)
    }
    if (!is.numeric(step) || length(step) != 1L || is.na(step)) {
      stop("`control$step` must be a single numeric value.", call. = FALSE)
    }

    if (lower >= upper) {
      stop(
        "`control$lower` (", lower, ") must be less than `control$upper` (", upper, ").",
        call. = FALSE
      )
    }

    if (step <= 0) {
      stop(
        "`control$step` must be positive. Received: ", step,
        call. = FALSE
      )
    }

    if (adaptive) {
      if (lower != floor(lower) || lower < 1L) {
        stop(
          "For adaptive bandwidth grid search, `control$lower` must be a positive integer. ",
          "Received: ", lower,
          call. = FALSE
        )
      }
    }
  }

  invisible(TRUE)
}

# ---------------------------------------------------------------------------
# validate_inputs  (top-level orchestrator)
# ---------------------------------------------------------------------------

#' Validate all inputs to the main GWPR functions
#'
#' This is the top-level validation entry point.  It calls the individual
#' `validate_*` helpers and also checks `model`, `effect`, and `kernel`.
#'
#' @param formula  A formula object.
#' @param data     A data frame.
#' @param spatial  An `sf` object.
#' @param id       Name of the unit ID column (character).
#' @param time     Name of the time column (character).
#' @param family   `"gaussian"` or `"binomial"`.
#' @param model    `"pooling"`, `"within"`, or `"random"`.
#' @param effect   `"individual"`, `"time"`, `"two-way"`, or `"nested"`.
#' @param kernel   One of `"bisquare"`, `"gaussian"`, `"exponential"`,
#'   `"tricube"`, `"boxcar"`.
#' @param adaptive Logical.
#' @param workers  Positive integer.
#'
#' @return Invisibly returns `TRUE` when all checks pass; stops otherwise.
#' @keywords internal
validate_inputs <- function(formula, data, spatial, id, time,
                            family  = c("gaussian", "binomial"),
                            model   = c("within", "pooling", "random"),
                            effect  = c("individual", "time", "two-way", "nested"),
                            kernel  = c("bisquare", "gaussian", "exponential",
                                        "tricube", "boxcar"),
                            adaptive = FALSE,
                            workers  = 1L) {
  # --- sub-validators ---
  validate_formula(formula, data)
  validate_panel_index(data, id, time)
  validate_spatial(spatial, id)
  validate_family_response(data, formula, family)
  validate_workers(workers)

  # --- model ---
  valid_models <- c("pooling", "within", "random")
  family_arg  <- match.arg(family,  c("gaussian", "binomial"))
  model_arg   <- match.arg(model,   valid_models)
  effect_arg  <- match.arg(effect,  c("individual", "time", "two-way", "nested"))
  kernel_arg  <- match.arg(kernel,  c("bisquare", "gaussian", "exponential",
                                       "tricube", "boxcar"))

  invisible(TRUE)
}

library(testthat)

# ---------------------------------------------------------------------------
# Helper: small simulated panel data and sf spatial object
# ---------------------------------------------------------------------------

make_test_data <- function() {
  data.frame(
    id   = rep(c("A", "B", "C"), each = 3),
    time = rep(1:3, times = 3),
    y    = c(1.1, 2.2, 3.3, 4.4, 5.5, 6.6, 7.7, 8.8, 9.9),
    x1   = rnorm(9, mean = 0, sd = 1),
    x2   = rnorm(9, mean = 5, sd = 2),
    stringsAsFactors = FALSE
  )
}

make_test_spatial <- function() {
  # Three simple point features aligned with ids A, B, C
  sf::st_as_sf(
    data.frame(
      id = c("A", "B", "C"),
      lon = c(0, 1, 2),
      lat = c(0, 1, 2)
    ),
    coords = c("lon", "lat"),
    crs = 4326
  )
}

# ---------------------------------------------------------------------------
# validate_formula
# ---------------------------------------------------------------------------

test_that("validate_formula: valid formula passes", {
  df <- make_test_data()
  expect_true(validate_formula(y ~ x1 + x2, df))
})

test_that("validate_formula: non-formula input raises error", {
  df <- make_test_data()
  expect_error(validate_formula("y ~ x1", df), "`formula` must be a formula object")
})

test_that("validate_formula: missing response variable raises error", {
  df <- make_test_data()
  expect_error(validate_formula(missing_y ~ x1, df), "Response variable")
})

test_that("validate_formula: missing predictor variable raises error", {
  df <- make_test_data()
  expect_error(validate_formula(y ~ x1 + x_missing, df), "Predictor variable")
})

test_that("validate_formula: non-data.frame raises error", {
  expect_error(validate_formula(y ~ x1, list(y = 1, x1 = 2)), "`data` must be a data.frame")
})

# ---------------------------------------------------------------------------
# validate_panel_index
# ---------------------------------------------------------------------------

test_that("validate_panel_index: valid columns pass", {
  df <- make_test_data()
  expect_true(validate_panel_index(df, "id", "time"))
})

test_that("validate_panel_index: missing id column raises error", {
  df <- make_test_data()
  expect_error(validate_panel_index(df, "no_such_id", "time"), "not found in `data`")
})

test_that("validate_panel_index: missing time column raises error", {
  df <- make_test_data()
  expect_error(validate_panel_index(df, "id", "no_such_time"), "not found in `data`")
})

test_that("validate_panel_index: non-character id raises error", {
  df <- make_test_data()
  expect_error(validate_panel_index(df, 1L, "time"), "`id` must be a single character string")
})

# ---------------------------------------------------------------------------
# validate_spatial
# ---------------------------------------------------------------------------

test_that("validate_spatial: valid sf object with id column passes", {
  sp <- make_test_spatial()
  expect_true(validate_spatial(sp, "id"))
})

test_that("validate_spatial: non-sf object raises error", {
  df <- make_test_data()
  expect_error(validate_spatial(df, "id"), "must be an sf object")
})

test_that("validate_spatial: missing id column in sf raises error", {
  sp <- make_test_spatial()
  expect_error(validate_spatial(sp, "no_such_id"), "not found in `spatial`")
})

test_that("validate_spatial: sp object raises informative error", {
  # A plain data.frame (not sp, not sf) should fail with the sf-required message
  fake_sp <- structure(list(), class = "SpatialPointsDataFrame")
  expect_error(validate_spatial(fake_sp, "id"), "must be an sf object")
})

# ---------------------------------------------------------------------------
# validate_family_response
# ---------------------------------------------------------------------------

test_that("validate_family_response: gaussian with numeric response passes", {
  df <- make_test_data()
  expect_true(validate_family_response(df, y ~ x1, "gaussian"))
})

test_that("validate_family_response: gaussian with non-numeric response raises error", {
  df <- make_test_data()
  df$y <- as.character(df$y)
  expect_error(validate_family_response(df, y ~ x1, "gaussian"), "must be numeric")
})

test_that("validate_family_response: binomial with 0/1 numeric passes", {
  df <- make_test_data()
  df$y <- rep(c(0L, 1L, 0L), 3)
  expect_true(validate_family_response(df, y ~ x1, "binomial"))
})

test_that("validate_family_response: binomial with logical passes", {
  df <- make_test_data()
  df$y <- rep(c(TRUE, FALSE, TRUE), 3)
  expect_true(validate_family_response(df, y ~ x1, "binomial"))
})

test_that("validate_family_response: binomial with two-level factor passes", {
  df <- make_test_data()
  df$y <- factor(rep(c("yes", "no", "yes"), 3))
  expect_true(validate_family_response(df, y ~ x1, "binomial"))
})

test_that("validate_family_response: binomial with multi-level factor raises error with 1.0.0 message", {
  df <- make_test_data()
  df$y <- factor(rep(c("a", "b", "c"), 3))
  expect_error(
    validate_family_response(df, y ~ x1, "binomial"),
    "1.0.0 does not support multinomial"
  )
})

test_that("validate_family_response: binomial with numeric not 0/1 raises error", {
  df <- make_test_data()
  df$y <- c(0, 1, 2, 0, 1, 0, 1, 0, 1)
  expect_error(validate_family_response(df, y ~ x1, "binomial"), "must contain only 0 and 1")
})

test_that("validate_family_response: unknown family raises error", {
  df <- make_test_data()
  expect_error(validate_family_response(df, y ~ x1, "poisson"))
})

# ---------------------------------------------------------------------------
# validate_workers
# ---------------------------------------------------------------------------

test_that("validate_workers: workers = 1 passes", {
  expect_true(validate_workers(1L))
})

test_that("validate_workers: workers = 4 passes", {
  expect_true(validate_workers(4L))
})

test_that("validate_workers: workers = 0 raises error", {
  expect_error(validate_workers(0), "positive integer")
})

test_that("validate_workers: workers = -1 raises error", {
  expect_error(validate_workers(-1), "positive integer")
})

test_that("validate_workers: non-integer numeric raises error", {
  expect_error(validate_workers(1.5), "positive integer")
})

test_that("validate_workers: character raises error", {
  expect_error(validate_workers("2"), "positive integer")
})

test_that("validate_workers: NA raises error", {
  expect_error(validate_workers(NA_real_), "positive integer")
})

# ---------------------------------------------------------------------------
# validate_bandwidth_control
# ---------------------------------------------------------------------------

test_that("validate_bandwidth_control: valid grid control passes", {
  ctrl <- list(lower = 10, upper = 100, step = 10)
  expect_true(validate_bandwidth_control("grid", ctrl, adaptive = FALSE))
})

test_that("validate_bandwidth_control: grid missing lower raises error", {
  ctrl <- list(upper = 100, step = 10)
  expect_error(validate_bandwidth_control("grid", ctrl, FALSE), "lower")
})

test_that("validate_bandwidth_control: grid missing upper raises error", {
  ctrl <- list(lower = 10, step = 10)
  expect_error(validate_bandwidth_control("grid", ctrl, FALSE), "upper")
})

test_that("validate_bandwidth_control: grid missing step raises error", {
  ctrl <- list(lower = 10, upper = 100)
  expect_error(validate_bandwidth_control("grid", ctrl, FALSE), "step")
})

test_that("validate_bandwidth_control: lower >= upper raises error", {
  ctrl <- list(lower = 100, upper = 10, step = 5)
  expect_error(validate_bandwidth_control("grid", ctrl, FALSE), "must be less than")
})

test_that("validate_bandwidth_control: step <= 0 raises error", {
  ctrl <- list(lower = 10, upper = 100, step = -1)
  expect_error(validate_bandwidth_control("grid", ctrl, FALSE), "must be positive")
})

test_that("validate_bandwidth_control: adaptive grid with non-integer lower raises error", {
  ctrl <- list(lower = 1.5, upper = 10, step = 1)
  expect_error(validate_bandwidth_control("grid", ctrl, adaptive = TRUE), "positive integer")
})

test_that("validate_bandwidth_control: adaptive grid with valid integer lower passes", {
  ctrl <- list(lower = 2, upper = 20, step = 2)
  expect_true(validate_bandwidth_control("grid", ctrl, adaptive = TRUE))
})

test_that("validate_bandwidth_control: sgd method with empty control passes", {
  expect_true(validate_bandwidth_control("sgd", list(), adaptive = FALSE))
})

test_that("validate_bandwidth_control: random method with empty control passes", {
  expect_true(validate_bandwidth_control("random", list(), adaptive = FALSE))
})

test_that("validate_bandwidth_control: non-list control raises error", {
  expect_error(validate_bandwidth_control("grid", "not_a_list", FALSE), "`control` must be a list")
})

# ---------------------------------------------------------------------------
# validate_inputs (top-level)
# ---------------------------------------------------------------------------

test_that("validate_inputs: all valid inputs pass", {
  df <- make_test_data()
  sp <- make_test_spatial()
  expect_true(
    validate_inputs(
      formula  = y ~ x1 + x2,
      data     = df,
      spatial  = sp,
      id       = "id",
      time     = "time",
      family   = "gaussian",
      model    = "within",
      effect   = "individual",
      kernel   = "bisquare",
      adaptive = FALSE,
      workers  = 1L
    )
  )
})

test_that("validate_inputs: invalid model raises error", {
  df <- make_test_data()
  sp <- make_test_spatial()
  expect_error(
    validate_inputs(
      formula  = y ~ x1,
      data     = df,
      spatial  = sp,
      id       = "id",
      time     = "time",
      family   = "gaussian",
      model    = "fixed",      # not a valid model
      effect   = "individual",
      kernel   = "bisquare",
      adaptive = FALSE,
      workers  = 1L
    )
  )
})

test_that("validate_inputs: invalid effect raises error", {
  df <- make_test_data()
  sp <- make_test_spatial()
  expect_error(
    validate_inputs(
      formula  = y ~ x1,
      data     = df,
      spatial  = sp,
      id       = "id",
      time     = "time",
      family   = "gaussian",
      model    = "within",
      effect   = "bad_effect",
      kernel   = "bisquare",
      adaptive = FALSE,
      workers  = 1L
    )
  )
})

test_that("validate_inputs: invalid kernel raises error", {
  df <- make_test_data()
  sp <- make_test_spatial()
  expect_error(
    validate_inputs(
      formula  = y ~ x1,
      data     = df,
      spatial  = sp,
      id       = "id",
      time     = "time",
      family   = "gaussian",
      model    = "within",
      effect   = "individual",
      kernel   = "linear",     # not supported
      adaptive = FALSE,
      workers  = 1L
    )
  )
})

test_that("validate_inputs: workers = 0 raises error", {
  df <- make_test_data()
  sp <- make_test_spatial()
  expect_error(
    validate_inputs(
      formula  = y ~ x1,
      data     = df,
      spatial  = sp,
      id       = "id",
      time     = "time",
      family   = "gaussian",
      model    = "within",
      effect   = "individual",
      kernel   = "bisquare",
      adaptive = FALSE,
      workers  = 0L
    ),
    "positive integer"
  )
})

test_that("validate_inputs: non-sf spatial raises error", {
  df <- make_test_data()
  expect_error(
    validate_inputs(
      formula  = y ~ x1,
      data     = df,
      spatial  = df,           # data.frame, not sf
      id       = "id",
      time     = "time",
      family   = "gaussian",
      model    = "within",
      effect   = "individual",
      kernel   = "bisquare",
      adaptive = FALSE,
      workers  = 1L
    ),
    "must be an sf object"
  )
})

test_that("validate_inputs: binomial multi-level factor raises 1.0.0 error", {
  df <- make_test_data()
  df$y <- factor(rep(c("a", "b", "c"), 3))
  sp <- make_test_spatial()
  expect_error(
    validate_inputs(
      formula  = y ~ x1,
      data     = df,
      spatial  = sp,
      id       = "id",
      time     = "time",
      family   = "binomial",
      model    = "pooling",
      effect   = "individual",
      kernel   = "bisquare",
      adaptive = FALSE,
      workers  = 1L
    ),
    "1.0.0 does not support multinomial"
  )
})

#' @title Memory Estimation Module for GWPR.light 1.0.0
#'
#' @description
#' Functions for estimating memory usage before running GWPR models.
#' The module provides warnings about memory risk levels to help users
#' avoid out-of-memory errors. These functions only warn; they never
#' stop execution.
#'
#' @name memory
#' @keywords internal
NULL

# ---------------------------------------------------------------------------
# estimate_memory
# ---------------------------------------------------------------------------

#' Estimate memory usage for a GWPR run
#'
#' Calculates an approximate memory requirement (in bytes) based on the
#' data dimensions stored in a `gwpr_context` object, the number of
#' parallel workers, and whether the full distance matrix will be cached.
#'
#' The two main cost components are:
#' \itemize{
#'   \item Distance matrix (when \code{cache_distance = TRUE}):
#'         \eqn{n\_units^2 \times 8} bytes.
#'   \item Local model working copies (one per worker):
#'         \eqn{n\_rows \times n\_vars \times 8 \times workers} bytes.
#' }
#'
#' @param context A `gwpr_context` list containing at least
#'   \code{metadata$n_units}, \code{metadata$n_time}, and
#'   \code{metadata$n_vars}.  If these keys are absent, the function falls
#'   back to direct arguments \code{n_units}, \code{n_time}, and
#'   \code{n_vars} when supplied via \code{...}.  Alternatively, pass a
#'   plain list with the required scalar fields directly.
#' @param workers Positive integer.  Number of parallel workers.
#' @param cache_distance Logical or \code{NULL}.  When \code{TRUE} the full
#'   \eqn{n\_units \times n\_units} distance matrix is assumed to be kept in
#'   memory.  When \code{NULL} (default) the function assumes caching is
#'   enabled for a conservative estimate.
#'
#' @return A named list with class \code{"gwpr_memory_estimate"}:
#'   \describe{
#'     \item{n_units}{Number of spatial units.}
#'     \item{n_time}{Number of time periods.}
#'     \item{n_vars}{Number of explanatory variables.}
#'     \item{n_rows}{Total panel rows (\code{n_units * n_time}).}
#'     \item{workers}{Workers used for the estimate.}
#'     \item{cache_distance}{Whether distance caching was assumed.}
#'     \item{distance_bytes}{Bytes for the distance matrix (0 if not cached).}
#'     \item{model_bytes}{Bytes for local model copies across workers.}
#'     \item{total_bytes}{Total estimated bytes.}
#'     \item{risk}{Character risk level: \code{"low"}, \code{"medium"}, or
#'       \code{"high"}.}
#'   }
#' @keywords internal
estimate_memory <- function(context, workers = 1, cache_distance = NULL) {
  # ---- resolve dimensions ------------------------------------------------
  if (inherits(context, "gwpr_context") || is.list(context)) {
    # Try metadata sub-list first (populated by data_prepare)
    meta <- context[["metadata"]]
    n_units <- meta[["n_units"]] %||% context[["n_units"]]
    n_time  <- meta[["n_time"]]  %||% context[["n_time"]]
    n_vars  <- meta[["n_vars"]]  %||% context[["n_vars"]]

    # Fall back to model_matrix dimensions when available
    if (is.null(n_vars) && !is.null(context[["model_matrix"]])) {
      n_vars <- ncol(context[["model_matrix"]])
    }
    # Fall back to counting unique IDs / times from panel_data
    if (!is.null(context[["panel_data"]])) {
      pd <- context[["panel_data"]]
      id_col   <- context[["id"]]
      time_col <- context[["time"]]
      if (is.null(n_units) && !is.null(id_col) && id_col %in% names(pd)) {
        n_units <- length(unique(pd[[id_col]]))
      }
      if (is.null(n_time) && !is.null(time_col) && time_col %in% names(pd)) {
        n_time <- length(unique(pd[[time_col]]))
      }
    }
  } else {
    stop("`context` must be a list or gwpr_context object.", call. = FALSE)
  }

  # ---- validate resolved values ------------------------------------------
  if (is.null(n_units) || !is.numeric(n_units) || n_units <= 0) {
    stop(
      "Cannot resolve `n_units` from `context`. ",
      "Set `context$metadata$n_units` or `context$n_units`.",
      call. = FALSE
    )
  }
  if (is.null(n_time) || !is.numeric(n_time) || n_time <= 0) {
    stop(
      "Cannot resolve `n_time` from `context`. ",
      "Set `context$metadata$n_time` or `context$n_time`.",
      call. = FALSE
    )
  }
  if (is.null(n_vars) || !is.numeric(n_vars) || n_vars <= 0) {
    stop(
      "Cannot resolve `n_vars` from `context`. ",
      "Set `context$metadata$n_vars` or `context$n_vars`.",
      call. = FALSE
    )
  }

  if (!is.numeric(workers) || length(workers) != 1L || workers < 1) {
    stop("`workers` must be a positive integer.", call. = FALSE)
  }
  workers <- as.integer(workers)

  # Default: assume caching enabled (conservative)
  if (is.null(cache_distance)) cache_distance <- TRUE

  n_units <- as.numeric(n_units)
  n_time  <- as.numeric(n_time)
  n_vars  <- as.numeric(n_vars)
  n_rows  <- n_units * n_time

  bytes_per_double <- 8

  # Distance matrix: n_units x n_units doubles
  distance_bytes <- if (isTRUE(cache_distance)) {
    n_units^2 * bytes_per_double
  } else {
    0
  }

  # Local model working copies: one per worker
  model_bytes <- n_rows * n_vars * bytes_per_double * workers

  total_bytes <- distance_bytes + model_bytes

  result <- list(
    n_units        = n_units,
    n_time         = n_time,
    n_vars         = n_vars,
    n_rows         = n_rows,
    workers        = workers,
    cache_distance = cache_distance,
    distance_bytes = distance_bytes,
    model_bytes    = model_bytes,
    total_bytes    = total_bytes,
    risk           = classify_memory_risk(total_bytes)
  )
  structure(result, class = "gwpr_memory_estimate")
}

# ---------------------------------------------------------------------------
# classify_memory_risk
# ---------------------------------------------------------------------------

#' Classify memory risk level
#'
#' Maps an estimated byte count to a human-readable risk category.
#'
#' Thresholds:
#' \itemize{
#'   \item \strong{low}:    < 500 MB
#'   \item \strong{medium}: 500 MB -- 2 GB
#'   \item \strong{high}:   > 2 GB
#' }
#'
#' @param estimated_bytes Non-negative numeric; total estimated bytes.
#'
#' @return A character string: \code{"low"}, \code{"medium"}, or
#'   \code{"high"}.
#' @keywords internal
classify_memory_risk <- function(estimated_bytes) {
  if (!is.numeric(estimated_bytes) || length(estimated_bytes) != 1L) {
    stop("`estimated_bytes` must be a single numeric value.", call. = FALSE)
  }
  if (estimated_bytes < 0) {
    stop("`estimated_bytes` must be non-negative.", call. = FALSE)
  }

  mb_500 <- 500  * 1024^2   # 500 MB in bytes
  gb_2   <- 2    * 1024^3   #   2 GB in bytes

  if (estimated_bytes < mb_500) {
    "low"
  } else if (estimated_bytes <= gb_2) {
    "medium"
  } else {
    "high"
  }
}

# ---------------------------------------------------------------------------
# format_memory_warning
# ---------------------------------------------------------------------------

#' Format a human-readable memory warning message
#'
#' Converts a `gwpr_memory_estimate` object (produced by
#' \code{\link{estimate_memory}}) into a readable character string.  For
#' high-risk estimates the message also includes actionable suggestions.
#'
#' @param memory_estimate A `gwpr_memory_estimate` list, typically the
#'   return value of \code{\link{estimate_memory}}.
#'
#' @return A character string containing the warning text.  The string is
#'   suitable for passing to \code{message()} or \code{warning()}.
#' @keywords internal
format_memory_warning <- function(memory_estimate) {
  if (!inherits(memory_estimate, "gwpr_memory_estimate")) {
    stop(
      "`memory_estimate` must be a `gwpr_memory_estimate` object ",
      "(returned by `estimate_memory()`).",
      call. = FALSE
    )
  }

  total_gb  <- memory_estimate$total_bytes / 1024^3
  risk      <- memory_estimate$risk
  workers   <- memory_estimate$workers
  cache     <- memory_estimate$cache_distance
  n_units   <- memory_estimate$n_units

  risk_label <- switch(
    risk,
    low    = "LOW",
    medium = "MEDIUM",
    high   = "HIGH"
  )

  header <- sprintf(
    "[GWPR Memory Estimate] Risk: %s | Estimated: %.2f GB",
    risk_label, total_gb
  )

  detail <- sprintf(
    "  n_units=%g, n_time=%g, n_vars=%g, n_rows=%g, workers=%d, cache_distance=%s",
    memory_estimate$n_units,
    memory_estimate$n_time,
    memory_estimate$n_vars,
    memory_estimate$n_rows,
    workers,
    if (isTRUE(cache)) "TRUE" else "FALSE"
  )

  msg <- paste(header, detail, sep = "\n")

  if (risk == "high") {
    suggestions <- paste(
      "  Suggestions to reduce memory usage:",
      "    - Reduce `workers` (currently %d). Try workers = 1.",
      "    - Disable distance caching: set cache_distance = FALSE.",
      "    - Reduce the number of spatial units (currently %g).",
      "    - Subset the data to fewer time periods.",
      sep = "\n"
    )
    suggestions <- sprintf(suggestions, workers, n_units)
    msg <- paste(msg, suggestions, sep = "\n")
  } else if (risk == "medium") {
    msg <- paste(
      msg,
      "  Consider monitoring system memory during the run.",
      sep = "\n"
    )
  }

  msg
}

# ---------------------------------------------------------------------------
# Helper: null-coalescing operator (local, not exported)
# ---------------------------------------------------------------------------

`%||%` <- function(x, y) if (is.null(x)) y else x

test_that("estimate_memory returns a gwpr_memory_estimate with correct structure", {
  ctx <- list(
    n_units = 50,
    n_time  = 10,
    n_vars  = 3,
    metadata = list()
  )
  result <- estimate_memory(ctx, workers = 1, cache_distance = TRUE)

  expect_s3_class(result, "gwpr_memory_estimate")
  expect_named(
    result,
    c("n_units", "n_time", "n_vars", "n_rows", "workers",
      "cache_distance", "distance_bytes", "model_bytes",
      "total_bytes", "risk"),
    ignore.order = TRUE
  )
  expect_equal(result$n_rows, 50 * 10)
  expect_equal(result$workers, 1L)
  expect_true(result$cache_distance)
})

test_that("estimate_memory calculates distance_bytes correctly when caching", {
  ctx <- list(n_units = 100, n_time = 5, n_vars = 2, metadata = list())
  result <- estimate_memory(ctx, workers = 1, cache_distance = TRUE)

  expected_dist  <- 100^2 * 8
  expected_model <- 100 * 5 * 2 * 8 * 1
  expect_equal(result$distance_bytes, expected_dist)
  expect_equal(result$model_bytes, expected_model)
  expect_equal(result$total_bytes, expected_dist + expected_model)
})

test_that("estimate_memory sets distance_bytes to 0 when cache_distance = FALSE", {
  ctx <- list(n_units = 100, n_time = 5, n_vars = 2, metadata = list())
  result <- estimate_memory(ctx, workers = 1, cache_distance = FALSE)

  expect_equal(result$distance_bytes, 0)
  expect_gt(result$model_bytes, 0)
})

test_that("estimate_memory uses conservative default (cache_distance = TRUE) when NULL", {
  ctx <- list(n_units = 100, n_time = 5, n_vars = 2, metadata = list())
  result_null <- estimate_memory(ctx, workers = 1, cache_distance = NULL)
  result_true <- estimate_memory(ctx, workers = 1, cache_distance = TRUE)

  expect_equal(result_null$total_bytes, result_true$total_bytes)
  expect_true(result_null$cache_distance)
})

test_that("workers increase causes estimated memory to not decrease", {
  ctx <- list(n_units = 80, n_time = 8, n_vars = 4, metadata = list())

  mem1 <- estimate_memory(ctx, workers = 1)
  mem2 <- estimate_memory(ctx, workers = 2)
  mem4 <- estimate_memory(ctx, workers = 4)

  expect_gte(mem2$total_bytes, mem1$total_bytes)
  expect_gte(mem4$total_bytes, mem2$total_bytes)
})

test_that("workers increase scales model_bytes linearly", {
  ctx <- list(n_units = 50, n_time = 5, n_vars = 3, metadata = list())

  mem1 <- estimate_memory(ctx, workers = 1, cache_distance = FALSE)
  mem3 <- estimate_memory(ctx, workers = 3, cache_distance = FALSE)

  expect_equal(mem3$model_bytes, mem1$model_bytes * 3)
})

test_that("large n_units triggers high memory risk", {
  # n_units = 20000 -> distance matrix = 20000^2 * 8 = 3.2 GB -> high risk
  ctx <- list(n_units = 20000, n_time = 5, n_vars = 2, metadata = list())
  result <- estimate_memory(ctx, workers = 1, cache_distance = TRUE)

  expect_equal(result$risk, "high")
})

test_that("classify_memory_risk returns 'low' below 500 MB", {
  bytes_low <- 100 * 1024^2   # 100 MB
  expect_equal(classify_memory_risk(bytes_low), "low")
})

test_that("classify_memory_risk returns 'medium' between 500 MB and 2 GB", {
  bytes_med <- 1 * 1024^3     # 1 GB
  expect_equal(classify_memory_risk(bytes_med), "medium")
})

test_that("classify_memory_risk returns 'high' above 2 GB", {
  bytes_high <- 3 * 1024^3    # 3 GB
  expect_equal(classify_memory_risk(bytes_high), "high")
})

test_that("classify_memory_risk handles boundary values", {
  mb_500 <- 500 * 1024^2
  gb_2   <- 2   * 1024^3

  # Just below 500 MB -> low
  expect_equal(classify_memory_risk(mb_500 - 1), "low")
  # Exactly 500 MB -> medium
  expect_equal(classify_memory_risk(mb_500), "medium")
  # Exactly 2 GB -> medium
  expect_equal(classify_memory_risk(gb_2), "medium")
  # Just above 2 GB -> high
  expect_equal(classify_memory_risk(gb_2 + 1), "high")
})

test_that("classify_memory_risk rejects invalid input", {
  expect_error(classify_memory_risk("large"), regexp = "numeric")
  expect_error(classify_memory_risk(c(1, 2)), regexp = "single")
  expect_error(classify_memory_risk(-1), regexp = "non-negative")
})

test_that("format_memory_warning returns a character string", {
  ctx  <- list(n_units = 50, n_time = 5, n_vars = 3, metadata = list())
  mem  <- estimate_memory(ctx, workers = 1)
  warn <- format_memory_warning(mem)

  expect_type(warn, "character")
  expect_length(warn, 1L)
})

test_that("format_memory_warning message contains risk level", {
  ctx  <- list(n_units = 50, n_time = 5, n_vars = 3, metadata = list())
  mem  <- estimate_memory(ctx, workers = 1)
  warn <- format_memory_warning(mem)

  expect_true(grepl("LOW|MEDIUM|HIGH", warn))
})

test_that("format_memory_warning for high risk includes actionable suggestions", {
  # Build a high-risk scenario
  ctx  <- list(n_units = 20000, n_time = 5, n_vars = 2, metadata = list())
  mem  <- estimate_memory(ctx, workers = 4, cache_distance = TRUE)
  warn <- format_memory_warning(mem)

  expect_equal(mem$risk, "high")
  # Should mention workers
  expect_true(grepl("workers", warn, ignore.case = TRUE))
  # Should mention cache
  expect_true(grepl("cache_distance", warn, ignore.case = TRUE))
  # Should mention spatial units or n_units
  expect_true(grepl("spatial units|n_units", warn, ignore.case = TRUE))
})

test_that("format_memory_warning for medium risk contains a monitoring note", {
  # Build a medium-risk scenario via classify_memory_risk directly:
  # 800 MB is between 500 MB and 2 GB -> medium
  bytes_medium <- 800 * 1024^2
  ctx <- list(n_units = 10, n_time = 2, n_vars = 1, metadata = list())
  mem <- estimate_memory(ctx, workers = 1, cache_distance = FALSE)
  # Override risk to medium to test the formatting branch independently
  mem$total_bytes <- bytes_medium
  mem$risk        <- classify_memory_risk(bytes_medium)
  expect_equal(mem$risk, "medium")

  warn <- format_memory_warning(mem)
  expect_true(grepl("monitor", warn, ignore.case = TRUE))
})

test_that("format_memory_warning rejects non-estimate input", {
  expect_error(format_memory_warning(list()), regexp = "gwpr_memory_estimate")
  expect_error(format_memory_warning("text"), regexp = "gwpr_memory_estimate")
})

test_that("estimate_memory resolves dimensions from context metadata", {
  ctx <- list(
    metadata = list(n_units = 30, n_time = 4, n_vars = 2)
  )
  result <- estimate_memory(ctx, workers = 1)
  expect_equal(result$n_units, 30)
  expect_equal(result$n_time, 4)
  expect_equal(result$n_vars, 2)
})

test_that("estimate_memory works with gwpr_context object", {
  ctx <- structure(
    list(
      n_units  = 40,
      n_time   = 6,
      n_vars   = 3,
      metadata = list(),
      warnings = character()
    ),
    class = "gwpr_context"
  )
  result <- estimate_memory(ctx, workers = 1)
  expect_s3_class(result, "gwpr_memory_estimate")
  expect_equal(result$n_units, 40)
})

test_that("estimate_memory does not stop execution (only informs)", {
  ctx <- list(n_units = 20000, n_time = 10, n_vars = 5, metadata = list())
  # Must NOT throw an error even for very large (high risk) estimates
  expect_no_error(estimate_memory(ctx, workers = 1, cache_distance = TRUE))
})

test_that("estimate_memory stops with informative error for missing n_units", {
  ctx <- list(n_time = 5, n_vars = 3, metadata = list())
  expect_error(estimate_memory(ctx, workers = 1), regexp = "n_units")
})

test_that("estimate_memory stops with informative error for missing n_time", {
  ctx <- list(n_units = 50, n_vars = 3, metadata = list())
  expect_error(estimate_memory(ctx, workers = 1), regexp = "n_time")
})

test_that("estimate_memory stops with informative error for missing n_vars", {
  ctx <- list(n_units = 50, n_time = 5, metadata = list())
  expect_error(estimate_memory(ctx, workers = 1), regexp = "n_vars")
})

test_that("estimate_memory stops with error for invalid workers", {
  ctx <- list(n_units = 50, n_time = 5, n_vars = 3, metadata = list())
  expect_error(estimate_memory(ctx, workers = 0), regexp = "workers")
  expect_error(estimate_memory(ctx, workers = -1), regexp = "workers")
})

#' @title SGD Bandwidth Search (`bandwidth_sgd.R`)
#'
#' @description
#' Implements a stochastic gradient descent (SGD) based bandwidth search for
#' GWPR models.  A one-dimensional finite-difference gradient approximation is
#' used to iteratively update the bandwidth over a fixed number of epochs.
#' Mini-batch sampling and early stopping are supported.
#'
#' The search does **not** require the user to specify `lower`, `upper`, or
#' `step`; SGD starts from a single initialised bandwidth and follows the
#' (approximate) gradient.  When `lower` / `upper` are supplied they are used
#' as hard constraints.
#'
#' @name bandwidth_sgd
#' @keywords internal
NULL

# ---------------------------------------------------------------------------
# initialize_bandwidth
# ---------------------------------------------------------------------------

#' Initialise a bandwidth value for SGD
#'
#' Randomly initialises a starting bandwidth within `[lower, upper]`.  For
#' fixed bandwidth the initialisation is done on the log scale so the
#' distribution is spread across orders of magnitude.  For adaptive bandwidth
#' the result is rounded to a positive integer.
#'
#' Both `lower` and `upper` **must** be supplied when calling this function.
#'
#' @param control  A list that **must** contain `lower` and `upper`.
#' @param adaptive Logical.  When `TRUE` an integer bandwidth is returned.
#' @param seed     Integer random seed or `NULL`.
#'
#' @return A numeric scalar (positive, and an integer when `adaptive = TRUE`).
#'
#' @examples
#' initialize_bandwidth(list(lower = 1, upper = 100), adaptive = FALSE,
#'                      seed = 42L)
#' initialize_bandwidth(list(lower = 3, upper = 20), adaptive = TRUE,
#'                      seed = 1L)
#'
#' @noRd
initialize_bandwidth <- function(control, adaptive, seed = NULL) {
  lower <- control$lower
  upper <- control$upper

  if (is.null(lower)) {
    stop(
      "SGD bandwidth initialisation requires `control$lower` to be set ",
      "explicitly. Automatic boundary inference is not supported.",
      call. = FALSE
    )
  }
  if (is.null(upper)) {
    stop(
      "SGD bandwidth initialisation requires `control$upper` to be set ",
      "explicitly. Automatic boundary inference is not supported.",
      call. = FALSE
    )
  }
  if (!is.numeric(lower) || length(lower) != 1L || !is.finite(lower) || lower <= 0) {
    stop("`lower` must be a single finite positive numeric value.", call. = FALSE)
  }
  if (!is.numeric(upper) || length(upper) != 1L || !is.finite(upper)) {
    stop("`upper` must be a single finite numeric value.", call. = FALSE)
  }
  if (lower >= upper) {
    stop("`lower` must be strictly less than `upper`.", call. = FALSE)
  }

  if (!is.null(seed)) set.seed(seed)

  if (adaptive) {
    bw <- as.integer(round(stats::runif(1L, min = lower, max = upper)))
    bw <- max(1L, min(bw, as.integer(upper)))
  } else {
    # Log-scale initialisation
    log_bw <- stats::runif(1L, min = log(lower), max = log(upper))
    bw <- exp(log_bw)
    bw <- max(lower, min(bw, upper))
  }

  bw
}

# ---------------------------------------------------------------------------
# sample_minibatch
# ---------------------------------------------------------------------------

#' Sample a mini-batch of spatial units
#'
#' Draws a random subset of spatial unit IDs from the context.  The fraction
#' of units to include is controlled by `batch_fraction`.  When `seed` is
#' supplied the result is reproducible.
#'
#' The context must contain either `id_map` (a named vector whose names are
#' unit IDs) or `coords` (a matrix whose row count equals the number of spatial
#' units).  If neither is present the function stops with an informative error.
#'
#' @param context        A `gwpr_context` list.
#' @param batch_fraction Numeric in `(0, 1]`.  Fraction of units to include.
#' @param seed           Integer random seed or `NULL`.
#'
#' @return A character vector of unit IDs in the mini-batch.
#'
#' @examples
#' ctx <- list(id_map = c("A" = 1L, "B" = 2L, "C" = 3L, "D" = 4L, "E" = 5L))
#' sample_minibatch(ctx, batch_fraction = 0.6)
#'
#' @noRd
sample_minibatch <- function(context, batch_fraction = 1.0, seed = NULL) {
  if (!is.numeric(batch_fraction) || length(batch_fraction) != 1L ||
      !is.finite(batch_fraction) || batch_fraction <= 0 || batch_fraction > 1) {
    stop("`batch_fraction` must be a single numeric value in (0, 1].",
         call. = FALSE)
  }

  # Determine unit IDs
  if (!is.null(context$id_map)) {
    all_ids <- names(context$id_map)
  } else if (!is.null(context$coords)) {
    all_ids <- as.character(seq_len(nrow(context$coords)))
  } else if (!is.null(context$panel_data) && !is.null(context$id)) {
    all_ids <- as.character(unique(context$panel_data[[context$id]]))
  } else {
    stop(
      "Cannot determine spatial units from context. ",
      "Supply `id_map`, `coords`, or `panel_data` + `id`.",
      call. = FALSE
    )
  }

  n_total <- length(all_ids)
  n_batch <- max(1L, as.integer(ceiling(n_total * batch_fraction)))

  if (!is.null(seed)) set.seed(seed)

  sampled <- sample(all_ids, size = n_batch, replace = FALSE)
  sampled
}

# ---------------------------------------------------------------------------
# estimate_bandwidth_gradient
# ---------------------------------------------------------------------------

#' Estimate the bandwidth gradient via finite differences
#'
#' Approximates the gradient of the scorer objective with respect to bandwidth
#' using a central finite difference:
#'
#' ```
#' gradient = (f(b + delta) - f(b - delta)) / (2 * delta)
#' ```
#'
#' Both evaluations use the same `minibatch`.  When either candidate falls
#' outside `[lower, upper]` the function falls back to a one-sided difference
#' or clips the candidate to the boundary.
#'
#' @param context   A `gwpr_context` list.
#' @param bandwidth Numeric scalar; current bandwidth.
#' @param delta     Numeric scalar > 0; finite-difference step size.
#' @param minibatch Character vector of unit IDs (ignored by the default
#'   scorer; kept for future mini-batch-aware scorers).
#' @param scorer    A function `scorer(context, bandwidth)` that returns a
#'   named list with at minimum a `score` field.
#' @param lower     Numeric lower bound (or `NULL`).
#' @param upper     Numeric upper bound (or `NULL`).
#'
#' @return A numeric scalar gradient estimate.
#'
#' @examples
#' \donttest{
#' grad <- estimate_bandwidth_gradient(
#'   context   = ctx,
#'   bandwidth = 10,
#'   delta     = 1,
#'   minibatch = character(),
#'   scorer    = my_scorer
#' )
#' }
#'
#' @noRd
estimate_bandwidth_gradient <- function(context, bandwidth, delta, minibatch,
                                        scorer, lower = NULL, upper = NULL) {
  if (!is.numeric(delta) || length(delta) != 1L ||
      !is.finite(delta) || delta <= 0) {
    stop("`delta` must be a single finite positive numeric value.", call. = FALSE)
  }

  bw_plus  <- bandwidth + delta
  bw_minus <- bandwidth - delta

  # Clip to [lower, upper] if bounds are provided
  if (!is.null(lower)) {
    bw_plus  <- max(bw_plus,  lower)
    bw_minus <- max(bw_minus, lower)
  }
  if (!is.null(upper)) {
    bw_plus  <- min(bw_plus,  upper)
    bw_minus <- min(bw_minus, upper)
  }

  # Evaluate both sides; use tryCatch to handle scorer failures gracefully
  score_plus <- tryCatch(
    {
      res <- scorer(context, bw_plus)
      if (is.list(res)) res$score %||% NA_real_ else as.numeric(res)
    },
    error = function(e) NA_real_
  )
  score_minus <- tryCatch(
    {
      res <- scorer(context, bw_minus)
      if (is.list(res)) res$score %||% NA_real_ else as.numeric(res)
    },
    error = function(e) NA_real_
  )

  denom <- bw_plus - bw_minus

  if (!is.finite(score_plus) || !is.finite(score_minus) ||
      !is.finite(denom) || denom == 0) {
    return(0)
  }

  (score_plus - score_minus) / denom
}

# ---------------------------------------------------------------------------
# update_bandwidth_sgd
# ---------------------------------------------------------------------------

#' Update bandwidth with SGD step
#'
#' Applies a single gradient-descent update to the current bandwidth.
#'
#' - **Fixed bandwidth**: update is performed on the log scale so that
#'   bandwidth stays positive:
#'   `new_bw = exp(log(bw) - learning_rate * gradient)`.
#' - **Adaptive bandwidth**: update is performed on the integer scale,
#'   then rounded and clipped to `[lower, upper]`.
#'
#' @param bandwidth     Numeric scalar; current bandwidth.
#' @param gradient      Numeric scalar; gradient estimate.
#' @param learning_rate Numeric scalar > 0.
#' @param adaptive      Logical; `TRUE` for adaptive (integer) bandwidth.
#' @param lower         Numeric lower bound (or `NULL`; ignored when `NULL`).
#' @param upper         Numeric upper bound (or `NULL`; ignored when `NULL`).
#'
#' @return A numeric scalar (always positive, integer when `adaptive = TRUE`).
#'
#' @examples
#' update_bandwidth_sgd(10, gradient = 2, learning_rate = 0.5,
#'                      adaptive = FALSE)
#' update_bandwidth_sgd(8L, gradient = -3, learning_rate = 1,
#'                      adaptive = TRUE, lower = 3, upper = 20)
#'
#' @noRd
update_bandwidth_sgd <- function(bandwidth, gradient, learning_rate,
                                 adaptive = FALSE, lower = NULL, upper = NULL) {
  if (adaptive) {
    new_bw <- bandwidth - learning_rate * gradient
    new_bw <- round(new_bw)
    if (!is.null(lower)) new_bw <- max(new_bw, lower)
    if (!is.null(upper)) new_bw <- min(new_bw, upper)
    new_bw <- as.integer(max(1L, new_bw))
  } else {
    # Log-scale update ensures bandwidth > 0
    log_bw <- log(bandwidth) - learning_rate * gradient
    new_bw <- exp(log_bw)
    if (!is.null(lower)) new_bw <- max(new_bw, lower)
    if (!is.null(upper)) new_bw <- min(new_bw, upper)
    # Ensure bandwidth is strictly positive
    new_bw <- max(new_bw, .Machine$double.eps)
  }

  new_bw
}

# ---------------------------------------------------------------------------
# check_early_stopping
# ---------------------------------------------------------------------------

#' Check whether SGD early stopping should trigger
#'
#' Returns `TRUE` when the score has not improved (decreased) for `patience`
#' consecutive epochs.  Returns `FALSE` immediately when `patience = 0`
#' (early stopping disabled).
#'
#' @param history A data.frame of SGD history with at least a `score` column.
#'   Only the last `patience + 1` rows are examined.
#' @param patience Non-negative integer.  `0` disables early stopping.
#'
#' @return Logical scalar.
#'
#' @examples
#' h <- data.frame(score = c(1.0, 0.9, 0.85, 0.85, 0.86))
#' check_early_stopping(h, patience = 2L)  # TRUE
#' check_early_stopping(h, patience = 0L)  # FALSE
#'
#' @noRd
check_early_stopping <- function(history, patience) {
  patience <- as.integer(patience)
  if (patience == 0L) return(FALSE)
  if (!is.data.frame(history) || nrow(history) < patience) return(FALSE)

  # Look at the last `patience` scores
  n <- nrow(history)
  recent_scores <- history$score[(n - patience + 1L):n]

  # Also need the score just before this window
  if (n <= patience) return(FALSE)
  prev_score <- history$score[n - patience]

  # Stop if no improvement over `patience` steps
  # (score has not strictly decreased below the score before the window)
  all(recent_scores >= prev_score, na.rm = TRUE)
}

# ---------------------------------------------------------------------------
# search_bandwidth_sgd  (main entry point)
# ---------------------------------------------------------------------------

#' SGD bandwidth search
#'
#' Searches for an optimal bandwidth using a stochastic gradient descent
#' procedure.  A finite-difference gradient approximation updates the
#' bandwidth iteratively over `control$epoch` epochs.
#'
#' The complete epoch-level history is included in the returned
#' `gwpr_bandwidth` object.
#'
#' @param context  A `gwpr_context` list.  Must contain at minimum `adaptive`.
#' @param control  A list with SGD parameters.  **Required** fields:
#'   \describe{
#'     \item{`lower`}{Numeric. Lower bound of the search space (must be > 0).}
#'     \item{`upper`}{Numeric. Upper bound of the search space.}
#'   }
#'   Optional fields:
#'   \describe{
#'     \item{`epoch`}{Positive integer. Number of gradient steps. Default 10.}
#'     \item{`learning_rate`}{Numeric > 0. Default 0.001.}
#'     \item{`early_stopping_patience`}{Non-negative integer. 0 disables early
#'       stopping. Default 0.}
#'     \item{`batch_fraction`}{Numeric in (0,1]. Mini-batch fraction. Default 1.}
#'     \item{`delta`}{Numeric > 0. Finite-difference step. Default: 1\% of
#'       initial bandwidth.}
#'   }
#' @param scorer   A function `scorer(context, bandwidth)` returning a named
#'   list with at minimum `score`, `criterion`, `n_local_models`,
#'   `n_failed_local_models`, and `metrics`.
#' @param workers  Integer. Number of parallel workers (currently unused;
#'   SGD is inherently sequential).  Default 1.
#' @param seed     Integer random seed or `NULL`.  Applied at initialisation
#'   and stored in the returned object.
#'
#' @return A `gwpr_bandwidth` object (see `new_gwpr_bandwidth()`).  The
#'   `history` data.frame has one row per epoch with the following columns:
#'   `epoch`, `bandwidth_before`, `bandwidth_after`, `score`, `gradient`,
#'   `learning_rate`, `delta`, `batch_size`, `early_stop_counter`, `status`,
#'   `elapsed_time`, plus linear / logistic metric columns.
#'
#' @examples
#' \donttest{
#' result <- search_bandwidth_sgd(
#'   context = ctx,
#'   control = list(lower = 1, upper = 100, epoch = 10L,
#'                  learning_rate = 0.001),
#'   scorer  = my_scorer_fn,
#'   workers = 1L,
#'   seed    = 42L
#' )
#' result$best_bandwidth
#' }
#'
#' @noRd
search_bandwidth_sgd <- function(context, control, scorer,
                                 workers = 1L, seed = NULL) {

  t_total_start <- proc.time()[["elapsed"]]

  # --- Parameters -----------------------------------------------------------
  adaptive      <- context$adaptive %||% FALSE
  epoch         <- as.integer(control$epoch         %||% 10L)
  learning_rate <- control$learning_rate             %||% 0.001
  patience      <- as.integer(
    control$early_stopping_patience %||% 0L
  )
  batch_fraction <- control$batch_fraction %||% 1.0
  lower          <- control$lower
  upper          <- control$upper

  # --- Initialise bandwidth -------------------------------------------------
  bw <- initialize_bandwidth(control, adaptive = adaptive, seed = seed)

  # Determine delta default: 1% of initial bandwidth (minimum 1e-6)
  delta_default <- max(abs(bw) * 0.01, 1e-6)
  delta         <- control$delta %||% delta_default

  # --- History storage ------------------------------------------------------
  history_rows <- vector("list", epoch)

  early_stop_counter <- 0L
  criterion          <- NA_character_
  best_bandwidth     <- bw
  best_score         <- Inf

  # --- SGD loop -------------------------------------------------------------
  for (ep in seq_len(epoch)) {
    t_epoch_start <- proc.time()[["elapsed"]]

    bw_before <- bw

    # 1. Sample mini-batch
    batch_seed <- if (!is.null(seed)) seed + ep else NULL
    minibatch  <- tryCatch(
      sample_minibatch(context, batch_fraction = batch_fraction,
                       seed = batch_seed),
      error = function(e) character()
    )
    batch_size <- length(minibatch)

    # 2. Estimate gradient
    grad <- tryCatch(
      estimate_bandwidth_gradient(
        context   = context,
        bandwidth = bw,
        delta     = delta,
        minibatch = minibatch,
        scorer    = scorer,
        lower     = lower,
        upper     = upper
      ),
      error = function(e) 0
    )

    # 3. Update bandwidth
    bw_after <- update_bandwidth_sgd(
      bandwidth     = bw,
      gradient      = grad,
      learning_rate = learning_rate,
      adaptive      = adaptive,
      lower         = lower,
      upper         = upper
    )

    # 4. Score the new bandwidth
    t_score_start <- proc.time()[["elapsed"]]
    score_res <- tryCatch(
      withCallingHandlers(
        scorer(context, bw_after),
        warning = function(w) invokeRestart("muffleWarning")
      ),
      error = function(e) {
        list(
          score                 = NA_real_,
          criterion             = NA_character_,
          status                = "failed",
          n_local_models        = NA_integer_,
          n_failed_local_models = NA_integer_,
          metrics               = list()
        )
      }
    )
    epoch_elapsed <- proc.time()[["elapsed"]] - t_epoch_start

    score   <- score_res$score     %||% NA_real_
    crit    <- score_res$criterion %||% NA_character_
    status  <- score_res$status    %||% "ok"
    metrics <- score_res$metrics   %||% list()

    if (!is.na(crit) && is.na(criterion)) criterion <- crit

    # Update best
    if (is.finite(score) && score < best_score) {
      best_score     <- score
      best_bandwidth <- bw_after
    }

    # 5. Early stopping counter
    if (ep >= 2L) {
      prev_score <- history_rows[[ep - 1L]]$score %||% NA_real_
      if (is.finite(score) && is.finite(prev_score) && score >= prev_score) {
        early_stop_counter <- early_stop_counter + 1L
      } else {
        early_stop_counter <- 0L
      }
    }

    # Record history row
    history_rows[[ep]] <- list(
      epoch              = ep,
      bandwidth_before   = bw_before,
      bandwidth_after    = bw_after,
      score              = score,
      gradient           = grad,
      learning_rate      = learning_rate,
      delta              = delta,
      batch_size         = batch_size,
      early_stop_counter = early_stop_counter,
      status             = status,
      elapsed_time       = epoch_elapsed,
      # Linear metrics
      r2                 = metrics$R2        %||% NA_real_,
      mse                = metrics$MSE       %||% NA_real_,
      rmse               = metrics$RMSE      %||% NA_real_,
      mae                = metrics$MAE       %||% NA_real_,
      # Logistic metrics
      log_loss           = metrics$log_loss  %||% NA_real_,
      accuracy           = metrics$accuracy  %||% NA_real_,
      precision          = metrics$precision %||% NA_real_,
      recall             = metrics$recall    %||% NA_real_,
      f1_score           = metrics$f1_score  %||% NA_real_
    )

    # Advance bandwidth
    bw <- bw_after

    # 6. Check early stopping
    history_so_far <- .rows_to_df(history_rows[seq_len(ep)])
    if (check_early_stopping(history_so_far, patience = patience)) {
      # Fill remaining epochs with NA rows then break
      for (ep2 in seq(ep + 1L, epoch)) {
        history_rows[[ep2]] <- .make_na_history_row(ep2, bw,
                                                     learning_rate, delta,
                                                     batch_size)
      }
      break
    }
  }

  # --- Assemble history data.frame ------------------------------------------
  history <- .rows_to_df(history_rows)

  total_elapsed <- proc.time()[["elapsed"]] - t_total_start

  new_gwpr_bandwidth(
    method           = "sgd",
    best_bandwidth   = best_bandwidth,
    best_score       = if (is.finite(best_score)) best_score else NA_real_,
    criterion        = criterion,
    history          = history,
    metrics_history  = NULL,
    seed             = seed,
    convergence_info = list(
      epoch         = epoch,
      learning_rate = learning_rate,
      patience      = patience,
      delta         = delta
    ),
    elapsed_time     = total_elapsed,
    warnings         = character()
  )
}

# ---------------------------------------------------------------------------
# Internal helpers
# ---------------------------------------------------------------------------

#' Convert a list of history row lists to a data.frame
#' @keywords internal
.rows_to_df <- function(rows) {
  # Remove NULL entries
  rows <- Filter(Negate(is.null), rows)
  if (length(rows) == 0L) {
    return(data.frame(
      epoch = integer(), bandwidth_before = numeric(),
      bandwidth_after = numeric(), score = numeric(),
      gradient = numeric(), learning_rate = numeric(),
      delta = numeric(), batch_size = integer(),
      early_stop_counter = integer(), status = character(),
      elapsed_time = numeric(),
      r2 = numeric(), mse = numeric(), rmse = numeric(), mae = numeric(),
      log_loss = numeric(), accuracy = numeric(),
      precision = numeric(), recall = numeric(), f1_score = numeric(),
      stringsAsFactors = FALSE
    ))
  }

  data.frame(
    epoch              = vapply(rows, `[[`, integer(1L),  "epoch"),
    bandwidth_before   = vapply(rows, `[[`, numeric(1L),  "bandwidth_before"),
    bandwidth_after    = vapply(rows, `[[`, numeric(1L),  "bandwidth_after"),
    score              = vapply(rows, `[[`, numeric(1L),  "score"),
    gradient           = vapply(rows, `[[`, numeric(1L),  "gradient"),
    learning_rate      = vapply(rows, `[[`, numeric(1L),  "learning_rate"),
    delta              = vapply(rows, `[[`, numeric(1L),  "delta"),
    batch_size         = vapply(rows, `[[`, integer(1L),  "batch_size"),
    early_stop_counter = vapply(rows, `[[`, integer(1L),  "early_stop_counter"),
    status             = vapply(rows, `[[`, character(1L), "status"),
    elapsed_time       = vapply(rows, `[[`, numeric(1L),  "elapsed_time"),
    r2                 = vapply(rows, `[[`, numeric(1L),  "r2"),
    mse                = vapply(rows, `[[`, numeric(1L),  "mse"),
    rmse               = vapply(rows, `[[`, numeric(1L),  "rmse"),
    mae                = vapply(rows, `[[`, numeric(1L),  "mae"),
    log_loss           = vapply(rows, `[[`, numeric(1L),  "log_loss"),
    accuracy           = vapply(rows, `[[`, numeric(1L),  "accuracy"),
    precision          = vapply(rows, `[[`, numeric(1L),  "precision"),
    recall             = vapply(rows, `[[`, numeric(1L),  "recall"),
    f1_score           = vapply(rows, `[[`, numeric(1L),  "f1_score"),
    stringsAsFactors   = FALSE
  )
}

#' Create a placeholder NA history row for epochs after early stopping
#' @keywords internal
.make_na_history_row <- function(ep, bw, learning_rate, delta, batch_size) {
  list(
    epoch              = as.integer(ep),
    bandwidth_before   = bw,
    bandwidth_after    = bw,
    score              = NA_real_,
    gradient           = NA_real_,
    learning_rate      = learning_rate,
    delta              = delta,
    batch_size         = as.integer(batch_size),
    early_stop_counter = NA_integer_,
    status             = "early_stopped",
    elapsed_time       = 0,
    r2                 = NA_real_,
    mse                = NA_real_,
    rmse               = NA_real_,
    mae                = NA_real_,
    log_loss           = NA_real_,
    accuracy           = NA_real_,
    precision          = NA_real_,
    recall             = NA_real_,
    f1_score           = NA_real_
  )
}

#' @noRd
`%||%` <- function(x, y) if (!is.null(x)) x else y

## tests/testthat/test-bandwidth-sgd.R
## Tests for the SGD Bandwidth Search module (R/bandwidth_sgd.R)
##
## All tests use small simulated data and workers = 1 (no parallelism).

# ---------------------------------------------------------------------------
# Helper: minimal mock scorer
# ---------------------------------------------------------------------------
# Returns a score that is the absolute distance from a "best" bandwidth,
# so we can predict which direction gradient descent should move.

make_sgd_mock_scorer <- function(best_bw = 10, fail_bw = NULL) {
  function(context, bandwidth) {
    if (!is.null(fail_bw) && isTRUE(all.equal(bandwidth, fail_bw))) {
      stop("Intentional scorer failure for SGD testing.")
    }
    score <- abs(bandwidth - best_bw)^2 + 0.01
    list(
      score                 = score,
      criterion             = "MSE",
      status                = "ok",
      n_local_models        = 4L,
      n_failed_local_models = 0L,
      metrics               = list(
        R2   = 1 - min(score, 1),
        MSE  = score,
        RMSE = sqrt(score),
        MAE  = score * 0.9
      )
    )
  }
}

# Monotonically increasing scorer (score always increases each call)
make_worsening_scorer <- function() {
  call_count <- 0L
  function(context, bandwidth) {
    call_count <<- call_count + 1L
    score <- call_count * 1.5
    list(
      score                 = score,
      criterion             = "MSE",
      status                = "ok",
      n_local_models        = 2L,
      n_failed_local_models = 0L,
      metrics               = list(
        R2 = 0.5, MSE = score, RMSE = sqrt(score), MAE = score * 0.9
      )
    )
  }
}

# Minimal gwpr_context with an id_map for minibatch sampling
make_sgd_context <- function(adaptive = FALSE, n_units = 6L) {
  id_map <- stats::setNames(seq_len(n_units), paste0("u", seq_len(n_units)))
  new_gwpr_context(
    formula  = y ~ x1,
    family   = "gaussian",
    model    = "pooling",
    effect   = "individual",
    id       = "id",
    time     = "time",
    kernel   = "bisquare",
    adaptive = adaptive,
    threshold = 0.5,
    workers  = 1L,
    id_map   = id_map
  )
}

# ===========================================================================
# initialize_bandwidth
# ===========================================================================

test_that("initialize_bandwidth errors when lower is missing from control", {
  expect_error(
    initialize_bandwidth(list(upper = 100), adaptive = FALSE),
    regexp = "lower"
  )
})

test_that("initialize_bandwidth errors when upper is missing from control", {
  expect_error(
    initialize_bandwidth(list(lower = 1), adaptive = FALSE),
    regexp = "upper"
  )
})

test_that("initialize_bandwidth errors when lower is NULL", {
  expect_error(
    initialize_bandwidth(list(lower = NULL, upper = 100), adaptive = FALSE),
    regexp = "lower"
  )
})

test_that("initialize_bandwidth errors when upper is NULL", {
  expect_error(
    initialize_bandwidth(list(lower = 1, upper = NULL), adaptive = FALSE),
    regexp = "upper"
  )
})

test_that("initialize_bandwidth errors when lower >= upper", {
  expect_error(
    initialize_bandwidth(list(lower = 10, upper = 5), adaptive = FALSE),
    regexp = "lower"
  )
})

test_that("initialize_bandwidth returns value in [lower, upper] for fixed", {
  bw <- initialize_bandwidth(list(lower = 2, upper = 50), adaptive = FALSE,
                              seed = 42L)
  expect_true(is.numeric(bw))
  expect_true(length(bw) == 1L)
  expect_true(bw >= 2)
  expect_true(bw <= 50)
})

test_that("initialize_bandwidth returns positive value for fixed bandwidth", {
  for (s in 1:5) {
    bw <- initialize_bandwidth(list(lower = 0.5, upper = 200), adaptive = FALSE,
                                seed = s)
    expect_true(bw > 0)
  }
})

test_that("initialize_bandwidth returns integer in range for adaptive", {
  bw <- initialize_bandwidth(list(lower = 3, upper = 20), adaptive = TRUE,
                              seed = 7L)
  expect_true(is.integer(bw) || bw == as.integer(bw))
  expect_true(bw >= 3L)
  expect_true(bw <= 20L)
})

test_that("initialize_bandwidth is reproducible with fixed seed", {
  bw1 <- initialize_bandwidth(list(lower = 1, upper = 100), adaptive = FALSE,
                               seed = 99L)
  bw2 <- initialize_bandwidth(list(lower = 1, upper = 100), adaptive = FALSE,
                               seed = 99L)
  expect_equal(bw1, bw2)
})

# ===========================================================================
# sample_minibatch
# ===========================================================================

test_that("sample_minibatch returns a subset of unit IDs", {
  ctx  <- make_sgd_context(n_units = 10L)
  mini <- sample_minibatch(ctx, batch_fraction = 0.5, seed = 1L)
  all_ids <- names(ctx$id_map)
  expect_true(all(mini %in% all_ids))
})

test_that("sample_minibatch returns correct approximate size", {
  ctx  <- make_sgd_context(n_units = 10L)
  mini <- sample_minibatch(ctx, batch_fraction = 0.6, seed = 2L)
  expect_equal(length(mini), as.integer(ceiling(10L * 0.6)))
})

test_that("sample_minibatch is reproducible with fixed seed", {
  ctx  <- make_sgd_context(n_units = 8L)
  m1   <- sample_minibatch(ctx, batch_fraction = 0.5, seed = 42L)
  m2   <- sample_minibatch(ctx, batch_fraction = 0.5, seed = 42L)
  expect_equal(m1, m2)
})

test_that("sample_minibatch differs across seeds", {
  ctx <- make_sgd_context(n_units = 20L)
  m1  <- sample_minibatch(ctx, batch_fraction = 0.5, seed = 1L)
  m2  <- sample_minibatch(ctx, batch_fraction = 0.5, seed = 2L)
  # With 10 out of 20 units, it is extremely unlikely they match
  expect_false(identical(sort(m1), sort(m2)))
})

test_that("sample_minibatch with batch_fraction = 1 returns all units", {
  ctx  <- make_sgd_context(n_units = 5L)
  mini <- sample_minibatch(ctx, batch_fraction = 1.0, seed = 10L)
  expect_equal(sort(mini), sort(names(ctx$id_map)))
})

test_that("sample_minibatch errors on invalid batch_fraction", {
  ctx <- make_sgd_context()
  expect_error(sample_minibatch(ctx, batch_fraction = 0),   regexp = "batch_fraction")
  expect_error(sample_minibatch(ctx, batch_fraction = 1.5), regexp = "batch_fraction")
  expect_error(sample_minibatch(ctx, batch_fraction = -0.5), regexp = "batch_fraction")
})

# ===========================================================================
# estimate_bandwidth_gradient
# ===========================================================================

test_that("estimate_bandwidth_gradient returns a numeric scalar", {
  ctx    <- make_sgd_context()
  scorer <- make_sgd_mock_scorer(best_bw = 10)
  grad   <- estimate_bandwidth_gradient(ctx, bandwidth = 5, delta = 0.5,
                                         minibatch = character(), scorer = scorer)
  expect_true(is.numeric(grad))
  expect_length(grad, 1L)
})

test_that("estimate_bandwidth_gradient is negative when bw < best_bw", {
  # f(b) = (b - best_bw)^2 + 0.01; gradient at b < best_bw is negative
  # → increasing b decreases score → f(b + delta) < f(b - delta) → grad < 0
  ctx    <- make_sgd_context()
  scorer <- make_sgd_mock_scorer(best_bw = 10)
  grad   <- estimate_bandwidth_gradient(ctx, bandwidth = 5, delta = 1,
                                         minibatch = character(), scorer = scorer)
  expect_true(grad < 0)
})

test_that("estimate_bandwidth_gradient is positive when bw > best_bw", {
  ctx    <- make_sgd_context()
  scorer <- make_sgd_mock_scorer(best_bw = 10)
  grad   <- estimate_bandwidth_gradient(ctx, bandwidth = 20, delta = 1,
                                         minibatch = character(), scorer = scorer)
  expect_true(grad > 0)
})

test_that("estimate_bandwidth_gradient returns 0 on scorer failure", {
  ctx    <- make_sgd_context()
  bad_scorer <- function(context, bw) stop("always fails")
  grad <- estimate_bandwidth_gradient(ctx, bandwidth = 5, delta = 1,
                                       minibatch = character(),
                                       scorer = bad_scorer)
  expect_equal(grad, 0)
})

# ===========================================================================
# update_bandwidth_sgd
# ===========================================================================

test_that("update_bandwidth_sgd fixed: result is strictly positive", {
  for (bw in c(0.1, 1, 10, 100)) {
    new_bw <- update_bandwidth_sgd(bw, gradient = 5, learning_rate = 0.5,
                                   adaptive = FALSE)
    expect_true(new_bw > 0, label = paste("bw =", bw))
  }
})

test_that("update_bandwidth_sgd fixed: moves in the right direction", {
  # Positive gradient → bandwidth should decrease
  bw     <- 10
  new_bw <- update_bandwidth_sgd(bw, gradient = 2, learning_rate = 0.5,
                                 adaptive = FALSE)
  expect_true(new_bw < bw)

  # Negative gradient → bandwidth should increase
  new_bw2 <- update_bandwidth_sgd(bw, gradient = -2, learning_rate = 0.5,
                                  adaptive = FALSE)
  expect_true(new_bw2 > bw)
})

test_that("update_bandwidth_sgd fixed: respects lower and upper bounds", {
  # Force a very large step that would violate bounds
  new_bw <- update_bandwidth_sgd(10, gradient = 1e8, learning_rate = 1,
                                 adaptive = FALSE, lower = 5, upper = 20)
  expect_true(new_bw >= 5)
  expect_true(new_bw <= 20)
})

test_that("update_bandwidth_sgd adaptive: result is a positive integer", {
  new_bw <- update_bandwidth_sgd(8L, gradient = -1.5, learning_rate = 2,
                                 adaptive = TRUE, lower = 3L, upper = 20L)
  expect_true(new_bw == as.integer(new_bw))
  expect_true(new_bw >= 1L)
})

test_that("update_bandwidth_sgd adaptive: clips to [lower, upper]", {
  # Push well below lower
  new_bw <- update_bandwidth_sgd(4L, gradient = -1e6, learning_rate = 1,
                                 adaptive = TRUE, lower = 3L, upper = 20L)
  expect_true(new_bw >= 3L)

  # Push well above upper
  new_bw2 <- update_bandwidth_sgd(15L, gradient = 1e6, learning_rate = 1,
                                  adaptive = TRUE, lower = 3L, upper = 20L)
  expect_true(new_bw2 <= 20L)
})

# ===========================================================================
# check_early_stopping
# ===========================================================================

test_that("check_early_stopping returns FALSE when patience = 0", {
  h <- data.frame(score = c(1, 2, 3, 4, 5))
  expect_false(check_early_stopping(h, patience = 0L))
})

test_that("check_early_stopping returns FALSE with improving scores", {
  h <- data.frame(score = c(1.0, 0.9, 0.8, 0.7, 0.6))
  expect_false(check_early_stopping(h, patience = 2L))
})

test_that("check_early_stopping returns TRUE when score stops improving", {
  # Score improved at epoch 1-3 but stagnated for 2 consecutive epochs
  h <- data.frame(score = c(1.0, 0.9, 0.85, 0.85, 0.86))
  expect_true(check_early_stopping(h, patience = 2L))
})

test_that("check_early_stopping returns FALSE when history is too short", {
  h <- data.frame(score = c(0.5, 0.6))
  expect_false(check_early_stopping(h, patience = 3L))
})

test_that("check_early_stopping counter tracks non-improving epochs", {
  # With scores that are always worsening and patience=1
  h <- data.frame(score = c(0.5, 0.6, 0.7))
  expect_true(check_early_stopping(h, patience = 1L))
})

# ===========================================================================
# search_bandwidth_sgd
# ===========================================================================

test_that("search_bandwidth_sgd returns a gwpr_bandwidth object", {
  ctx    <- make_sgd_context()
  scorer <- make_sgd_mock_scorer(best_bw = 10)
  result <- search_bandwidth_sgd(
    ctx,
    control = list(lower = 1, upper = 50, epoch = 5L,
                   learning_rate = 0.01),
    scorer  = scorer,
    workers = 1L,
    seed    = 1L
  )
  expect_s3_class(result, "gwpr_bandwidth")
})

test_that("search_bandwidth_sgd method field is 'sgd'", {
  ctx    <- make_sgd_context()
  scorer <- make_sgd_mock_scorer()
  result <- search_bandwidth_sgd(
    ctx,
    control = list(lower = 1, upper = 50, epoch = 4L),
    scorer  = scorer,
    workers = 1L,
    seed    = 2L
  )
  expect_equal(result$method, "sgd")
})

test_that("search_bandwidth_sgd default epoch is 10", {
  ctx    <- make_sgd_context()
  scorer <- make_sgd_mock_scorer()
  result <- search_bandwidth_sgd(
    ctx,
    control = list(lower = 1, upper = 100),   # epoch not specified
    scorer  = scorer,
    workers = 1L,
    seed    = 3L
  )
  expect_equal(nrow(result$history), 10L)
})

test_that("search_bandwidth_sgd respects explicit epoch parameter", {
  ctx    <- make_sgd_context()
  scorer <- make_sgd_mock_scorer()
  result <- search_bandwidth_sgd(
    ctx,
    control = list(lower = 1, upper = 50, epoch = 7L),
    scorer  = scorer,
    workers = 1L,
    seed    = 4L
  )
  expect_equal(nrow(result$history), 7L)
})

test_that("search_bandwidth_sgd history epoch column is sequential", {
  ctx    <- make_sgd_context()
  scorer <- make_sgd_mock_scorer()
  result <- search_bandwidth_sgd(
    ctx,
    control = list(lower = 1, upper = 50, epoch = 6L),
    scorer  = scorer,
    workers = 1L,
    seed    = 5L
  )
  expect_equal(result$history$epoch, seq_len(6L))
})

test_that("search_bandwidth_sgd history has all required columns", {
  ctx    <- make_sgd_context()
  scorer <- make_sgd_mock_scorer()
  result <- search_bandwidth_sgd(
    ctx,
    control = list(lower = 1, upper = 50, epoch = 3L),
    scorer  = scorer,
    workers = 1L,
    seed    = 6L
  )
  required_cols <- c(
    "epoch", "bandwidth_before", "bandwidth_after", "score",
    "gradient", "learning_rate", "delta", "batch_size",
    "early_stop_counter", "status", "elapsed_time",
    "r2", "mse", "rmse", "mae",
    "log_loss", "accuracy", "precision", "recall", "f1_score"
  )
  expect_true(all(required_cols %in% names(result$history)))
})

test_that("search_bandwidth_sgd fixed bandwidth never goes below 0", {
  ctx    <- make_sgd_context(adaptive = FALSE)
  scorer <- make_sgd_mock_scorer(best_bw = 1)
  result <- search_bandwidth_sgd(
    ctx,
    control = list(lower = 0.1, upper = 100, epoch = 10L,
                   learning_rate = 0.1),
    scorer  = scorer,
    workers = 1L,
    seed    = 7L
  )
  expect_true(all(result$history$bandwidth_after > 0, na.rm = TRUE))
})

test_that("search_bandwidth_sgd adaptive bandwidth stays as integers", {
  ctx    <- make_sgd_context(adaptive = TRUE)
  scorer <- make_sgd_mock_scorer(best_bw = 8)
  result <- search_bandwidth_sgd(
    ctx,
    control = list(lower = 3L, upper = 20L, epoch = 8L,
                   learning_rate = 2),
    scorer  = scorer,
    workers = 1L,
    seed    = 8L
  )
  bw_after <- result$history$bandwidth_after
  expect_true(all(bw_after == as.integer(bw_after), na.rm = TRUE))
  expect_true(all(bw_after >= 3L, na.rm = TRUE))
  expect_true(all(bw_after <= 20L, na.rm = TRUE))
})

test_that("search_bandwidth_sgd patience=0 does not early-stop", {
  ctx    <- make_sgd_context()
  scorer <- make_worsening_scorer()
  result <- search_bandwidth_sgd(
    ctx,
    control = list(lower = 1, upper = 100, epoch = 6L,
                   early_stopping_patience = 0L),
    scorer  = scorer,
    workers = 1L,
    seed    = 9L
  )
  # All 6 epochs should run and have status "ok"
  expect_equal(nrow(result$history), 6L)
  ok_rows <- result$history[result$history$status == "ok", , drop = FALSE]
  expect_equal(nrow(ok_rows), 6L)
})

test_that("search_bandwidth_sgd early_stop_counter increases when score worsens", {
  ctx    <- make_sgd_context()
  scorer <- make_worsening_scorer()
  result <- search_bandwidth_sgd(
    ctx,
    control = list(lower = 1, upper = 100, epoch = 5L,
                   early_stopping_patience = 0L,  # don't actually stop
                   learning_rate = 0.0001),
    scorer  = scorer,
    workers = 1L,
    seed    = 10L
  )
  # From epoch 2 onward, score always increases → counter should be > 0
  counters <- result$history$early_stop_counter[result$history$epoch >= 2L]
  expect_true(any(counters > 0L, na.rm = TRUE))
})

test_that("search_bandwidth_sgd early stopping halts after patience exceeded", {
  ctx    <- make_sgd_context()
  scorer <- make_worsening_scorer()
  result <- search_bandwidth_sgd(
    ctx,
    control = list(lower = 1, upper = 100, epoch = 10L,
                   early_stopping_patience = 2L,
                   learning_rate = 0.0001),
    scorer  = scorer,
    workers = 1L,
    seed    = 11L
  )
  # After patience = 2 non-improving epochs we should have some early_stopped rows
  stopped_rows <- result$history[result$history$status == "early_stopped", , drop = FALSE]
  expect_true(nrow(stopped_rows) > 0L)
})

test_that("search_bandwidth_sgd records seed in returned object", {
  ctx    <- make_sgd_context()
  scorer <- make_sgd_mock_scorer()
  result <- search_bandwidth_sgd(
    ctx,
    control = list(lower = 1, upper = 50, epoch = 3L),
    scorer  = scorer,
    workers = 1L,
    seed    = 77L
  )
  expect_equal(result$seed, 77L)
})

test_that("search_bandwidth_sgd is reproducible with fixed seed", {
  ctx    <- make_sgd_context()
  scorer <- make_sgd_mock_scorer()

  res1 <- search_bandwidth_sgd(
    ctx,
    control = list(lower = 1, upper = 50, epoch = 5L, learning_rate = 0.01),
    scorer  = scorer,
    workers = 1L,
    seed    = 123L
  )
  res2 <- search_bandwidth_sgd(
    ctx,
    control = list(lower = 1, upper = 50, epoch = 5L, learning_rate = 0.01),
    scorer  = scorer,
    workers = 1L,
    seed    = 123L
  )
  expect_equal(res1$history$bandwidth_after, res2$history$bandwidth_after)
  expect_equal(res1$best_bandwidth, res2$best_bandwidth)
})

test_that("search_bandwidth_sgd best_bandwidth is positive for fixed", {
  ctx    <- make_sgd_context(adaptive = FALSE)
  scorer <- make_sgd_mock_scorer(best_bw = 15)
  result <- search_bandwidth_sgd(
    ctx,
    control = list(lower = 1, upper = 100, epoch = 10L,
                   learning_rate = 0.001),
    scorer  = scorer,
    workers = 1L,
    seed    = 14L
  )
  expect_true(is.numeric(result$best_bandwidth))
  expect_true(result$best_bandwidth > 0)
})

test_that("search_bandwidth_sgd best_bandwidth is integer for adaptive", {
  ctx    <- make_sgd_context(adaptive = TRUE)
  scorer <- make_sgd_mock_scorer(best_bw = 10)
  result <- search_bandwidth_sgd(
    ctx,
    control = list(lower = 3L, upper = 25L, epoch = 8L,
                   learning_rate = 1),
    scorer  = scorer,
    workers = 1L,
    seed    = 15L
  )
  expect_true(result$best_bandwidth == as.integer(result$best_bandwidth))
})

test_that("search_bandwidth_sgd elapsed_time is non-negative", {
  ctx    <- make_sgd_context()
  scorer <- make_sgd_mock_scorer()
  result <- search_bandwidth_sgd(
    ctx,
    control = list(lower = 1, upper = 50, epoch = 3L),
    scorer  = scorer,
    workers = 1L,
    seed    = 1L
  )
  expect_true(result$elapsed_time >= 0)
})

test_that("search_bandwidth_sgd convergence_info contains key fields", {
  ctx    <- make_sgd_context()
  scorer <- make_sgd_mock_scorer()
  result <- search_bandwidth_sgd(
    ctx,
    control = list(lower = 1, upper = 50, epoch = 4L,
                   learning_rate = 0.005),
    scorer  = scorer,
    workers = 1L,
    seed    = 2L
  )
  expect_true("epoch"         %in% names(result$convergence_info))
  expect_true("learning_rate" %in% names(result$convergence_info))
  expect_true("patience"      %in% names(result$convergence_info))
  expect_true("delta"         %in% names(result$convergence_info))
})

test_that("search_bandwidth_sgd errors when lower is missing", {
  ctx    <- make_sgd_context()
  scorer <- make_sgd_mock_scorer()
  expect_error(
    search_bandwidth_sgd(
      ctx,
      control = list(upper = 50, epoch = 3L),
      scorer  = scorer,
      workers = 1L,
      seed    = 1L
    ),
    regexp = "lower"
  )
})

test_that("search_bandwidth_sgd errors when upper is missing", {
  ctx    <- make_sgd_context()
  scorer <- make_sgd_mock_scorer()
  expect_error(
    search_bandwidth_sgd(
      ctx,
      control = list(lower = 1, epoch = 3L),
      scorer  = scorer,
      workers = 1L,
      seed    = 1L
    ),
    regexp = "upper"
  )
})

---
title: "The Introduction of GWPR: from the Scratch"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{introduction_of_GWPR}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
options(rmarkdown.html_vignette.check_title = FALSE)
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  eval = FALSE
)
```

# Author:
Chao Li <chaoli0394@gmail.com> Shunsuke Managi <managi@doc.kyushu-u.ac.jp>

**Note:** This vignette documents the legacy 0.2.x API. For the new 1.0.0 API,
see `vignette("gwpr-introduction")`.

# 1.Overview
This is a basic introduction. If you have been familiar with [geographically weighted regressions (GWR)](https://www.bristol.ac.uk/media-library/sites/cmpo/migrated/documents/gwr.pdf), you can skip this part and directly go to the sections about the Functions. If not? Well, you are going to get a funny story.

## 1.1 What is Geographically Weighted Panel Regression (GWPR)
Generally, this algorithm mainly solves the problem of residuals from panel regressions clustering spatially. Basically, the GWPR is a type of weighted regression, but the only difference between [Weighted Least Square (WLS)](https://en.wikipedia.org/wiki/Weighted_least_squares) and GWPR is where the weights come from. In following sections, we will provide more information about the weights.

## 1.2 Why we need to use GWPR
As we have mentioned that, sometimes, the residuals cluster spatially, but the widely used method,
[Ordinary least squares (OLS)](https://en.wikipedia.org/wiki/Ordinary_least_squares), assume that the residuals distribute randomly, even on the maps. We give an cross-sectional example using the data in the package "GWPR.light", since the cross-sectional example is easy to read.
```{r example_GWmodel, message=FALSE}
library(GWPR.light)
library(sf)
library(tmap)
data("California")
data("TransAirPolCalif")
formula.GWPR <- pm25 ~ co2_mean + Developed_Open_Space_perc + Developed_Low_Intensity_perc +
   Developed_Medium_Intensity_perc + Developed_High_Intensity_perc +
   Open_Water_perc + Woody_Wetlands_perc + Emergent_Herbaceous_Wetlands_perc +
   Deciduous_Forest_perc + Evergreen_Forest_perc + Mixed_Forest_perc +
   Shrub_perc + Grassland_perc + Pasture_perc + Cultivated_Crops_perc +
   pop_density + summer_tmmx + winter_tmmx + summer_rmax + winter_rmax
TransAirPolCalif <- dplyr::filter(TransAirPolCalif, year == 2001)
lm.2001 <- lm(formula = formula.GWPR, TransAirPolCalif)
```
```{r}
summary(lm.2001)
resid.lm <- cbind(lm.2001$residuals, TransAirPolCalif$GEOID)
colnames(resid.lm) <- c("resid", "GEOID")
head(resid.lm)
```
```{r cross_sectional_residual, fig.align='center', out.height= '80%', out.width='80%'}
to_show <- merge(California, resid.lm, by = "GEOID")
tm_shape(to_show) + tm_polygons(col = 'resid')
```

<strong>Note:</strong> This is just an example about the residuals spatially clustering. Maybe, not so clear to convince you, so we need use Moran's I test to statistically confirm this situation. We will discuss that part in the Function section.


## 1.3 How we can perform GWPR
Before the package "GWPR.light" had been created, it was a little bit complicated. You need to know how to transform the raw data. Of course, among the well-known models, between panel regression, pooling regression, first difference and fixed effects model are not difficult to apply the data transformation. Then, the algorithm divides the whole sample into numerous subsamples according to a specific rules (we will discuss in the following section). The individual regressions of the subsamples are performed to estimate the local coefficients. However, the random effects model makes us have a headache. Theta $\theta$ is really hard to confirm. We do not know whether they change in each subsample. If we remove the individual effects based on the global $\theta$, then it may reduce accuracy, though it works. If we try to remove the individual effects in each subsample manually, it is impossible unless you are willing to pay several months or more. So:
```{r setup, message=FALSE}
library(GWPR.light)
```

# 2 Functions

## 2.1 bw.GWPR(): who is your neighborhoods?
Who is the neighborhoods of the specific point is really hard to define. In the Beenstock's and Fotheringham's books(ref. 1 & 2), if the distance ($d$) between point A and point B is within a certain distance ($h$), they could be deemed as the neighborhood of each other, because they have a spatial relationship ($a$). The Equations are as follows:
$$a_{ik}=
\begin{cases}
1& d_{ik} < h\\
0& \text{otherwise}
\end{cases}$$
where $d_{ik}$ is the distance between the locations of observations $i$ and $k$. It is also possible to relate $d_{ik}$ to $a_{ik}$ with a continuous function. Currently, this equation is widely used:
$$a_{ik} =
\begin{cases}
[1 - (d_{ik}/h)^2]^2 & d_{ik} < h\\
0& \text{otherwise}
\end{cases}
$$
In fact, numerous research that the closer the objects are, the more similar they are. Since the relationship have been normalized from 0 to 1. Therefore, here, we can say the weights ($W$) of those relationships are the same:
$$W_{ik} = a_{ik}$$

### 2.1.1 Choosing $h$: the kernel bandwidth
Before talking about this part, please look at our bw.GWPR() function to have an impression.
```{r}
bw.GWPR(formula = formula, data = data, index = index, SDF = SDF,
        adaptive = F, p = 2,bigdata = F, upperratio = 0.25,
        effect = "individual", model = c("pooling", "within", "random"),
        random.method = "swar", approach = c("CV","AIC"), kernel = "bisquare",
        longlat = F, doParallel = F, human.set.range = F,
        h.upper = NULL, h.lower = NULL)
```

There is a setting named "approach", including two selections: one is "CV" and the other is "AIC". They are two criteria to calibrate the bandwidth. In "CV" method, when the sum of squared errors arrives at the minimum, the bandwidth is optimal. Basically, the sum of squared errors is written as:
$$SS(h)  = \sum_i\{y_i - \hat{y_i}(h)\}^2$$
However, when $h$ varies, $i$ also changes. Possibly, this function is monotonically increasing.
```{r, out.height="60%", out.width="60%", fig.align="center"}
h = 1:10 # just an example, not truth.
SS = h^2
plot(h, SS, col = "red")
```

To solve this issue, the $SS$ function changes as follows (ref. 2):
$$CVSS(h) = \frac{n  \sum_i\{y_i - \hat{y_i}(h)\}^2} {(n-p+1)^2}$$
where $n$ is the numbers of observations, $p$ is the number of the estimated parameters.
Furthermore, the "AIC" criterion also has the same issue. It has changed in our calculation:
$$AIC(h) = 2nln(sd) + 2ln(2\pi) + \frac{n\times [tr(hat) + n]}{[n - 2 - tr(hat)]}$$
where $sd$ is the standard deviation of residuals, and $hat$ is the hat matrix.
<strong>Now, you get the fixed distance bandwidth!:)</strong>

### 2.1.2 Adaptive and Fixed Distance Bandwidth
Sometimes, some researches assume the numbers of neighborhoods is fixed. In other words, a specific point always has the $n$ neighborhoods, no matter how far they are. Here, we are going to use the adaptive distances bandwidth. The $h$ is not a fixed number, but depends on the furthest neighborhood. If you have this request, you need to select the "adaptive" as TRUE.

### 2.1.3 "bigdata" Setting
Our calibration process normally start from the golden ratio of range. In this way, if your dataset is huge, in the beginning, the calculation is really time-consuming. However, according to our experience, the bandwidth is seldom large. Therefore, you could set "bigdata" as TRUE and reset upper boundary.

### 2.1.4 "human.set.range" Setting
If you have run this function for several times and you know the potential optimal range of bandwidth, you can set "human.set.range" as TRUE to cut down the calculation time. Then, you must also set h.upper and h.lower both. Otherwise, errors.

### 2.1.5 "gradientIncrement" Setting
**Note:** This setting is super important.
In some remote sensing research, we find the relationship between bandwidth and CV or AIC score is not U-shape. So, we recommend to use gradient increment selection to calibrate the optimal bandwidth. The user need to set the upper boundary of the gradient increment selection (GI.upper), the lower boundary (GI.lower), and the step length of the increment (GI.step). If it is an analysis on grid data, we recommend the GI.step is the spatial resolution.

### 2.1.6 "model" Setting
Now, this version of "GWPR.light" can figure our fixed effects model ("within"), pooling regression ("pooling"), and random effects models ("random"), internally. However, if you perform the basic transformation, we can also perform first difference model and between model.
For first difference model, you need to confirm your data is balanced panel. Then, please calculate the difference between two continuous period of each variable in every individual. Reset the "time" index, put them into the function and set the model = "pooling". Here we go!
For between model, you need to directly calculation the mean of each variables in every individual. Then, reset the "time" index, put them into the function and set the model = "pooling". You got it.
Unfortunately, the authors of this package still have statistic concern about Hausman-Toylar model, Maybe, in future verion, we will add it into the package. Thank you!

## 2.2 GWPR.moran.test(): are the residuals really clustering
```{r, fig.align='center', out.height= '80%', out.width='80%'}
tm_shape(to_show) +
  tm_polygons(col = 'resid')
```

You may be familiar with [Moran's I](https://en.wikipedia.org/wiki/Moran%27s_I). If not, click it. This version only works with the balanced panel data. Here, we directly go with the math:
$$I=\frac{\sum_i\sum_kW_{ik}\hat{u}_i\hat{u}_k}{\sum_i\hat{u}_i^2}$$
where $u_i$ and $u_k$ are the residuals of individuals $i$ and $k$. If $u_i$ and $u_k$ are perfectly positively correlated, the $I$ would 1. Reversely, if $u_i$ and $u_k$ are totally unrelated, the $I$ should be 0. bw.GWPR() must perform before this test. Otherwise, we do not know who is whose neighborhood. In the global panel model, Moran's I may be calculated for each time period and averaged:
$$\bar{I} = \frac{1}{T}\sum_{t=1}^TI_t$$
where $T$ is the number of periods. Since the data is balanced panel, the expected value of $\bar{I}$ is:
$$E(\bar{I}) = -\frac{1}{N-1}$$
where $N$ is the number of observations. And the variance ($V^2$) should be:
$$V^2 = \frac{N^2\sum_i\sum_kW_{ik}^2 + 3(\sum_i\sum_kW_{ik})^2 - N\sum_i(\sum_kW_{ik})^2}{T(N^2-1)(\sum_i\sum_kW_{ik})}$$
Now, to test for spatial autocorrelation in panel data the standardized panel average has a standard normal distribution:
$$\frac{\bar{I}}{V} \sim N(0,1)$$
Let us try, using the "GWPR.light::GWPR.moran.test()"
```{r}
data(TransAirPolCalif)
pdata <- plm::pdata.frame(TransAirPolCalif, index = c("GEOID", "year"))
moran.plm.model <- plm::plm(formula = formula.GWPR, data = pdata, model = "within")
summary(moran.plm.model)
```
```{r}
bw.AIC.F <- bw.GWPR(formula = formula.GWPR, data = TransAirPolCalif, index = c("GEOID", "year"),
                    SDF = California, adaptive = F, p = 2, bigdata = F, effect = "individual",
                    model = "within", approach = "AIC", kernel = "bisquare", longlat = F,
                    doParallel = F)
```
```{r}
bw.AIC.F <- 2.010529 #precomputed results from #150:155
GWPR.moran.test(moran.plm.model, SDF = California, bw = bw.AIC.F, kernel = "bisquare",
                 adaptive = F, p = 2, longlat=F, alternative = "greater")
```
The statistic is significantly greater than 0. Therefore, the residuals are spatially clustered.
Because the spatial relationship among residuals, using GWPR is reasonable. :)

## 2.3 Model Selection: which model is better?
If you know the models very well, including fixed effects model, random effects model, pooling regression, among others, you could skip this part. You know why you should use the model, and the tests, provided by the package, may only help you confirm your selection.
However, for people are not very familiar with the panel model, we provide three local tests to help them select model. Which are <strong>GWPR.pFtest(), GWPR.phtest(), GWPR.plmtest()</strong>
We give several examples here.

### 2.3.1 GWPR.pFtest(): Locally F test
```{r eg.GWPR.pFtest,  fig.align='center', out.height= '80%', out.width='80%'}
GWPR.pFtest.resu.F <- GWPR.pFtest(formula = formula.GWPR, data = TransAirPolCalif, index = c("GEOID", "year"),
                                  SDF = California, bw = bw.AIC.F, adaptive = F, p = 2, effect = "individual",
                                  kernel = "bisquare", longlat = F)
tm_shape(GWPR.pFtest.resu.F$SDF) +
     tm_polygons(col = "p.value", breaks = c(0, 0.05, 1))
```

If the p-value is lower than the specific level (0.01, 0.05, etc.), significant effects exist.

### 2.3.2 GWPR.plmtest(): Locally Breusch-Pagan Lagrange Multiplier test
```{r eg.GWPR.plmtest, fig.align='center', out.height= '80%', out.width='80%'}
GWPR.plmtest.resu.F <- GWPR.plmtest(formula = formula.GWPR, data = TransAirPolCalif, index = c("GEOID", "year"),
                                    SDF = California, bw = bw.AIC.F, adaptive = F, p = 2,
                                    kernel = "bisquare", longlat = F)
tm_shape(GWPR.plmtest.resu.F$SDF) +
     tm_polygons(col = "p.value", breaks = c(0, 0.05, 1))
```

If the p-value is lower than the specific level (0.01, 0.05, etc.), significant effects exist.

### 2.3.3 GWPR.phtest(): Locally Hausman test
```{r eg.GWPR.phtest, fig.align='center', out.height= '80%', out.width='80%'}
GWPR.phtest.resu.F <- GWPR.phtest(formula = formula.GWPR, data = TransAirPolCalif, index = c("GEOID", "year"),
                                  SDF = California, bw = bw.AIC.F, adaptive = F, p = 2, effect = "individual",
                                  kernel = "bisquare", longlat = F, random.method = "amemiya")
tm_shape(GWPR.phtest.resu.F$SDF) +
     tm_polygons(col = "p.value", breaks = c(0, 0.05, 1))
```

If the p-value is lower than the specific level (0.01, 0.05, etc.), one model is inconsistent.
<strong>Note:</strong> Here, we use the "amemiya" random method. Normally, we should use "swar", but need to change the bandwidth. Because it is just example so we do not do that.

## 2.4 GWPR(): the regression process
Now, we come to the most important funciton in this package, GWPR(). Let us look it at first.
```{r}
GWPR(formula = formula, data = data, index = index, SDF = SDF, bw = NULL, adaptive = F, p = 2,
            effect = "individual", model = c("pooling", "within", "random"),
            random.method = "swar", kernel = "bisquare", longlat = F)
```
You might know more things about weight least square.
Similarly, the GWR equation is as follows ()ref. 3:
$$y_i = \sum_jX_{ij}\beta_j(p_i)+\epsilon_i$$
Here $p_i$ is the geographical location of the $i$th case. These $\beta_j(p_i)$ would themselves contain
coefficients to be estimated. If without $(p_i)$, the $\beta$ should be estimated as follows:
$$\hat\beta = (X^TX)^{-1}X^Ty$$
Now, it should be as follows:
$$\hat\beta_i = (X^TW_iX)^{-1}X^TW_iy$$
This $W_i$ is the spatial weights we mention in the beginning of this documents. Now, you know GWR based on OLS. All panel model is just a data transformation, then perform the model use almost the same process.
$$X' =
\begin{cases}
X & \text{if model = pooling}\\
X - \bar{X} & \text{if model = FEM}\\
X - \theta\bar{X} & \text{if model = REM}\\
\bar{X} & \text{if model = between}\\
X_t - X_{t-1} & \text{if model = first difference}\\
\end{cases}
$$
$$y' =
\begin{cases}
y & \text{if model = pooling}\\
y - \bar{y} & \text{if model = FEM}\\
y - \theta\bar{y} & \text{if model = REM}\\
\bar{y} & \text{if model = between}\\
y_t - y_{t-1} & \text{if model = first difference}\\
\end{cases}
$$
Where $X'$ is the data after transformation, $X$ is a matrix of original independent variables, $\bar{X}$ is the means of all independent variables during all periods.
Our GWPR() can solve first three type within the function. "between" and "first difference" can be easily performed, then use our GWPR() with (model = "pooling") to do.
Here, we give an example about GWPR()
```{r eg.GWPR, fig.align='center', out.height= '80%', out.width='80%'}
result.F.AIC <- GWPR(bw = bw.AIC.F, formula = formula.GWPR, data = TransAirPolCalif, index = c("GEOID", "year"),
                     SDF = California, adaptive = F, p = 2, effect = "individual", model = "within",
                     kernel = "bisquare", longlat = F)
summary(result.F.AIC$SDF$Local_R2)
tm_shape(result.F.AIC$SDF) +
  tm_polygons(col = "Local_R2", pal = "Reds",auto.palette.mapping = F,
              style = 'cont')
```
Look, the $R^2$ is 0.8969606, while the panel liner regression is 0.48772. BAM!!!
<strong>Note:</strong>Our model include SE and t-value of local coefficients, you can check the significance.

# END
Thank for your cooperation, we are going to write an article about this package. After it is publsihed, we will update the reference of all. If you could cite our article we would really appreciate.

## Archive in GitHub:
https://github.com/MichaelChaoLi-cpu/GWPR.light

## BugReports:
If you find bugs or have other ideas, be free to contact us.
https://github.com/MichaelChaoLi-cpu/GWPR.light/issues


# Reference
1. Beenstock, M., Felsenstein, D., 2019. The econometric analysis of non-stationary spatial panel data. Springer.
2. Fotheringham, A., Brunsdon, C., Charlton, M., 2002. Geographically Weighted Regression: The Analysis of Spatially Varying Relationships. John Wiley & Sons 13.
3. Brunsdon, C., Fotheringham, S., Charlton, M., 1998. Geographically Weighted Regression. Journal of the Royal Statistical Society: Series D (The Statistician) 47, 431-443.

---
title: "Getting Started with GWPR.light 1.0.0"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Getting Started with GWPR.light 1.0.0}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

## Overview

`GWPR.light` 1.0.0 provides a modern, `sf`-first API for Geographically Weighted
Panel Regression (GWPR). The public interface consists of four functions:

- `gwpr()` — full pipeline (bandwidth search + fitting + diagnostics)
- `select_bandwidth()` — standalone bandwidth optimisation
- `fit_gwpr()` — fit with a known bandwidth
- `diagnose_gwpr()` — diagnostic tests on a fitted model

All functions accept panel data as a plain `data.frame` and spatial information
as an `sf` object. The `workers` argument controls parallel execution; the
default `workers = 1` runs serially and is safe in all environments.

## Minimal linear GWPR example

```{r minimal-linear}
library(GWPR.light)
library(sf)

set.seed(42)

# Simulate a tiny spatial panel: 6 units, 4 time periods
n_units <- 6
n_time  <- 4

pts <- sf::st_as_sf(
  data.frame(
    id = 1:n_units,
    X  = c(0, 1, 2, 0, 1, 2),
    Y  = c(0, 0, 0, 1, 1, 1)
  ),
  coords = c("X", "Y"),
  crs    = NA_integer_
)

dat <- data.frame(
  id   = rep(1:n_units, each = n_time),
  time = rep(1:n_time,  n_units),
  x1   = rnorm(n_units * n_time),
  x2   = rnorm(n_units * n_time)
)
dat$y <- 1.5 * dat$x1 - 0.8 * dat$x2 + rnorm(n_units * n_time, sd = 0.3)

# Fit with a known bandwidth (skip automatic search for speed)
fit <- fit_gwpr(
  formula   = y ~ x1 + x2,
  data      = dat,
  spatial   = pts,
  id        = "id",
  time      = "time",
  bandwidth = 2,
  family    = "gaussian",
  model     = "pooling",
  workers   = 1
)

print(fit)
```

## Accessing results

```{r results}
# Overall goodness-of-fit metrics
str(fit$metrics)

# Per-unit spatial coefficients (one row per spatial unit)
if (!is.null(fit$spatial_results)) {
  head(fit$spatial_results)
}
```

## Bandwidth selection (grid search)

```{r bandwidth-search}
bw <- select_bandwidth(
  formula  = y ~ x1 + x2,
  data     = dat,
  spatial  = pts,
  id       = "id",
  time     = "time",
  family   = "gaussian",
  model    = "pooling",
  method   = "grid",
  control  = list(lower = 0.5, upper = 3, step = 0.5),
  workers  = 1
)

print(bw)
cat("Best bandwidth:", bw$best_bandwidth, "\n")
```

## Full pipeline with gwpr()

```{r full-pipeline}
# Use the best bandwidth found above to avoid re-running search
full_fit <- gwpr(
  formula     = y ~ x1 + x2,
  data        = dat,
  spatial     = pts,
  id          = "id",
  time        = "time",
  bandwidth   = bw$best_bandwidth,
  family      = "gaussian",
  model       = "pooling",
  diagnostics = FALSE,   # skip diagnostics for speed
  workers     = 1
)

print(full_fit)
```

## Diagnostics

```{r diagnostics}
diag_result <- diagnose_gwpr(
  full_fit,
  diagnostics = c("f_test", "hausman", "lm_test")
)

print(diag_result)
```

## Long-running examples

The following code illustrates automatic bandwidth search via SGD and a
binomial (logistic) GWPR. These are wrapped in `\donttest{}` in the
function documentation because they may take more than a few seconds on
larger datasets.

```{r long-examples, eval=FALSE}
# Automatic SGD bandwidth search + fit (may take several seconds)
fit_auto <- gwpr(
  formula          = y ~ x1 + x2,
  data             = dat,
  spatial          = pts,
  id               = "id",
  time             = "time",
  bandwidth_method = "sgd",
  bandwidth_control = list(n_iter = 20, step_size = 0.1),
  workers          = 1,
  seed             = 123
)

# Binomial GWPR
dat$y_bin <- as.integer(dat$y > 0)
fit_logit <- fit_gwpr(
  formula   = y_bin ~ x1 + x2,
  data      = dat,
  spatial   = pts,
  id        = "id",
  time      = "time",
  bandwidth = 2,
  family    = "binomial",
  model     = "pooling",
  workers   = 1
)
```

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/validation.R
\name{validate_formula}
\alias{validate_formula}
\title{Validate a model formula against a data frame}
\usage{
validate_formula(formula, data)
}
\arguments{
\item{formula}{A formula object.}

\item{data}{A data frame containing the model variables.}
}
\value{
Invisibly returns `TRUE` when valid; stops with an informative message
  when a problem is detected.
}
\description{
Validate a model formula against a data frame
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/model_logistic.R
\name{fit_logistic_pooling}
\alias{fit_logistic_pooling}
\title{Fit a pooled logistic model via stats::glm}
\usage{
fit_logistic_pooling(local_data, formula, weights, family = "binomial")
}
\arguments{
\item{local_data}{A \code{data.frame} containing all formula variables.}

\item{formula}{A formula object with a binary response.}

\item{weights}{Numeric vector of kernel weights (same length as
\code{nrow(local_data)}).}

\item{family}{Character scalar reserved for future extension;
currently only \code{"binomial"} is supported.}
}
\value{
A \code{glm} object.
}
\description{
Fit a pooled logistic model via stats::glm
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/validation.R
\name{validate_family_response}
\alias{validate_family_response}
\title{Validate the response variable against the specified family}
\usage{
validate_family_response(data, formula, family)
}
\arguments{
\item{data}{A data frame.}

\item{formula}{A formula; the left-hand side is the response variable.}

\item{family}{Character string: `"gaussian"` or `"binomial"`.}
}
\value{
Invisibly returns `TRUE` when valid; stops otherwise.
}
\description{
Validate the response variable against the specified family
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/model_logistic.R
\name{fit_logistic_local_model}
\alias{fit_logistic_local_model}
\title{Fit a single local binary logistic panel model}
\usage{
fit_logistic_local_model(
  local_data,
  formula,
  model,
  effect,
  weights,
  index,
  threshold = 0.5,
  family = "binomial"
)
}
\arguments{
\item{local_data}{A \code{data.frame} with all formula variables and panel
index columns.}

\item{formula}{A formula object.}

\item{model}{Character scalar: \code{"pooling"}, \code{"fixed"}, or
\code{"random"}.}

\item{effect}{Character scalar: \code{"individual"}, \code{"time"},
\code{"two-way"}, or \code{"nested"}.}

\item{weights}{Numeric vector of kernel weights aligned with
\code{nrow(local_data)}.}

\item{index}{Character vector of length 2: \code{c(id_col, time_col)}.}

\item{threshold}{Numeric scalar; classification threshold (default 0.5).}

\item{family}{Character scalar reserved for future extension (currently
only \code{"binomial"} is implemented).}
}
\value{
A list with elements:
\describe{
  \item{\code{fit}}{The fitted model object, or \code{NULL} on failure.}
  \item{\code{status}}{\code{"ok"} or \code{"failed"}.}
  \item{\code{error}}{\code{NULL} or character error message.}
  \item{\code{metadata}}{Named list with additional fitting metadata.}
}
}
\description{
Dispatches to the correct backend (\code{pooling}, \code{fixed}, or
\code{random}) based on \code{model}, wrapping execution in a
\code{tryCatch} so that convergence failures or complete-separation errors
are captured rather than propagated.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/model_logistic.R
\name{standardize_logistic_response}
\alias{standardize_logistic_response}
\title{Coerce a binary response to a numeric 0/1 integer vector}
\usage{
standardize_logistic_response(y)
}
\arguments{
\item{y}{A numeric, logical, or factor vector.}
}
\value{
An integer vector of 0s and 1s.
}
\description{
Converts logical and two-level factor responses to 0/1 integer.
Numeric 0/1 vectors are returned unchanged.
Factors with more than two levels raise an error via
\code{validate_binary_response()}.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bandwidth_sgd.R
\name{bandwidth_sgd}
\alias{bandwidth_sgd}
\title{SGD Bandwidth Search (`bandwidth_sgd.R`)}
\description{
Implements a stochastic gradient descent (SGD) based bandwidth search for
GWPR models.  A one-dimensional finite-difference gradient approximation is
used to iteratively update the bandwidth over a fixed number of epochs.
Mini-batch sampling and early stopping are supported.

The search does **not** require the user to specify `lower`, `upper`, or
`step`; SGD starts from a single initialised bandwidth and follows the
(approximate) gradient.  When `lower` / `upper` are supplied they are used
as hard constraints.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/spatial_sf.R
\name{extract_geometry}
\alias{extract_geometry}
\title{Extract the geometry column from an sf object}
\usage{
extract_geometry(spatial)
}
\arguments{
\item{spatial}{An `sf` object.}
}
\value{
An `sfc` geometry column.
}
\description{
Returns the `sfc` geometry column of the supplied `sf` object.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/diagnostics.R
\name{diagnose_hausman}
\alias{diagnose_hausman}
\title{Local Hausman test diagnostic on a gwpr_fit object}
\usage{
diagnose_hausman(object, ...)
}
\arguments{
\item{object}{A `gwpr_fit` object.}

\item{...}{Currently ignored.}
}
\value{
A named list with elements:
\describe{
  \item{`local_hausman`}{Data frame with columns `unit_id`, `statistic`,
    `p_value`, `df`, `status`.}
  \item{`n_tested`}{Number of units tested.}
  \item{`n_failed`}{Number of units where the test could not be computed.}
  \item{`status`}{Overall status: `"ok"`, `"not_applicable"`, or
    `"no_local_results"`.}
}
}
\description{
Performs a local Hausman test (within vs. random) for each spatial unit
using test statistics pre-computed during model fitting or stored in
`local_results`.
}
\details{
**Applicable models**: gaussian with `model = "random"`.  For pooling
models the Hausman test is not meaningful; the function returns a
`status = "not_applicable"` result.  For logistic models the function also
returns `status = "not_applicable"`.

**Panel balance requirement**: No constraint at the unit level.

**Failure conditions**: Returns `status = "missing_hausman_data"` for any
unit where the required statistics are absent from `local_results`.

**Logistic interpretation limit**: Not applicable.
}
\examples{
\donttest{
library(sf)
pts <- sf::st_as_sf(
  data.frame(id = 1:4, X = c(0,1,0,1), Y = c(0,0,1,1)),
  coords = c("X", "Y"), crs = NA_integer_
)
dat <- data.frame(
  id   = rep(1:4, each = 5),
  time = rep(1:5, 4),
  y    = rnorm(20),
  x1   = rnorm(20)
)
fit <- fit_gwpr(y ~ x1, data = dat, spatial = pts, id = "id",
                time = "time", bandwidth = 2, workers = 1)
diagnose_hausman(fit)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/model_linear.R
\name{predict_linear_local_model}
\alias{predict_linear_local_model}
\title{Predict response values for a local linear model}
\usage{
predict_linear_local_model(local_result, local_data)
}
\arguments{
\item{local_result}{A list as returned by `fit_linear_local_model()`.}

\item{local_data}{The data frame used for prediction (same structure as
the training data).}
}
\value{
Numeric vector of predicted values (same length as `nrow(local_data)`).
  Returns a vector of `NA_real_` on failure.
}
\description{
Returns predicted values for the rows of `local_data`, using the fitted
model object stored in `local_result`.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/result.R
\name{summary.gwpr_fit}
\alias{summary.gwpr_fit}
\title{Summarise a gwpr_fit object}
\usage{
\method{summary}{gwpr_fit}(object, ...)
}
\arguments{
\item{object}{A `gwpr_fit` object.}

\item{...}{Currently ignored.}
}
\value{
Invisibly returns `object`.
}
\description{
Prints the global model overview, quantile summary of local coefficients
(when available), and goodness-of-fit metrics.
}
\examples{
\donttest{
library(sf)
pts <- sf::st_as_sf(
  data.frame(id = 1:4, X = c(0,1,0,1), Y = c(0,0,1,1)),
  coords = c("X", "Y"), crs = NA_integer_
)
dat <- data.frame(id = rep(1:4, each = 5), time = rep(1:5, 4),
                  y = rnorm(20), x1 = rnorm(20))
fit <- fit_gwpr(y ~ x1, data = dat, spatial = pts, id = "id",
                time = "time", bandwidth = 2, workers = 1)
summary(fit)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/result.R
\name{summary.gwpr_bandwidth}
\alias{summary.gwpr_bandwidth}
\title{Summarise a gwpr_bandwidth object}
\usage{
\method{summary}{gwpr_bandwidth}(object, ...)
}
\arguments{
\item{object}{A `gwpr_bandwidth` object.}

\item{...}{Currently ignored.}
}
\value{
Invisibly returns `object`.
}
\description{
Prints the search method, best bandwidth, and a brief history overview.
}
\examples{
\donttest{
library(sf)
pts <- sf::st_as_sf(
  data.frame(id = 1:4, X = c(0,1,0,1), Y = c(0,0,1,1)),
  coords = c("X", "Y"), crs = NA_integer_
)
dat <- data.frame(id = rep(1:4, each = 5), time = rep(1:5, 4),
                  y = rnorm(20), x1 = rnorm(20))
bw <- select_bandwidth(y ~ x1, data = dat, spatial = pts, id = "id",
  time = "time", method = "grid",
  control = list(lower = 1, upper = 3, step = 1), workers = 1)
summary(bw)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/memory.R
\name{memory}
\alias{memory}
\title{Memory Estimation Module for GWPR.light 1.0.0}
\description{
Functions for estimating memory usage before running GWPR models.
The module provides warnings about memory risk levels to help users
avoid out-of-memory errors. These functions only warn; they never
stop execution.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/api.R
\name{fit_gwpr}
\alias{fit_gwpr}
\title{Fit GWPR with a given bandwidth}
\usage{
fit_gwpr(
  formula,
  data,
  spatial,
  id,
  time,
  bandwidth,
  family = c("gaussian", "binomial"),
  model = c("within", "pooling", "random"),
  effect = c("individual", "time", "two-way", "nested"),
  kernel = c("bisquare", "gaussian", "exponential", "tricube", "boxcar"),
  adaptive = FALSE,
  threshold = 0.5,
  workers = 1L,
  seed = NULL,
  ...
)
}
\arguments{
\item{formula}{A \code{formula} object.}

\item{data}{A \code{data.frame} with panel data.}

\item{spatial}{An \code{sf} object.}

\item{id}{Character scalar; unit ID column name.}

\item{time}{Character scalar; time column name.}

\item{bandwidth}{Numeric scalar.  The bandwidth to use (fixed distance or
number of neighbours when \code{adaptive = TRUE}).}

\item{family}{\code{"gaussian"} (default) or \code{"binomial"}.}

\item{model}{\code{"within"} (default), \code{"pooling"}, or
\code{"random"}.}

\item{effect}{\code{"individual"} (default), \code{"time"},
\code{"two-way"}, or \code{"nested"}.}

\item{kernel}{Kernel function name (default \code{"bisquare"}).}

\item{adaptive}{Logical; \code{FALSE} (default) for fixed bandwidth.}

\item{threshold}{Numeric; classification threshold (binomial only,
default \code{0.5}).}

\item{workers}{Positive integer; number of parallel workers (default 1).}

\item{seed}{Integer random seed, or \code{NULL}.}

\item{...}{Currently unused.}
}
\value{
A \code{gwpr_fit} object.
}
\description{
Validates inputs, prepares data, builds spatial weights, and fits the
Geographically Weighted Panel Regression for the specified bandwidth.
Returns a \code{gwpr_fit} object.
}
\examples{
\donttest{
library(sf)
pts <- sf::st_as_sf(
  data.frame(id = 1:4, X = c(0,1,0,1), Y = c(0,0,1,1)),
  coords = c("X", "Y"), crs = NA_integer_
)
dat <- data.frame(
  id   = rep(1:4, each = 5),
  time = rep(1:5, 4),
  y    = rnorm(20),
  x1   = rnorm(20)
)
fit <- fit_gwpr(y ~ x1, data = dat, spatial = pts, id = "id",
                time = "time", bandwidth = 2, workers = 1)
print(fit)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data_prepare.R
\name{standardize_binary_response}
\alias{standardize_binary_response}
\title{Standardise a binary response variable to 0/1 numeric}
\usage{
standardize_binary_response(y)
}
\arguments{
\item{y}{A vector that is 0/1 numeric, logical, or a two-level factor.}
}
\value{
A numeric vector of 0s and 1s.
}
\description{
Converts logical and two-level factor responses to 0/1 integer. Numeric
0/1 vectors are returned unchanged (as numeric). Other inputs raise an
error.
}
\details{
For factor inputs the first level is mapped to 0 and the second level is
mapped to 1.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/model_logistic.R
\name{model_logistic}
\alias{model_logistic}
\title{Binary Panel Logistic Engine for GWPR.light 1.0.0}
\description{
Internal functions for fitting Geographically Weighted Panel Regression
with a binary (binomial) response.  Supports pooling (stats::glm),
fixed effects (fixest::feglm), and random effects (glmmTMB::glmmTMB)
panel models, plus individual, time, two-way, and nested effects.

The \code{family} parameter is reserved for future multi-class extension;
in version 1.0.0 only \code{"binomial"} is supported.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/weights.R
\name{validate_bandwidth}
\alias{validate_bandwidth}
\title{Validate bandwidth parameter}
\usage{
validate_bandwidth(bandwidth, adaptive)
}
\arguments{
\item{bandwidth}{Numeric scalar.  The bandwidth value to validate.}

\item{adaptive}{Logical scalar.  `TRUE` for adaptive (k-nearest-neighbour)
bandwidth, `FALSE` for fixed distance bandwidth.}
}
\value{
Invisibly returns `TRUE` when validation passes.
}
\description{
Checks that the supplied bandwidth is legal:
* fixed (adaptive = FALSE): must be a finite positive numeric scalar.
* adaptive (adaptive = TRUE): must be a finite positive integer scalar.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/api.R
\name{select_bandwidth}
\alias{select_bandwidth}
\title{Select an optimal bandwidth for GWPR}
\usage{
select_bandwidth(
  formula,
  data,
  spatial,
  id,
  time,
  family = c("gaussian", "binomial"),
  model = c("within", "pooling", "random"),
  effect = c("individual", "time", "two-way", "nested"),
  method = c("sgd", "grid", "random"),
  control = list(),
  kernel = c("bisquare", "gaussian", "exponential", "tricube", "boxcar"),
  adaptive = FALSE,
  threshold = 0.5,
  workers = 1L,
  seed = NULL,
  ...
)
}
\arguments{
\item{formula}{A \code{formula} object.}

\item{data}{A \code{data.frame} with panel data.}

\item{spatial}{An \code{sf} object.}

\item{id}{Character scalar; unit ID column name.}

\item{time}{Character scalar; time column name.}

\item{family}{\code{"gaussian"} or \code{"binomial"}.}

\item{model}{\code{"within"}, \code{"pooling"}, or \code{"random"}.}

\item{effect}{\code{"individual"}, \code{"time"}, \code{"two-way"}, or
\code{"nested"}.}

\item{method}{Bandwidth search method: \code{"sgd"} (default),
\code{"grid"}, or \code{"random"}.}

\item{control}{Named list of search control parameters.}

\item{kernel}{Kernel function name.}

\item{adaptive}{Logical; \code{TRUE} for adaptive (k-NN) bandwidth.}

\item{threshold}{Numeric; classification threshold (binomial only).}

\item{workers}{Positive integer; number of parallel workers.}

\item{seed}{Integer random seed, or \code{NULL}.}

\item{...}{Additional arguments (currently unused).}
}
\value{
A \code{gwpr_bandwidth} object with fields:
\describe{
  \item{\code{best_bandwidth}}{The selected bandwidth value.}
  \item{\code{best_score}}{The criterion value at the best bandwidth.}
  \item{\code{method}}{The search method used.}
  \item{\code{history}}{Search history data frame.}
}
}
\description{
Validates inputs, prepares data, and dispatches to the appropriate bandwidth
search algorithm: grid search, stochastic gradient descent (sgd), or random
search, depending on the \code{method} argument.
}
\examples{
\donttest{
library(sf)
pts <- sf::st_as_sf(
  data.frame(id = 1:4, X = c(0,1,0,1), Y = c(0,0,1,1)),
  coords = c("X", "Y"), crs = NA_integer_
)
dat <- data.frame(
  id   = rep(1:4, each = 5),
  time = rep(1:5, 4),
  y    = rnorm(20),
  x1   = rnorm(20)
)
bw <- select_bandwidth(
  y ~ x1, data = dat, spatial = pts, id = "id", time = "time",
  method  = "grid",
  control = list(lower = 0.5, upper = 2, step = 0.5),
  workers = 1
)
bw$best_bandwidth
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bandwidth_grid.R
\name{bandwidth_grid}
\alias{bandwidth_grid}
\title{Grid Search for Bandwidth Selection}
\description{
Functions implementing an exhaustive grid search over a user-specified
range of bandwidth candidates.  Each candidate is evaluated by a user-
supplied scorer function; the full search history and the best bandwidth
are returned as a `gwpr_bandwidth` object.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/spatial_sf.R
\name{assert_sf}
\alias{assert_sf}
\title{Assert that an object is an sf object}
\usage{
assert_sf(spatial)
}
\arguments{
\item{spatial}{Any R object.}
}
\value{
Invisibly returns `TRUE` when the check passes.
}
\description{
Stops with an informative error when `spatial` does not inherit from `"sf"`.
sp objects are not supported.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/model_linear.R
\name{effect_to_plm}
\alias{effect_to_plm}
\title{Map user-facing effect string to plm effect parameter}
\usage{
effect_to_plm(effect)
}
\arguments{
\item{effect}{Character scalar: one of `"individual"`, `"time"`,
`"two-way"`, `"nested"`.}
}
\value{
Character scalar accepted by `plm::plm()` for its `effect`
  argument: `"individual"`, `"time"`, or `"twoways"`.  For `"nested"`,
  the function stops with an informative message because `plm` does not
  support nested effects without additional data conventions.
}
\description{
Map user-facing effect string to plm effect parameter
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/context.R
\name{context}
\alias{context}
\title{Internal Context Object for GWPR.light 1.0.0}
\description{
These internal functions construct and validate the standardised
`gwpr_context` list that is passed between modules, eliminating
repetitive argument passing.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/diagnostics.R
\name{diagnose_gwpr}
\alias{diagnose_gwpr}
\title{Run diagnostic tests on a fitted GWPR model}
\usage{
diagnose_gwpr(
  object,
  diagnostics = c("moran", "f_test", "hausman", "lm_test"),
  spatial_weights = NULL,
  panel_index = NULL,
  ...
)
}
\arguments{
\item{object}{A `gwpr_fit` object returned by `fit_gwpr()` or similar.}

\item{diagnostics}{Character vector naming the tests to run.  Any subset of
`c("moran", "f_test", "hausman", "lm_test")`.  Default is all four.}

\item{spatial_weights}{Required when `"moran"` is in `diagnostics`.
A row-standardised n x n spatial weights matrix.}

\item{panel_index}{Required when `"moran"` is in `diagnostics`.
A data.frame with columns `id` and `time` identifying each element of
`object$residuals`.}

\item{...}{Additional arguments passed to individual diagnostic functions.}
}
\value{
A `gwpr_diagnostics` object (list with class `"gwpr_diagnostics"`)
  whose `diagnostics` slot contains the result of each requested test.
}
\description{
Top-level interface that dispatches to individual diagnostic sub-functions.
Returns a `gwpr_diagnostics` object containing all requested test results.
}
\details{
Tests that are not applicable to the fitted model (e.g., Hausman test on a
pooling model) return a list with `status = "not_applicable"` and an
explanatory `message`, rather than an error.
}
\examples{
\donttest{
library(sf)
pts <- sf::st_as_sf(
  data.frame(id = 1:4, X = c(0,1,0,1), Y = c(0,0,1,1)),
  coords = c("X", "Y"), crs = NA_integer_
)
dat <- data.frame(
  id   = rep(1:4, each = 5),
  time = rep(1:5, 4),
  y    = rnorm(20),
  x1   = rnorm(20)
)
fit <- fit_gwpr(y ~ x1, data = dat, spatial = pts, id = "id",
                time = "time", bandwidth = 2, workers = 1)
diag_result <- diagnose_gwpr(fit, diagnostics = c("f_test", "hausman"))
print(diag_result)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/memory.R
\name{estimate_memory}
\alias{estimate_memory}
\title{Estimate memory usage for a GWPR run}
\usage{
estimate_memory(context, workers = 1, cache_distance = NULL)
}
\arguments{
\item{context}{A `gwpr_context` list containing at least
\code{metadata$n_units}, \code{metadata$n_time}, and
\code{metadata$n_vars}.  If these keys are absent, the function falls
back to direct arguments \code{n_units}, \code{n_time}, and
\code{n_vars} when supplied via \code{...}.  Alternatively, pass a
plain list with the required scalar fields directly.}

\item{workers}{Positive integer.  Number of parallel workers.}

\item{cache_distance}{Logical or \code{NULL}.  When \code{TRUE} the full
\eqn{n\_units \times n\_units} distance matrix is assumed to be kept in
memory.  When \code{NULL} (default) the function assumes caching is
enabled for a conservative estimate.}
}
\value{
A named list with class \code{"gwpr_memory_estimate"}:
  \describe{
    \item{n_units}{Number of spatial units.}
    \item{n_time}{Number of time periods.}
    \item{n_vars}{Number of explanatory variables.}
    \item{n_rows}{Total panel rows (\code{n_units * n_time}).}
    \item{workers}{Workers used for the estimate.}
    \item{cache_distance}{Whether distance caching was assumed.}
    \item{distance_bytes}{Bytes for the distance matrix (0 if not cached).}
    \item{model_bytes}{Bytes for local model copies across workers.}
    \item{total_bytes}{Total estimated bytes.}
    \item{risk}{Character risk level: \code{"low"}, \code{"medium"}, or
      \code{"high"}.}
  }
}
\description{
Calculates an approximate memory requirement (in bytes) based on the
data dimensions stored in a `gwpr_context` object, the number of
parallel workers, and whether the full distance matrix will be cached.
}
\details{
The two main cost components are:
\itemize{
  \item Distance matrix (when \code{cache_distance = TRUE}):
        \eqn{n\_units^2 \times 8} bytes.
  \item Local model working copies (one per worker):
        \eqn{n\_rows \times n\_vars \times 8 \times workers} bytes.
}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data_prepare.R
\name{prepare_spatial_data}
\alias{prepare_spatial_data}
\title{Extract and align spatial data for GWPR fitting}
\usage{
prepare_spatial_data(context)
}
\arguments{
\item{context}{A `gwpr_context` with `raw_spatial`, `panel_data`, and `id`
populated.}
}
\value{
Updated context with `spatial_data` and `coords` populated.
}
\description{
Extracts geometry and representative coordinates from the `raw_spatial` sf
object. Aligns spatial rows to the unique individual IDs present in
`panel_data` via `id_map`.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/spatial_sf.R
\name{align_spatial_to_panel}
\alias{align_spatial_to_panel}
\title{Reorder spatial rows to match panel ID order}
\usage{
align_spatial_to_panel(spatial, id_map)
}
\arguments{
\item{spatial}{An `sf` object containing at minimum the rows referenced in
`id_map`.}

\item{id_map}{A named integer vector as produced by `build_id_map()`.
Names are unit IDs (character); values are 1-based row indices into
`spatial`.}
}
\value{
An `sf` object with rows reordered to match `id_map`.
}
\description{
Given a named integer vector `id_map` (names = unit IDs, values = row
indices in `spatial`) produced by `build_id_map()`, this function reorders
the rows of `spatial` so that they align with the panel data ordering.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/weights.R
\name{weights}
\alias{weights}
\title{Distance and Kernel Weights Module for GWPR.light 1.0.0}
\description{
Internal functions for computing spatial distances and geographically
weighted kernel weights. Supports fixed and adaptive bandwidths, and five
kernel functions: bisquare, gaussian, exponential, tricube, and boxcar.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/result.R
\name{result}
\alias{result}
\title{Result Object Module for GWPR.light 1.0.0}
\description{
S3 classes and constructor functions for the three result objects returned
by the GWPR.light public API: `gwpr_fit`, `gwpr_bandwidth`, and
`gwpr_diagnostics`.  Also provides `build_spatial_results()` for assembling
a data.frame that can be aligned with an `sf` geometry column.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data_prepare.R
\name{build_model_matrix}
\alias{build_model_matrix}
\title{Build the design matrix and response vector}
\usage{
build_model_matrix(context)
}
\arguments{
\item{context}{A `gwpr_context` with `model_frame`, `formula`, and `family`
populated.}
}
\value{
Updated context with `model_matrix` and `response` populated.
}
\description{
Extracts the response variable `y` and the design matrix `X` from the
model frame.  For `binomial` family, the response is standardised to 0/1
via `standardize_binary_response()`.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/validation.R
\name{validate_spatial}
\alias{validate_spatial}
\title{Validate a spatial sf object}
\usage{
validate_spatial(spatial, id)
}
\arguments{
\item{spatial}{An `sf` object representing spatial units.}

\item{id}{Name of the ID column that must be present in `spatial`.}
}
\value{
Invisibly returns `TRUE` when valid; stops otherwise.
}
\description{
Validate a spatial sf object
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/model_linear.R
\name{fit_linear_local_model}
\alias{fit_linear_local_model}
\title{Fit a single local panel linear model}
\usage{
fit_linear_local_model(
  local_data,
  formula,
  model,
  effect,
  weights,
  index,
  random_method = "swar"
)
}
\arguments{
\item{local_data}{A `pdata.frame` or plain `data.frame` with columns for
all formula variables plus the panel indices.}

\item{formula}{A formula object.}

\item{model}{Character scalar: `"pooling"`, `"within"`, or
`"random"`.}

\item{effect}{Character scalar: `"individual"`, `"time"`, `"two-way"`,
or `"nested"`.}

\item{weights}{Numeric vector of kernel weights aligned with the rows of
`local_data`.}

\item{index}{Character vector of length 2 giving the panel index
column names: `c(id_col, time_col)`.}

\item{random_method}{Character scalar; estimation method for variance
components when `model = "random"` (default `"swar"`).}
}
\value{
A list with elements:
\describe{
  \item{`fit`}{The fitted model object, or `NULL` on failure.}
  \item{`status`}{`"ok"` or `"failed"`.}
  \item{`error`}{`NULL` or character string with the error message.}
  \item{`metadata`}{Named list with additional model-fitting metadata,
    e.g. flagging single-observation individuals for within models.}
}
}
\description{
Fits a geographically weighted panel linear model for one focal spatial
unit.  Errors are caught and returned as a structured failure result rather
than propagating to the caller.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data_prepare.R
\name{build_id_map}
\alias{build_id_map}
\title{Build a mapping from panel unit IDs to spatial row indices}
\usage{
build_id_map(panel_data, spatial_data, id)
}
\arguments{
\item{panel_data}{A data frame with a column named `id`.}

\item{spatial_data}{An `sf` data frame with a column named `id`.}

\item{id}{Character; name of the shared ID column.}
}
\value{
Named integer vector mapping unit ID to spatial row index.
}
\description{
Returns a named integer vector where each name is a panel unit ID (as a
character string) and each value is the corresponding row index in
`spatial_data` (1-based).
}
\details{
Rules:
- Every panel ID must have a spatial match; missing IDs cause an error.
- Extra spatial rows (not in panel) are silently ignored.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/diagnostics.R
\name{diagnose_local_f}
\alias{diagnose_local_f}
\title{Local F test diagnostic on a gwpr_fit object}
\usage{
diagnose_local_f(object, ...)
}
\arguments{
\item{object}{A `gwpr_fit` object.}

\item{...}{Currently ignored.}
}
\value{
A named list with elements:
\describe{
  \item{`local_f`}{Data frame with columns `unit_id`, `statistic`, `p_value`,
    `df1`, `df2`, `status`.}
  \item{`n_tested`}{Number of units tested.}
  \item{`n_failed`}{Number of units where the test could not be computed.}
}
}
\description{
Performs a local F test (fixed effects vs. pooling) using per-unit local
residuals stored in the fitted model object.
}
\details{
**Applicable models**: gaussian (linear).  Not applicable to logistic models;
returns a `status = "not_applicable"` result when `family = "binomial"`.

**Panel balance requirement**: No constraint; the test uses per-unit local
residuals already computed during fitting.

**Failure conditions**: If `local_results` is empty or missing, all units
are reported as failed.  If a unit's local result does not contain the
information needed (within and pooling residuals), that unit is reported as
failed with an informative `status`.

**Logistic interpretation limit**: Not applicable; see above.
}
\examples{
\donttest{
library(sf)
pts <- sf::st_as_sf(
  data.frame(id = 1:4, X = c(0,1,0,1), Y = c(0,0,1,1)),
  coords = c("X", "Y"), crs = NA_integer_
)
dat <- data.frame(
  id   = rep(1:4, each = 5),
  time = rep(1:5, 4),
  y    = rnorm(20),
  x1   = rnorm(20)
)
fit <- fit_gwpr(y ~ x1, data = dat, spatial = pts, id = "id",
                time = "time", bandwidth = 2, workers = 1)
diagnose_local_f(fit)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/diagnostics.R
\name{diagnose_moran}
\alias{diagnose_moran}
\title{Run Moran's I diagnostic on a gwpr_fit object}
\usage{
diagnose_moran(object, spatial_weights, panel_index, ...)
}
\arguments{
\item{object}{A `gwpr_fit` object returned by `fit_gwpr()` or similar.}

\item{spatial_weights}{A row-standardised n x n spatial weights matrix.
`n` must equal the number of spatial individuals in the fitted model.}

\item{panel_index}{A data.frame or list with columns/elements `id` and
`time` that identify each element of `object$residuals`.}

\item{...}{Currently ignored.}
}
\value{
A named list compatible with the `diagnostics` slot of a
  `gwpr_diagnostics` object.  Contains the elements returned by
  `compute_panel_moran()` plus `residual_type`.
}
\description{
Extracts residuals from a fitted GWPR model (Pearson residuals for logistic
models, raw residuals for linear models) and computes the panel Moran's I
statistic.
}
\details{
**Applicable models**: gaussian (linear residuals) and binomial (Pearson
residuals).

**Panel balance**: See `compute_panel_moran()`.

**Failure conditions**: Fails if `object` is not a `gwpr_fit`, if
`object$residuals` is `NULL`, or if `spatial_weights` dimensions do not
match the number of individuals.

**Logistic interpretation limit**: Moran's I computed on Pearson residuals
is exploratory; the asymptotic distribution differs from the linear case.
}
\examples{
\donttest{
library(sf)
pts <- sf::st_as_sf(
  data.frame(id = 1:4, X = c(0,1,0,1), Y = c(0,0,1,1)),
  coords = c("X", "Y"), crs = NA_integer_
)
dat <- data.frame(
  id   = rep(1:4, each = 5),
  time = rep(1:5, 4),
  y    = rnorm(20),
  x1   = rnorm(20)
)
fit <- fit_gwpr(y ~ x1, data = dat, spatial = pts, id = "id",
                time = "time", bandwidth = 2, workers = 1)
W <- matrix(1/3, nrow = 4, ncol = 4); diag(W) <- 0
idx <- dat[, c("id", "time")]
diagnose_moran(fit, W, idx)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/validation.R
\name{validate_inputs}
\alias{validate_inputs}
\title{Validate all inputs to the main GWPR functions}
\usage{
validate_inputs(
  formula,
  data,
  spatial,
  id,
  time,
  family = c("gaussian", "binomial"),
  model = c("within", "pooling", "random"),
  effect = c("individual", "time", "two-way", "nested"),
  kernel = c("bisquare", "gaussian", "exponential", "tricube", "boxcar"),
  adaptive = FALSE,
  workers = 1L
)
}
\arguments{
\item{formula}{A formula object.}

\item{data}{A data frame.}

\item{spatial}{An `sf` object.}

\item{id}{Name of the unit ID column (character).}

\item{time}{Name of the time column (character).}

\item{family}{`"gaussian"` or `"binomial"`.}

\item{model}{`"pooling"`, `"within"`, or `"random"`.}

\item{effect}{`"individual"`, `"time"`, `"two-way"`, or `"nested"`.}

\item{kernel}{One of `"bisquare"`, `"gaussian"`, `"exponential"`,
`"tricube"`, `"boxcar"`.}

\item{adaptive}{Logical.}

\item{workers}{Positive integer.}
}
\value{
Invisibly returns `TRUE` when all checks pass; stops otherwise.
}
\description{
This is the top-level validation entry point.  It calls the individual
`validate_*` helpers and also checks `model`, `effect`, and `kernel`.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/spatial_sf.R
\name{build_neighbor_structure}
\alias{build_neighbor_structure}
\title{Build a neighbour structure from an sf object}
\usage{
build_neighbor_structure(spatial, type = c("distance", "contiguity"))
}
\arguments{
\item{spatial}{An `sf` object.}

\item{type}{Character scalar: `"distance"` (default) or `"contiguity"`.}
}
\value{
* For `"distance"`: a numeric matrix with columns `X` and `Y`.
* For `"contiguity"`: a named list of integer vectors.
}
\description{
Returns different structures depending on `type`:
}
\details{
* `"distance"` — a numeric coordinate matrix (columns `X` and `Y`) suitable
  for pairwise distance computation.
* `"contiguity"` — a named list where each element is the integer vector of
  neighbour row indices (1-based, Queen contiguity via
  `sf::st_relate()`). Used for spatial diagnostics such as Moran's I.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/context.R
\name{new_gwpr_context}
\alias{new_gwpr_context}
\title{Construct a new gwpr_context object}
\usage{
new_gwpr_context(
  call = NULL,
  formula = NULL,
  family = NULL,
  model = NULL,
  effect = NULL,
  id = NULL,
  time = NULL,
  kernel = NULL,
  adaptive = NULL,
  threshold = NULL,
  workers = NULL,
  seed = NULL,
  raw_data = NULL,
  raw_spatial = NULL,
  panel_data = NULL,
  spatial_data = NULL,
  id_map = NULL,
  coords = NULL,
  model_frame = NULL,
  model_matrix = NULL,
  response = NULL,
  metadata = list(),
  warnings = character(),
  ...
)
}
\arguments{
\item{call}{The matched call from the top-level API function.}

\item{formula}{A formula object.}

\item{family}{Character: `"gaussian"` or `"binomial"`.}

\item{model}{Character: `"pooling"`, `"within"`, or `"random"`.}

\item{effect}{Character: `"individual"`, `"time"`, `"two-way"`, or
`"nested"`.}

\item{id}{Name of the unit ID column.}

\item{time}{Name of the time column.}

\item{kernel}{Kernel name.}

\item{adaptive}{Logical; `TRUE` for adaptive bandwidth.}

\item{threshold}{Numeric classification threshold (Logistic).}

\item{workers}{Number of parallel workers.}

\item{seed}{Integer random seed or `NULL`.}

\item{raw_data}{The original user-supplied data frame.}

\item{raw_spatial}{The original user-supplied sf object.}

\item{panel_data}{Processed panel data frame.}

\item{spatial_data}{Processed sf object.}

\item{id_map}{Named integer vector mapping unit IDs to row indices.}

\item{coords}{Matrix of spatial coordinates.}

\item{model_frame}{Model frame derived from formula and panel_data.}

\item{model_matrix}{Design matrix.}

\item{response}{Numeric response vector.}

\item{metadata}{Named list of supplementary information.}

\item{warnings}{Character vector of accumulated warnings.}

\item{...}{Additional named fields stored in the context list.}
}
\value{
A named list with class `"gwpr_context"`.
}
\description{
Creates the standardised internal context list used to pass state between
GWPR modules.  Any field not supplied defaults to `NULL` (or, for
`metadata` and `warnings`, to their appropriate empty types).
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bandwidth_sgd.R
\name{.make_na_history_row}
\alias{.make_na_history_row}
\title{Create a placeholder NA history row for epochs after early stopping}
\usage{
.make_na_history_row(ep, bw, learning_rate, delta, batch_size)
}
\description{
Create a placeholder NA history row for epochs after early stopping
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/memory.R
\name{classify_memory_risk}
\alias{classify_memory_risk}
\title{Classify memory risk level}
\usage{
classify_memory_risk(estimated_bytes)
}
\arguments{
\item{estimated_bytes}{Non-negative numeric; total estimated bytes.}
}
\value{
A character string: \code{"low"}, \code{"medium"}, or
  \code{"high"}.
}
\description{
Maps an estimated byte count to a human-readable risk category.
}
\details{
Thresholds:
\itemize{
  \item \strong{low}:    < 500 MB
  \item \strong{medium}: 500 MB -- 2 GB
  \item \strong{high}:   > 2 GB
}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data_prepare.R
\name{extract_coords_from_sf}
\alias{extract_coords_from_sf}
\title{Extract representative XY coordinates from an sf object}
\usage{
extract_coords_from_sf(spatial)
}
\arguments{
\item{spatial}{An `sf` object.}
}
\value{
A numeric matrix with columns `X` and `Y`.
}
\description{
Uses point coordinates for POINT geometries and centroids for other
geometry types (POLYGON, MULTIPOLYGON, etc.).
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data_prepare.R
\name{prepare_data}
\alias{prepare_data}
\title{Prepare all data structures for GWPR fitting}
\usage{
prepare_data(context)
}
\arguments{
\item{context}{A `gwpr_context` object with at least `formula`, `family`,
`id`, `time`, `model`, `raw_data`, and `raw_spatial` populated.}
}
\value{
An updated `gwpr_context` with `panel_data`, `spatial_data`,
  `id_map`, `coords`, `model_frame`, `model_matrix`, `response`, and
  `metadata` filled in.
}
\description{
Orchestrates the full data preparation pipeline: panel data extraction,
spatial data extraction, ID mapping, model frame and model matrix
construction. Updates and returns a `gwpr_context`.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/model_logistic.R
\name{predict_logistic_local_model}
\alias{predict_logistic_local_model}
\title{Predict probabilities for a local logistic model}
\usage{
predict_logistic_local_model(local_result, local_data)
}
\arguments{
\item{local_result}{A list as returned by \code{fit_logistic_local_model()}.}

\item{local_data}{A \code{data.frame} used for prediction.}
}
\value{
Numeric vector of probabilities in \code{[0, 1]}.  Returns a vector
  of \code{NA_real_} on failure.
}
\description{
Returns predicted probabilities (type = \code{"response"}) for the rows of
\code{local_data}, using the fitted model stored in \code{local_result}.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data_prepare.R
\name{data_prepare}
\alias{data_prepare}
\title{Data Preparation Module for GWPR.light 1.0.0}
\description{
Internal functions that convert user inputs into the internal data structures
required by the model engine: panel indices, spatial alignment, model frame,
and model matrix.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/model_logistic.R
\name{fit_logistic_random}
\alias{fit_logistic_random}
\title{Fit a random-effects logistic model via glmmTMB::glmmTMB}
\usage{
fit_logistic_random(
  local_data,
  formula,
  effect,
  weights,
  id_col,
  time_col,
  family = "binomial"
)
}
\arguments{
\item{local_data}{A \code{data.frame} containing all formula variables plus
the panel index columns.}

\item{formula}{A formula object with a binary response (no random-effect
terms; those are added automatically from \code{effect}).}

\item{effect}{Character scalar: one of \code{"individual"},
\code{"time"}, \code{"two-way"}, \code{"nested"}.}

\item{weights}{Numeric vector of kernel weights.}

\item{id_col}{Name of the individual ID column.}

\item{time_col}{Name of the time column.}

\item{family}{Character scalar reserved for future extension.}
}
\value{
A \code{glmmTMB} object.
}
\description{
Fit a random-effects logistic model via glmmTMB::glmmTMB
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/weights.R
\name{compute_distance}
\alias{compute_distance}
\title{Compute a full pairwise distance matrix from a coordinate matrix}
\usage{
compute_distance(coords, longlat = FALSE)
}
\arguments{
\item{coords}{A numeric matrix with at least two columns (`X` and `Y`, or
the first two columns if names are absent).  Each row is one spatial unit.}

\item{longlat}{Logical.  If `TRUE`, great-circle distances (in kilometres)
are calculated using the Haversine formula.  If `FALSE` (default),
Euclidean distance is used.}
}
\value{
An n x n symmetric numeric matrix where element [i, j] is the
  distance between unit i and unit j.  Diagonal is 0.
}
\description{
Compute a full pairwise distance matrix from a coordinate matrix
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bandwidth_sgd.R
\name{.rows_to_df}
\alias{.rows_to_df}
\title{Convert a list of history row lists to a data.frame}
\usage{
.rows_to_df(rows)
}
\description{
Convert a list of history row lists to a data.frame
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/model_logistic.R
\name{validate_binary_response}
\alias{validate_binary_response}
\title{Validate that a response vector is suitable for binary logistic regression}
\usage{
validate_binary_response(y)
}
\arguments{
\item{y}{A numeric, logical, or factor vector.}
}
\value{
Invisibly \code{TRUE} when validation passes.
}
\description{
Stops with an informative error when \code{y} is a factor with more than two
levels.  Passes through numeric 0/1 vectors and two-level factors unchanged.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/parallel.R
\name{parallel_module}
\alias{parallel_module}
\title{Parallel Execution Module}
\description{
Unified parallel execution interface for bandwidth search,
  local model fitting, and diagnostics. Shields backend differences and
  ensures CRAN-friendly behaviour.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/validation.R
\name{validate_workers}
\alias{validate_workers}
\title{Validate the workers argument}
\usage{
validate_workers(workers)
}
\arguments{
\item{workers}{Number of parallel workers. Must be a positive integer.}
}
\value{
Invisibly returns `TRUE` when valid; stops otherwise.
}
\description{
Validate the workers argument
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bandwidth_grid.R
\name{score_bandwidth_candidate}
\alias{score_bandwidth_candidate}
\title{Score a single bandwidth candidate}
\usage{
score_bandwidth_candidate(context, bandwidth, scorer)
}
\arguments{
\item{context}{A `gwpr_context` list.}

\item{bandwidth}{Numeric scalar; the candidate bandwidth to evaluate.}

\item{scorer}{A function with signature `scorer(context, bandwidth)`
returning a named list with at minimum:
\describe{
  \item{`score`}{Numeric scalar.  The criterion value (lower is better).}
  \item{`criterion`}{Character scalar.  Name of the criterion.}
  \item{`n_local_models`}{Integer.  Number of local models attempted.}
  \item{`n_failed_local_models`}{Integer.  Number of local models that
    failed.}
  \item{`metrics`}{Named list of aggregate metrics (e.g. R2, MSE, etc.).}
}}
}
\value{
A named list describing the candidate result:
\describe{
  \item{`bandwidth`}{The evaluated bandwidth.}
  \item{`score`}{Numeric criterion score, or `NA_real_` on failure.}
  \item{`criterion`}{Name of the scoring criterion.}
  \item{`status`}{`"ok"` or `"failed"`.}
  \item{`error_message`}{`NA_character_` or error text.}
  \item{`warning_message`}{`NA_character_` or warning text.}
  \item{`elapsed_time`}{Elapsed wall-clock time in seconds.}
  \item{`n_local_models`}{Number of local models attempted.}
  \item{`n_failed_local_models`}{Number of local models that failed.}
  \item{`r2`, `mse`, `rmse`, `mae`}{Linear metrics, or `NA_real_`.}
  \item{`log_loss`, `accuracy`, `precision`, `recall`, `f1_score`}{Logistic
    metrics, or `NA_real_`.}
}
}
\description{
Calls the `scorer` function for a given bandwidth and records the result
together with timing information, model counts, and metric values.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/api.R
\name{gwpr}
\alias{gwpr}
\title{Fit a Geographically Weighted Panel Regression (main entry point)}
\usage{
gwpr(
  formula,
  data,
  spatial,
  id,
  time,
  family = c("gaussian", "binomial"),
  model = c("within", "pooling", "random"),
  effect = c("individual", "time", "two-way", "nested"),
  bandwidth = NULL,
  bandwidth_method = c("sgd", "grid", "random"),
  bandwidth_control = list(),
  kernel = c("bisquare", "gaussian", "exponential", "tricube", "boxcar"),
  adaptive = FALSE,
  threshold = 0.5,
  workers = 1L,
  seed = NULL,
  diagnostics = TRUE,
  ...
)
}
\arguments{
\item{formula}{A \code{formula} object specifying the model (e.g.
\code{y ~ x1 + x2}).}

\item{data}{A \code{data.frame} containing the panel data.  Must include
the columns referenced by \code{id} and \code{time}.}

\item{spatial}{An \code{sf} object with one row per spatial unit.  Must
include the column referenced by \code{id}.}

\item{id}{Character scalar.  Name of the unit (individual) ID column shared
by \code{data} and \code{spatial}.}

\item{time}{Character scalar.  Name of the time-period column in \code{data}.}

\item{family}{Character scalar.  Model family: \code{"gaussian"} (default,
linear GWPR) or \code{"binomial"} (binary panel logistic GWPR).}

\item{model}{Character scalar.  Panel model type: \code{"within"} (default),
\code{"pooling"}, or \code{"random"}.}

\item{effect}{Character scalar.  Panel effect: \code{"individual"} (default),
\code{"time"}, \code{"two-way"}, or \code{"nested"}.}

\item{bandwidth}{Numeric scalar or \code{NULL} (default).  When \code{NULL}
the bandwidth is selected automatically via \code{select_bandwidth()}.}

\item{bandwidth_method}{Character scalar.  Method for automatic bandwidth
search: \code{"sgd"} (default), \code{"grid"}, or \code{"random"}.
Ignored when \code{bandwidth} is supplied.}

\item{bandwidth_control}{Named list of control parameters passed to the
bandwidth search function.  For \code{"grid"}: \code{lower}, \code{upper},
\code{step}.  For \code{"sgd"}: \code{lower}, \code{upper}, learning rate
etc.  For \code{"random"}: \code{lower}, \code{upper}, \code{n_samples}.}

\item{kernel}{Character scalar.  Kernel function: \code{"bisquare"}
(default), \code{"gaussian"}, \code{"exponential"}, \code{"tricube"}, or
\code{"boxcar"}.}

\item{adaptive}{Logical scalar.  \code{FALSE} (default) uses a fixed
distance bandwidth; \code{TRUE} uses an adaptive (k-nearest-neighbour)
bandwidth.}

\item{threshold}{Numeric scalar.  Classification threshold for
\code{family = "binomial"} (default \code{0.5}).}

\item{workers}{Positive integer.  Number of parallel workers.  \code{1}
(default) uses serial execution; values \code{> 1} enable explicit
parallelism.}

\item{seed}{Integer or \code{NULL}.  Random seed for reproducibility of any
stochastic steps (bandwidth search, parallel RNG).}

\item{diagnostics}{Logical scalar.  When \code{TRUE} (default),
\code{diagnose_gwpr()} is called after fitting and its results are stored
in the returned object.}

\item{...}{Additional arguments passed to the bandwidth search or fitting
functions.}
}
\value{
A \code{gwpr_fit} object.  Key fields:
\describe{
  \item{\code{local_results}}{Per-unit local model results.}
  \item{\code{predictions}}{In-sample predicted values / probabilities.}
  \item{\code{residuals}}{Residuals or Pearson residuals.}
  \item{\code{metrics}}{Overall goodness-of-fit metrics.}
  \item{\code{spatial_results}}{Data frame of per-unit coefficients.}
  \item{\code{search}}{Bandwidth search result (\code{gwpr_bandwidth}), or
    \code{NULL} when bandwidth was supplied directly.}
  \item{\code{diagnostics}}{A \code{gwpr_diagnostics} object, or
    \code{NULL}.}
}
}
\description{
Orchestrates the complete GWPR pipeline: input validation, data preparation,
optional memory estimation, optional bandwidth search, model fitting, and
optional diagnostics.
}
\examples{
\donttest{
# Minimal linear GWPR with a fixed bandwidth
library(sf)
pts <- sf::st_as_sf(
  data.frame(id = 1:4, X = c(0,1,0,1), Y = c(0,0,1,1)),
  coords = c("X", "Y"), crs = NA_integer_
)
dat <- data.frame(
  id   = rep(1:4, each = 5),
  time = rep(1:5, 4),
  y    = rnorm(20),
  x1   = rnorm(20)
)
fit <- gwpr(y ~ x1, data = dat, spatial = pts, id = "id", time = "time",
            bandwidth = 2, diagnostics = FALSE, workers = 1)
print(fit)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/TransAirPolCalif.R
\docType{data}
\name{TransAirPolCalif}
\alias{TransAirPolCalif}
\title{Panel Dataset for Testing GWPR}
\format{
A \code{data.frame} with 23 variables, and 928 observations, which are:
\describe{
\item{GEOID}{a numeric vector, fips IDs of the counties}
\item{year}{a numeric vector, year}
\item{pm25}{a numeric vector, annually average PM2.5 concentration in the counties}
\item{co2_mean}{a numeric vector, geographically average CO2 emission from on-road transportation in each year, million tons/km2}
\item{Developed_Open_Space_perc}{a numeric vector, percentage of developed open space of total area in each county}
\item{Developed_Low_Intensity_perc}{a numeric vector, percentage of low-intensity developed area of total area in each county}
\item{Developed_Medium_Intensity_perc}{a numeric vector, percentage of medium-intensity developed area of total area in each county}
\item{Developed_High_Intensity_perc}{a numeric vector, percentage of high-intensity develope area of total area in each county}
\item{Open_Water_perc}{a numeric vector, percentage of open water of total area in each county}
\item{Woody_Wetlands_perc}{a numeric vector, percentage of woody wetland of total area in each county}
\item{Emergent_Herbaceous_Wetlands_perc}{a numeric vector, percentage of emergent herbaceous wetland of total area in each county}
\item{Deciduous_Forest_perc}{a numeric vector, percentage of deciduous forest of total area in each county}
\item{Evergreen_Forest_perc}{a numeric vector, percentage of evergreen forest of total area in each county}
\item{Mixed_Forest_perc}{a numeric vector, percentage of mixed forest of total area in each county}
\item{Shrub_perc}{a numeric vector, percentage of shrub of total area in each county}
\item{Grassland_perc}{a numeric vector, percentage of grassland of total area in each county}
\item{Pasture_perc}{a numeric vector, percentage of pasture of total area in each county}
\item{Cultivated_Crops_perc}{a numeric vector, percentage of cultivated crops of total area in each county}
\item{pop_density}{a numeric vector, average population density in each county}
\item{summer_tmmx}{a numeric vector, average temperature in summer}
\item{winter_tmmx}{a numeric vector, average temperature in winter}
\item{summer_rmax}{a numeric vector, average humidity in summer}
\item{winter_rmax}{a numeric vector, average humidity in winter}
}
}
\usage{
data(TransAirPolCalif)
}
\description{
Panel dataset to estimate the relationship between county-level PM2.5 concentration and on-road transporation in California.
}
\examples{
\donttest{
data(TransAirPolCalif)
head(TransAirPolCalif)
}
}
\author{
Chao Li <chaoli0394@gmail.com> Shunsuke Managi <managi.s@gmail.com>
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/diagnostics.R
\name{diagnostics}
\alias{diagnostics}
\title{Diagnostics Module for GWPR.light 1.0.0}
\description{
Unified diagnostic interface for Geographically Weighted Panel Regression
models.  Provides Moran's I (spatial autocorrelation), local F test
(fixed vs. pooling), local Hausman test (fixed vs. random), and local
Breusch-Pagan LM test.
}
\details{
**Model applicability**

| Diagnostic   | Linear | Logistic | Notes                                      |
|--------------|--------|----------|--------------------------------------------|
| moran        | yes    | yes      | Logistic uses Pearson residual              |
| f_test       | yes    | no       | Requires within and pooling models          |
| hausman      | yes    | no       | Only meaningful for random-effect models    |
| lm_test      | yes    | no       | Pooling or random-effect models             |

**Panel balance**

`compute_panel_moran()` is fully supported for balanced panels.  For
unbalanced panels a `warning()` is issued and the function attempts
computation using only the time periods present in every individual;
results may be unreliable.

**Logistic interpretation**

Moran's I computed from Pearson residuals of a Logistic model does not
follow the same asymptotic distribution as for linear models.  Treat the
test result as an exploratory heuristic, not a formal test.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/validation.R
\name{validation}
\alias{validation}
\title{Input Validation Functions for GWPR.light 1.0.0}
\description{
These internal functions validate user inputs before any expensive computation,
providing early and consistent error messages.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/California.R
\docType{data}
\name{California}
\alias{California}
\title{California (sf)}
\format{
An \code{sf} object with 58 rows (one per county) and two columns:
\describe{
\item{GEOID}{a numeric vector, fips IDs of the counties}
\item{geometry}{sfc_MULTIPOLYGON, county boundary polygons (CRS: NAD83 longlat)}
}
}
\usage{
data(California)
}
\description{
The counties' boundary in California
}
\examples{
data(California)
class(California)
}
\author{
Chao Li <chaoli0394@gmail.com> Shunsuke Managi <managi.s@gmail.com>
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/metrics.R
\name{metrics}
\alias{metrics}
\title{Metrics Module}
\description{
Functions for computing evaluation metrics for linear and
  logistic panel regression models.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/validation.R
\name{validate_panel_index}
\alias{validate_panel_index}
\title{Validate panel index columns in a data frame}
\usage{
validate_panel_index(data, id, time)
}
\arguments{
\item{data}{A data frame.}

\item{id}{Name of the unit (individual) index column.}

\item{time}{Name of the time index column.}
}
\value{
Invisibly returns `TRUE` when valid; stops otherwise.
}
\description{
Validate panel index columns in a data frame
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/model_linear.R
\name{model_linear}
\alias{model_linear}
\title{Linear GWPR Engine for GWPR.light 1.0.0}
\description{
Internal functions for fitting Geographically Weighted Panel Regression
with a Gaussian (linear) response.  Supports pooling, within, and random
panel models, plus individual, time, two-way, and nested effects.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/parallel.R
\name{with_reproducible_seed}
\alias{with_reproducible_seed}
\title{Execute an expression with a reproducible seed}
\usage{
with_reproducible_seed(seed, expr)
}
\arguments{
\item{seed}{Integer scalar. Seed value passed to \code{set.seed}.}

\item{expr}{An R expression to evaluate.}
}
\value{
The value of \code{expr}.
}
\description{
Sets \code{set.seed(seed)} before evaluating \code{expr}, then restores the
prior RNG state so the caller's random stream is unaffected.
}
\examples{
r1 <- with_reproducible_seed(42, runif(3))
r2 <- with_reproducible_seed(42, runif(3))
stopifnot(identical(r1, r2))

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/model_linear.R
\name{extract_linear_local_result}
\alias{extract_linear_local_result}
\title{Extract coefficients and diagnostics from a local linear model}
\usage{
extract_linear_local_result(local_result)
}
\arguments{
\item{local_result}{A list as returned by `fit_linear_local_model()`.}
}
\value{
A named list with elements:
\describe{
  \item{`coefficients`}{Named numeric vector of local coefficient estimates.}
  \item{`se`}{Named numeric vector of standard errors (same names).}
  \item{`tvalues`}{Named numeric vector of t-statistics.}
  \item{`local_r2`}{Numeric scalar or `NA_real_`.}
  \item{`local_aic`}{Numeric scalar or `NA_real_`.}
  \item{`status`}{`"ok"` or `"failed"`.}
  \item{`error`}{`NULL` or character error message.}
}
}
\description{
Extract coefficients and diagnostics from a local linear model
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bandwidth_random.R
\name{bandwidth_random}
\alias{bandwidth_random}
\title{Random Bandwidth Optimizer (`bandwidth_random.R`)}
\description{
Implements a bounded random search for bandwidth selection.  A user-
specified number of candidate bandwidths are drawn uniformly at random from
`[lower, upper]`, scored by a user-supplied scorer function, and the
candidate with the lowest score is returned as the best bandwidth.

The search boundaries (`lower`, `upper`) **must** be set explicitly by the
caller; automatic inference is intentionally not supported.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/api.R
\name{.make_scorer}
\alias{.make_scorer}
\title{Build a scorer function for bandwidth search}
\usage{
.make_scorer(family = "gaussian", threshold = 0.5)
}
\arguments{
\item{family}{Character; `"gaussian"` or `"binomial"`.}

\item{threshold}{Numeric; classification threshold (binomial only).}
}
\value{
A function with signature `scorer(context, bandwidth)`.
}
\description{
Returns a closure that fits the full GWPR engine for a given bandwidth and
returns the score (MSE for linear, log_loss for logistic), together with
aggregate metrics.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/weights.R
\name{build_distance_context}
\alias{build_distance_context}
\title{Build a distance context object}
\usage{
build_distance_context(coords, ids, longlat = FALSE, cache = TRUE)
}
\arguments{
\item{coords}{A numeric matrix with columns `X` and `Y`.}

\item{ids}{Character or numeric vector of unit IDs (length = nrow(coords)).}

\item{longlat}{Logical.  Passed to `compute_distance()`.}

\item{cache}{Logical.  If `TRUE` (default), pre-computes and caches the
full n x n distance matrix.  Set `FALSE` for very large datasets to avoid
memory pressure; `get_local_distances()` will then compute rows on demand.}
}
\value{
A list with class `"gwpr_distance_context"` containing:
* `ids` — character vector of unit IDs.
* `distance_matrix` — n x n matrix (or `NULL` if `cache = FALSE`).
* `coords` — the original coordinate matrix.
* `longlat` — logical flag.
}
\description{
Wraps a distance matrix together with unit IDs into a list suitable for
passing to `get_local_distances()`.  Optionally pre-computes the full
distance matrix (recommended for small data) or stores only the coordinate
matrix for on-the-fly computation (recommended for large data).
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/weights.R
\name{compute_kernel_weights}
\alias{compute_kernel_weights}
\title{Compute geographically weighted kernel weights}
\usage{
compute_kernel_weights(distance, bandwidth, kernel, adaptive)
}
\arguments{
\item{distance}{Numeric vector of distances from the focal unit to all
spatial units (length n).}

\item{bandwidth}{Numeric scalar.  For fixed bandwidth: distance scale.
For adaptive bandwidth: positive integer number of neighbours.}

\item{kernel}{Character scalar, one of `"bisquare"`, `"gaussian"`,
`"exponential"`, `"tricube"`, `"boxcar"`.}

\item{adaptive}{Logical.  `TRUE` for adaptive (kNN) bandwidth.}
}
\value{
A numeric vector of length n with non-negative kernel weights.
}
\description{
Given a numeric vector of distances from one focal unit to all others,
returns a weight for each unit according to the chosen kernel and bandwidth.
}
\details{
For **fixed** bandwidth (`adaptive = FALSE`) the bandwidth parameter is a
distance threshold or scale parameter used directly in the kernel formula.

For **adaptive** bandwidth (`adaptive = TRUE`) the bandwidth parameter is
the number of nearest neighbours k.  The function first identifies the
k-th smallest distance (among all units, including the focal unit itself at
distance 0) and uses that distance as the effective bandwidth in the kernel
formula.

Kernel formulae (d = distance, bw = effective bandwidth):
* `bisquare`:    `(1 - (d/bw)^2)^2` for `d <= bw`, else 0.
* `gaussian`:    `exp(-0.5 * (d/bw)^2)`.
* `exponential`: `exp(-d/bw)`.
* `tricube`:     `(1 - (d/bw)^3)^3` for `d <= bw`, else 0.
* `boxcar`:      `1` for `d <= bw`, else 0.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/memory.R
\name{format_memory_warning}
\alias{format_memory_warning}
\title{Format a human-readable memory warning message}
\usage{
format_memory_warning(memory_estimate)
}
\arguments{
\item{memory_estimate}{A `gwpr_memory_estimate` list, typically the
return value of \code{\link{estimate_memory}}.}
}
\value{
A character string containing the warning text.  The string is
  suitable for passing to \code{message()} or \code{warning()}.
}
\description{
Converts a `gwpr_memory_estimate` object (produced by
\code{\link{estimate_memory}}) into a readable character string.  For
high-risk estimates the message also includes actionable suggestions.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/model_logistic.R
\name{extract_logistic_local_result}
\alias{extract_logistic_local_result}
\title{Extract coefficients and diagnostics from a local logistic model}
\usage{
extract_logistic_local_result(
  local_result,
  local_data,
  formula,
  threshold = 0.5,
  eps = 1e-15
)
}
\arguments{
\item{local_result}{A list as returned by \code{fit_logistic_local_model()}.}

\item{local_data}{The \code{data.frame} used to fit the model (needed to
extract \code{y} and to predict).}

\item{formula}{The model formula (used to extract the response name).}

\item{threshold}{Numeric scalar; classification threshold (default 0.5).}

\item{eps}{Numeric scalar; clipping bound for probability (default
\code{1e-15}).}
}
\value{
A named list with elements:
\describe{
  \item{\code{coefficients}}{Named numeric vector of local coefficient
    estimates, or \code{NA_real_} on failure.}
  \item{\code{prob}}{Numeric vector of predicted probabilities for the local
    data rows, or \code{NA_real_} on failure.}
  \item{\code{class_pred}}{Integer vector (0/1) of predicted classes, or
    \code{NA_real_} on failure.}
  \item{\code{pearson_resid}}{Numeric vector of Pearson residuals, or
    \code{NA_real_} on failure.}
  \item{\code{status}}{\code{"ok"} or \code{"failed"}.}
  \item{\code{error}}{\code{NULL} or character error message.}
}
}
\description{
Computes predicted probabilities, predicted classes, Pearson residuals,
and extracts coefficient estimates.  Pearson residuals are defined as
\code{(y - p) / sqrt(p * (1 - p))} with \code{p} clipped to
\code{[eps, 1 - eps]} to avoid division by zero.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/api.R
\name{api}
\alias{api}
\title{Public API for GWPR.light 1.0.0}
\description{
High-level user-facing functions for Geographically Weighted Panel
Regression.  These four functions form the complete public interface;
all internal complexity is hidden behind them.

\itemize{
  \item \code{\link{gwpr}} -- full pipeline (bandwidth search + fitting +
    optional diagnostics).
  \item \code{\link{select_bandwidth}} -- standalone bandwidth search.
  \item \code{\link{fit_gwpr}} -- fit with a known bandwidth.
  \item \code{\link{diagnose_gwpr}} -- run diagnostics on a fitted model.
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/weights.R
\name{get_local_distances}
\alias{get_local_distances}
\title{Extract distances from one focus unit to all others}
\usage{
get_local_distances(distance_context, focus_id)
}
\arguments{
\item{distance_context}{A list as returned by `build_distance_context()`, or
a plain n x n numeric distance matrix.  If a plain matrix is supplied, the
rows and columns must already be in the same order as the spatial units.}

\item{focus_id}{Integer scalar (1-based) or character matching a
row/column name of the distance matrix.  The focal unit whose distances
are extracted.}
}
\value{
A numeric vector of length n giving the distance from the focus unit
  to every unit (including itself, which is 0).
}
\description{
Extract distances from one focus unit to all others
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/model_logistic.R
\name{effect_to_feglm_fml}
\alias{effect_to_feglm_fml}
\title{Build a fixest::feglm formula string from the user-facing effect string}
\usage{
effect_to_feglm_fml(base_formula, effect, id_col, time_col)
}
\arguments{
\item{base_formula}{A formula object (without fixed-effect terms).}

\item{effect}{Character scalar: one of \code{"individual"},
\code{"time"}, \code{"two-way"}, \code{"nested"}.}

\item{id_col}{Name of the individual ID column.}

\item{time_col}{Name of the time column.}
}
\value{
A formula suitable for \code{fixest::feglm()}.
}
\description{
Build a fixest::feglm formula string from the user-facing effect string
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data_prepare.R
\name{prepare_panel_data}
\alias{prepare_panel_data}
\title{Prepare panel data from the raw data in context}
\usage{
prepare_panel_data(context)
}
\arguments{
\item{context}{A `gwpr_context` with `raw_data`, `formula`, `id`, `time`,
and `model` populated.}
}
\value{
Updated context with `panel_data` and `metadata` populated.
}
\description{
Extracts the `id`, `time`, and formula variables from `raw_data`. Adds a
`raw_row_id` column preserving the original row positions. Sorts by id then
time. Records panel balance information in `metadata`. For `within` models,
records single-observation individuals in `metadata`.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/spatial_sf.R
\name{spatial_sf}
\alias{spatial_sf}
\title{Spatial SF Module for GWPR.light 1.0.0}
\description{
Internal functions providing the `sf`-first spatial interface. These
functions abstract geometry extraction, coordinate derivation, spatial
alignment to panel data, and neighbour-structure construction. They are used
by the data preparation and diagnostics modules.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data_prepare.R
\name{build_model_frame}
\alias{build_model_frame}
\title{Build the model frame from the formula and panel data}
\usage{
build_model_frame(context)
}
\arguments{
\item{context}{A `gwpr_context` with `formula` and `panel_data` populated.}
}
\value{
Updated context with `model_frame` populated.
}
\description{
Build the model frame from the formula and panel data
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/spatial_sf.R
\name{extract_coordinates}
\alias{extract_coordinates}
\title{Extract representative XY coordinates from an sf object}
\usage{
extract_coordinates(spatial)
}
\arguments{
\item{spatial}{An `sf` object.}
}
\value{
A numeric matrix with columns `X` and `Y`, one row per feature.
}
\description{
For POINT geometries the point coordinates are returned directly. For
all other geometry types (POLYGON, MULTIPOLYGON, etc.) `sf::st_centroid()`
is used to derive a representative point, with warnings suppressed (they
are typically non-actionable geographic-CRS notes).
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/validation.R
\name{validate_bandwidth_control}
\alias{validate_bandwidth_control}
\title{Validate bandwidth search control parameters}
\usage{
validate_bandwidth_control(method, control, adaptive)
}
\arguments{
\item{method}{Character: `"grid"`, `"sgd"`, or `"random"`.}

\item{control}{Named list of control parameters.}

\item{adaptive}{Logical; if `TRUE`, bandwidth is a neighbour count.}
}
\value{
Invisibly returns `TRUE` when valid; stops otherwise.
}
\description{
Validate bandwidth search control parameters
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bandwidth_grid.R, R/parallel.R
\name{parallel_map}
\alias{parallel_map}
\title{Minimal parallel_map implementation}
\usage{
parallel_map(x, fn, workers = 1, seed = NULL, ..., packages = NULL)

parallel_map(x, fn, workers = 1, seed = NULL, ..., packages = NULL)
}
\arguments{
\item{x}{A list (or vector) of inputs to iterate over.}

\item{fn}{A function to apply to each element of \code{x}. The function
receives the element as its first argument; additional arguments are
passed via \code{...}.}

\item{workers}{Integer scalar. Number of parallel workers. \code{1}
(default) uses serial \code{lapply} and never touches \code{future}.}

\item{seed}{Integer scalar or \code{NULL}. Random seed for reproducibility.
In serial mode the R RNG is seeded with \code{set.seed(seed)}. In
parallel mode the seed is forwarded to \code{future_lapply} via the
\code{future.seed} argument using L'Ecuyer-CMRG streams.}

\item{...}{Additional arguments forwarded to \code{fn}.}

\item{packages}{Character vector of package names that workers need to load,
or \code{NULL} (default). Ignored in serial mode.}
}
\value{
A list of the same length as \code{x}. Elements where \code{fn}
  threw an error are replaced with a character string \code{"ERROR: <msg>"}.
}
\description{
Wraps `lapply` when `workers = 1` and `parallel::mclapply` when
`workers > 1`.  This stub is superseded once `parallel.R` is available.

A thin wrapper that uses plain \code{lapply} when \code{workers = 1} and
switches to \code{future.apply::future_lapply} with a \code{multisession}
plan for \code{workers > 1}. The global \code{future} plan is always
restored to \code{sequential} after the call, preventing side-effects.
}
\details{
Worker-level errors are caught and returned as character strings (prefixed
with \code{"ERROR: "}) rather than aborting the entire call.
}
\examples{
result <- parallel_map(1:3, function(x) x^2, workers = 1)
stopifnot(identical(result, list(1, 4, 9)))

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/context.R
\name{validate_gwpr_context}
\alias{validate_gwpr_context}
\title{Validate a gwpr_context object}
\usage{
validate_gwpr_context(context)
}
\arguments{
\item{context}{A list (typically of class `"gwpr_context"`) to validate.}
}
\value{
Invisibly returns `TRUE` when all core fields are present.
}
\description{
Checks that all core fields required for model fitting are non-`NULL`.
Stops with an informative message listing every missing field.
}
\details{
Core fields: `formula`, `family`, `id`, `time`, `model`, `effect`,
`kernel`, `adaptive`, `threshold`, `workers`.
}
\keyword{internal}
